;;; footnote-init.el --- Reads in existing footnotes 

;; Version 1.7.1
;; Copyright (C) 2007, 2008, 2009  Andreas Roehler

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>
;; Keywords: mail, news, editing

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Major changes to previous version:
;;  `fi-solve-conflicts' removed


;; Source: https://code.launchpad.net/s-x-emacs-werkstatt/

;;; Commentary: `footnote-init' reads in existing
;; footnodes.

;; To make it work, footnote.el has to be patched: use
;; GNU-Emacs-footnode.diff resp. XEmacs-footnode.diff.

;; Please use `fi-Footnote-add-footnote' instead of
;; `Footnote-add-footnote', which will avoid some bugs.

;; Hint: In case of errors check the values of
;; `footnote-section-tag-regexp' and `footnote-section-tag'

;; Bugs: If `footnote-section-tag-regexp' and/or
;; `footnote-section-tag' are used in common text, text
;; may mistakenly be changed by footnote-renumbering


;; At the request of Uwe Brauer, oub at mat.ucm.es, it
;; provides commented footnodes, which might be used
;; inside programs, see screeshots `commented.elisp.footnote2.png' and
;; commented.tex.footnote.png 

;; Uwe Brauer too sent his settings:
;; 
;; (defun turn-on-my-footnote-mode () 
;; (interactive)
;; (footnote-mode 1))
;; 
;; (add-hook  'mail-mode-hook      'turn-on-my-footnote-mode)
;; (add-hook  'message-mode-hook   'turn-on-my-footnote-mode)
;; (add-hook  'matlab-mode-hook   'turn-on-my-footnote-mode)
;; (add-hook  'LaTeX-mode-hook   'turn-on-my-footnote-mode)
;; (add-hook  'text-mode-hook   'turn-on-my-footnote-mode)
;; (add-hook  'emacs-lisp-mode   'turn-on-my-footnote-mode)
;; (setq footnote-narrow-to-footnotes-when-editing t)  ;does not work
;; 
(defun fi-set-footnote-keys ()
  (interactive)
  (local-set-key [(control right)] 'fi-Footnote-add-footnote)
  (local-set-key [(control c) right] 'Footnote-cycle-style)
  (local-set-key [(control c) left] 'footnote-init)
  (local-set-key [(control c) f31] 'footnote-delete-footnote)
  (local-set-key [(control f30)] 'Footnote-goto-footnote)
  (local-set-key [(control left)] 'Footnote-back-to-message)
  (local-set-key [(super q)] 'fi-Footnote-what-style)
)
;; 
;; 
;; (add-hook 'footnote-mode-hook 'my-set-footnote-keys)
;; (add-hook 'footnote-mode-hook 'footnote-init)

;; (defun my-set-footnote-style ()
;; (interactive)
;; (Footnote-set-style 'numeric))
;; 

;; (setq footnote-prefix [(control c) (control n)])
;; (require 'footnote-init)

;; http://calypso.tux.org/pipermail/xemacs-beta/2009-May/016497.html

;; Changes to previous version: Bugs fixed which
;; occurred if `footnote-init' was called with buffer
;; without footnotes.

;;; Code:

(require 'footnote)

(make-variable-buffer-local 'footnote-start-tag)
(make-variable-buffer-local 'footnote-end-tag)

(setq footnote-comment-function 'comment-region)

(defvar footnote-current-regexp nil
  "Regexp of the index of the current style. 
Returned by fi-Footnote-current-regexp:
(setq footnote-current-regexp (assq footnote-style footnote-style-alist))
For intern use. ")
(make-variable-buffer-local 'footnote-current-regexp)

(defvar footnote-section-tag-pos nil
  "Intern use only. ")
(make-variable-buffer-local 'footnote-section-tag-pos)

(defvar footnote-initialised nil
       "Footnote init should only called once in a buffer")
(make-variable-buffer-local 'footnote-initialised)

(defcustom fi-solve-by-augment t
  "If footnote conflicts shall be solved by adding chars used by `footnote-start-tag' and `footnote-end-tag' to this. 
I.e. [1] might change to [[1]]. "
  :type 'boolean
  :group 'footnote)

(defcustom footnote-start-tag-2 "\<[["
  "Alternative string used to denote start of numbered
footnote. Customizing this variable may have no effect on buffers
already displaying footnotes."
  :type 'string
  :group 'footnote)
(make-variable-buffer-local 'footnote-start-tag-2)

(defcustom footnote-end-tag-2 "]]\>"
  "Alternative string used to denote end of numbered
footnote. Customizing this variable may have no effect on buffers
already displaying footnotes."
  :type 'string
  :group 'footnote)
(make-variable-buffer-local 'footnote-end-tag-2)

(defcustom footnote-comment-function 'comment-region
  "If non-nil, prompt before deleting a footnote.
There is currently no way to undo deletions."
  :type 'symbol
  :group 'footnote)
(make-variable-buffer-local 'footnote-comment-function)

(defcustom footnote-comment-function 'comment-region
  "If non-nil, prompt before deleting a footnote.
There is currently no way to undo deletions."
  :type 'symbol
  :group 'footnote)
(make-variable-buffer-local 'footnote-comment-function)

(defun fi-Footnote-restore-default-tags () 
  (interactive)
  ;; allow new footnote-init
  (setq footnote-initialised nil)
  (setq footnote-start-tag "[")
  (setq footnote-end-tag "]")
    (when (interactive-p) (message "footnote start end tags: %s %s" (concat "\"" footnote-start-tag "\"") (concat "\"" footnote-end-tag "\""))))

(defun fi-Footnote-set-clumsy-tags () 
  "Example function to set footnote-start- and end-tag. "
  (interactive)
  (setq footnote-start-tag "\<[[")
  (setq footnote-end-tag "]]\>")
    (when (interactive-p) (message "footnote start end tags: %s %s" (concat "\"" footnote-start-tag "\"") (concat "\"" footnote-end-tag "\""))))

(defun fi-Footnote-set-my-cats-preffered-tags () 
  "Example function to set footnote-start- and end-tag. "
  (interactive)
  (setq footnote-start-tag "«")
  (setq footnote-end-tag "»")
    (when (interactive-p) (message "footnote start end tags: %s %s" (concat "\"" footnote-start-tag "\"") (concat "\"" footnote-end-tag "\""))))

(defun fi-Footnote-set-latin () 
  "Set latin style with empty footnote-start- and end-tag. "
  (interactive)
  (Footnote-set-style 'latin)
  (setq footnote-start-tag "")
  (setq footnote-end-tag "")
    (when (interactive-p) (message "style: %s %s %s" footnote-style (concat "\"" footnote-start-tag "\"") (concat "\"" footnote-end-tag "\""))))

(defun fi-Footnote-what-style () 
  "Returns footnote-style footnote-start-tag footnote-end-tag fn-section-tag-beg.
footnote-style in effect according footnote-style-alist:
default is (numeric Footnote-numeric ,footnote-numeric-regexp)"
  (interactive)
  (let (fn-section-tag-beg fn-tag-beg fn-tag-end found done index-regexp)
    (save-excursion
      (goto-char (point-min))
      (when (re-search-forward (concat "^.*" "\\(" footnote-section-tag "\\)") nil t 1)
        (setq fn-section-tag-beg (match-beginning 1)) 
        (end-of-line) 
        (skip-chars-forward " \t\r\n\f")
        (when (stringp comment-start)
          (while (looking-at comment-start)
            (goto-char (match-end 0))))
        (skip-chars-forward " \t\r\n\f")
        (setq fn-tag-beg (point)) 
        (skip-chars-forward "^ \t\r\n\f")
        (setq fn-tag-end (point)) 
	(setq found (buffer-substring-no-properties fn-tag-beg fn-tag-end))
	(dolist (styles (reverse footnote-style-alist))
          (unless done
            (when
                (string-match (nth 2 styles) found)
              (Footnote-set-style (nth 0 styles))
              (setq index-regexp (nth 2 (assq footnote-style footnote-style-alist)))
              (setq done t))))
        (setq footnote-start-tag (buffer-substring-no-properties fn-tag-beg (+ fn-tag-beg (string-match index-regexp (buffer-substring-no-properties fn-tag-beg fn-tag-end)))))
        (setq footnote-end-tag (buffer-substring-no-properties (+ fn-tag-beg (match-end 0)) fn-tag-end)))
      (when (interactive-p) (message "Fstyle: %s \"%s\"  \"%s\""  footnote-style footnote-start-tag footnote-end-tag index-regexp))
      (list footnote-style footnote-start-tag footnote-end-tag fn-section-tag-beg index-regexp))))

(defun footnote-init (&optional beg end) 
  "Let's footnote take notice of already existing footnotes"
  (interactive)
  (unless footnote-mode (footnote-mode))
  (let* ((beg (cond (beg beg)
                    ((region-exists-p)
                     (region-beginning))
                    (t (point-min)))) 
         (end (cond (end (copy-marker end))
                    ((region-exists-p)
                     (copy-marker (region-end)))
                    (t (copy-marker (point-max)))))
         (stil (fi-Footnote-what-style))
         (footnote-style (car stil))
         (footnote-start-tag (nth 1 stil))
         (footnote-end-tag (nth 2 stil))
         (fn-section-tag-beg (nth 3 stil))
         (index-regexp (nth 4 stil)))
    ;; Check for conflicting chars/regexp at beginning
    ;; (unless footnote-initialised
      ;; (goto-char (point-min)) 
      ;; (when (re-search-forward (concat (if footnote-start-tag (regexp-quote footnote-start-tag)(regexp-quote "[")) ".*?" (if footnote-end-tag (regexp-quote footnote-end-tag)(regexp-quote "]"))) nil t 1)
        ;; (fi-solve-conflicts)))
    (fi-footnote-init-intern beg end footnote-style footnote-start-tag footnote-end-tag fn-section-tag-beg index-regexp)
    (unless (featurep 'xemacs) (Footnote-refresh-footnotes index-regexp))
    (setq footnote-initialised t)
    ))
  
(defun fi-footnote-init-intern (beg end footnote-style footnote-start-tag footnote-end-tag fn-section-tag-beg index-regexp) 
  (set (make-local-variable 'footnote-pointer-marker-alist) '())
  (set (make-local-variable 'footnote-text-marker-alist) '())
  (save-excursion
    (let* ((count 0)
           note)
      (when fn-section-tag-beg
        (goto-char beg) 
        (while (re-search-forward
                (concat (when footnote-start-tag (regexp-quote footnote-start-tag)) index-regexp (when footnote-end-tag (regexp-quote footnote-end-tag))) fn-section-tag-beg t 1)
          (setq count (1+ count))
          (delete-region (match-beginning 0) (match-end 0))
          (Footnote-insert-numbered-footnote count 1)
          (Footnote-insert-pointer-marker count (point)))
        (setq count 0)
        (while (re-search-forward (concat (when footnote-start-tag (regexp-quote footnote-start-tag)) index-regexp (when footnote-end-tag (regexp-quote footnote-end-tag))) nil t 1)
          (setq count (1+ count))
          (delete-region (match-beginning 0) (match-end 0))
          (Footnote-insert-numbered-footnote count 1)
          (Footnote-insert-text-marker count (point)))
        (when (interactive-p)(message "footnote-pointer-marker-alist: %s\n footnote-text-marker-alist:    %s" footnote-pointer-marker-alist footnote-text-marker-alist))))))

(defun fi-Footnote-add-footnote ()
  "Add footnote at point, update footnodes. "
  (interactive "*")
  (footnote-init)
  (Footnote-add-footnote)
  (when (featurep 'xemacs))(footnote-init))

(defun fi-Footnote-current-regexp ()
  (interactive)
  "Return the regexp of the index of the current style."
  (let ((erg (nth 2 (or (assq footnote-style footnote-style-alist)
		     (nth 0 footnote-style-alist)))))
    (when (interactive-p) (message "Footnote-current-regexp: %s " erg))
    erg))

(defun fi-Footnote-add-commented-footnote ()
  "Put added footnotes and textes inside a comment. "
  (interactive "*")
  (unless (string-match comment-start footnote-section-tag) 
    (setq footnote-section-tag (concat comment-start " Footnotes:" (unless (string= "" comment-end) comment-end))))
  (unless (string-match  "\\.\\*" footnote-section-tag-regexp)
    (setq footnote-section-tag-regexp (concat ".*" footnote-section-tag-regexp ".*$")))
  (let ((beg (point)) textend) 
    (save-excursion
      (fi-Footnote-add-footnote)
      (fi-comment-footnote-preserving-point (line-beginning-position) (point))
      (setq textend (copy-marker (point))))
    (fi-comment-footnote-preserving-point beg (progn (skip-chars-forward "^ \t\r\n\f") (point)))
    (goto-char textend)
    ;; (end-of-line)
    ))

(defun fi-comment-footnote-preserving-point (&optional beg end) 
  "Comment footnote unless already in comment. "
  (let ((beg (cond (beg beg)
		   ((region-exists-p)
		    (region-beginning))
		   (t (line-beginning-position)))) 
	(end (cond (end (copy-marker end))
		   ((region-exists-p)
		    (copy-marker (region-end)))
		   (t (copy-marker (line-end-position))))))
    (unless (or (if (functionp 'in-comment-p-lor)
                    (in-comment-p-lor)
                  (progn
                    (save-excursion
                      (unless (eolp) (forward-char 1))
                      (nth 4 (parse-partial-sexp (point-min) (point))))))) 
      (save-excursion
        (funcall footnote-comment-function beg (point))))))

(defun fi-conflicting-chars-p (&optional beg end) 
  "Return `t' if chars used as footnote delimiters occur in common text.
Assumes no footnote is set already, false positiv otherwise. "
  (interactive)
  (let* ((beg (cond (beg beg)
                    ((region-exists-p)
                     (region-beginning))
                    (t (point-min)))) 
         (end (cond (end (copy-marker end))
                    ((region-exists-p)
                     (copy-marker (region-end)))
                    (t (copy-marker (point-max)))))
         (conflict
          (progn
            (save-excursion 
              (goto-char beg)
              (when
                  (re-search-forward (concat (when footnote-start-tag (regexp-quote footnote-start-tag))
                                             (Footnote-current-regexp)
                                             (when footnote-end-tag (regexp-quote footnote-end-tag))) nil t 1)
                (point))))))
    (when conflict (goto-char conflict))
    (when (interactive-p)
      (if conflict
          (message "Expression used for footnote at point %d %s" conflict ", maybe footnote-start-tag, footnote-end-tag")
        (message "Okay!: %s \"%s\"  \"%s\""  footnote-style footnote-start-tag footnote-end-tag)))
    conflict))

(defun fi-solve-conflicts (&optional beg end) 
  " "
  (interactive)
  (let ((beg (cond (beg beg)
		   ((region-exists-p)
		    (region-beginning))
		   (t (point-min)))) 
	(end (cond (end (copy-marker end))
		   ((region-exists-p)
		    (copy-marker (region-end)))
		   (t (copy-marker (point-max)))))
        done)
    (save-excursion
      (goto-char beg)
      (if (fi-conflicting-chars-p beg end)
          (progn 
            (if fi-solve-by-augment
                (progn 
                  (setq footnote-start-tag (concat footnote-start-tag footnote-start-tag))
                  (setq footnote-end-tag (concat footnote-end-tag footnote-end-tag))
                  (setq done t)
                  (message "Footnote tags and regexp set to: %s \"%s\"  \"%s\""  footnote-style footnote-start-tag footnote-end-tag))
              (progn
                (if (string= footnote-start-tag footnote-start-tag-2)
                    (message "%s" "footnote-start-tag-2 already in use, please set footnote-tags manually, via `fi-Footnote-set-my-preferred-tags', or set `fi-solve-by-augment' to `t'")
                  (setq footnote-start-tag footnote-start-tag-2)
                  (setq footnote-end-tag footnote-end-tag-2)
                  (setq done t)
                  (message "Footnote tags and regexp set to: %s \"%s\"  \"%s\""  footnote-style footnote-start-tag footnote-end-tag))))
            (when done (fi-solve-conflicts beg end)))
        (message "No conflicts detected!: %s \"%s\"  \"%s\""  footnote-style footnote-start-tag footnote-end-tag)))))

;; helpers from misc-utils

(defcustom strip-chars-before  " \t\r\n\f*"
 "Regexp indicating which chars shall be stripped before STRING - which is defined by `string-chars-preserve'."

:type 'string
:group 'convenience)

;; (setq strip-chars-after  " \t\r\n\f*")
(defcustom strip-chars-after  " \t\r\n\f*\\'"
  "Regexp indicating which chars shall be stripped after STRING - which is defined by `string-chars-preserve'."

:type 'string
:group 'convenience)

(defcustom string-chars-preserve "\\(.*?\\)"
  "Chars preserved of STRING.
`strip-chars-after' and
`strip-chars-before' indicate what class of chars to strip."
  :type 'string
  :group 'convenience)

;; This string-strip results from a discussion at
;; http://lists.gnu.org/archive/html/emacs-devel/2006-06/msg00428.html
;; Thanks all contributors.
(defun string-strip (str &optional chars-before chars-after chars-preserve)
  "Return a copy of STR, CHARS removed, like
\(string-strip \" foo   \") -> \"foo\" 
Spezify chars to remove via
`CHARS-BEFORE' and `CHARS-AFTER' default is \"[ \\t\\r\\n\\f]*\",
i.e. spaces, tabs, carriage returns, newlines and newpages.
`CHARS-PRESERVE' must be a parentized expression,
it defaults to \"\\(.*?\\)\"
"
  (let ((s-c-b (or chars-before
		   strip-chars-before))
	(s-c-a (or chars-after
		   strip-chars-after))
	(s-c-p (or chars-preserve
		   string-chars-preserve)))
    (string-match
     (concat "\\`[" s-c-b"]*" s-c-p "[" s-c-a "]*\\'") str)
    (match-string 1 str)))

;;;###autoload
(unless (functionp 'in-string-p)
  (defun in-string-p (&optional pos)
    (let ((orig (or pos (point))))
      (save-excursion
        (save-restriction
          (widen)
          (beginning-of-defun)
          (numberp (nth 3 (parse-partial-sexp (point) orig))))))))

(unless (functionp 'empty-stringp)
  (defun empty-stringp (string)
    (string= "" string)))

(unless (boundp 'empty-line-p-chars)
  (defcustom empty-line-p-chars "^[ \t\f\r]*$"
    "Empty-line-p-chars."
    :type 'regexp
    :group 'convenience))

(unless (functionp 'empty-line-p) 
  (defun empty-line-p (&optional iact)
    "Returns t if cursor is at an empty line, nil otherwise."
    (interactive "p")
    (save-excursion
      (beginning-of-line)
      (when iact
        (message "%s" (looking-at empty-line-p-chars)))
      (looking-at empty-line-p-chars))))

(unless (featurep 'xemacs)
  (unless (functionp 'region-exists-p)
    (defun region-exists-p ()
      "and mark-active transient-mark-mode
 (not (eq (region-beginning) (region-end)"
      (and mark-active transient-mark-mode
           (not (eq (region-beginning) (region-end)))))))

(unless (functionp 'indent-with-spaces)
  (defun indent-with-spaces (column)
    "Indent with spaces to column "
    (interactive "*")
    (while (< (current-column) column)
      (insert " "))
    (forward-char 1)))

(when (featurep 'xemacs) 
  (unless (functionp 'looking-back)
    ;; from GNU Emacs subr.el
    (defun looking-back (regexp &optional limit greedy)
      "Return non-nil if text before point matches regular expression REGEXP.
    Like `looking-at' except matches before point, and is slower.
    LIMIT if non-nil speeds up the search by specifying a minimum
    starting position, to avoid checking matches that would start
    before LIMIT.
    If GREEDY is non-nil, extend the match backwards as far as possible,
    stopping when a single additional previous character cannot be part
    of a match for REGEXP."
      (let ((start (point))
            (pos
             (save-excursion
               (and (re-search-backward (concat "\\(?:" regexp "\\)\\=") limit t)
                    (point)))))
        (if (and greedy pos)
            (save-restriction
              (narrow-to-region (point-min) start)
              (while (and (> pos (point-min))
                          (save-excursion
                            (goto-char pos)
                            (backward-char 1)
                            (looking-at (concat "\\(?:"  regexp "\\)\\'"))))
                (setq pos (1- pos)))
              (save-excursion
                (goto-char pos)
                (looking-at (concat "\\(?:"  regexp "\\)\\'")))))
        (not (null pos))))))

(provide 'footnote-init)
;;; footnote-init.el ends here
