;;; thing-at-point-utils.el   --- more th-at-point edit functions

;; Version: 1.13

;; Copyright (C) 2006, 2007, 2008, 2009 Andreas Roehler, unless
;; indicated otherwise

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>, unless
;; indicated otherwise

;; Keywords: convenience

;; This file is free software; you can redistribute it

;; and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2, or (at your option) any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Commentary:

;; Information is given with thingatpt-utils-base.el

(require 'thingatpt-utils-base)



;;;###autoload
(defun alnum-atpt (&optional arg) 
  "Returns alnum at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'alnum arg))

(defalias 'bounds-of-alnum-atpt 'alnum-bounds-atpt)
;;;###autoload
(defun alnum-bounds-atpt (&optional arg) 
  "Returns borders of alnum at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'alnum arg))

;;;###autoload
(defun alnum-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'alnum arg))

;;;###autoload
(defun alnum-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-end 'alnum arg))

(defalias 'beginning-of-alnum-atpt 'alnum-beginning-atpt)
;;;###autoload
(defun alnum-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'alnum arg))

(defalias 'end-of-alnum-atpt 'alnum-end-atpt)
;;;###autoload
(defun alnum-end-atpt (&optional arg) 
  "Goto end of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'alnum arg))

(defalias 'length-of-alnum-atpt 'alnum-length-atpt)
;;;###autoload
(defun alnum-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-length 'alnum arg))

(defalias 'copy-alnum-atpt 'alnum-copy-atpt)
;;;###autoload
(defun alnum-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'alnum arg))

(defalias 'escape-alnum-atpt 'alnum-escape-atpt)
;;;###autoload
(defun alnum-escape-atpt () 
  "Makes a copy of symbol or char-class alnum at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'alnum))

(defalias 'doublequote-alnum-atpt 'alnum-doublequote-atpt)
;;;###autoload
(defun alnum-doublequote-atpt (&optional arg) 
  "Doublequotes 'alnum at point if any. "
  (interactive "*p")
    (th-doublequote 'alnum arg))

(defalias 'dollar-alnum-atpt 'alnum-dollar-atpt)
;;;###autoload
(defun alnum-dollar-atpt (&optional arg) 
  "Doublequotes 'alnum at point if any. "
  (interactive "*p")
    (th-dollar 'alnum arg))

(defalias 'angle-greater-alnum-atpt 'alnum-greater-angle-atpt)
;;;###autoload
(defun alnum-greater-angle-atpt (&optional arg) 
  "Sets angles for 'alnum after point if any. "
  (interactive "*p")
    (th-greater-angle 'alnum arg))

(defalias 'angle-lower-alnum-atpt 'alnum-lower-angle-atpt)
;;;###autoload
(defun alnum-lower-angle-atpt (&optional arg) 
  "Sets angles for 'alnum after point if any. "
  (interactive "*p")
    (th-lower-angle 'alnum arg))

(defalias 'backslash-alnum-atpt 'alnum-backslash-atpt)
;;;###autoload
(defun alnum-backslash-atpt (&optional arg) 
  "Backslash 'alnum at point if any. "
  (interactive "*p")
    (th-backslash 'alnum arg))   

(defalias 'brace-alnum-atpt 'alnum-brace-atpt)
;;;###autoload
(defun alnum-brace-atpt (&optional arg) 
  "Braces 'alnum at point if any. "
  (interactive "*p")
    (th-brace 'alnum arg))

(defalias 'bracket-alnum-atpt 'alnum-bracket-atpt)
;;;###autoload
(defun alnum-bracket-atpt (&optional arg) 
  "Brackets 'alnum after point if any. "
  (interactive "*p")
    (th-bracket 'alnum arg))

;; (defalias 'comment-alnum-atpt 'alnum-comment-atpt)
;;;###autoload
(defun alnum-comment-atpt (&optional arg) 
  "Comments 'alnum at point if any. "
  (interactive "*p")
  (th-comment 'alnum arg))

(defalias 'mark-alnum-atpt 'alnum-mark-atpt)
;;;###autoload
(defun alnum-mark-atpt () 
  "Marks 'alnum at point if any. "
  (interactive)
  (th-mark 'alnum))

(defalias 'hide-alnum-atpt 'alnum-hide-atpt)
;;;###autoload
(defun alnum-hide-atpt () 
  "Toggles hide 'alnum in region or buffer if any. "
  (interactive)
  (th-hide 'alnum))

(defalias 'highlight-alnum-atpt-mode 'alnum-highlight-atpt-mode)
(defun alnum-highlight-atpt-mode (&optional arg) 
  "Toggles alnum-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'alnum arg))

(defalias 'kill-alnum-atpt 'alnum-kill-atpt)
;;;###autoload
(defun alnum-kill-atpt (&optional arg) 
  "Kills alnum at point if any. "
  (interactive "*P")
  (th-kill 'alnum arg))

(defalias 'parentize-alnum-atpt 'alnum-parentize-atpt)
;;;###autoload
(defun alnum-parentize-atpt (&optional arg) 
  "Parentizes 'alnum at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'alnum arg))

(defalias 'separate-alnum-atpt 'alnum-separate-atpt)
;;;###autoload
(defun alnum-separate-atpt (&optional arg) 
  "Separates 'alnum at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'alnum arg))

(defalias 'singlequote-alnum-atpt 'alnum-singlequote-atpt)
;;;###autoload
(defun alnum-singlequote-atpt (&optional arg) 
  "Singlequotes 'alnum at point if any. "
  (interactive "*p")
    (th-singlequote 'alnum arg))

(defalias 'slash-alnum-atpt 'alnum-slash-atpt)
;;;###autoload
(defun alnum-slash-atpt (&optional arg) 
  "Slash 'alnum at point if any. "
  (interactive "*p")
    (th-slash 'alnum arg))

(defalias 'trim-alnum-atpt 'alnum-trim-atpt)
;;;###autoload
(defun alnum-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alnum t t))

(defalias 'trim-alnum-left-atpt 'alnum-left-trim-atpt)
;;;###autoload
(defun alnum-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alnum t nil))

(defalias 'trim-alnum-right-atpt 'alnum-right-trim-atpt)
;;;###autoload
(defun alnum-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alnum nil t))

(defalias 'forward-alnum-atpt 'alnum-forward-atpt)
;;;###autoload
(defun alnum-forward-atpt (&optional arg) 
  "Moves forward over alnum at point if any, does nothing otherwise.
Returns bounds of alnum as a list. "
  (interactive "p")
  (th-forward 'alnum arg))

(defalias 'backward-alnum-atpt 'alnum-backward-atpt)
;;;###autoload
(defun alnum-backward-atpt (&optional arg) 
  "Moves backward over alnum before point if any, does nothing otherwise.
Returns bounds of alnum as a list. "
  (interactive "p")
  (th-backward 'alnum arg))

(defalias 'transpose-alnum-atpt 'alnum-transpose-atpt)
;;;###autoload
(defun alnum-transpose-atpt (&optional arg) 
  "Transposes alnum with alnum before point if any. "
  (interactive "*P")
  (th-transpose 'alnum arg))

(defalias 'sort-alnum-atpt 'alnum-sort-atpt)
;;;###autoload
(defun alnum-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts alnum's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'alnum reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-alnum-atpt 'alnum-check-atpt)
;;;###autoload
(defun alnum-check-atpt (&optional arg) 
  " alnum at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "alnum" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "alnum" "-end-position-atpt")))))
    (th-check 'alnum arg)))
   

;;;###autoload
(defun count-alnum-atpt (&optional beg end) 
    "Counts matches of alnum at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'alnum beg end (interactive-p)))


;;;###autoload
(defun count-alpha-in-alnum-atpt (&optional beg end) 
    "Counts matches of alpha in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'alnum beg end (interactive-p)))

(defun ratio-alpha-in-alnum-atpt (&optional beg end) 
    "Relation of alpha in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'alnum beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-alnum-atpt (&optional beg end) 
    "Counts matches of ascii in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'alnum beg end (interactive-p)))

(defun ratio-ascii-in-alnum-atpt (&optional beg end) 
    "Relation of ascii in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'alnum beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-alnum-atpt (&optional beg end) 
    "Counts matches of digit in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'alnum beg end (interactive-p)))

(defun ratio-digit-in-alnum-atpt (&optional beg end) 
    "Relation of digit in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'alnum beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-alnum-atpt (&optional beg end) 
    "Counts matches of graph in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'alnum beg end (interactive-p)))

(defun ratio-graph-in-alnum-atpt (&optional beg end) 
    "Relation of graph in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'alnum beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-alnum-atpt (&optional beg end) 
    "Counts matches of lower in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'alnum beg end (interactive-p)))

(defun ratio-lower-in-alnum-atpt (&optional beg end) 
    "Relation of lower in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'alnum beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-alnum-atpt (&optional beg end) 
    "Counts matches of upper in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'alnum beg end (interactive-p)))

(defun ratio-upper-in-alnum-atpt (&optional beg end) 
    "Relation of upper in alnum if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'alnum beg end t (interactive-p)))



;;;###autoload
(defun alpha-atpt (&optional arg) 
  "Returns alpha at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'alpha arg))

(defalias 'bounds-of-alpha-atpt 'alpha-bounds-atpt)
;;;###autoload
(defun alpha-bounds-atpt (&optional arg) 
  "Returns borders of alpha at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'alpha arg))

;;;###autoload
(defun alpha-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'alpha arg))

;;;###autoload
(defun alpha-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-end 'alpha arg))

(defalias 'beginning-of-alpha-atpt 'alpha-beginning-atpt)
;;;###autoload
(defun alpha-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'alpha arg))

(defalias 'end-of-alpha-atpt 'alpha-end-atpt)
;;;###autoload
(defun alpha-end-atpt (&optional arg) 
  "Goto end of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'alpha arg))

(defalias 'length-of-alpha-atpt 'alpha-length-atpt)
;;;###autoload
(defun alpha-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-length 'alpha arg))

(defalias 'copy-alpha-atpt 'alpha-copy-atpt)
;;;###autoload
(defun alpha-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'alpha arg))

(defalias 'escape-alpha-atpt 'alpha-escape-atpt)
;;;###autoload
(defun alpha-escape-atpt () 
  "Makes a copy of symbol or char-class alpha at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'alpha))

(defalias 'doublequote-alpha-atpt 'alpha-doublequote-atpt)
;;;###autoload
(defun alpha-doublequote-atpt (&optional arg) 
  "Doublequotes 'alpha at point if any. "
  (interactive "*p")
    (th-doublequote 'alpha arg))

(defalias 'dollar-alpha-atpt 'alpha-dollar-atpt)
;;;###autoload
(defun alpha-dollar-atpt (&optional arg) 
  "Doublequotes 'alpha at point if any. "
  (interactive "*p")
    (th-dollar 'alpha arg))

(defalias 'angle-greater-alpha-atpt 'alpha-greater-angle-atpt)
;;;###autoload
(defun alpha-greater-angle-atpt (&optional arg) 
  "Sets angles for 'alpha after point if any. "
  (interactive "*p")
    (th-greater-angle 'alpha arg))

(defalias 'angle-lower-alpha-atpt 'alpha-lower-angle-atpt)
;;;###autoload
(defun alpha-lower-angle-atpt (&optional arg) 
  "Sets angles for 'alpha after point if any. "
  (interactive "*p")
    (th-lower-angle 'alpha arg))

(defalias 'backslash-alpha-atpt 'alpha-backslash-atpt)
;;;###autoload
(defun alpha-backslash-atpt (&optional arg) 
  "Backslash 'alpha at point if any. "
  (interactive "*p")
    (th-backslash 'alpha arg))   

(defalias 'brace-alpha-atpt 'alpha-brace-atpt)
;;;###autoload
(defun alpha-brace-atpt (&optional arg) 
  "Braces 'alpha at point if any. "
  (interactive "*p")
    (th-brace 'alpha arg))

(defalias 'bracket-alpha-atpt 'alpha-bracket-atpt)
;;;###autoload
(defun alpha-bracket-atpt (&optional arg) 
  "Brackets 'alpha after point if any. "
  (interactive "*p")
    (th-bracket 'alpha arg))

;; (defalias 'comment-alpha-atpt 'alpha-comment-atpt)
;;;###autoload
(defun alpha-comment-atpt (&optional arg) 
  "Comments 'alpha at point if any. "
  (interactive "*p")
  (th-comment 'alpha arg))

(defalias 'mark-alpha-atpt 'alpha-mark-atpt)
;;;###autoload
(defun alpha-mark-atpt () 
  "Marks 'alpha at point if any. "
  (interactive)
  (th-mark 'alpha))

(defalias 'hide-alpha-atpt 'alpha-hide-atpt)
;;;###autoload
(defun alpha-hide-atpt () 
  "Toggles hide 'alpha in region or buffer if any. "
  (interactive)
  (th-hide 'alpha))

(defalias 'highlight-alpha-atpt-mode 'alpha-highlight-atpt-mode)
(defun alpha-highlight-atpt-mode (&optional arg) 
  "Toggles alpha-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'alpha arg))

(defalias 'kill-alpha-atpt 'alpha-kill-atpt)
;;;###autoload
(defun alpha-kill-atpt (&optional arg) 
  "Kills alpha at point if any. "
  (interactive "*P")
  (th-kill 'alpha arg))

(defalias 'parentize-alpha-atpt 'alpha-parentize-atpt)
;;;###autoload
(defun alpha-parentize-atpt (&optional arg) 
  "Parentizes 'alpha at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'alpha arg))

(defalias 'separate-alpha-atpt 'alpha-separate-atpt)
;;;###autoload
(defun alpha-separate-atpt (&optional arg) 
  "Separates 'alpha at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'alpha arg))

(defalias 'singlequote-alpha-atpt 'alpha-singlequote-atpt)
;;;###autoload
(defun alpha-singlequote-atpt (&optional arg) 
  "Singlequotes 'alpha at point if any. "
  (interactive "*p")
    (th-singlequote 'alpha arg))

(defalias 'slash-alpha-atpt 'alpha-slash-atpt)
;;;###autoload
(defun alpha-slash-atpt (&optional arg) 
  "Slash 'alpha at point if any. "
  (interactive "*p")
    (th-slash 'alpha arg))

(defalias 'trim-alpha-atpt 'alpha-trim-atpt)
;;;###autoload
(defun alpha-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alpha t t))

(defalias 'trim-alpha-left-atpt 'alpha-left-trim-atpt)
;;;###autoload
(defun alpha-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alpha t nil))

(defalias 'trim-alpha-right-atpt 'alpha-right-trim-atpt)
;;;###autoload
(defun alpha-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'alpha nil t))

(defalias 'forward-alpha-atpt 'alpha-forward-atpt)
;;;###autoload
(defun alpha-forward-atpt (&optional arg) 
  "Moves forward over alpha at point if any, does nothing otherwise.
Returns bounds of alpha as a list. "
  (interactive "p")
  (th-forward 'alpha arg))

(defalias 'backward-alpha-atpt 'alpha-backward-atpt)
;;;###autoload
(defun alpha-backward-atpt (&optional arg) 
  "Moves backward over alpha before point if any, does nothing otherwise.
Returns bounds of alpha as a list. "
  (interactive "p")
  (th-backward 'alpha arg))

(defalias 'transpose-alpha-atpt 'alpha-transpose-atpt)
;;;###autoload
(defun alpha-transpose-atpt (&optional arg) 
  "Transposes alpha with alpha before point if any. "
  (interactive "*P")
  (th-transpose 'alpha arg))

(defalias 'sort-alpha-atpt 'alpha-sort-atpt)
;;;###autoload
(defun alpha-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts alpha's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'alpha reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-alpha-atpt 'alpha-check-atpt)
;;;###autoload
(defun alpha-check-atpt (&optional arg) 
  " alpha at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "alpha" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "alpha" "-end-position-atpt")))))
    (th-check 'alpha arg)))
   

;;;###autoload
(defun count-alpha-atpt (&optional beg end) 
    "Counts matches of alpha at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'alpha beg end (interactive-p)))


;;;###autoload
(defun count-ascii-in-alpha-atpt (&optional beg end) 
    "Counts matches of ascii in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'alpha beg end (interactive-p)))

(defun ratio-ascii-in-alpha-atpt (&optional beg end) 
    "Relation of ascii in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'alpha beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-alpha-atpt (&optional beg end) 
    "Counts matches of lower in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'alpha beg end (interactive-p)))

(defun ratio-lower-in-alpha-atpt (&optional beg end) 
    "Relation of lower in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'alpha beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-alpha-atpt (&optional beg end) 
    "Counts matches of upper in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'alpha beg end (interactive-p)))

(defun ratio-upper-in-alpha-atpt (&optional beg end) 
    "Relation of upper in alpha if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'alpha beg end t (interactive-p)))



;;;###autoload
(defun angled-no-nest-atpt (&optional arg) 
  "Returns angled-no-nest at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'angled-no-nest arg))

(defalias 'bounds-of-angled-no-nest-atpt 'angled-no-nest-bounds-atpt)
;;;###autoload
(defun angled-no-nest-bounds-atpt (&optional arg) 
  "Returns borders of angled-no-nest at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'angled-no-nest arg))

;;;###autoload
(defun angled-no-nest-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'angled-no-nest arg))

;;;###autoload
(defun angled-no-nest-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-end 'angled-no-nest arg))

(defalias 'beginning-of-angled-no-nest-atpt 'angled-no-nest-beginning-atpt)
;;;###autoload
(defun angled-no-nest-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'angled-no-nest arg))

(defalias 'end-of-angled-no-nest-atpt 'angled-no-nest-end-atpt)
;;;###autoload
(defun angled-no-nest-end-atpt (&optional arg) 
  "Goto end of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'angled-no-nest arg))

(defalias 'length-of-angled-no-nest-atpt 'angled-no-nest-length-atpt)
;;;###autoload
(defun angled-no-nest-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-length 'angled-no-nest arg))

(defalias 'copy-angled-no-nest-atpt 'angled-no-nest-copy-atpt)
;;;###autoload
(defun angled-no-nest-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'angled-no-nest arg))

(defalias 'escape-angled-no-nest-atpt 'angled-no-nest-escape-atpt)
;;;###autoload
(defun angled-no-nest-escape-atpt () 
  "Makes a copy of symbol or char-class angled-no-nest at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'angled-no-nest))

(defalias 'doublequote-angled-no-nest-atpt 'angled-no-nest-doublequote-atpt)
;;;###autoload
(defun angled-no-nest-doublequote-atpt (&optional arg) 
  "Doublequotes 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-doublequote 'angled-no-nest arg))

(defalias 'dollar-angled-no-nest-atpt 'angled-no-nest-dollar-atpt)
;;;###autoload
(defun angled-no-nest-dollar-atpt (&optional arg) 
  "Doublequotes 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-dollar 'angled-no-nest arg))

(defalias 'angle-greater-angled-no-nest-atpt 'angled-no-nest-greater-angle-atpt)
;;;###autoload
(defun angled-no-nest-greater-angle-atpt (&optional arg) 
  "Sets angles for 'angled-no-nest after point if any. "
  (interactive "*p")
    (th-greater-angle 'angled-no-nest arg))

(defalias 'angle-lower-angled-no-nest-atpt 'angled-no-nest-lower-angle-atpt)
;;;###autoload
(defun angled-no-nest-lower-angle-atpt (&optional arg) 
  "Sets angles for 'angled-no-nest after point if any. "
  (interactive "*p")
    (th-lower-angle 'angled-no-nest arg))

(defalias 'backslash-angled-no-nest-atpt 'angled-no-nest-backslash-atpt)
;;;###autoload
(defun angled-no-nest-backslash-atpt (&optional arg) 
  "Backslash 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-backslash 'angled-no-nest arg))   

(defalias 'brace-angled-no-nest-atpt 'angled-no-nest-brace-atpt)
;;;###autoload
(defun angled-no-nest-brace-atpt (&optional arg) 
  "Braces 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-brace 'angled-no-nest arg))

(defalias 'bracket-angled-no-nest-atpt 'angled-no-nest-bracket-atpt)
;;;###autoload
(defun angled-no-nest-bracket-atpt (&optional arg) 
  "Brackets 'angled-no-nest after point if any. "
  (interactive "*p")
    (th-bracket 'angled-no-nest arg))

;; (defalias 'comment-angled-no-nest-atpt 'angled-no-nest-comment-atpt)
;;;###autoload
(defun angled-no-nest-comment-atpt (&optional arg) 
  "Comments 'angled-no-nest at point if any. "
  (interactive "*p")
  (th-comment 'angled-no-nest arg))

(defalias 'mark-angled-no-nest-atpt 'angled-no-nest-mark-atpt)
;;;###autoload
(defun angled-no-nest-mark-atpt () 
  "Marks 'angled-no-nest at point if any. "
  (interactive)
  (th-mark 'angled-no-nest))

(defalias 'hide-angled-no-nest-atpt 'angled-no-nest-hide-atpt)
;;;###autoload
(defun angled-no-nest-hide-atpt () 
  "Toggles hide 'angled-no-nest in region or buffer if any. "
  (interactive)
  (th-hide 'angled-no-nest))

(defalias 'highlight-angled-no-nest-atpt-mode 'angled-no-nest-highlight-atpt-mode)
(defun angled-no-nest-highlight-atpt-mode (&optional arg) 
  "Toggles angled-no-nest-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'angled-no-nest arg))

(defalias 'kill-angled-no-nest-atpt 'angled-no-nest-kill-atpt)
;;;###autoload
(defun angled-no-nest-kill-atpt (&optional arg) 
  "Kills angled-no-nest at point if any. "
  (interactive "*P")
  (th-kill 'angled-no-nest arg))

(defalias 'parentize-angled-no-nest-atpt 'angled-no-nest-parentize-atpt)
;;;###autoload
(defun angled-no-nest-parentize-atpt (&optional arg) 
  "Parentizes 'angled-no-nest at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'angled-no-nest arg))

(defalias 'separate-angled-no-nest-atpt 'angled-no-nest-separate-atpt)
;;;###autoload
(defun angled-no-nest-separate-atpt (&optional arg) 
  "Separates 'angled-no-nest at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'angled-no-nest arg))

(defalias 'singlequote-angled-no-nest-atpt 'angled-no-nest-singlequote-atpt)
;;;###autoload
(defun angled-no-nest-singlequote-atpt (&optional arg) 
  "Singlequotes 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-singlequote 'angled-no-nest arg))

(defalias 'slash-angled-no-nest-atpt 'angled-no-nest-slash-atpt)
;;;###autoload
(defun angled-no-nest-slash-atpt (&optional arg) 
  "Slash 'angled-no-nest at point if any. "
  (interactive "*p")
    (th-slash 'angled-no-nest arg))

(defalias 'trim-angled-no-nest-atpt 'angled-no-nest-trim-atpt)
;;;###autoload
(defun angled-no-nest-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-no-nest t t))

(defalias 'trim-angled-no-nest-left-atpt 'angled-no-nest-left-trim-atpt)
;;;###autoload
(defun angled-no-nest-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-no-nest t nil))

(defalias 'trim-angled-no-nest-right-atpt 'angled-no-nest-right-trim-atpt)
;;;###autoload
(defun angled-no-nest-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-no-nest nil t))

(defalias 'forward-angled-no-nest-atpt 'angled-no-nest-forward-atpt)
;;;###autoload
(defun angled-no-nest-forward-atpt (&optional arg) 
  "Moves forward over angled-no-nest at point if any, does nothing otherwise.
Returns bounds of angled-no-nest as a list. "
  (interactive "p")
  (th-forward 'angled-no-nest arg))

(defalias 'backward-angled-no-nest-atpt 'angled-no-nest-backward-atpt)
;;;###autoload
(defun angled-no-nest-backward-atpt (&optional arg) 
  "Moves backward over angled-no-nest before point if any, does nothing otherwise.
Returns bounds of angled-no-nest as a list. "
  (interactive "p")
  (th-backward 'angled-no-nest arg))

(defalias 'transpose-angled-no-nest-atpt 'angled-no-nest-transpose-atpt)
;;;###autoload
(defun angled-no-nest-transpose-atpt (&optional arg) 
  "Transposes angled-no-nest with angled-no-nest before point if any. "
  (interactive "*P")
  (th-transpose 'angled-no-nest arg))

(defalias 'sort-angled-no-nest-atpt 'angled-no-nest-sort-atpt)
;;;###autoload
(defun angled-no-nest-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts angled-no-nest's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'angled-no-nest reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-angled-no-nest-atpt 'angled-no-nest-check-atpt)
;;;###autoload
(defun angled-no-nest-check-atpt (&optional arg) 
  " angled-no-nest at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "angled-no-nest" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "angled-no-nest" "-end-position-atpt")))))
    (th-check 'angled-no-nest arg)))
   

;;;###autoload
(defun count-alnum-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of alnum in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-no-nest beg end (interactive-p)))

(defun ratio-alnum-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of alnum in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of alpha in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-no-nest beg end (interactive-p)))

(defun ratio-alpha-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of alpha in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of ascii in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-no-nest beg end (interactive-p)))

(defun ratio-ascii-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of ascii in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of digit in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-no-nest beg end (interactive-p)))

(defun ratio-digit-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of digit in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of graph in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-no-nest beg end (interactive-p)))

(defun ratio-graph-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of graph in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of lower in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-no-nest beg end (interactive-p)))

(defun ratio-lower-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of lower in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of punct in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-no-nest beg end (interactive-p)))

(defun ratio-punct-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of punct in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-angled-no-nest-atpt (&optional beg end) 
    "Counts matches of upper in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-no-nest beg end (interactive-p)))

(defun ratio-upper-in-angled-no-nest-atpt (&optional beg end) 
    "Relation of upper in angled-no-nest if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-no-nest beg end t (interactive-p)))



;;;###autoload
(defun angled-greater-atpt (&optional arg) 
  "Returns angled-greater at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'angled-greater arg))

(defalias 'bounds-of-angled-greater-atpt 'angled-greater-bounds-atpt)
;;;###autoload
(defun angled-greater-bounds-atpt (&optional arg) 
  "Returns borders of angled-greater at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'angled-greater arg))

;;;###autoload
(defun angled-greater-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'angled-greater arg))

;;;###autoload
(defun angled-greater-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-end 'angled-greater arg))

(defalias 'beginning-of-angled-greater-atpt 'angled-greater-beginning-atpt)
;;;###autoload
(defun angled-greater-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'angled-greater arg))

(defalias 'end-of-angled-greater-atpt 'angled-greater-end-atpt)
;;;###autoload
(defun angled-greater-end-atpt (&optional arg) 
  "Goto end of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'angled-greater arg))

(defalias 'length-of-angled-greater-atpt 'angled-greater-length-atpt)
;;;###autoload
(defun angled-greater-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-length 'angled-greater arg))

(defalias 'copy-angled-greater-atpt 'angled-greater-copy-atpt)
;;;###autoload
(defun angled-greater-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'angled-greater arg))

(defalias 'escape-angled-greater-atpt 'angled-greater-escape-atpt)
;;;###autoload
(defun angled-greater-escape-atpt () 
  "Makes a copy of symbol or char-class angled-greater at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'angled-greater))

(defalias 'doublequote-angled-greater-atpt 'angled-greater-doublequote-atpt)
;;;###autoload
(defun angled-greater-doublequote-atpt (&optional arg) 
  "Doublequotes 'angled-greater at point if any. "
  (interactive "*p")
    (th-doublequote 'angled-greater arg))

(defalias 'dollar-angled-greater-atpt 'angled-greater-dollar-atpt)
;;;###autoload
(defun angled-greater-dollar-atpt (&optional arg) 
  "Doublequotes 'angled-greater at point if any. "
  (interactive "*p")
    (th-dollar 'angled-greater arg))

(defalias 'angle-greater-angled-greater-atpt 'angled-greater-greater-angle-atpt)
;;;###autoload
(defun angled-greater-greater-angle-atpt (&optional arg) 
  "Sets angles for 'angled-greater after point if any. "
  (interactive "*p")
    (th-greater-angle 'angled-greater arg))

(defalias 'angle-lower-angled-greater-atpt 'angled-greater-lower-angle-atpt)
;;;###autoload
(defun angled-greater-lower-angle-atpt (&optional arg) 
  "Sets angles for 'angled-greater after point if any. "
  (interactive "*p")
    (th-lower-angle 'angled-greater arg))

(defalias 'backslash-angled-greater-atpt 'angled-greater-backslash-atpt)
;;;###autoload
(defun angled-greater-backslash-atpt (&optional arg) 
  "Backslash 'angled-greater at point if any. "
  (interactive "*p")
    (th-backslash 'angled-greater arg))   

(defalias 'brace-angled-greater-atpt 'angled-greater-brace-atpt)
;;;###autoload
(defun angled-greater-brace-atpt (&optional arg) 
  "Braces 'angled-greater at point if any. "
  (interactive "*p")
    (th-brace 'angled-greater arg))

(defalias 'bracket-angled-greater-atpt 'angled-greater-bracket-atpt)
;;;###autoload
(defun angled-greater-bracket-atpt (&optional arg) 
  "Brackets 'angled-greater after point if any. "
  (interactive "*p")
    (th-bracket 'angled-greater arg))

;; (defalias 'comment-angled-greater-atpt 'angled-greater-comment-atpt)
;;;###autoload
(defun angled-greater-comment-atpt (&optional arg) 
  "Comments 'angled-greater at point if any. "
  (interactive "*p")
  (th-comment 'angled-greater arg))

(defalias 'mark-angled-greater-atpt 'angled-greater-mark-atpt)
;;;###autoload
(defun angled-greater-mark-atpt () 
  "Marks 'angled-greater at point if any. "
  (interactive)
  (th-mark 'angled-greater))

(defalias 'hide-angled-greater-atpt 'angled-greater-hide-atpt)
;;;###autoload
(defun angled-greater-hide-atpt () 
  "Toggles hide 'angled-greater in region or buffer if any. "
  (interactive)
  (th-hide 'angled-greater))

(defalias 'highlight-angled-greater-atpt-mode 'angled-greater-highlight-atpt-mode)
(defun angled-greater-highlight-atpt-mode (&optional arg) 
  "Toggles angled-greater-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'angled-greater arg))

(defalias 'kill-angled-greater-atpt 'angled-greater-kill-atpt)
;;;###autoload
(defun angled-greater-kill-atpt (&optional arg) 
  "Kills angled-greater at point if any. "
  (interactive "*P")
  (th-kill 'angled-greater arg))

(defalias 'parentize-angled-greater-atpt 'angled-greater-parentize-atpt)
;;;###autoload
(defun angled-greater-parentize-atpt (&optional arg) 
  "Parentizes 'angled-greater at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'angled-greater arg))

(defalias 'separate-angled-greater-atpt 'angled-greater-separate-atpt)
;;;###autoload
(defun angled-greater-separate-atpt (&optional arg) 
  "Separates 'angled-greater at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'angled-greater arg))

(defalias 'singlequote-angled-greater-atpt 'angled-greater-singlequote-atpt)
;;;###autoload
(defun angled-greater-singlequote-atpt (&optional arg) 
  "Singlequotes 'angled-greater at point if any. "
  (interactive "*p")
    (th-singlequote 'angled-greater arg))

(defalias 'slash-angled-greater-atpt 'angled-greater-slash-atpt)
;;;###autoload
(defun angled-greater-slash-atpt (&optional arg) 
  "Slash 'angled-greater at point if any. "
  (interactive "*p")
    (th-slash 'angled-greater arg))

(defalias 'trim-angled-greater-atpt 'angled-greater-trim-atpt)
;;;###autoload
(defun angled-greater-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-greater t t))

(defalias 'trim-angled-greater-left-atpt 'angled-greater-left-trim-atpt)
;;;###autoload
(defun angled-greater-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-greater t nil))

(defalias 'trim-angled-greater-right-atpt 'angled-greater-right-trim-atpt)
;;;###autoload
(defun angled-greater-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-greater nil t))

(defalias 'forward-angled-greater-atpt 'angled-greater-forward-atpt)
;;;###autoload
(defun angled-greater-forward-atpt (&optional arg) 
  "Moves forward over angled-greater at point if any, does nothing otherwise.
Returns bounds of angled-greater as a list. "
  (interactive "p")
  (th-forward 'angled-greater arg))

(defalias 'backward-angled-greater-atpt 'angled-greater-backward-atpt)
;;;###autoload
(defun angled-greater-backward-atpt (&optional arg) 
  "Moves backward over angled-greater before point if any, does nothing otherwise.
Returns bounds of angled-greater as a list. "
  (interactive "p")
  (th-backward 'angled-greater arg))

(defalias 'transpose-angled-greater-atpt 'angled-greater-transpose-atpt)
;;;###autoload
(defun angled-greater-transpose-atpt (&optional arg) 
  "Transposes angled-greater with angled-greater before point if any. "
  (interactive "*P")
  (th-transpose 'angled-greater arg))

(defalias 'sort-angled-greater-atpt 'angled-greater-sort-atpt)
;;;###autoload
(defun angled-greater-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts angled-greater's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'angled-greater reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-angled-greater-atpt 'angled-greater-check-atpt)
;;;###autoload
(defun angled-greater-check-atpt (&optional arg) 
  " angled-greater at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "angled-greater" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "angled-greater" "-end-position-atpt")))))
    (th-check 'angled-greater arg)))
   

;;;###autoload
(defun count-alnum-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of alnum in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-greater beg end (interactive-p)))

(defun ratio-alnum-in-angled-greater-atpt (&optional beg end) 
    "Relation of alnum in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of alpha in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-greater beg end (interactive-p)))

(defun ratio-alpha-in-angled-greater-atpt (&optional beg end) 
    "Relation of alpha in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of ascii in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-greater beg end (interactive-p)))

(defun ratio-ascii-in-angled-greater-atpt (&optional beg end) 
    "Relation of ascii in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of digit in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-greater beg end (interactive-p)))

(defun ratio-digit-in-angled-greater-atpt (&optional beg end) 
    "Relation of digit in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of graph in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-greater beg end (interactive-p)))

(defun ratio-graph-in-angled-greater-atpt (&optional beg end) 
    "Relation of graph in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of lower in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-greater beg end (interactive-p)))

(defun ratio-lower-in-angled-greater-atpt (&optional beg end) 
    "Relation of lower in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of punct in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-greater beg end (interactive-p)))

(defun ratio-punct-in-angled-greater-atpt (&optional beg end) 
    "Relation of punct in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-angled-greater-atpt (&optional beg end) 
    "Counts matches of upper in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-greater beg end (interactive-p)))

(defun ratio-upper-in-angled-greater-atpt (&optional beg end) 
    "Relation of upper in angled-greater if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-greater beg end t (interactive-p)))



;;;###autoload
(defun angled-lower-atpt (&optional arg) 
  "Returns angled-lower at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'angled-lower arg))

(defalias 'bounds-of-angled-lower-atpt 'angled-lower-bounds-atpt)
;;;###autoload
(defun angled-lower-bounds-atpt (&optional arg) 
  "Returns borders of angled-lower at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'angled-lower arg))

;;;###autoload
(defun angled-lower-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'angled-lower arg))

;;;###autoload
(defun angled-lower-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-end 'angled-lower arg))

(defalias 'beginning-of-angled-lower-atpt 'angled-lower-beginning-atpt)
;;;###autoload
(defun angled-lower-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'angled-lower arg))

(defalias 'end-of-angled-lower-atpt 'angled-lower-end-atpt)
;;;###autoload
(defun angled-lower-end-atpt (&optional arg) 
  "Goto end of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'angled-lower arg))

(defalias 'length-of-angled-lower-atpt 'angled-lower-length-atpt)
;;;###autoload
(defun angled-lower-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-length 'angled-lower arg))

(defalias 'copy-angled-lower-atpt 'angled-lower-copy-atpt)
;;;###autoload
(defun angled-lower-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'angled-lower arg))

(defalias 'escape-angled-lower-atpt 'angled-lower-escape-atpt)
;;;###autoload
(defun angled-lower-escape-atpt () 
  "Makes a copy of symbol or char-class angled-lower at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'angled-lower))

(defalias 'doublequote-angled-lower-atpt 'angled-lower-doublequote-atpt)
;;;###autoload
(defun angled-lower-doublequote-atpt (&optional arg) 
  "Doublequotes 'angled-lower at point if any. "
  (interactive "*p")
    (th-doublequote 'angled-lower arg))

(defalias 'dollar-angled-lower-atpt 'angled-lower-dollar-atpt)
;;;###autoload
(defun angled-lower-dollar-atpt (&optional arg) 
  "Doublequotes 'angled-lower at point if any. "
  (interactive "*p")
    (th-dollar 'angled-lower arg))

(defalias 'angle-greater-angled-lower-atpt 'angled-lower-greater-angle-atpt)
;;;###autoload
(defun angled-lower-greater-angle-atpt (&optional arg) 
  "Sets angles for 'angled-lower after point if any. "
  (interactive "*p")
    (th-greater-angle 'angled-lower arg))

(defalias 'angle-lower-angled-lower-atpt 'angled-lower-lower-angle-atpt)
;;;###autoload
(defun angled-lower-lower-angle-atpt (&optional arg) 
  "Sets angles for 'angled-lower after point if any. "
  (interactive "*p")
    (th-lower-angle 'angled-lower arg))

(defalias 'backslash-angled-lower-atpt 'angled-lower-backslash-atpt)
;;;###autoload
(defun angled-lower-backslash-atpt (&optional arg) 
  "Backslash 'angled-lower at point if any. "
  (interactive "*p")
    (th-backslash 'angled-lower arg))   

(defalias 'brace-angled-lower-atpt 'angled-lower-brace-atpt)
;;;###autoload
(defun angled-lower-brace-atpt (&optional arg) 
  "Braces 'angled-lower at point if any. "
  (interactive "*p")
    (th-brace 'angled-lower arg))

(defalias 'bracket-angled-lower-atpt 'angled-lower-bracket-atpt)
;;;###autoload
(defun angled-lower-bracket-atpt (&optional arg) 
  "Brackets 'angled-lower after point if any. "
  (interactive "*p")
    (th-bracket 'angled-lower arg))

;; (defalias 'comment-angled-lower-atpt 'angled-lower-comment-atpt)
;;;###autoload
(defun angled-lower-comment-atpt (&optional arg) 
  "Comments 'angled-lower at point if any. "
  (interactive "*p")
  (th-comment 'angled-lower arg))

(defalias 'mark-angled-lower-atpt 'angled-lower-mark-atpt)
;;;###autoload
(defun angled-lower-mark-atpt () 
  "Marks 'angled-lower at point if any. "
  (interactive)
  (th-mark 'angled-lower))

(defalias 'hide-angled-lower-atpt 'angled-lower-hide-atpt)
;;;###autoload
(defun angled-lower-hide-atpt () 
  "Toggles hide 'angled-lower in region or buffer if any. "
  (interactive)
  (th-hide 'angled-lower))

(defalias 'highlight-angled-lower-atpt-mode 'angled-lower-highlight-atpt-mode)
(defun angled-lower-highlight-atpt-mode (&optional arg) 
  "Toggles angled-lower-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'angled-lower arg))

(defalias 'kill-angled-lower-atpt 'angled-lower-kill-atpt)
;;;###autoload
(defun angled-lower-kill-atpt (&optional arg) 
  "Kills angled-lower at point if any. "
  (interactive "*P")
  (th-kill 'angled-lower arg))

(defalias 'parentize-angled-lower-atpt 'angled-lower-parentize-atpt)
;;;###autoload
(defun angled-lower-parentize-atpt (&optional arg) 
  "Parentizes 'angled-lower at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'angled-lower arg))

(defalias 'separate-angled-lower-atpt 'angled-lower-separate-atpt)
;;;###autoload
(defun angled-lower-separate-atpt (&optional arg) 
  "Separates 'angled-lower at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'angled-lower arg))

(defalias 'singlequote-angled-lower-atpt 'angled-lower-singlequote-atpt)
;;;###autoload
(defun angled-lower-singlequote-atpt (&optional arg) 
  "Singlequotes 'angled-lower at point if any. "
  (interactive "*p")
    (th-singlequote 'angled-lower arg))

(defalias 'slash-angled-lower-atpt 'angled-lower-slash-atpt)
;;;###autoload
(defun angled-lower-slash-atpt (&optional arg) 
  "Slash 'angled-lower at point if any. "
  (interactive "*p")
    (th-slash 'angled-lower arg))

(defalias 'trim-angled-lower-atpt 'angled-lower-trim-atpt)
;;;###autoload
(defun angled-lower-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-lower t t))

(defalias 'trim-angled-lower-left-atpt 'angled-lower-left-trim-atpt)
;;;###autoload
(defun angled-lower-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-lower t nil))

(defalias 'trim-angled-lower-right-atpt 'angled-lower-right-trim-atpt)
;;;###autoload
(defun angled-lower-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'angled-lower nil t))

(defalias 'forward-angled-lower-atpt 'angled-lower-forward-atpt)
;;;###autoload
(defun angled-lower-forward-atpt (&optional arg) 
  "Moves forward over angled-lower at point if any, does nothing otherwise.
Returns bounds of angled-lower as a list. "
  (interactive "p")
  (th-forward 'angled-lower arg))

(defalias 'backward-angled-lower-atpt 'angled-lower-backward-atpt)
;;;###autoload
(defun angled-lower-backward-atpt (&optional arg) 
  "Moves backward over angled-lower before point if any, does nothing otherwise.
Returns bounds of angled-lower as a list. "
  (interactive "p")
  (th-backward 'angled-lower arg))

(defalias 'transpose-angled-lower-atpt 'angled-lower-transpose-atpt)
;;;###autoload
(defun angled-lower-transpose-atpt (&optional arg) 
  "Transposes angled-lower with angled-lower before point if any. "
  (interactive "*P")
  (th-transpose 'angled-lower arg))

(defalias 'sort-angled-lower-atpt 'angled-lower-sort-atpt)
;;;###autoload
(defun angled-lower-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts angled-lower's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'angled-lower reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-angled-lower-atpt 'angled-lower-check-atpt)
;;;###autoload
(defun angled-lower-check-atpt (&optional arg) 
  " angled-lower at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "angled-lower" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "angled-lower" "-end-position-atpt")))))
    (th-check 'angled-lower arg)))
   

;;;###autoload
(defun count-alnum-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of alnum in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-lower beg end (interactive-p)))

(defun ratio-alnum-in-angled-lower-atpt (&optional beg end) 
    "Relation of alnum in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of alpha in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-lower beg end (interactive-p)))

(defun ratio-alpha-in-angled-lower-atpt (&optional beg end) 
    "Relation of alpha in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of ascii in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-lower beg end (interactive-p)))

(defun ratio-ascii-in-angled-lower-atpt (&optional beg end) 
    "Relation of ascii in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of digit in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-lower beg end (interactive-p)))

(defun ratio-digit-in-angled-lower-atpt (&optional beg end) 
    "Relation of digit in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of graph in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-lower beg end (interactive-p)))

(defun ratio-graph-in-angled-lower-atpt (&optional beg end) 
    "Relation of graph in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of lower in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-lower beg end (interactive-p)))

(defun ratio-lower-in-angled-lower-atpt (&optional beg end) 
    "Relation of lower in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of punct in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-lower beg end (interactive-p)))

(defun ratio-punct-in-angled-lower-atpt (&optional beg end) 
    "Relation of punct in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-angled-lower-atpt (&optional beg end) 
    "Counts matches of upper in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-lower beg end (interactive-p)))

(defun ratio-upper-in-angled-lower-atpt (&optional beg end) 
    "Relation of upper in angled-lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'angled-lower beg end t (interactive-p)))



;;;###autoload
(defun ascii-atpt (&optional arg) 
  "Returns ascii at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'ascii arg))

(defalias 'bounds-of-ascii-atpt 'ascii-bounds-atpt)
;;;###autoload
(defun ascii-bounds-atpt (&optional arg) 
  "Returns borders of ascii at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'ascii arg))

;;;###autoload
(defun ascii-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'ascii arg))

;;;###autoload
(defun ascii-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-end 'ascii arg))

(defalias 'beginning-of-ascii-atpt 'ascii-beginning-atpt)
;;;###autoload
(defun ascii-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'ascii arg))

(defalias 'end-of-ascii-atpt 'ascii-end-atpt)
;;;###autoload
(defun ascii-end-atpt (&optional arg) 
  "Goto end of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'ascii arg))

(defalias 'length-of-ascii-atpt 'ascii-length-atpt)
;;;###autoload
(defun ascii-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-length 'ascii arg))

(defalias 'copy-ascii-atpt 'ascii-copy-atpt)
;;;###autoload
(defun ascii-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'ascii arg))

(defalias 'escape-ascii-atpt 'ascii-escape-atpt)
;;;###autoload
(defun ascii-escape-atpt () 
  "Makes a copy of symbol or char-class ascii at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'ascii))

(defalias 'doublequote-ascii-atpt 'ascii-doublequote-atpt)
;;;###autoload
(defun ascii-doublequote-atpt (&optional arg) 
  "Doublequotes 'ascii at point if any. "
  (interactive "*p")
    (th-doublequote 'ascii arg))

(defalias 'dollar-ascii-atpt 'ascii-dollar-atpt)
;;;###autoload
(defun ascii-dollar-atpt (&optional arg) 
  "Doublequotes 'ascii at point if any. "
  (interactive "*p")
    (th-dollar 'ascii arg))

(defalias 'angle-greater-ascii-atpt 'ascii-greater-angle-atpt)
;;;###autoload
(defun ascii-greater-angle-atpt (&optional arg) 
  "Sets angles for 'ascii after point if any. "
  (interactive "*p")
    (th-greater-angle 'ascii arg))

(defalias 'angle-lower-ascii-atpt 'ascii-lower-angle-atpt)
;;;###autoload
(defun ascii-lower-angle-atpt (&optional arg) 
  "Sets angles for 'ascii after point if any. "
  (interactive "*p")
    (th-lower-angle 'ascii arg))

(defalias 'backslash-ascii-atpt 'ascii-backslash-atpt)
;;;###autoload
(defun ascii-backslash-atpt (&optional arg) 
  "Backslash 'ascii at point if any. "
  (interactive "*p")
    (th-backslash 'ascii arg))   

(defalias 'brace-ascii-atpt 'ascii-brace-atpt)
;;;###autoload
(defun ascii-brace-atpt (&optional arg) 
  "Braces 'ascii at point if any. "
  (interactive "*p")
    (th-brace 'ascii arg))

(defalias 'bracket-ascii-atpt 'ascii-bracket-atpt)
;;;###autoload
(defun ascii-bracket-atpt (&optional arg) 
  "Brackets 'ascii after point if any. "
  (interactive "*p")
    (th-bracket 'ascii arg))

;; (defalias 'comment-ascii-atpt 'ascii-comment-atpt)
;;;###autoload
(defun ascii-comment-atpt (&optional arg) 
  "Comments 'ascii at point if any. "
  (interactive "*p")
  (th-comment 'ascii arg))

(defalias 'mark-ascii-atpt 'ascii-mark-atpt)
;;;###autoload
(defun ascii-mark-atpt () 
  "Marks 'ascii at point if any. "
  (interactive)
  (th-mark 'ascii))

(defalias 'hide-ascii-atpt 'ascii-hide-atpt)
;;;###autoload
(defun ascii-hide-atpt () 
  "Toggles hide 'ascii in region or buffer if any. "
  (interactive)
  (th-hide 'ascii))

(defalias 'highlight-ascii-atpt-mode 'ascii-highlight-atpt-mode)
(defun ascii-highlight-atpt-mode (&optional arg) 
  "Toggles ascii-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'ascii arg))

(defalias 'kill-ascii-atpt 'ascii-kill-atpt)
;;;###autoload
(defun ascii-kill-atpt (&optional arg) 
  "Kills ascii at point if any. "
  (interactive "*P")
  (th-kill 'ascii arg))

(defalias 'parentize-ascii-atpt 'ascii-parentize-atpt)
;;;###autoload
(defun ascii-parentize-atpt (&optional arg) 
  "Parentizes 'ascii at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'ascii arg))

(defalias 'separate-ascii-atpt 'ascii-separate-atpt)
;;;###autoload
(defun ascii-separate-atpt (&optional arg) 
  "Separates 'ascii at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'ascii arg))

(defalias 'singlequote-ascii-atpt 'ascii-singlequote-atpt)
;;;###autoload
(defun ascii-singlequote-atpt (&optional arg) 
  "Singlequotes 'ascii at point if any. "
  (interactive "*p")
    (th-singlequote 'ascii arg))

(defalias 'slash-ascii-atpt 'ascii-slash-atpt)
;;;###autoload
(defun ascii-slash-atpt (&optional arg) 
  "Slash 'ascii at point if any. "
  (interactive "*p")
    (th-slash 'ascii arg))

(defalias 'trim-ascii-atpt 'ascii-trim-atpt)
;;;###autoload
(defun ascii-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ascii t t))

(defalias 'trim-ascii-left-atpt 'ascii-left-trim-atpt)
;;;###autoload
(defun ascii-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ascii t nil))

(defalias 'trim-ascii-right-atpt 'ascii-right-trim-atpt)
;;;###autoload
(defun ascii-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ascii nil t))

(defalias 'forward-ascii-atpt 'ascii-forward-atpt)
;;;###autoload
(defun ascii-forward-atpt (&optional arg) 
  "Moves forward over ascii at point if any, does nothing otherwise.
Returns bounds of ascii as a list. "
  (interactive "p")
  (th-forward 'ascii arg))

(defalias 'backward-ascii-atpt 'ascii-backward-atpt)
;;;###autoload
(defun ascii-backward-atpt (&optional arg) 
  "Moves backward over ascii before point if any, does nothing otherwise.
Returns bounds of ascii as a list. "
  (interactive "p")
  (th-backward 'ascii arg))

(defalias 'transpose-ascii-atpt 'ascii-transpose-atpt)
;;;###autoload
(defun ascii-transpose-atpt (&optional arg) 
  "Transposes ascii with ascii before point if any. "
  (interactive "*P")
  (th-transpose 'ascii arg))

(defalias 'sort-ascii-atpt 'ascii-sort-atpt)
;;;###autoload
(defun ascii-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts ascii's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'ascii reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-ascii-atpt 'ascii-check-atpt)
;;;###autoload
(defun ascii-check-atpt (&optional arg) 
  " ascii at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "ascii" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "ascii" "-end-position-atpt")))))
    (th-check 'ascii arg)))
   

;;;###autoload
(defun count-alnum-in-ascii-atpt (&optional beg end) 
    "Counts matches of alnum in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ascii beg end (interactive-p)))

(defun ratio-alnum-in-ascii-atpt (&optional beg end) 
    "Relation of alnum in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-ascii-atpt (&optional beg end) 
    "Counts matches of alpha in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ascii beg end (interactive-p)))

(defun ratio-alpha-in-ascii-atpt (&optional beg end) 
    "Relation of alpha in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-atpt (&optional beg end) 
    "Counts matches of ascii at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'ascii beg end (interactive-p)))


;;;###autoload
(defun count-digit-in-ascii-atpt (&optional beg end) 
    "Counts matches of digit in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ascii beg end (interactive-p)))

(defun ratio-digit-in-ascii-atpt (&optional beg end) 
    "Relation of digit in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-ascii-atpt (&optional beg end) 
    "Counts matches of graph in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ascii beg end (interactive-p)))

(defun ratio-graph-in-ascii-atpt (&optional beg end) 
    "Relation of graph in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-ascii-atpt (&optional beg end) 
    "Counts matches of lower in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ascii beg end (interactive-p)))

(defun ratio-lower-in-ascii-atpt (&optional beg end) 
    "Relation of lower in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-ascii-atpt (&optional beg end) 
    "Counts matches of punct in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ascii beg end (interactive-p)))

(defun ratio-punct-in-ascii-atpt (&optional beg end) 
    "Relation of punct in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ascii beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-ascii-atpt (&optional beg end) 
    "Counts matches of upper in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ascii beg end (interactive-p)))

(defun ratio-upper-in-ascii-atpt (&optional beg end) 
    "Relation of upper in ascii if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ascii beg end t (interactive-p)))



;;;###autoload
(defun backslashed-atpt (&optional arg) 
  "Returns backslashed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'backslashed arg))

(defalias 'bounds-of-backslashed-atpt 'backslashed-bounds-atpt)
;;;###autoload
(defun backslashed-bounds-atpt (&optional arg) 
  "Returns borders of backslashed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'backslashed arg))

;;;###autoload
(defun backslashed-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'backslashed arg))

;;;###autoload
(defun backslashed-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-end 'backslashed arg))

(defalias 'beginning-of-backslashed-atpt 'backslashed-beginning-atpt)
;;;###autoload
(defun backslashed-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'backslashed arg))

(defalias 'end-of-backslashed-atpt 'backslashed-end-atpt)
;;;###autoload
(defun backslashed-end-atpt (&optional arg) 
  "Goto end of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'backslashed arg))

(defalias 'length-of-backslashed-atpt 'backslashed-length-atpt)
;;;###autoload
(defun backslashed-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-length 'backslashed arg))

(defalias 'copy-backslashed-atpt 'backslashed-copy-atpt)
;;;###autoload
(defun backslashed-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'backslashed arg))

(defalias 'escape-backslashed-atpt 'backslashed-escape-atpt)
;;;###autoload
(defun backslashed-escape-atpt () 
  "Makes a copy of symbol or char-class backslashed at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'backslashed))

(defalias 'doublequote-backslashed-atpt 'backslashed-doublequote-atpt)
;;;###autoload
(defun backslashed-doublequote-atpt (&optional arg) 
  "Doublequotes 'backslashed at point if any. "
  (interactive "*p")
    (th-doublequote 'backslashed arg))

(defalias 'dollar-backslashed-atpt 'backslashed-dollar-atpt)
;;;###autoload
(defun backslashed-dollar-atpt (&optional arg) 
  "Doublequotes 'backslashed at point if any. "
  (interactive "*p")
    (th-dollar 'backslashed arg))

(defalias 'angle-greater-backslashed-atpt 'backslashed-greater-angle-atpt)
;;;###autoload
(defun backslashed-greater-angle-atpt (&optional arg) 
  "Sets angles for 'backslashed after point if any. "
  (interactive "*p")
    (th-greater-angle 'backslashed arg))

(defalias 'angle-lower-backslashed-atpt 'backslashed-lower-angle-atpt)
;;;###autoload
(defun backslashed-lower-angle-atpt (&optional arg) 
  "Sets angles for 'backslashed after point if any. "
  (interactive "*p")
    (th-lower-angle 'backslashed arg))

(defalias 'backslash-backslashed-atpt 'backslashed-backslash-atpt)
;;;###autoload
(defun backslashed-backslash-atpt (&optional arg) 
  "Backslash 'backslashed at point if any. "
  (interactive "*p")
    (th-backslash 'backslashed arg))   

(defalias 'brace-backslashed-atpt 'backslashed-brace-atpt)
;;;###autoload
(defun backslashed-brace-atpt (&optional arg) 
  "Braces 'backslashed at point if any. "
  (interactive "*p")
    (th-brace 'backslashed arg))

(defalias 'bracket-backslashed-atpt 'backslashed-bracket-atpt)
;;;###autoload
(defun backslashed-bracket-atpt (&optional arg) 
  "Brackets 'backslashed after point if any. "
  (interactive "*p")
    (th-bracket 'backslashed arg))

;; (defalias 'comment-backslashed-atpt 'backslashed-comment-atpt)
;;;###autoload
(defun backslashed-comment-atpt (&optional arg) 
  "Comments 'backslashed at point if any. "
  (interactive "*p")
  (th-comment 'backslashed arg))

(defalias 'mark-backslashed-atpt 'backslashed-mark-atpt)
;;;###autoload
(defun backslashed-mark-atpt () 
  "Marks 'backslashed at point if any. "
  (interactive)
  (th-mark 'backslashed))

(defalias 'hide-backslashed-atpt 'backslashed-hide-atpt)
;;;###autoload
(defun backslashed-hide-atpt () 
  "Toggles hide 'backslashed in region or buffer if any. "
  (interactive)
  (th-hide 'backslashed))

(defalias 'highlight-backslashed-atpt-mode 'backslashed-highlight-atpt-mode)
(defun backslashed-highlight-atpt-mode (&optional arg) 
  "Toggles backslashed-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'backslashed arg))

(defalias 'kill-backslashed-atpt 'backslashed-kill-atpt)
;;;###autoload
(defun backslashed-kill-atpt (&optional arg) 
  "Kills backslashed at point if any. "
  (interactive "*P")
  (th-kill 'backslashed arg))

(defalias 'parentize-backslashed-atpt 'backslashed-parentize-atpt)
;;;###autoload
(defun backslashed-parentize-atpt (&optional arg) 
  "Parentizes 'backslashed at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'backslashed arg))

(defalias 'separate-backslashed-atpt 'backslashed-separate-atpt)
;;;###autoload
(defun backslashed-separate-atpt (&optional arg) 
  "Separates 'backslashed at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'backslashed arg))

(defalias 'singlequote-backslashed-atpt 'backslashed-singlequote-atpt)
;;;###autoload
(defun backslashed-singlequote-atpt (&optional arg) 
  "Singlequotes 'backslashed at point if any. "
  (interactive "*p")
    (th-singlequote 'backslashed arg))

(defalias 'slash-backslashed-atpt 'backslashed-slash-atpt)
;;;###autoload
(defun backslashed-slash-atpt (&optional arg) 
  "Slash 'backslashed at point if any. "
  (interactive "*p")
    (th-slash 'backslashed arg))

(defalias 'trim-backslashed-atpt 'backslashed-trim-atpt)
;;;###autoload
(defun backslashed-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'backslashed t t))

(defalias 'trim-backslashed-left-atpt 'backslashed-left-trim-atpt)
;;;###autoload
(defun backslashed-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'backslashed t nil))

(defalias 'trim-backslashed-right-atpt 'backslashed-right-trim-atpt)
;;;###autoload
(defun backslashed-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'backslashed nil t))

(defalias 'forward-backslashed-atpt 'backslashed-forward-atpt)
;;;###autoload
(defun backslashed-forward-atpt (&optional arg) 
  "Moves forward over backslashed at point if any, does nothing otherwise.
Returns bounds of backslashed as a list. "
  (interactive "p")
  (th-forward 'backslashed arg))

(defalias 'backward-backslashed-atpt 'backslashed-backward-atpt)
;;;###autoload
(defun backslashed-backward-atpt (&optional arg) 
  "Moves backward over backslashed before point if any, does nothing otherwise.
Returns bounds of backslashed as a list. "
  (interactive "p")
  (th-backward 'backslashed arg))

(defalias 'transpose-backslashed-atpt 'backslashed-transpose-atpt)
;;;###autoload
(defun backslashed-transpose-atpt (&optional arg) 
  "Transposes backslashed with backslashed before point if any. "
  (interactive "*P")
  (th-transpose 'backslashed arg))

(defalias 'sort-backslashed-atpt 'backslashed-sort-atpt)
;;;###autoload
(defun backslashed-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts backslashed's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'backslashed reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-backslashed-atpt 'backslashed-check-atpt)
;;;###autoload
(defun backslashed-check-atpt (&optional arg) 
  " backslashed at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "backslashed" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "backslashed" "-end-position-atpt")))))
    (th-check 'backslashed arg)))
   

;;;###autoload
(defun count-alnum-in-backslashed-atpt (&optional beg end) 
    "Counts matches of alnum in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'backslashed beg end (interactive-p)))

(defun ratio-alnum-in-backslashed-atpt (&optional beg end) 
    "Relation of alnum in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-backslashed-atpt (&optional beg end) 
    "Counts matches of alpha in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'backslashed beg end (interactive-p)))

(defun ratio-alpha-in-backslashed-atpt (&optional beg end) 
    "Relation of alpha in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-backslashed-atpt (&optional beg end) 
    "Counts matches of ascii in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'backslashed beg end (interactive-p)))

(defun ratio-ascii-in-backslashed-atpt (&optional beg end) 
    "Relation of ascii in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-backslashed-atpt (&optional beg end) 
    "Counts matches of digit in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'backslashed beg end (interactive-p)))

(defun ratio-digit-in-backslashed-atpt (&optional beg end) 
    "Relation of digit in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-backslashed-atpt (&optional beg end) 
    "Counts matches of graph in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'backslashed beg end (interactive-p)))

(defun ratio-graph-in-backslashed-atpt (&optional beg end) 
    "Relation of graph in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-backslashed-atpt (&optional beg end) 
    "Counts matches of lower in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'backslashed beg end (interactive-p)))

(defun ratio-lower-in-backslashed-atpt (&optional beg end) 
    "Relation of lower in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-backslashed-atpt (&optional beg end) 
    "Counts matches of punct in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'backslashed beg end (interactive-p)))

(defun ratio-punct-in-backslashed-atpt (&optional beg end) 
    "Relation of punct in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-backslashed-atpt (&optional beg end) 
    "Counts matches of upper in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'backslashed beg end (interactive-p)))

(defun ratio-upper-in-backslashed-atpt (&optional beg end) 
    "Relation of upper in backslashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'backslashed beg end t (interactive-p)))



;;;###autoload
(defun blank-atpt (&optional arg) 
  "Returns blank at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'blank arg))

(defalias 'bounds-of-blank-atpt 'blank-bounds-atpt)
;;;###autoload
(defun blank-bounds-atpt (&optional arg) 
  "Returns borders of blank at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'blank arg))

;;;###autoload
(defun blank-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'blank arg))

;;;###autoload
(defun blank-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-end 'blank arg))

(defalias 'beginning-of-blank-atpt 'blank-beginning-atpt)
;;;###autoload
(defun blank-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'blank arg))

(defalias 'end-of-blank-atpt 'blank-end-atpt)
;;;###autoload
(defun blank-end-atpt (&optional arg) 
  "Goto end of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'blank arg))

(defalias 'length-of-blank-atpt 'blank-length-atpt)
;;;###autoload
(defun blank-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-length 'blank arg))

(defalias 'copy-blank-atpt 'blank-copy-atpt)
;;;###autoload
(defun blank-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class blank at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'blank arg))

(defalias 'escape-blank-atpt 'blank-escape-atpt)
;;;###autoload
(defun blank-escape-atpt () 
  "Makes a copy of symbol or char-class blank at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'blank))

(defalias 'doublequote-blank-atpt 'blank-doublequote-atpt)
;;;###autoload
(defun blank-doublequote-atpt (&optional arg) 
  "Doublequotes 'blank at point if any. "
  (interactive "*p")
    (th-doublequote 'blank arg))

(defalias 'dollar-blank-atpt 'blank-dollar-atpt)
;;;###autoload
(defun blank-dollar-atpt (&optional arg) 
  "Doublequotes 'blank at point if any. "
  (interactive "*p")
    (th-dollar 'blank arg))

(defalias 'angle-greater-blank-atpt 'blank-greater-angle-atpt)
;;;###autoload
(defun blank-greater-angle-atpt (&optional arg) 
  "Sets angles for 'blank after point if any. "
  (interactive "*p")
    (th-greater-angle 'blank arg))

(defalias 'angle-lower-blank-atpt 'blank-lower-angle-atpt)
;;;###autoload
(defun blank-lower-angle-atpt (&optional arg) 
  "Sets angles for 'blank after point if any. "
  (interactive "*p")
    (th-lower-angle 'blank arg))

(defalias 'backslash-blank-atpt 'blank-backslash-atpt)
;;;###autoload
(defun blank-backslash-atpt (&optional arg) 
  "Backslash 'blank at point if any. "
  (interactive "*p")
    (th-backslash 'blank arg))   

(defalias 'brace-blank-atpt 'blank-brace-atpt)
;;;###autoload
(defun blank-brace-atpt (&optional arg) 
  "Braces 'blank at point if any. "
  (interactive "*p")
    (th-brace 'blank arg))

(defalias 'bracket-blank-atpt 'blank-bracket-atpt)
;;;###autoload
(defun blank-bracket-atpt (&optional arg) 
  "Brackets 'blank after point if any. "
  (interactive "*p")
    (th-bracket 'blank arg))

;; (defalias 'comment-blank-atpt 'blank-comment-atpt)
;;;###autoload
(defun blank-comment-atpt (&optional arg) 
  "Comments 'blank at point if any. "
  (interactive "*p")
  (th-comment 'blank arg))

(defalias 'mark-blank-atpt 'blank-mark-atpt)
;;;###autoload
(defun blank-mark-atpt () 
  "Marks 'blank at point if any. "
  (interactive)
  (th-mark 'blank))

(defalias 'hide-blank-atpt 'blank-hide-atpt)
;;;###autoload
(defun blank-hide-atpt () 
  "Toggles hide 'blank in region or buffer if any. "
  (interactive)
  (th-hide 'blank))

(defalias 'highlight-blank-atpt-mode 'blank-highlight-atpt-mode)
(defun blank-highlight-atpt-mode (&optional arg) 
  "Toggles blank-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'blank arg))

(defalias 'kill-blank-atpt 'blank-kill-atpt)
;;;###autoload
(defun blank-kill-atpt (&optional arg) 
  "Kills blank at point if any. "
  (interactive "*P")
  (th-kill 'blank arg))

(defalias 'parentize-blank-atpt 'blank-parentize-atpt)
;;;###autoload
(defun blank-parentize-atpt (&optional arg) 
  "Parentizes 'blank at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'blank arg))

(defalias 'separate-blank-atpt 'blank-separate-atpt)
;;;###autoload
(defun blank-separate-atpt (&optional arg) 
  "Separates 'blank at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'blank arg))

(defalias 'singlequote-blank-atpt 'blank-singlequote-atpt)
;;;###autoload
(defun blank-singlequote-atpt (&optional arg) 
  "Singlequotes 'blank at point if any. "
  (interactive "*p")
    (th-singlequote 'blank arg))

(defalias 'slash-blank-atpt 'blank-slash-atpt)
;;;###autoload
(defun blank-slash-atpt (&optional arg) 
  "Slash 'blank at point if any. "
  (interactive "*p")
    (th-slash 'blank arg))

(defalias 'trim-blank-atpt 'blank-trim-atpt)
;;;###autoload
(defun blank-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'blank t t))

(defalias 'trim-blank-left-atpt 'blank-left-trim-atpt)
;;;###autoload
(defun blank-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'blank t nil))

(defalias 'trim-blank-right-atpt 'blank-right-trim-atpt)
;;;###autoload
(defun blank-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'blank nil t))

(defalias 'forward-blank-atpt 'blank-forward-atpt)
;;;###autoload
(defun blank-forward-atpt (&optional arg) 
  "Moves forward over blank at point if any, does nothing otherwise.
Returns bounds of blank as a list. "
  (interactive "p")
  (th-forward 'blank arg))

(defalias 'backward-blank-atpt 'blank-backward-atpt)
;;;###autoload
(defun blank-backward-atpt (&optional arg) 
  "Moves backward over blank before point if any, does nothing otherwise.
Returns bounds of blank as a list. "
  (interactive "p")
  (th-backward 'blank arg))

(defalias 'transpose-blank-atpt 'blank-transpose-atpt)
;;;###autoload
(defun blank-transpose-atpt (&optional arg) 
  "Transposes blank with blank before point if any. "
  (interactive "*P")
  (th-transpose 'blank arg))

(defalias 'sort-blank-atpt 'blank-sort-atpt)
;;;###autoload
(defun blank-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts blank's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'blank reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-blank-atpt 'blank-check-atpt)
;;;###autoload
(defun blank-check-atpt (&optional arg) 
  " blank at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "blank" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "blank" "-end-position-atpt")))))
    (th-check 'blank arg)))
   

;;;###autoload
(defun count-alnum-in-blank-atpt (&optional beg end) 
    "Counts matches of alnum in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'blank beg end (interactive-p)))

(defun ratio-alnum-in-blank-atpt (&optional beg end) 
    "Relation of alnum in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-blank-atpt (&optional beg end) 
    "Counts matches of alpha in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'blank beg end (interactive-p)))

(defun ratio-alpha-in-blank-atpt (&optional beg end) 
    "Relation of alpha in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-blank-atpt (&optional beg end) 
    "Counts matches of ascii in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'blank beg end (interactive-p)))

(defun ratio-ascii-in-blank-atpt (&optional beg end) 
    "Relation of ascii in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-blank-atpt (&optional beg end) 
    "Counts matches of digit in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'blank beg end (interactive-p)))

(defun ratio-digit-in-blank-atpt (&optional beg end) 
    "Relation of digit in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-blank-atpt (&optional beg end) 
    "Counts matches of graph in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'blank beg end (interactive-p)))

(defun ratio-graph-in-blank-atpt (&optional beg end) 
    "Relation of graph in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-blank-atpt (&optional beg end) 
    "Counts matches of lower in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'blank beg end (interactive-p)))

(defun ratio-lower-in-blank-atpt (&optional beg end) 
    "Relation of lower in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-blank-atpt (&optional beg end) 
    "Counts matches of punct in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'blank beg end (interactive-p)))

(defun ratio-punct-in-blank-atpt (&optional beg end) 
    "Relation of punct in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'blank beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-blank-atpt (&optional beg end) 
    "Counts matches of upper in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'blank beg end (interactive-p)))

(defun ratio-upper-in-blank-atpt (&optional beg end) 
    "Relation of upper in blank if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'blank beg end t (interactive-p)))



;;;###autoload
(defun braced-atpt (&optional arg) 
  "Returns braced at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'braced arg))

(defalias 'bounds-of-braced-atpt 'braced-bounds-atpt)
;;;###autoload
(defun braced-bounds-atpt (&optional arg) 
  "Returns borders of braced at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'braced arg))

;;;###autoload
(defun braced-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'braced arg))

;;;###autoload
(defun braced-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-end 'braced arg))

(defalias 'beginning-of-braced-atpt 'braced-beginning-atpt)
;;;###autoload
(defun braced-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'braced arg))

(defalias 'end-of-braced-atpt 'braced-end-atpt)
;;;###autoload
(defun braced-end-atpt (&optional arg) 
  "Goto end of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'braced arg))

(defalias 'length-of-braced-atpt 'braced-length-atpt)
;;;###autoload
(defun braced-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-length 'braced arg))

(defalias 'copy-braced-atpt 'braced-copy-atpt)
;;;###autoload
(defun braced-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class braced at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'braced arg))

(defalias 'escape-braced-atpt 'braced-escape-atpt)
;;;###autoload
(defun braced-escape-atpt () 
  "Makes a copy of symbol or char-class braced at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'braced))

(defalias 'doublequote-braced-atpt 'braced-doublequote-atpt)
;;;###autoload
(defun braced-doublequote-atpt (&optional arg) 
  "Doublequotes 'braced at point if any. "
  (interactive "*p")
    (th-doublequote 'braced arg))

(defalias 'dollar-braced-atpt 'braced-dollar-atpt)
;;;###autoload
(defun braced-dollar-atpt (&optional arg) 
  "Doublequotes 'braced at point if any. "
  (interactive "*p")
    (th-dollar 'braced arg))

(defalias 'angle-greater-braced-atpt 'braced-greater-angle-atpt)
;;;###autoload
(defun braced-greater-angle-atpt (&optional arg) 
  "Sets angles for 'braced after point if any. "
  (interactive "*p")
    (th-greater-angle 'braced arg))

(defalias 'angle-lower-braced-atpt 'braced-lower-angle-atpt)
;;;###autoload
(defun braced-lower-angle-atpt (&optional arg) 
  "Sets angles for 'braced after point if any. "
  (interactive "*p")
    (th-lower-angle 'braced arg))

(defalias 'backslash-braced-atpt 'braced-backslash-atpt)
;;;###autoload
(defun braced-backslash-atpt (&optional arg) 
  "Backslash 'braced at point if any. "
  (interactive "*p")
    (th-backslash 'braced arg))   

(defalias 'brace-braced-atpt 'braced-brace-atpt)
;;;###autoload
(defun braced-brace-atpt (&optional arg) 
  "Braces 'braced at point if any. "
  (interactive "*p")
    (th-brace 'braced arg))

(defalias 'bracket-braced-atpt 'braced-bracket-atpt)
;;;###autoload
(defun braced-bracket-atpt (&optional arg) 
  "Brackets 'braced after point if any. "
  (interactive "*p")
    (th-bracket 'braced arg))

;; (defalias 'comment-braced-atpt 'braced-comment-atpt)
;;;###autoload
(defun braced-comment-atpt (&optional arg) 
  "Comments 'braced at point if any. "
  (interactive "*p")
  (th-comment 'braced arg))

(defalias 'mark-braced-atpt 'braced-mark-atpt)
;;;###autoload
(defun braced-mark-atpt () 
  "Marks 'braced at point if any. "
  (interactive)
  (th-mark 'braced))

(defalias 'hide-braced-atpt 'braced-hide-atpt)
;;;###autoload
(defun braced-hide-atpt () 
  "Toggles hide 'braced in region or buffer if any. "
  (interactive)
  (th-hide 'braced))

(defalias 'highlight-braced-atpt-mode 'braced-highlight-atpt-mode)
(defun braced-highlight-atpt-mode (&optional arg) 
  "Toggles braced-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'braced arg))

(defalias 'kill-braced-atpt 'braced-kill-atpt)
;;;###autoload
(defun braced-kill-atpt (&optional arg) 
  "Kills braced at point if any. "
  (interactive "*P")
  (th-kill 'braced arg))

(defalias 'parentize-braced-atpt 'braced-parentize-atpt)
;;;###autoload
(defun braced-parentize-atpt (&optional arg) 
  "Parentizes 'braced at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'braced arg))

(defalias 'separate-braced-atpt 'braced-separate-atpt)
;;;###autoload
(defun braced-separate-atpt (&optional arg) 
  "Separates 'braced at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'braced arg))

(defalias 'singlequote-braced-atpt 'braced-singlequote-atpt)
;;;###autoload
(defun braced-singlequote-atpt (&optional arg) 
  "Singlequotes 'braced at point if any. "
  (interactive "*p")
    (th-singlequote 'braced arg))

(defalias 'slash-braced-atpt 'braced-slash-atpt)
;;;###autoload
(defun braced-slash-atpt (&optional arg) 
  "Slash 'braced at point if any. "
  (interactive "*p")
    (th-slash 'braced arg))

(defalias 'trim-braced-atpt 'braced-trim-atpt)
;;;###autoload
(defun braced-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'braced t t))

(defalias 'trim-braced-left-atpt 'braced-left-trim-atpt)
;;;###autoload
(defun braced-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'braced t nil))

(defalias 'trim-braced-right-atpt 'braced-right-trim-atpt)
;;;###autoload
(defun braced-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'braced nil t))

(defalias 'forward-braced-atpt 'braced-forward-atpt)
;;;###autoload
(defun braced-forward-atpt (&optional arg) 
  "Moves forward over braced at point if any, does nothing otherwise.
Returns bounds of braced as a list. "
  (interactive "p")
  (th-forward 'braced arg))

(defalias 'backward-braced-atpt 'braced-backward-atpt)
;;;###autoload
(defun braced-backward-atpt (&optional arg) 
  "Moves backward over braced before point if any, does nothing otherwise.
Returns bounds of braced as a list. "
  (interactive "p")
  (th-backward 'braced arg))

(defalias 'transpose-braced-atpt 'braced-transpose-atpt)
;;;###autoload
(defun braced-transpose-atpt (&optional arg) 
  "Transposes braced with braced before point if any. "
  (interactive "*P")
  (th-transpose 'braced arg))

(defalias 'sort-braced-atpt 'braced-sort-atpt)
;;;###autoload
(defun braced-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts braced's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'braced reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-braced-atpt 'braced-check-atpt)
;;;###autoload
(defun braced-check-atpt (&optional arg) 
  " braced at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "braced" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "braced" "-end-position-atpt")))))
    (th-check 'braced arg)))
   

;;;###autoload
(defun count-alnum-in-braced-atpt (&optional beg end) 
    "Counts matches of alnum in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'braced beg end (interactive-p)))

(defun ratio-alnum-in-braced-atpt (&optional beg end) 
    "Relation of alnum in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-braced-atpt (&optional beg end) 
    "Counts matches of alpha in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'braced beg end (interactive-p)))

(defun ratio-alpha-in-braced-atpt (&optional beg end) 
    "Relation of alpha in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-braced-atpt (&optional beg end) 
    "Counts matches of ascii in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'braced beg end (interactive-p)))

(defun ratio-ascii-in-braced-atpt (&optional beg end) 
    "Relation of ascii in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-braced-atpt (&optional beg end) 
    "Counts matches of digit in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'braced beg end (interactive-p)))

(defun ratio-digit-in-braced-atpt (&optional beg end) 
    "Relation of digit in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-braced-atpt (&optional beg end) 
    "Counts matches of graph in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'braced beg end (interactive-p)))

(defun ratio-graph-in-braced-atpt (&optional beg end) 
    "Relation of graph in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-braced-atpt (&optional beg end) 
    "Counts matches of lower in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'braced beg end (interactive-p)))

(defun ratio-lower-in-braced-atpt (&optional beg end) 
    "Relation of lower in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-braced-atpt (&optional beg end) 
    "Counts matches of punct in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'braced beg end (interactive-p)))

(defun ratio-punct-in-braced-atpt (&optional beg end) 
    "Relation of punct in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'braced beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-braced-atpt (&optional beg end) 
    "Counts matches of upper in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'braced beg end (interactive-p)))

(defun ratio-upper-in-braced-atpt (&optional beg end) 
    "Relation of upper in braced if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'braced beg end t (interactive-p)))



;;;###autoload
(defun bracketed-atpt (&optional arg) 
  "Returns bracketed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'bracketed arg))

(defalias 'bounds-of-bracketed-atpt 'bracketed-bounds-atpt)
;;;###autoload
(defun bracketed-bounds-atpt (&optional arg) 
  "Returns borders of bracketed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'bracketed arg))

;;;###autoload
(defun bracketed-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'bracketed arg))

;;;###autoload
(defun bracketed-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-end 'bracketed arg))

(defalias 'beginning-of-bracketed-atpt 'bracketed-beginning-atpt)
;;;###autoload
(defun bracketed-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'bracketed arg))

(defalias 'end-of-bracketed-atpt 'bracketed-end-atpt)
;;;###autoload
(defun bracketed-end-atpt (&optional arg) 
  "Goto end of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'bracketed arg))

(defalias 'length-of-bracketed-atpt 'bracketed-length-atpt)
;;;###autoload
(defun bracketed-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-length 'bracketed arg))

(defalias 'copy-bracketed-atpt 'bracketed-copy-atpt)
;;;###autoload
(defun bracketed-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'bracketed arg))

(defalias 'escape-bracketed-atpt 'bracketed-escape-atpt)
;;;###autoload
(defun bracketed-escape-atpt () 
  "Makes a copy of symbol or char-class bracketed at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'bracketed))

(defalias 'doublequote-bracketed-atpt 'bracketed-doublequote-atpt)
;;;###autoload
(defun bracketed-doublequote-atpt (&optional arg) 
  "Doublequotes 'bracketed at point if any. "
  (interactive "*p")
    (th-doublequote 'bracketed arg))

(defalias 'dollar-bracketed-atpt 'bracketed-dollar-atpt)
;;;###autoload
(defun bracketed-dollar-atpt (&optional arg) 
  "Doublequotes 'bracketed at point if any. "
  (interactive "*p")
    (th-dollar 'bracketed arg))

(defalias 'angle-greater-bracketed-atpt 'bracketed-greater-angle-atpt)
;;;###autoload
(defun bracketed-greater-angle-atpt (&optional arg) 
  "Sets angles for 'bracketed after point if any. "
  (interactive "*p")
    (th-greater-angle 'bracketed arg))

(defalias 'angle-lower-bracketed-atpt 'bracketed-lower-angle-atpt)
;;;###autoload
(defun bracketed-lower-angle-atpt (&optional arg) 
  "Sets angles for 'bracketed after point if any. "
  (interactive "*p")
    (th-lower-angle 'bracketed arg))

(defalias 'backslash-bracketed-atpt 'bracketed-backslash-atpt)
;;;###autoload
(defun bracketed-backslash-atpt (&optional arg) 
  "Backslash 'bracketed at point if any. "
  (interactive "*p")
    (th-backslash 'bracketed arg))   

(defalias 'brace-bracketed-atpt 'bracketed-brace-atpt)
;;;###autoload
(defun bracketed-brace-atpt (&optional arg) 
  "Braces 'bracketed at point if any. "
  (interactive "*p")
    (th-brace 'bracketed arg))

(defalias 'bracket-bracketed-atpt 'bracketed-bracket-atpt)
;;;###autoload
(defun bracketed-bracket-atpt (&optional arg) 
  "Brackets 'bracketed after point if any. "
  (interactive "*p")
    (th-bracket 'bracketed arg))

;; (defalias 'comment-bracketed-atpt 'bracketed-comment-atpt)
;;;###autoload
(defun bracketed-comment-atpt (&optional arg) 
  "Comments 'bracketed at point if any. "
  (interactive "*p")
  (th-comment 'bracketed arg))

(defalias 'mark-bracketed-atpt 'bracketed-mark-atpt)
;;;###autoload
(defun bracketed-mark-atpt () 
  "Marks 'bracketed at point if any. "
  (interactive)
  (th-mark 'bracketed))

(defalias 'hide-bracketed-atpt 'bracketed-hide-atpt)
;;;###autoload
(defun bracketed-hide-atpt () 
  "Toggles hide 'bracketed in region or buffer if any. "
  (interactive)
  (th-hide 'bracketed))

(defalias 'highlight-bracketed-atpt-mode 'bracketed-highlight-atpt-mode)
(defun bracketed-highlight-atpt-mode (&optional arg) 
  "Toggles bracketed-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'bracketed arg))

(defalias 'kill-bracketed-atpt 'bracketed-kill-atpt)
;;;###autoload
(defun bracketed-kill-atpt (&optional arg) 
  "Kills bracketed at point if any. "
  (interactive "*P")
  (th-kill 'bracketed arg))

(defalias 'parentize-bracketed-atpt 'bracketed-parentize-atpt)
;;;###autoload
(defun bracketed-parentize-atpt (&optional arg) 
  "Parentizes 'bracketed at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'bracketed arg))

(defalias 'separate-bracketed-atpt 'bracketed-separate-atpt)
;;;###autoload
(defun bracketed-separate-atpt (&optional arg) 
  "Separates 'bracketed at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'bracketed arg))

(defalias 'singlequote-bracketed-atpt 'bracketed-singlequote-atpt)
;;;###autoload
(defun bracketed-singlequote-atpt (&optional arg) 
  "Singlequotes 'bracketed at point if any. "
  (interactive "*p")
    (th-singlequote 'bracketed arg))

(defalias 'slash-bracketed-atpt 'bracketed-slash-atpt)
;;;###autoload
(defun bracketed-slash-atpt (&optional arg) 
  "Slash 'bracketed at point if any. "
  (interactive "*p")
    (th-slash 'bracketed arg))

(defalias 'trim-bracketed-atpt 'bracketed-trim-atpt)
;;;###autoload
(defun bracketed-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'bracketed t t))

(defalias 'trim-bracketed-left-atpt 'bracketed-left-trim-atpt)
;;;###autoload
(defun bracketed-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'bracketed t nil))

(defalias 'trim-bracketed-right-atpt 'bracketed-right-trim-atpt)
;;;###autoload
(defun bracketed-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'bracketed nil t))

(defalias 'forward-bracketed-atpt 'bracketed-forward-atpt)
;;;###autoload
(defun bracketed-forward-atpt (&optional arg) 
  "Moves forward over bracketed at point if any, does nothing otherwise.
Returns bounds of bracketed as a list. "
  (interactive "p")
  (th-forward 'bracketed arg))

(defalias 'backward-bracketed-atpt 'bracketed-backward-atpt)
;;;###autoload
(defun bracketed-backward-atpt (&optional arg) 
  "Moves backward over bracketed before point if any, does nothing otherwise.
Returns bounds of bracketed as a list. "
  (interactive "p")
  (th-backward 'bracketed arg))

(defalias 'transpose-bracketed-atpt 'bracketed-transpose-atpt)
;;;###autoload
(defun bracketed-transpose-atpt (&optional arg) 
  "Transposes bracketed with bracketed before point if any. "
  (interactive "*P")
  (th-transpose 'bracketed arg))

(defalias 'sort-bracketed-atpt 'bracketed-sort-atpt)
;;;###autoload
(defun bracketed-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts bracketed's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'bracketed reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-bracketed-atpt 'bracketed-check-atpt)
;;;###autoload
(defun bracketed-check-atpt (&optional arg) 
  " bracketed at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "bracketed" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "bracketed" "-end-position-atpt")))))
    (th-check 'bracketed arg)))
   

;;;###autoload
(defun count-alnum-in-bracketed-atpt (&optional beg end) 
    "Counts matches of alnum in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'bracketed beg end (interactive-p)))

(defun ratio-alnum-in-bracketed-atpt (&optional beg end) 
    "Relation of alnum in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-bracketed-atpt (&optional beg end) 
    "Counts matches of alpha in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'bracketed beg end (interactive-p)))

(defun ratio-alpha-in-bracketed-atpt (&optional beg end) 
    "Relation of alpha in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-bracketed-atpt (&optional beg end) 
    "Counts matches of ascii in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'bracketed beg end (interactive-p)))

(defun ratio-ascii-in-bracketed-atpt (&optional beg end) 
    "Relation of ascii in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-bracketed-atpt (&optional beg end) 
    "Counts matches of digit in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'bracketed beg end (interactive-p)))

(defun ratio-digit-in-bracketed-atpt (&optional beg end) 
    "Relation of digit in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-bracketed-atpt (&optional beg end) 
    "Counts matches of graph in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'bracketed beg end (interactive-p)))

(defun ratio-graph-in-bracketed-atpt (&optional beg end) 
    "Relation of graph in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-bracketed-atpt (&optional beg end) 
    "Counts matches of lower in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'bracketed beg end (interactive-p)))

(defun ratio-lower-in-bracketed-atpt (&optional beg end) 
    "Relation of lower in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-bracketed-atpt (&optional beg end) 
    "Counts matches of punct in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'bracketed beg end (interactive-p)))

(defun ratio-punct-in-bracketed-atpt (&optional beg end) 
    "Relation of punct in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-bracketed-atpt (&optional beg end) 
    "Counts matches of upper in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'bracketed beg end (interactive-p)))

(defun ratio-upper-in-bracketed-atpt (&optional beg end) 
    "Relation of upper in bracketed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'bracketed beg end t (interactive-p)))



;;;###autoload
(defun cntrl-atpt (&optional arg) 
  "Returns cntrl at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'cntrl arg))

(defalias 'bounds-of-cntrl-atpt 'cntrl-bounds-atpt)
;;;###autoload
(defun cntrl-bounds-atpt (&optional arg) 
  "Returns borders of cntrl at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'cntrl arg))

;;;###autoload
(defun cntrl-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'cntrl arg))

;;;###autoload
(defun cntrl-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-end 'cntrl arg))

(defalias 'beginning-of-cntrl-atpt 'cntrl-beginning-atpt)
;;;###autoload
(defun cntrl-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'cntrl arg))

(defalias 'end-of-cntrl-atpt 'cntrl-end-atpt)
;;;###autoload
(defun cntrl-end-atpt (&optional arg) 
  "Goto end of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'cntrl arg))

(defalias 'length-of-cntrl-atpt 'cntrl-length-atpt)
;;;###autoload
(defun cntrl-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-length 'cntrl arg))

(defalias 'copy-cntrl-atpt 'cntrl-copy-atpt)
;;;###autoload
(defun cntrl-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'cntrl arg))

(defalias 'escape-cntrl-atpt 'cntrl-escape-atpt)
;;;###autoload
(defun cntrl-escape-atpt () 
  "Makes a copy of symbol or char-class cntrl at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'cntrl))

(defalias 'doublequote-cntrl-atpt 'cntrl-doublequote-atpt)
;;;###autoload
(defun cntrl-doublequote-atpt (&optional arg) 
  "Doublequotes 'cntrl at point if any. "
  (interactive "*p")
    (th-doublequote 'cntrl arg))

(defalias 'dollar-cntrl-atpt 'cntrl-dollar-atpt)
;;;###autoload
(defun cntrl-dollar-atpt (&optional arg) 
  "Doublequotes 'cntrl at point if any. "
  (interactive "*p")
    (th-dollar 'cntrl arg))

(defalias 'angle-greater-cntrl-atpt 'cntrl-greater-angle-atpt)
;;;###autoload
(defun cntrl-greater-angle-atpt (&optional arg) 
  "Sets angles for 'cntrl after point if any. "
  (interactive "*p")
    (th-greater-angle 'cntrl arg))

(defalias 'angle-lower-cntrl-atpt 'cntrl-lower-angle-atpt)
;;;###autoload
(defun cntrl-lower-angle-atpt (&optional arg) 
  "Sets angles for 'cntrl after point if any. "
  (interactive "*p")
    (th-lower-angle 'cntrl arg))

(defalias 'backslash-cntrl-atpt 'cntrl-backslash-atpt)
;;;###autoload
(defun cntrl-backslash-atpt (&optional arg) 
  "Backslash 'cntrl at point if any. "
  (interactive "*p")
    (th-backslash 'cntrl arg))   

(defalias 'brace-cntrl-atpt 'cntrl-brace-atpt)
;;;###autoload
(defun cntrl-brace-atpt (&optional arg) 
  "Braces 'cntrl at point if any. "
  (interactive "*p")
    (th-brace 'cntrl arg))

(defalias 'bracket-cntrl-atpt 'cntrl-bracket-atpt)
;;;###autoload
(defun cntrl-bracket-atpt (&optional arg) 
  "Brackets 'cntrl after point if any. "
  (interactive "*p")
    (th-bracket 'cntrl arg))

;; (defalias 'comment-cntrl-atpt 'cntrl-comment-atpt)
;;;###autoload
(defun cntrl-comment-atpt (&optional arg) 
  "Comments 'cntrl at point if any. "
  (interactive "*p")
  (th-comment 'cntrl arg))

(defalias 'mark-cntrl-atpt 'cntrl-mark-atpt)
;;;###autoload
(defun cntrl-mark-atpt () 
  "Marks 'cntrl at point if any. "
  (interactive)
  (th-mark 'cntrl))

(defalias 'hide-cntrl-atpt 'cntrl-hide-atpt)
;;;###autoload
(defun cntrl-hide-atpt () 
  "Toggles hide 'cntrl in region or buffer if any. "
  (interactive)
  (th-hide 'cntrl))

(defalias 'highlight-cntrl-atpt-mode 'cntrl-highlight-atpt-mode)
(defun cntrl-highlight-atpt-mode (&optional arg) 
  "Toggles cntrl-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'cntrl arg))

(defalias 'kill-cntrl-atpt 'cntrl-kill-atpt)
;;;###autoload
(defun cntrl-kill-atpt (&optional arg) 
  "Kills cntrl at point if any. "
  (interactive "*P")
  (th-kill 'cntrl arg))

(defalias 'parentize-cntrl-atpt 'cntrl-parentize-atpt)
;;;###autoload
(defun cntrl-parentize-atpt (&optional arg) 
  "Parentizes 'cntrl at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'cntrl arg))

(defalias 'separate-cntrl-atpt 'cntrl-separate-atpt)
;;;###autoload
(defun cntrl-separate-atpt (&optional arg) 
  "Separates 'cntrl at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'cntrl arg))

(defalias 'singlequote-cntrl-atpt 'cntrl-singlequote-atpt)
;;;###autoload
(defun cntrl-singlequote-atpt (&optional arg) 
  "Singlequotes 'cntrl at point if any. "
  (interactive "*p")
    (th-singlequote 'cntrl arg))

(defalias 'slash-cntrl-atpt 'cntrl-slash-atpt)
;;;###autoload
(defun cntrl-slash-atpt (&optional arg) 
  "Slash 'cntrl at point if any. "
  (interactive "*p")
    (th-slash 'cntrl arg))

(defalias 'trim-cntrl-atpt 'cntrl-trim-atpt)
;;;###autoload
(defun cntrl-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'cntrl t t))

(defalias 'trim-cntrl-left-atpt 'cntrl-left-trim-atpt)
;;;###autoload
(defun cntrl-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'cntrl t nil))

(defalias 'trim-cntrl-right-atpt 'cntrl-right-trim-atpt)
;;;###autoload
(defun cntrl-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'cntrl nil t))

(defalias 'forward-cntrl-atpt 'cntrl-forward-atpt)
;;;###autoload
(defun cntrl-forward-atpt (&optional arg) 
  "Moves forward over cntrl at point if any, does nothing otherwise.
Returns bounds of cntrl as a list. "
  (interactive "p")
  (th-forward 'cntrl arg))

(defalias 'backward-cntrl-atpt 'cntrl-backward-atpt)
;;;###autoload
(defun cntrl-backward-atpt (&optional arg) 
  "Moves backward over cntrl before point if any, does nothing otherwise.
Returns bounds of cntrl as a list. "
  (interactive "p")
  (th-backward 'cntrl arg))

(defalias 'transpose-cntrl-atpt 'cntrl-transpose-atpt)
;;;###autoload
(defun cntrl-transpose-atpt (&optional arg) 
  "Transposes cntrl with cntrl before point if any. "
  (interactive "*P")
  (th-transpose 'cntrl arg))

(defalias 'sort-cntrl-atpt 'cntrl-sort-atpt)
;;;###autoload
(defun cntrl-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts cntrl's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'cntrl reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-cntrl-atpt 'cntrl-check-atpt)
;;;###autoload
(defun cntrl-check-atpt (&optional arg) 
  " cntrl at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "cntrl" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "cntrl" "-end-position-atpt")))))
    (th-check 'cntrl arg)))
   

;;;###autoload
(defun comment-atpt (&optional arg) 
  "Returns comment at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'comment arg))

(defalias 'bounds-of-comment-atpt 'comment-bounds-atpt)
;;;###autoload
(defun comment-bounds-atpt (&optional arg) 
  "Returns borders of comment at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'comment arg))

;;;###autoload
(defun comment-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'comment arg))

;;;###autoload
(defun comment-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-end 'comment arg))

(defalias 'beginning-of-comment-atpt 'comment-beginning-atpt)
;;;###autoload
(defun comment-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'comment arg))

(defalias 'end-of-comment-atpt 'comment-end-atpt)
;;;###autoload
(defun comment-end-atpt (&optional arg) 
  "Goto end of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'comment arg))

(defalias 'length-of-comment-atpt 'comment-length-atpt)
;;;###autoload
(defun comment-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-length 'comment arg))

(defalias 'copy-comment-atpt 'comment-copy-atpt)
;;;###autoload
(defun comment-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class comment at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'comment arg))

(defalias 'escape-comment-atpt 'comment-escape-atpt)
;;;###autoload
(defun comment-escape-atpt () 
  "Makes a copy of symbol or char-class comment at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'comment))

(defalias 'doublequote-comment-atpt 'comment-doublequote-atpt)
;;;###autoload
(defun comment-doublequote-atpt (&optional arg) 
  "Doublequotes 'comment at point if any. "
  (interactive "*p")
    (th-doublequote 'comment arg))

(defalias 'dollar-comment-atpt 'comment-dollar-atpt)
;;;###autoload
(defun comment-dollar-atpt (&optional arg) 
  "Doublequotes 'comment at point if any. "
  (interactive "*p")
    (th-dollar 'comment arg))

(defalias 'angle-greater-comment-atpt 'comment-greater-angle-atpt)
;;;###autoload
(defun comment-greater-angle-atpt (&optional arg) 
  "Sets angles for 'comment after point if any. "
  (interactive "*p")
    (th-greater-angle 'comment arg))

(defalias 'angle-lower-comment-atpt 'comment-lower-angle-atpt)
;;;###autoload
(defun comment-lower-angle-atpt (&optional arg) 
  "Sets angles for 'comment after point if any. "
  (interactive "*p")
    (th-lower-angle 'comment arg))

(defalias 'backslash-comment-atpt 'comment-backslash-atpt)
;;;###autoload
(defun comment-backslash-atpt (&optional arg) 
  "Backslash 'comment at point if any. "
  (interactive "*p")
    (th-backslash 'comment arg))   

(defalias 'brace-comment-atpt 'comment-brace-atpt)
;;;###autoload
(defun comment-brace-atpt (&optional arg) 
  "Braces 'comment at point if any. "
  (interactive "*p")
    (th-brace 'comment arg))

(defalias 'bracket-comment-atpt 'comment-bracket-atpt)
;;;###autoload
(defun comment-bracket-atpt (&optional arg) 
  "Brackets 'comment after point if any. "
  (interactive "*p")
    (th-bracket 'comment arg))

;; (defalias 'comment-comment-atpt 'comment-comment-atpt)
;;;###autoload
(defun comment-comment-atpt (&optional arg) 
  "Comments 'comment at point if any. "
  (interactive "*p")
  (th-comment 'comment arg))

(defalias 'mark-comment-atpt 'comment-mark-atpt)
;;;###autoload
(defun comment-mark-atpt () 
  "Marks 'comment at point if any. "
  (interactive)
  (th-mark 'comment))

(defalias 'hide-comment-atpt 'comment-hide-atpt)
;;;###autoload
(defun comment-hide-atpt () 
  "Toggles hide 'comment in region or buffer if any. "
  (interactive)
  (th-hide 'comment))

(defalias 'highlight-comment-atpt-mode 'comment-highlight-atpt-mode)
(defun comment-highlight-atpt-mode (&optional arg) 
  "Toggles comment-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'comment arg))

(defalias 'kill-comment-atpt 'comment-kill-atpt)
;;;###autoload
(defun comment-kill-atpt (&optional arg) 
  "Kills comment at point if any. "
  (interactive "*P")
  (th-kill 'comment arg))

(defalias 'parentize-comment-atpt 'comment-parentize-atpt)
;;;###autoload
(defun comment-parentize-atpt (&optional arg) 
  "Parentizes 'comment at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'comment arg))

(defalias 'separate-comment-atpt 'comment-separate-atpt)
;;;###autoload
(defun comment-separate-atpt (&optional arg) 
  "Separates 'comment at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'comment arg))

(defalias 'singlequote-comment-atpt 'comment-singlequote-atpt)
;;;###autoload
(defun comment-singlequote-atpt (&optional arg) 
  "Singlequotes 'comment at point if any. "
  (interactive "*p")
    (th-singlequote 'comment arg))

(defalias 'slash-comment-atpt 'comment-slash-atpt)
;;;###autoload
(defun comment-slash-atpt (&optional arg) 
  "Slash 'comment at point if any. "
  (interactive "*p")
    (th-slash 'comment arg))

(defalias 'trim-comment-atpt 'comment-trim-atpt)
;;;###autoload
(defun comment-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'comment t t))

(defalias 'trim-comment-left-atpt 'comment-left-trim-atpt)
;;;###autoload
(defun comment-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'comment t nil))

(defalias 'trim-comment-right-atpt 'comment-right-trim-atpt)
;;;###autoload
(defun comment-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'comment nil t))

(defalias 'forward-comment-atpt 'comment-forward-atpt)
;;;###autoload
(defun comment-forward-atpt (&optional arg) 
  "Moves forward over comment at point if any, does nothing otherwise.
Returns bounds of comment as a list. "
  (interactive "p")
  (th-forward 'comment arg))

(defalias 'backward-comment-atpt 'comment-backward-atpt)
;;;###autoload
(defun comment-backward-atpt (&optional arg) 
  "Moves backward over comment before point if any, does nothing otherwise.
Returns bounds of comment as a list. "
  (interactive "p")
  (th-backward 'comment arg))

(defalias 'transpose-comment-atpt 'comment-transpose-atpt)
;;;###autoload
(defun comment-transpose-atpt (&optional arg) 
  "Transposes comment with comment before point if any. "
  (interactive "*P")
  (th-transpose 'comment arg))

(defalias 'sort-comment-atpt 'comment-sort-atpt)
;;;###autoload
(defun comment-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts comment's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'comment reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-comment-atpt 'comment-check-atpt)
;;;###autoload
(defun comment-check-atpt (&optional arg) 
  " comment at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "comment" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "comment" "-end-position-atpt")))))
    (th-check 'comment arg)))
   

;;;###autoload
(defun count-alnum-in-comment-atpt (&optional beg end) 
    "Counts matches of alnum in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'comment beg end (interactive-p)))

(defun ratio-alnum-in-comment-atpt (&optional beg end) 
    "Relation of alnum in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-comment-atpt (&optional beg end) 
    "Counts matches of alpha in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'comment beg end (interactive-p)))

(defun ratio-alpha-in-comment-atpt (&optional beg end) 
    "Relation of alpha in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-comment-atpt (&optional beg end) 
    "Counts matches of ascii in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'comment beg end (interactive-p)))

(defun ratio-ascii-in-comment-atpt (&optional beg end) 
    "Relation of ascii in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-comment-atpt (&optional beg end) 
    "Counts matches of digit in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'comment beg end (interactive-p)))

(defun ratio-digit-in-comment-atpt (&optional beg end) 
    "Relation of digit in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-comment-atpt (&optional beg end) 
    "Counts matches of graph in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'comment beg end (interactive-p)))

(defun ratio-graph-in-comment-atpt (&optional beg end) 
    "Relation of graph in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-comment-atpt (&optional beg end) 
    "Counts matches of lower in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'comment beg end (interactive-p)))

(defun ratio-lower-in-comment-atpt (&optional beg end) 
    "Relation of lower in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-comment-atpt (&optional beg end) 
    "Counts matches of punct in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'comment beg end (interactive-p)))

(defun ratio-punct-in-comment-atpt (&optional beg end) 
    "Relation of punct in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'comment beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-comment-atpt (&optional beg end) 
    "Counts matches of upper in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'comment beg end (interactive-p)))

(defun ratio-upper-in-comment-atpt (&optional beg end) 
    "Relation of upper in comment if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'comment beg end t (interactive-p)))



;;;###autoload
(defun csv-atpt (&optional arg) 
  "Returns csv at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'csv arg))

(defalias 'bounds-of-csv-atpt 'csv-bounds-atpt)
;;;###autoload
(defun csv-bounds-atpt (&optional arg) 
  "Returns borders of csv at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'csv arg))

;;;###autoload
(defun csv-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'csv arg))

;;;###autoload
(defun csv-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-end 'csv arg))

(defalias 'beginning-of-csv-atpt 'csv-beginning-atpt)
;;;###autoload
(defun csv-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'csv arg))

(defalias 'end-of-csv-atpt 'csv-end-atpt)
;;;###autoload
(defun csv-end-atpt (&optional arg) 
  "Goto end of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'csv arg))

(defalias 'length-of-csv-atpt 'csv-length-atpt)
;;;###autoload
(defun csv-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-length 'csv arg))

(defalias 'copy-csv-atpt 'csv-copy-atpt)
;;;###autoload
(defun csv-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class csv at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'csv arg))

(defalias 'escape-csv-atpt 'csv-escape-atpt)
;;;###autoload
(defun csv-escape-atpt () 
  "Makes a copy of symbol or char-class csv at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'csv))

(defalias 'doublequote-csv-atpt 'csv-doublequote-atpt)
;;;###autoload
(defun csv-doublequote-atpt (&optional arg) 
  "Doublequotes 'csv at point if any. "
  (interactive "*p")
    (th-doublequote 'csv arg))

(defalias 'dollar-csv-atpt 'csv-dollar-atpt)
;;;###autoload
(defun csv-dollar-atpt (&optional arg) 
  "Doublequotes 'csv at point if any. "
  (interactive "*p")
    (th-dollar 'csv arg))

(defalias 'angle-greater-csv-atpt 'csv-greater-angle-atpt)
;;;###autoload
(defun csv-greater-angle-atpt (&optional arg) 
  "Sets angles for 'csv after point if any. "
  (interactive "*p")
    (th-greater-angle 'csv arg))

(defalias 'angle-lower-csv-atpt 'csv-lower-angle-atpt)
;;;###autoload
(defun csv-lower-angle-atpt (&optional arg) 
  "Sets angles for 'csv after point if any. "
  (interactive "*p")
    (th-lower-angle 'csv arg))

(defalias 'backslash-csv-atpt 'csv-backslash-atpt)
;;;###autoload
(defun csv-backslash-atpt (&optional arg) 
  "Backslash 'csv at point if any. "
  (interactive "*p")
    (th-backslash 'csv arg))   

(defalias 'brace-csv-atpt 'csv-brace-atpt)
;;;###autoload
(defun csv-brace-atpt (&optional arg) 
  "Braces 'csv at point if any. "
  (interactive "*p")
    (th-brace 'csv arg))

(defalias 'bracket-csv-atpt 'csv-bracket-atpt)
;;;###autoload
(defun csv-bracket-atpt (&optional arg) 
  "Brackets 'csv after point if any. "
  (interactive "*p")
    (th-bracket 'csv arg))

;; (defalias 'comment-csv-atpt 'csv-comment-atpt)
;;;###autoload
(defun csv-comment-atpt (&optional arg) 
  "Comments 'csv at point if any. "
  (interactive "*p")
  (th-comment 'csv arg))

(defalias 'mark-csv-atpt 'csv-mark-atpt)
;;;###autoload
(defun csv-mark-atpt () 
  "Marks 'csv at point if any. "
  (interactive)
  (th-mark 'csv))

(defalias 'hide-csv-atpt 'csv-hide-atpt)
;;;###autoload
(defun csv-hide-atpt () 
  "Toggles hide 'csv in region or buffer if any. "
  (interactive)
  (th-hide 'csv))

(defalias 'highlight-csv-atpt-mode 'csv-highlight-atpt-mode)
(defun csv-highlight-atpt-mode (&optional arg) 
  "Toggles csv-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'csv arg))

(defalias 'kill-csv-atpt 'csv-kill-atpt)
;;;###autoload
(defun csv-kill-atpt (&optional arg) 
  "Kills csv at point if any. "
  (interactive "*P")
  (th-kill 'csv arg))

(defalias 'parentize-csv-atpt 'csv-parentize-atpt)
;;;###autoload
(defun csv-parentize-atpt (&optional arg) 
  "Parentizes 'csv at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'csv arg))

(defalias 'separate-csv-atpt 'csv-separate-atpt)
;;;###autoload
(defun csv-separate-atpt (&optional arg) 
  "Separates 'csv at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'csv arg))

(defalias 'singlequote-csv-atpt 'csv-singlequote-atpt)
;;;###autoload
(defun csv-singlequote-atpt (&optional arg) 
  "Singlequotes 'csv at point if any. "
  (interactive "*p")
    (th-singlequote 'csv arg))

(defalias 'slash-csv-atpt 'csv-slash-atpt)
;;;###autoload
(defun csv-slash-atpt (&optional arg) 
  "Slash 'csv at point if any. "
  (interactive "*p")
    (th-slash 'csv arg))

(defalias 'trim-csv-atpt 'csv-trim-atpt)
;;;###autoload
(defun csv-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'csv t t))

(defalias 'trim-csv-left-atpt 'csv-left-trim-atpt)
;;;###autoload
(defun csv-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'csv t nil))

(defalias 'trim-csv-right-atpt 'csv-right-trim-atpt)
;;;###autoload
(defun csv-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'csv nil t))

(defalias 'forward-csv-atpt 'csv-forward-atpt)
;;;###autoload
(defun csv-forward-atpt (&optional arg) 
  "Moves forward over csv at point if any, does nothing otherwise.
Returns bounds of csv as a list. "
  (interactive "p")
  (th-forward 'csv arg))

(defalias 'backward-csv-atpt 'csv-backward-atpt)
;;;###autoload
(defun csv-backward-atpt (&optional arg) 
  "Moves backward over csv before point if any, does nothing otherwise.
Returns bounds of csv as a list. "
  (interactive "p")
  (th-backward 'csv arg))

(defalias 'transpose-csv-atpt 'csv-transpose-atpt)
;;;###autoload
(defun csv-transpose-atpt (&optional arg) 
  "Transposes csv with csv before point if any. "
  (interactive "*P")
  (th-transpose 'csv arg))

(defalias 'sort-csv-atpt 'csv-sort-atpt)
;;;###autoload
(defun csv-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts csv's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'csv reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-csv-atpt 'csv-check-atpt)
;;;###autoload
(defun csv-check-atpt (&optional arg) 
  " csv at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "csv" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "csv" "-end-position-atpt")))))
    (th-check 'csv arg)))
   

;;;###autoload
(defun count-alnum-in-csv-atpt (&optional beg end) 
    "Counts matches of alnum in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'csv beg end (interactive-p)))

(defun ratio-alnum-in-csv-atpt (&optional beg end) 
    "Relation of alnum in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-csv-atpt (&optional beg end) 
    "Counts matches of alpha in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'csv beg end (interactive-p)))

(defun ratio-alpha-in-csv-atpt (&optional beg end) 
    "Relation of alpha in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-csv-atpt (&optional beg end) 
    "Counts matches of ascii in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'csv beg end (interactive-p)))

(defun ratio-ascii-in-csv-atpt (&optional beg end) 
    "Relation of ascii in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-csv-atpt (&optional beg end) 
    "Counts matches of digit in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'csv beg end (interactive-p)))

(defun ratio-digit-in-csv-atpt (&optional beg end) 
    "Relation of digit in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-csv-atpt (&optional beg end) 
    "Counts matches of graph in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'csv beg end (interactive-p)))

(defun ratio-graph-in-csv-atpt (&optional beg end) 
    "Relation of graph in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-csv-atpt (&optional beg end) 
    "Counts matches of lower in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'csv beg end (interactive-p)))

(defun ratio-lower-in-csv-atpt (&optional beg end) 
    "Relation of lower in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-csv-atpt (&optional beg end) 
    "Counts matches of punct in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'csv beg end (interactive-p)))

(defun ratio-punct-in-csv-atpt (&optional beg end) 
    "Relation of punct in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'csv beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-csv-atpt (&optional beg end) 
    "Counts matches of upper in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'csv beg end (interactive-p)))

(defun ratio-upper-in-csv-atpt (&optional beg end) 
    "Relation of upper in csv if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'csv beg end t (interactive-p)))



;;;###autoload
(defun date-atpt (&optional arg) 
  "Returns date at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'date arg))

(defalias 'bounds-of-date-atpt 'date-bounds-atpt)
;;;###autoload
(defun date-bounds-atpt (&optional arg) 
  "Returns borders of date at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'date arg))

;;;###autoload
(defun date-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'date arg))

;;;###autoload
(defun date-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-end 'date arg))

(defalias 'beginning-of-date-atpt 'date-beginning-atpt)
;;;###autoload
(defun date-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'date arg))

(defalias 'end-of-date-atpt 'date-end-atpt)
;;;###autoload
(defun date-end-atpt (&optional arg) 
  "Goto end of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'date arg))

(defalias 'length-of-date-atpt 'date-length-atpt)
;;;###autoload
(defun date-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-length 'date arg))

(defalias 'copy-date-atpt 'date-copy-atpt)
;;;###autoload
(defun date-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class date at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'date arg))

(defalias 'escape-date-atpt 'date-escape-atpt)
;;;###autoload
(defun date-escape-atpt () 
  "Makes a copy of symbol or char-class date at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'date))

(defalias 'doublequote-date-atpt 'date-doublequote-atpt)
;;;###autoload
(defun date-doublequote-atpt (&optional arg) 
  "Doublequotes 'date at point if any. "
  (interactive "*p")
    (th-doublequote 'date arg))

(defalias 'dollar-date-atpt 'date-dollar-atpt)
;;;###autoload
(defun date-dollar-atpt (&optional arg) 
  "Doublequotes 'date at point if any. "
  (interactive "*p")
    (th-dollar 'date arg))

(defalias 'angle-greater-date-atpt 'date-greater-angle-atpt)
;;;###autoload
(defun date-greater-angle-atpt (&optional arg) 
  "Sets angles for 'date after point if any. "
  (interactive "*p")
    (th-greater-angle 'date arg))

(defalias 'angle-lower-date-atpt 'date-lower-angle-atpt)
;;;###autoload
(defun date-lower-angle-atpt (&optional arg) 
  "Sets angles for 'date after point if any. "
  (interactive "*p")
    (th-lower-angle 'date arg))

(defalias 'backslash-date-atpt 'date-backslash-atpt)
;;;###autoload
(defun date-backslash-atpt (&optional arg) 
  "Backslash 'date at point if any. "
  (interactive "*p")
    (th-backslash 'date arg))   

(defalias 'brace-date-atpt 'date-brace-atpt)
;;;###autoload
(defun date-brace-atpt (&optional arg) 
  "Braces 'date at point if any. "
  (interactive "*p")
    (th-brace 'date arg))

(defalias 'bracket-date-atpt 'date-bracket-atpt)
;;;###autoload
(defun date-bracket-atpt (&optional arg) 
  "Brackets 'date after point if any. "
  (interactive "*p")
    (th-bracket 'date arg))

;; (defalias 'comment-date-atpt 'date-comment-atpt)
;;;###autoload
(defun date-comment-atpt (&optional arg) 
  "Comments 'date at point if any. "
  (interactive "*p")
  (th-comment 'date arg))

(defalias 'mark-date-atpt 'date-mark-atpt)
;;;###autoload
(defun date-mark-atpt () 
  "Marks 'date at point if any. "
  (interactive)
  (th-mark 'date))

(defalias 'hide-date-atpt 'date-hide-atpt)
;;;###autoload
(defun date-hide-atpt () 
  "Toggles hide 'date in region or buffer if any. "
  (interactive)
  (th-hide 'date))

(defalias 'highlight-date-atpt-mode 'date-highlight-atpt-mode)
(defun date-highlight-atpt-mode (&optional arg) 
  "Toggles date-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'date arg))

(defalias 'kill-date-atpt 'date-kill-atpt)
;;;###autoload
(defun date-kill-atpt (&optional arg) 
  "Kills date at point if any. "
  (interactive "*P")
  (th-kill 'date arg))

(defalias 'parentize-date-atpt 'date-parentize-atpt)
;;;###autoload
(defun date-parentize-atpt (&optional arg) 
  "Parentizes 'date at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'date arg))

(defalias 'separate-date-atpt 'date-separate-atpt)
;;;###autoload
(defun date-separate-atpt (&optional arg) 
  "Separates 'date at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'date arg))

(defalias 'singlequote-date-atpt 'date-singlequote-atpt)
;;;###autoload
(defun date-singlequote-atpt (&optional arg) 
  "Singlequotes 'date at point if any. "
  (interactive "*p")
    (th-singlequote 'date arg))

(defalias 'slash-date-atpt 'date-slash-atpt)
;;;###autoload
(defun date-slash-atpt (&optional arg) 
  "Slash 'date at point if any. "
  (interactive "*p")
    (th-slash 'date arg))

(defalias 'trim-date-atpt 'date-trim-atpt)
;;;###autoload
(defun date-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'date t t))

(defalias 'trim-date-left-atpt 'date-left-trim-atpt)
;;;###autoload
(defun date-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'date t nil))

(defalias 'trim-date-right-atpt 'date-right-trim-atpt)
;;;###autoload
(defun date-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'date nil t))

(defalias 'forward-date-atpt 'date-forward-atpt)
;;;###autoload
(defun date-forward-atpt (&optional arg) 
  "Moves forward over date at point if any, does nothing otherwise.
Returns bounds of date as a list. "
  (interactive "p")
  (th-forward 'date arg))

(defalias 'backward-date-atpt 'date-backward-atpt)
;;;###autoload
(defun date-backward-atpt (&optional arg) 
  "Moves backward over date before point if any, does nothing otherwise.
Returns bounds of date as a list. "
  (interactive "p")
  (th-backward 'date arg))

(defalias 'transpose-date-atpt 'date-transpose-atpt)
;;;###autoload
(defun date-transpose-atpt (&optional arg) 
  "Transposes date with date before point if any. "
  (interactive "*P")
  (th-transpose 'date arg))

(defalias 'sort-date-atpt 'date-sort-atpt)
;;;###autoload
(defun date-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts date's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'date reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-date-atpt 'date-check-atpt)
;;;###autoload
(defun date-check-atpt (&optional arg) 
  " date at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "date" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "date" "-end-position-atpt")))))
    (th-check 'date arg)))
   

;;;###autoload
(defun count-alnum-in-date-atpt (&optional beg end) 
    "Counts matches of alnum in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'date beg end (interactive-p)))

(defun ratio-alnum-in-date-atpt (&optional beg end) 
    "Relation of alnum in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'date beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-date-atpt (&optional beg end) 
    "Counts matches of alpha in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'date beg end (interactive-p)))

(defun ratio-alpha-in-date-atpt (&optional beg end) 
    "Relation of alpha in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'date beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-date-atpt (&optional beg end) 
    "Counts matches of ascii in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'date beg end (interactive-p)))

(defun ratio-ascii-in-date-atpt (&optional beg end) 
    "Relation of ascii in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'date beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-date-atpt (&optional beg end) 
    "Counts matches of digit in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'date beg end (interactive-p)))

(defun ratio-digit-in-date-atpt (&optional beg end) 
    "Relation of digit in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'date beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-date-atpt (&optional beg end) 
    "Counts matches of graph in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'date beg end (interactive-p)))

(defun ratio-graph-in-date-atpt (&optional beg end) 
    "Relation of graph in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'date beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-date-atpt (&optional beg end) 
    "Counts matches of lower in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'date beg end (interactive-p)))

(defun ratio-lower-in-date-atpt (&optional beg end) 
    "Relation of lower in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'date beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-date-atpt (&optional beg end) 
    "Counts matches of punct in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'date beg end (interactive-p)))

(defun ratio-punct-in-date-atpt (&optional beg end) 
    "Relation of punct in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'date beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-date-atpt (&optional beg end) 
    "Counts matches of upper in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'date beg end (interactive-p)))

(defun ratio-upper-in-date-atpt (&optional beg end) 
    "Relation of upper in date if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'date beg end t (interactive-p)))



;;;###autoload
(defun defun-atpt (&optional arg) 
  "Returns defun at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'defun arg))

(defalias 'bounds-of-defun-atpt 'defun-bounds-atpt)
;;;###autoload
(defun defun-bounds-atpt (&optional arg) 
  "Returns borders of defun at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'defun arg))

;;;###autoload
(defun defun-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'defun arg))

;;;###autoload
(defun defun-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-end 'defun arg))

(defalias 'beginning-of-defun-atpt 'defun-beginning-atpt)
;;;###autoload
(defun defun-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'defun arg))

(defalias 'end-of-defun-atpt 'defun-end-atpt)
;;;###autoload
(defun defun-end-atpt (&optional arg) 
  "Goto end of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'defun arg))

(defalias 'length-of-defun-atpt 'defun-length-atpt)
;;;###autoload
(defun defun-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-length 'defun arg))

(defalias 'copy-defun-atpt 'defun-copy-atpt)
;;;###autoload
(defun defun-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class defun at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'defun arg))

(defalias 'escape-defun-atpt 'defun-escape-atpt)
;;;###autoload
(defun defun-escape-atpt () 
  "Makes a copy of symbol or char-class defun at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'defun))

(defalias 'doublequote-defun-atpt 'defun-doublequote-atpt)
;;;###autoload
(defun defun-doublequote-atpt (&optional arg) 
  "Doublequotes 'defun at point if any. "
  (interactive "*p")
    (th-doublequote 'defun arg))

(defalias 'dollar-defun-atpt 'defun-dollar-atpt)
;;;###autoload
(defun defun-dollar-atpt (&optional arg) 
  "Doublequotes 'defun at point if any. "
  (interactive "*p")
    (th-dollar 'defun arg))

(defalias 'angle-greater-defun-atpt 'defun-greater-angle-atpt)
;;;###autoload
(defun defun-greater-angle-atpt (&optional arg) 
  "Sets angles for 'defun after point if any. "
  (interactive "*p")
    (th-greater-angle 'defun arg))

(defalias 'angle-lower-defun-atpt 'defun-lower-angle-atpt)
;;;###autoload
(defun defun-lower-angle-atpt (&optional arg) 
  "Sets angles for 'defun after point if any. "
  (interactive "*p")
    (th-lower-angle 'defun arg))

(defalias 'backslash-defun-atpt 'defun-backslash-atpt)
;;;###autoload
(defun defun-backslash-atpt (&optional arg) 
  "Backslash 'defun at point if any. "
  (interactive "*p")
    (th-backslash 'defun arg))   

(defalias 'brace-defun-atpt 'defun-brace-atpt)
;;;###autoload
(defun defun-brace-atpt (&optional arg) 
  "Braces 'defun at point if any. "
  (interactive "*p")
    (th-brace 'defun arg))

(defalias 'bracket-defun-atpt 'defun-bracket-atpt)
;;;###autoload
(defun defun-bracket-atpt (&optional arg) 
  "Brackets 'defun after point if any. "
  (interactive "*p")
    (th-bracket 'defun arg))

;; (defalias 'comment-defun-atpt 'defun-comment-atpt)
;;;###autoload
(defun defun-comment-atpt (&optional arg) 
  "Comments 'defun at point if any. "
  (interactive "*p")
  (th-comment 'defun arg))

(defalias 'mark-defun-atpt 'defun-mark-atpt)
;;;###autoload
(defun defun-mark-atpt () 
  "Marks 'defun at point if any. "
  (interactive)
  (th-mark 'defun))

(defalias 'hide-defun-atpt 'defun-hide-atpt)
;;;###autoload
(defun defun-hide-atpt () 
  "Toggles hide 'defun in region or buffer if any. "
  (interactive)
  (th-hide 'defun))

(defalias 'highlight-defun-atpt-mode 'defun-highlight-atpt-mode)
(defun defun-highlight-atpt-mode (&optional arg) 
  "Toggles defun-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'defun arg))

(defalias 'kill-defun-atpt 'defun-kill-atpt)
;;;###autoload
(defun defun-kill-atpt (&optional arg) 
  "Kills defun at point if any. "
  (interactive "*P")
  (th-kill 'defun arg))

(defalias 'parentize-defun-atpt 'defun-parentize-atpt)
;;;###autoload
(defun defun-parentize-atpt (&optional arg) 
  "Parentizes 'defun at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'defun arg))

(defalias 'separate-defun-atpt 'defun-separate-atpt)
;;;###autoload
(defun defun-separate-atpt (&optional arg) 
  "Separates 'defun at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'defun arg))

(defalias 'singlequote-defun-atpt 'defun-singlequote-atpt)
;;;###autoload
(defun defun-singlequote-atpt (&optional arg) 
  "Singlequotes 'defun at point if any. "
  (interactive "*p")
    (th-singlequote 'defun arg))

(defalias 'slash-defun-atpt 'defun-slash-atpt)
;;;###autoload
(defun defun-slash-atpt (&optional arg) 
  "Slash 'defun at point if any. "
  (interactive "*p")
    (th-slash 'defun arg))

(defalias 'trim-defun-atpt 'defun-trim-atpt)
;;;###autoload
(defun defun-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'defun t t))

(defalias 'trim-defun-left-atpt 'defun-left-trim-atpt)
;;;###autoload
(defun defun-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'defun t nil))

(defalias 'trim-defun-right-atpt 'defun-right-trim-atpt)
;;;###autoload
(defun defun-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'defun nil t))

(defalias 'forward-defun-atpt 'defun-forward-atpt)
;;;###autoload
(defun defun-forward-atpt (&optional arg) 
  "Moves forward over defun at point if any, does nothing otherwise.
Returns bounds of defun as a list. "
  (interactive "p")
  (th-forward 'defun arg))

(defalias 'backward-defun-atpt 'defun-backward-atpt)
;;;###autoload
(defun defun-backward-atpt (&optional arg) 
  "Moves backward over defun before point if any, does nothing otherwise.
Returns bounds of defun as a list. "
  (interactive "p")
  (th-backward 'defun arg))

(defalias 'transpose-defun-atpt 'defun-transpose-atpt)
;;;###autoload
(defun defun-transpose-atpt (&optional arg) 
  "Transposes defun with defun before point if any. "
  (interactive "*P")
  (th-transpose 'defun arg))

(defalias 'sort-defun-atpt 'defun-sort-atpt)
;;;###autoload
(defun defun-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts defun's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'defun reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-defun-atpt 'defun-check-atpt)
;;;###autoload
(defun defun-check-atpt (&optional arg) 
  " defun at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "defun" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "defun" "-end-position-atpt")))))
    (th-check 'defun arg)))
   

;;;###autoload
(defun count-alnum-in-defun-atpt (&optional beg end) 
    "Counts matches of alnum in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'defun beg end (interactive-p)))

(defun ratio-alnum-in-defun-atpt (&optional beg end) 
    "Relation of alnum in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-defun-atpt (&optional beg end) 
    "Counts matches of alpha in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'defun beg end (interactive-p)))

(defun ratio-alpha-in-defun-atpt (&optional beg end) 
    "Relation of alpha in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-defun-atpt (&optional beg end) 
    "Counts matches of ascii in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'defun beg end (interactive-p)))

(defun ratio-ascii-in-defun-atpt (&optional beg end) 
    "Relation of ascii in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-defun-atpt (&optional beg end) 
    "Counts matches of digit in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'defun beg end (interactive-p)))

(defun ratio-digit-in-defun-atpt (&optional beg end) 
    "Relation of digit in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-defun-atpt (&optional beg end) 
    "Counts matches of graph in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'defun beg end (interactive-p)))

(defun ratio-graph-in-defun-atpt (&optional beg end) 
    "Relation of graph in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-defun-atpt (&optional beg end) 
    "Counts matches of lower in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'defun beg end (interactive-p)))

(defun ratio-lower-in-defun-atpt (&optional beg end) 
    "Relation of lower in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-defun-atpt (&optional beg end) 
    "Counts matches of punct in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'defun beg end (interactive-p)))

(defun ratio-punct-in-defun-atpt (&optional beg end) 
    "Relation of punct in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'defun beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-defun-atpt (&optional beg end) 
    "Counts matches of upper in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'defun beg end (interactive-p)))

(defun ratio-upper-in-defun-atpt (&optional beg end) 
    "Relation of upper in defun if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'defun beg end t (interactive-p)))



;;;###autoload
(defun delimited-atpt (&optional arg) 
  "Returns delimited at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'delimited arg))

(defalias 'bounds-of-delimited-atpt 'delimited-bounds-atpt)
;;;###autoload
(defun delimited-bounds-atpt (&optional arg) 
  "Returns borders of delimited at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'delimited arg))

;;;###autoload
(defun delimited-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'delimited arg))

;;;###autoload
(defun delimited-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-end 'delimited arg))

(defalias 'beginning-of-delimited-atpt 'delimited-beginning-atpt)
;;;###autoload
(defun delimited-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'delimited arg))

(defalias 'end-of-delimited-atpt 'delimited-end-atpt)
;;;###autoload
(defun delimited-end-atpt (&optional arg) 
  "Goto end of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'delimited arg))

(defalias 'length-of-delimited-atpt 'delimited-length-atpt)
;;;###autoload
(defun delimited-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-length 'delimited arg))

(defalias 'copy-delimited-atpt 'delimited-copy-atpt)
;;;###autoload
(defun delimited-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'delimited arg))

(defalias 'escape-delimited-atpt 'delimited-escape-atpt)
;;;###autoload
(defun delimited-escape-atpt () 
  "Makes a copy of symbol or char-class delimited at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'delimited))

(defalias 'doublequote-delimited-atpt 'delimited-doublequote-atpt)
;;;###autoload
(defun delimited-doublequote-atpt (&optional arg) 
  "Doublequotes 'delimited at point if any. "
  (interactive "*p")
    (th-doublequote 'delimited arg))

(defalias 'dollar-delimited-atpt 'delimited-dollar-atpt)
;;;###autoload
(defun delimited-dollar-atpt (&optional arg) 
  "Doublequotes 'delimited at point if any. "
  (interactive "*p")
    (th-dollar 'delimited arg))

(defalias 'angle-greater-delimited-atpt 'delimited-greater-angle-atpt)
;;;###autoload
(defun delimited-greater-angle-atpt (&optional arg) 
  "Sets angles for 'delimited after point if any. "
  (interactive "*p")
    (th-greater-angle 'delimited arg))

(defalias 'angle-lower-delimited-atpt 'delimited-lower-angle-atpt)
;;;###autoload
(defun delimited-lower-angle-atpt (&optional arg) 
  "Sets angles for 'delimited after point if any. "
  (interactive "*p")
    (th-lower-angle 'delimited arg))

(defalias 'backslash-delimited-atpt 'delimited-backslash-atpt)
;;;###autoload
(defun delimited-backslash-atpt (&optional arg) 
  "Backslash 'delimited at point if any. "
  (interactive "*p")
    (th-backslash 'delimited arg))   

(defalias 'brace-delimited-atpt 'delimited-brace-atpt)
;;;###autoload
(defun delimited-brace-atpt (&optional arg) 
  "Braces 'delimited at point if any. "
  (interactive "*p")
    (th-brace 'delimited arg))

(defalias 'bracket-delimited-atpt 'delimited-bracket-atpt)
;;;###autoload
(defun delimited-bracket-atpt (&optional arg) 
  "Brackets 'delimited after point if any. "
  (interactive "*p")
    (th-bracket 'delimited arg))

;; (defalias 'comment-delimited-atpt 'delimited-comment-atpt)
;;;###autoload
(defun delimited-comment-atpt (&optional arg) 
  "Comments 'delimited at point if any. "
  (interactive "*p")
  (th-comment 'delimited arg))

(defalias 'mark-delimited-atpt 'delimited-mark-atpt)
;;;###autoload
(defun delimited-mark-atpt () 
  "Marks 'delimited at point if any. "
  (interactive)
  (th-mark 'delimited))

(defalias 'hide-delimited-atpt 'delimited-hide-atpt)
;;;###autoload
(defun delimited-hide-atpt () 
  "Toggles hide 'delimited in region or buffer if any. "
  (interactive)
  (th-hide 'delimited))

(defalias 'highlight-delimited-atpt-mode 'delimited-highlight-atpt-mode)
(defun delimited-highlight-atpt-mode (&optional arg) 
  "Toggles delimited-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'delimited arg))

(defalias 'kill-delimited-atpt 'delimited-kill-atpt)
;;;###autoload
(defun delimited-kill-atpt (&optional arg) 
  "Kills delimited at point if any. "
  (interactive "*P")
  (th-kill 'delimited arg))

(defalias 'parentize-delimited-atpt 'delimited-parentize-atpt)
;;;###autoload
(defun delimited-parentize-atpt (&optional arg) 
  "Parentizes 'delimited at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'delimited arg))

(defalias 'separate-delimited-atpt 'delimited-separate-atpt)
;;;###autoload
(defun delimited-separate-atpt (&optional arg) 
  "Separates 'delimited at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'delimited arg))

(defalias 'singlequote-delimited-atpt 'delimited-singlequote-atpt)
;;;###autoload
(defun delimited-singlequote-atpt (&optional arg) 
  "Singlequotes 'delimited at point if any. "
  (interactive "*p")
    (th-singlequote 'delimited arg))

(defalias 'slash-delimited-atpt 'delimited-slash-atpt)
;;;###autoload
(defun delimited-slash-atpt (&optional arg) 
  "Slash 'delimited at point if any. "
  (interactive "*p")
    (th-slash 'delimited arg))

(defalias 'trim-delimited-atpt 'delimited-trim-atpt)
;;;###autoload
(defun delimited-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'delimited t t))

(defalias 'trim-delimited-left-atpt 'delimited-left-trim-atpt)
;;;###autoload
(defun delimited-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'delimited t nil))

(defalias 'trim-delimited-right-atpt 'delimited-right-trim-atpt)
;;;###autoload
(defun delimited-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'delimited nil t))

(defalias 'forward-delimited-atpt 'delimited-forward-atpt)
;;;###autoload
(defun delimited-forward-atpt (&optional arg) 
  "Moves forward over delimited at point if any, does nothing otherwise.
Returns bounds of delimited as a list. "
  (interactive "p")
  (th-forward 'delimited arg))

(defalias 'backward-delimited-atpt 'delimited-backward-atpt)
;;;###autoload
(defun delimited-backward-atpt (&optional arg) 
  "Moves backward over delimited before point if any, does nothing otherwise.
Returns bounds of delimited as a list. "
  (interactive "p")
  (th-backward 'delimited arg))

(defalias 'transpose-delimited-atpt 'delimited-transpose-atpt)
;;;###autoload
(defun delimited-transpose-atpt (&optional arg) 
  "Transposes delimited with delimited before point if any. "
  (interactive "*P")
  (th-transpose 'delimited arg))

(defalias 'sort-delimited-atpt 'delimited-sort-atpt)
;;;###autoload
(defun delimited-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts delimited's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'delimited reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-delimited-atpt 'delimited-check-atpt)
;;;###autoload
(defun delimited-check-atpt (&optional arg) 
  " delimited at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "delimited" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "delimited" "-end-position-atpt")))))
    (th-check 'delimited arg)))
   

;;;###autoload
(defun count-alnum-in-delimited-atpt (&optional beg end) 
    "Counts matches of alnum in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'delimited beg end (interactive-p)))

(defun ratio-alnum-in-delimited-atpt (&optional beg end) 
    "Relation of alnum in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-delimited-atpt (&optional beg end) 
    "Counts matches of alpha in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'delimited beg end (interactive-p)))

(defun ratio-alpha-in-delimited-atpt (&optional beg end) 
    "Relation of alpha in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-delimited-atpt (&optional beg end) 
    "Counts matches of ascii in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'delimited beg end (interactive-p)))

(defun ratio-ascii-in-delimited-atpt (&optional beg end) 
    "Relation of ascii in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-delimited-atpt (&optional beg end) 
    "Counts matches of digit in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'delimited beg end (interactive-p)))

(defun ratio-digit-in-delimited-atpt (&optional beg end) 
    "Relation of digit in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-delimited-atpt (&optional beg end) 
    "Counts matches of graph in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'delimited beg end (interactive-p)))

(defun ratio-graph-in-delimited-atpt (&optional beg end) 
    "Relation of graph in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-delimited-atpt (&optional beg end) 
    "Counts matches of lower in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'delimited beg end (interactive-p)))

(defun ratio-lower-in-delimited-atpt (&optional beg end) 
    "Relation of lower in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-delimited-atpt (&optional beg end) 
    "Counts matches of punct in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'delimited beg end (interactive-p)))

(defun ratio-punct-in-delimited-atpt (&optional beg end) 
    "Relation of punct in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'delimited beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-delimited-atpt (&optional beg end) 
    "Counts matches of upper in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'delimited beg end (interactive-p)))

(defun ratio-upper-in-delimited-atpt (&optional beg end) 
    "Relation of upper in delimited if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'delimited beg end t (interactive-p)))



;;;###autoload
(defun digit-atpt (&optional arg) 
  "Returns digit at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'digit arg))

(defalias 'bounds-of-digit-atpt 'digit-bounds-atpt)
;;;###autoload
(defun digit-bounds-atpt (&optional arg) 
  "Returns borders of digit at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'digit arg))

;;;###autoload
(defun digit-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'digit arg))

;;;###autoload
(defun digit-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-end 'digit arg))

(defalias 'beginning-of-digit-atpt 'digit-beginning-atpt)
;;;###autoload
(defun digit-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'digit arg))

(defalias 'end-of-digit-atpt 'digit-end-atpt)
;;;###autoload
(defun digit-end-atpt (&optional arg) 
  "Goto end of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'digit arg))

(defalias 'length-of-digit-atpt 'digit-length-atpt)
;;;###autoload
(defun digit-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-length 'digit arg))

(defalias 'copy-digit-atpt 'digit-copy-atpt)
;;;###autoload
(defun digit-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class digit at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'digit arg))

(defalias 'escape-digit-atpt 'digit-escape-atpt)
;;;###autoload
(defun digit-escape-atpt () 
  "Makes a copy of symbol or char-class digit at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'digit))

(defalias 'doublequote-digit-atpt 'digit-doublequote-atpt)
;;;###autoload
(defun digit-doublequote-atpt (&optional arg) 
  "Doublequotes 'digit at point if any. "
  (interactive "*p")
    (th-doublequote 'digit arg))

(defalias 'dollar-digit-atpt 'digit-dollar-atpt)
;;;###autoload
(defun digit-dollar-atpt (&optional arg) 
  "Doublequotes 'digit at point if any. "
  (interactive "*p")
    (th-dollar 'digit arg))

(defalias 'angle-greater-digit-atpt 'digit-greater-angle-atpt)
;;;###autoload
(defun digit-greater-angle-atpt (&optional arg) 
  "Sets angles for 'digit after point if any. "
  (interactive "*p")
    (th-greater-angle 'digit arg))

(defalias 'angle-lower-digit-atpt 'digit-lower-angle-atpt)
;;;###autoload
(defun digit-lower-angle-atpt (&optional arg) 
  "Sets angles for 'digit after point if any. "
  (interactive "*p")
    (th-lower-angle 'digit arg))

(defalias 'backslash-digit-atpt 'digit-backslash-atpt)
;;;###autoload
(defun digit-backslash-atpt (&optional arg) 
  "Backslash 'digit at point if any. "
  (interactive "*p")
    (th-backslash 'digit arg))   

(defalias 'brace-digit-atpt 'digit-brace-atpt)
;;;###autoload
(defun digit-brace-atpt (&optional arg) 
  "Braces 'digit at point if any. "
  (interactive "*p")
    (th-brace 'digit arg))

(defalias 'bracket-digit-atpt 'digit-bracket-atpt)
;;;###autoload
(defun digit-bracket-atpt (&optional arg) 
  "Brackets 'digit after point if any. "
  (interactive "*p")
    (th-bracket 'digit arg))

;; (defalias 'comment-digit-atpt 'digit-comment-atpt)
;;;###autoload
(defun digit-comment-atpt (&optional arg) 
  "Comments 'digit at point if any. "
  (interactive "*p")
  (th-comment 'digit arg))

(defalias 'mark-digit-atpt 'digit-mark-atpt)
;;;###autoload
(defun digit-mark-atpt () 
  "Marks 'digit at point if any. "
  (interactive)
  (th-mark 'digit))

(defalias 'hide-digit-atpt 'digit-hide-atpt)
;;;###autoload
(defun digit-hide-atpt () 
  "Toggles hide 'digit in region or buffer if any. "
  (interactive)
  (th-hide 'digit))

(defalias 'highlight-digit-atpt-mode 'digit-highlight-atpt-mode)
(defun digit-highlight-atpt-mode (&optional arg) 
  "Toggles digit-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'digit arg))

(defalias 'kill-digit-atpt 'digit-kill-atpt)
;;;###autoload
(defun digit-kill-atpt (&optional arg) 
  "Kills digit at point if any. "
  (interactive "*P")
  (th-kill 'digit arg))

(defalias 'parentize-digit-atpt 'digit-parentize-atpt)
;;;###autoload
(defun digit-parentize-atpt (&optional arg) 
  "Parentizes 'digit at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'digit arg))

(defalias 'separate-digit-atpt 'digit-separate-atpt)
;;;###autoload
(defun digit-separate-atpt (&optional arg) 
  "Separates 'digit at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'digit arg))

(defalias 'singlequote-digit-atpt 'digit-singlequote-atpt)
;;;###autoload
(defun digit-singlequote-atpt (&optional arg) 
  "Singlequotes 'digit at point if any. "
  (interactive "*p")
    (th-singlequote 'digit arg))

(defalias 'slash-digit-atpt 'digit-slash-atpt)
;;;###autoload
(defun digit-slash-atpt (&optional arg) 
  "Slash 'digit at point if any. "
  (interactive "*p")
    (th-slash 'digit arg))

(defalias 'trim-digit-atpt 'digit-trim-atpt)
;;;###autoload
(defun digit-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'digit t t))

(defalias 'trim-digit-left-atpt 'digit-left-trim-atpt)
;;;###autoload
(defun digit-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'digit t nil))

(defalias 'trim-digit-right-atpt 'digit-right-trim-atpt)
;;;###autoload
(defun digit-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'digit nil t))

(defalias 'forward-digit-atpt 'digit-forward-atpt)
;;;###autoload
(defun digit-forward-atpt (&optional arg) 
  "Moves forward over digit at point if any, does nothing otherwise.
Returns bounds of digit as a list. "
  (interactive "p")
  (th-forward 'digit arg))

(defalias 'backward-digit-atpt 'digit-backward-atpt)
;;;###autoload
(defun digit-backward-atpt (&optional arg) 
  "Moves backward over digit before point if any, does nothing otherwise.
Returns bounds of digit as a list. "
  (interactive "p")
  (th-backward 'digit arg))

(defalias 'transpose-digit-atpt 'digit-transpose-atpt)
;;;###autoload
(defun digit-transpose-atpt (&optional arg) 
  "Transposes digit with digit before point if any. "
  (interactive "*P")
  (th-transpose 'digit arg))

(defalias 'sort-digit-atpt 'digit-sort-atpt)
;;;###autoload
(defun digit-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts digit's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'digit reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-digit-atpt 'digit-check-atpt)
;;;###autoload
(defun digit-check-atpt (&optional arg) 
  " digit at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "digit" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "digit" "-end-position-atpt")))))
    (th-check 'digit arg)))
   

;;;###autoload
(defun dollared-atpt (&optional arg) 
  "Returns dollared at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'dollared arg))

(defalias 'bounds-of-dollared-atpt 'dollared-bounds-atpt)
;;;###autoload
(defun dollared-bounds-atpt (&optional arg) 
  "Returns borders of dollared at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'dollared arg))

;;;###autoload
(defun dollared-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'dollared arg))

;;;###autoload
(defun dollared-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-end 'dollared arg))

(defalias 'beginning-of-dollared-atpt 'dollared-beginning-atpt)
;;;###autoload
(defun dollared-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'dollared arg))

(defalias 'end-of-dollared-atpt 'dollared-end-atpt)
;;;###autoload
(defun dollared-end-atpt (&optional arg) 
  "Goto end of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'dollared arg))

(defalias 'length-of-dollared-atpt 'dollared-length-atpt)
;;;###autoload
(defun dollared-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-length 'dollared arg))

(defalias 'copy-dollared-atpt 'dollared-copy-atpt)
;;;###autoload
(defun dollared-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'dollared arg))

(defalias 'escape-dollared-atpt 'dollared-escape-atpt)
;;;###autoload
(defun dollared-escape-atpt () 
  "Makes a copy of symbol or char-class dollared at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'dollared))

(defalias 'doublequote-dollared-atpt 'dollared-doublequote-atpt)
;;;###autoload
(defun dollared-doublequote-atpt (&optional arg) 
  "Doublequotes 'dollared at point if any. "
  (interactive "*p")
    (th-doublequote 'dollared arg))

(defalias 'dollar-dollared-atpt 'dollared-dollar-atpt)
;;;###autoload
(defun dollared-dollar-atpt (&optional arg) 
  "Doublequotes 'dollared at point if any. "
  (interactive "*p")
    (th-dollar 'dollared arg))

(defalias 'angle-greater-dollared-atpt 'dollared-greater-angle-atpt)
;;;###autoload
(defun dollared-greater-angle-atpt (&optional arg) 
  "Sets angles for 'dollared after point if any. "
  (interactive "*p")
    (th-greater-angle 'dollared arg))

(defalias 'angle-lower-dollared-atpt 'dollared-lower-angle-atpt)
;;;###autoload
(defun dollared-lower-angle-atpt (&optional arg) 
  "Sets angles for 'dollared after point if any. "
  (interactive "*p")
    (th-lower-angle 'dollared arg))

(defalias 'backslash-dollared-atpt 'dollared-backslash-atpt)
;;;###autoload
(defun dollared-backslash-atpt (&optional arg) 
  "Backslash 'dollared at point if any. "
  (interactive "*p")
    (th-backslash 'dollared arg))   

(defalias 'brace-dollared-atpt 'dollared-brace-atpt)
;;;###autoload
(defun dollared-brace-atpt (&optional arg) 
  "Braces 'dollared at point if any. "
  (interactive "*p")
    (th-brace 'dollared arg))

(defalias 'bracket-dollared-atpt 'dollared-bracket-atpt)
;;;###autoload
(defun dollared-bracket-atpt (&optional arg) 
  "Brackets 'dollared after point if any. "
  (interactive "*p")
    (th-bracket 'dollared arg))

;; (defalias 'comment-dollared-atpt 'dollared-comment-atpt)
;;;###autoload
(defun dollared-comment-atpt (&optional arg) 
  "Comments 'dollared at point if any. "
  (interactive "*p")
  (th-comment 'dollared arg))

(defalias 'mark-dollared-atpt 'dollared-mark-atpt)
;;;###autoload
(defun dollared-mark-atpt () 
  "Marks 'dollared at point if any. "
  (interactive)
  (th-mark 'dollared))

(defalias 'hide-dollared-atpt 'dollared-hide-atpt)
;;;###autoload
(defun dollared-hide-atpt () 
  "Toggles hide 'dollared in region or buffer if any. "
  (interactive)
  (th-hide 'dollared))

(defalias 'highlight-dollared-atpt-mode 'dollared-highlight-atpt-mode)
(defun dollared-highlight-atpt-mode (&optional arg) 
  "Toggles dollared-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'dollared arg))

(defalias 'kill-dollared-atpt 'dollared-kill-atpt)
;;;###autoload
(defun dollared-kill-atpt (&optional arg) 
  "Kills dollared at point if any. "
  (interactive "*P")
  (th-kill 'dollared arg))

(defalias 'parentize-dollared-atpt 'dollared-parentize-atpt)
;;;###autoload
(defun dollared-parentize-atpt (&optional arg) 
  "Parentizes 'dollared at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'dollared arg))

(defalias 'separate-dollared-atpt 'dollared-separate-atpt)
;;;###autoload
(defun dollared-separate-atpt (&optional arg) 
  "Separates 'dollared at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'dollared arg))

(defalias 'singlequote-dollared-atpt 'dollared-singlequote-atpt)
;;;###autoload
(defun dollared-singlequote-atpt (&optional arg) 
  "Singlequotes 'dollared at point if any. "
  (interactive "*p")
    (th-singlequote 'dollared arg))

(defalias 'slash-dollared-atpt 'dollared-slash-atpt)
;;;###autoload
(defun dollared-slash-atpt (&optional arg) 
  "Slash 'dollared at point if any. "
  (interactive "*p")
    (th-slash 'dollared arg))

(defalias 'trim-dollared-atpt 'dollared-trim-atpt)
;;;###autoload
(defun dollared-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'dollared t t))

(defalias 'trim-dollared-left-atpt 'dollared-left-trim-atpt)
;;;###autoload
(defun dollared-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'dollared t nil))

(defalias 'trim-dollared-right-atpt 'dollared-right-trim-atpt)
;;;###autoload
(defun dollared-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'dollared nil t))

(defalias 'forward-dollared-atpt 'dollared-forward-atpt)
;;;###autoload
(defun dollared-forward-atpt (&optional arg) 
  "Moves forward over dollared at point if any, does nothing otherwise.
Returns bounds of dollared as a list. "
  (interactive "p")
  (th-forward 'dollared arg))

(defalias 'backward-dollared-atpt 'dollared-backward-atpt)
;;;###autoload
(defun dollared-backward-atpt (&optional arg) 
  "Moves backward over dollared before point if any, does nothing otherwise.
Returns bounds of dollared as a list. "
  (interactive "p")
  (th-backward 'dollared arg))

(defalias 'transpose-dollared-atpt 'dollared-transpose-atpt)
;;;###autoload
(defun dollared-transpose-atpt (&optional arg) 
  "Transposes dollared with dollared before point if any. "
  (interactive "*P")
  (th-transpose 'dollared arg))

(defalias 'sort-dollared-atpt 'dollared-sort-atpt)
;;;###autoload
(defun dollared-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts dollared's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'dollared reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-dollared-atpt 'dollared-check-atpt)
;;;###autoload
(defun dollared-check-atpt (&optional arg) 
  " dollared at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "dollared" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "dollared" "-end-position-atpt")))))
    (th-check 'dollared arg)))
   

;;;###autoload
(defun count-alnum-in-dollared-atpt (&optional beg end) 
    "Counts matches of alnum in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'dollared beg end (interactive-p)))

(defun ratio-alnum-in-dollared-atpt (&optional beg end) 
    "Relation of alnum in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-dollared-atpt (&optional beg end) 
    "Counts matches of alpha in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'dollared beg end (interactive-p)))

(defun ratio-alpha-in-dollared-atpt (&optional beg end) 
    "Relation of alpha in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-dollared-atpt (&optional beg end) 
    "Counts matches of ascii in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'dollared beg end (interactive-p)))

(defun ratio-ascii-in-dollared-atpt (&optional beg end) 
    "Relation of ascii in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-dollared-atpt (&optional beg end) 
    "Counts matches of digit in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'dollared beg end (interactive-p)))

(defun ratio-digit-in-dollared-atpt (&optional beg end) 
    "Relation of digit in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-dollared-atpt (&optional beg end) 
    "Counts matches of graph in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'dollared beg end (interactive-p)))

(defun ratio-graph-in-dollared-atpt (&optional beg end) 
    "Relation of graph in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-dollared-atpt (&optional beg end) 
    "Counts matches of lower in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'dollared beg end (interactive-p)))

(defun ratio-lower-in-dollared-atpt (&optional beg end) 
    "Relation of lower in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-dollared-atpt (&optional beg end) 
    "Counts matches of punct in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'dollared beg end (interactive-p)))

(defun ratio-punct-in-dollared-atpt (&optional beg end) 
    "Relation of punct in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'dollared beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-dollared-atpt (&optional beg end) 
    "Counts matches of upper in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'dollared beg end (interactive-p)))

(defun ratio-upper-in-dollared-atpt (&optional beg end) 
    "Relation of upper in dollared if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'dollared beg end t (interactive-p)))



;;;###autoload
(defun doublequoted-atpt (&optional arg) 
  "Returns doublequoted at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'doublequoted arg))

(defalias 'bounds-of-doublequoted-atpt 'doublequoted-bounds-atpt)
;;;###autoload
(defun doublequoted-bounds-atpt (&optional arg) 
  "Returns borders of doublequoted at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'doublequoted arg))

;;;###autoload
(defun doublequoted-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'doublequoted arg))

;;;###autoload
(defun doublequoted-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-end 'doublequoted arg))

(defalias 'beginning-of-doublequoted-atpt 'doublequoted-beginning-atpt)
;;;###autoload
(defun doublequoted-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'doublequoted arg))

(defalias 'end-of-doublequoted-atpt 'doublequoted-end-atpt)
;;;###autoload
(defun doublequoted-end-atpt (&optional arg) 
  "Goto end of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'doublequoted arg))

(defalias 'length-of-doublequoted-atpt 'doublequoted-length-atpt)
;;;###autoload
(defun doublequoted-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-length 'doublequoted arg))

(defalias 'copy-doublequoted-atpt 'doublequoted-copy-atpt)
;;;###autoload
(defun doublequoted-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'doublequoted arg))

(defalias 'escape-doublequoted-atpt 'doublequoted-escape-atpt)
;;;###autoload
(defun doublequoted-escape-atpt () 
  "Makes a copy of symbol or char-class doublequoted at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'doublequoted))

(defalias 'doublequote-doublequoted-atpt 'doublequoted-doublequote-atpt)
;;;###autoload
(defun doublequoted-doublequote-atpt (&optional arg) 
  "Doublequotes 'doublequoted at point if any. "
  (interactive "*p")
    (th-doublequote 'doublequoted arg))

(defalias 'dollar-doublequoted-atpt 'doublequoted-dollar-atpt)
;;;###autoload
(defun doublequoted-dollar-atpt (&optional arg) 
  "Doublequotes 'doublequoted at point if any. "
  (interactive "*p")
    (th-dollar 'doublequoted arg))

(defalias 'angle-greater-doublequoted-atpt 'doublequoted-greater-angle-atpt)
;;;###autoload
(defun doublequoted-greater-angle-atpt (&optional arg) 
  "Sets angles for 'doublequoted after point if any. "
  (interactive "*p")
    (th-greater-angle 'doublequoted arg))

(defalias 'angle-lower-doublequoted-atpt 'doublequoted-lower-angle-atpt)
;;;###autoload
(defun doublequoted-lower-angle-atpt (&optional arg) 
  "Sets angles for 'doublequoted after point if any. "
  (interactive "*p")
    (th-lower-angle 'doublequoted arg))

(defalias 'backslash-doublequoted-atpt 'doublequoted-backslash-atpt)
;;;###autoload
(defun doublequoted-backslash-atpt (&optional arg) 
  "Backslash 'doublequoted at point if any. "
  (interactive "*p")
    (th-backslash 'doublequoted arg))   

(defalias 'brace-doublequoted-atpt 'doublequoted-brace-atpt)
;;;###autoload
(defun doublequoted-brace-atpt (&optional arg) 
  "Braces 'doublequoted at point if any. "
  (interactive "*p")
    (th-brace 'doublequoted arg))

(defalias 'bracket-doublequoted-atpt 'doublequoted-bracket-atpt)
;;;###autoload
(defun doublequoted-bracket-atpt (&optional arg) 
  "Brackets 'doublequoted after point if any. "
  (interactive "*p")
    (th-bracket 'doublequoted arg))

;; (defalias 'comment-doublequoted-atpt 'doublequoted-comment-atpt)
;;;###autoload
(defun doublequoted-comment-atpt (&optional arg) 
  "Comments 'doublequoted at point if any. "
  (interactive "*p")
  (th-comment 'doublequoted arg))

(defalias 'mark-doublequoted-atpt 'doublequoted-mark-atpt)
;;;###autoload
(defun doublequoted-mark-atpt () 
  "Marks 'doublequoted at point if any. "
  (interactive)
  (th-mark 'doublequoted))

(defalias 'hide-doublequoted-atpt 'doublequoted-hide-atpt)
;;;###autoload
(defun doublequoted-hide-atpt () 
  "Toggles hide 'doublequoted in region or buffer if any. "
  (interactive)
  (th-hide 'doublequoted))

(defalias 'highlight-doublequoted-atpt-mode 'doublequoted-highlight-atpt-mode)
(defun doublequoted-highlight-atpt-mode (&optional arg) 
  "Toggles doublequoted-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'doublequoted arg))

(defalias 'kill-doublequoted-atpt 'doublequoted-kill-atpt)
;;;###autoload
(defun doublequoted-kill-atpt (&optional arg) 
  "Kills doublequoted at point if any. "
  (interactive "*P")
  (th-kill 'doublequoted arg))

(defalias 'parentize-doublequoted-atpt 'doublequoted-parentize-atpt)
;;;###autoload
(defun doublequoted-parentize-atpt (&optional arg) 
  "Parentizes 'doublequoted at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'doublequoted arg))

(defalias 'separate-doublequoted-atpt 'doublequoted-separate-atpt)
;;;###autoload
(defun doublequoted-separate-atpt (&optional arg) 
  "Separates 'doublequoted at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'doublequoted arg))

(defalias 'singlequote-doublequoted-atpt 'doublequoted-singlequote-atpt)
;;;###autoload
(defun doublequoted-singlequote-atpt (&optional arg) 
  "Singlequotes 'doublequoted at point if any. "
  (interactive "*p")
    (th-singlequote 'doublequoted arg))

(defalias 'slash-doublequoted-atpt 'doublequoted-slash-atpt)
;;;###autoload
(defun doublequoted-slash-atpt (&optional arg) 
  "Slash 'doublequoted at point if any. "
  (interactive "*p")
    (th-slash 'doublequoted arg))

(defalias 'trim-doublequoted-atpt 'doublequoted-trim-atpt)
;;;###autoload
(defun doublequoted-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'doublequoted t t))

(defalias 'trim-doublequoted-left-atpt 'doublequoted-left-trim-atpt)
;;;###autoload
(defun doublequoted-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'doublequoted t nil))

(defalias 'trim-doublequoted-right-atpt 'doublequoted-right-trim-atpt)
;;;###autoload
(defun doublequoted-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'doublequoted nil t))

(defalias 'forward-doublequoted-atpt 'doublequoted-forward-atpt)
;;;###autoload
(defun doublequoted-forward-atpt (&optional arg) 
  "Moves forward over doublequoted at point if any, does nothing otherwise.
Returns bounds of doublequoted as a list. "
  (interactive "p")
  (th-forward 'doublequoted arg))

(defalias 'backward-doublequoted-atpt 'doublequoted-backward-atpt)
;;;###autoload
(defun doublequoted-backward-atpt (&optional arg) 
  "Moves backward over doublequoted before point if any, does nothing otherwise.
Returns bounds of doublequoted as a list. "
  (interactive "p")
  (th-backward 'doublequoted arg))

(defalias 'transpose-doublequoted-atpt 'doublequoted-transpose-atpt)
;;;###autoload
(defun doublequoted-transpose-atpt (&optional arg) 
  "Transposes doublequoted with doublequoted before point if any. "
  (interactive "*P")
  (th-transpose 'doublequoted arg))

(defalias 'sort-doublequoted-atpt 'doublequoted-sort-atpt)
;;;###autoload
(defun doublequoted-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts doublequoted's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'doublequoted reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-doublequoted-atpt 'doublequoted-check-atpt)
;;;###autoload
(defun doublequoted-check-atpt (&optional arg) 
  " doublequoted at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "doublequoted" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "doublequoted" "-end-position-atpt")))))
    (th-check 'doublequoted arg)))
   

;;;###autoload
(defun count-alnum-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of alnum in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'doublequoted beg end (interactive-p)))

(defun ratio-alnum-in-doublequoted-atpt (&optional beg end) 
    "Relation of alnum in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of alpha in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'doublequoted beg end (interactive-p)))

(defun ratio-alpha-in-doublequoted-atpt (&optional beg end) 
    "Relation of alpha in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of ascii in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'doublequoted beg end (interactive-p)))

(defun ratio-ascii-in-doublequoted-atpt (&optional beg end) 
    "Relation of ascii in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of digit in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'doublequoted beg end (interactive-p)))

(defun ratio-digit-in-doublequoted-atpt (&optional beg end) 
    "Relation of digit in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of graph in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'doublequoted beg end (interactive-p)))

(defun ratio-graph-in-doublequoted-atpt (&optional beg end) 
    "Relation of graph in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of lower in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'doublequoted beg end (interactive-p)))

(defun ratio-lower-in-doublequoted-atpt (&optional beg end) 
    "Relation of lower in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of punct in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'doublequoted beg end (interactive-p)))

(defun ratio-punct-in-doublequoted-atpt (&optional beg end) 
    "Relation of punct in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-doublequoted-atpt (&optional beg end) 
    "Counts matches of upper in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'doublequoted beg end (interactive-p)))

(defun ratio-upper-in-doublequoted-atpt (&optional beg end) 
    "Relation of upper in doublequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'doublequoted beg end t (interactive-p)))



;;;###autoload
(defun email-atpt (&optional arg) 
  "Returns email at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'email arg))

(defalias 'bounds-of-email-atpt 'email-bounds-atpt)
;;;###autoload
(defun email-bounds-atpt (&optional arg) 
  "Returns borders of email at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'email arg))

;;;###autoload
(defun email-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'email arg))

;;;###autoload
(defun email-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-end 'email arg))

(defalias 'beginning-of-email-atpt 'email-beginning-atpt)
;;;###autoload
(defun email-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'email arg))

(defalias 'end-of-email-atpt 'email-end-atpt)
;;;###autoload
(defun email-end-atpt (&optional arg) 
  "Goto end of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'email arg))

(defalias 'length-of-email-atpt 'email-length-atpt)
;;;###autoload
(defun email-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-length 'email arg))

(defalias 'copy-email-atpt 'email-copy-atpt)
;;;###autoload
(defun email-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class email at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'email arg))

(defalias 'escape-email-atpt 'email-escape-atpt)
;;;###autoload
(defun email-escape-atpt () 
  "Makes a copy of symbol or char-class email at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'email))

(defalias 'doublequote-email-atpt 'email-doublequote-atpt)
;;;###autoload
(defun email-doublequote-atpt (&optional arg) 
  "Doublequotes 'email at point if any. "
  (interactive "*p")
    (th-doublequote 'email arg))

(defalias 'dollar-email-atpt 'email-dollar-atpt)
;;;###autoload
(defun email-dollar-atpt (&optional arg) 
  "Doublequotes 'email at point if any. "
  (interactive "*p")
    (th-dollar 'email arg))

(defalias 'angle-greater-email-atpt 'email-greater-angle-atpt)
;;;###autoload
(defun email-greater-angle-atpt (&optional arg) 
  "Sets angles for 'email after point if any. "
  (interactive "*p")
    (th-greater-angle 'email arg))

(defalias 'angle-lower-email-atpt 'email-lower-angle-atpt)
;;;###autoload
(defun email-lower-angle-atpt (&optional arg) 
  "Sets angles for 'email after point if any. "
  (interactive "*p")
    (th-lower-angle 'email arg))

(defalias 'backslash-email-atpt 'email-backslash-atpt)
;;;###autoload
(defun email-backslash-atpt (&optional arg) 
  "Backslash 'email at point if any. "
  (interactive "*p")
    (th-backslash 'email arg))   

(defalias 'brace-email-atpt 'email-brace-atpt)
;;;###autoload
(defun email-brace-atpt (&optional arg) 
  "Braces 'email at point if any. "
  (interactive "*p")
    (th-brace 'email arg))

(defalias 'bracket-email-atpt 'email-bracket-atpt)
;;;###autoload
(defun email-bracket-atpt (&optional arg) 
  "Brackets 'email after point if any. "
  (interactive "*p")
    (th-bracket 'email arg))

;; (defalias 'comment-email-atpt 'email-comment-atpt)
;;;###autoload
(defun email-comment-atpt (&optional arg) 
  "Comments 'email at point if any. "
  (interactive "*p")
  (th-comment 'email arg))

(defalias 'mark-email-atpt 'email-mark-atpt)
;;;###autoload
(defun email-mark-atpt () 
  "Marks 'email at point if any. "
  (interactive)
  (th-mark 'email))

(defalias 'hide-email-atpt 'email-hide-atpt)
;;;###autoload
(defun email-hide-atpt () 
  "Toggles hide 'email in region or buffer if any. "
  (interactive)
  (th-hide 'email))

(defalias 'highlight-email-atpt-mode 'email-highlight-atpt-mode)
(defun email-highlight-atpt-mode (&optional arg) 
  "Toggles email-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'email arg))

(defalias 'kill-email-atpt 'email-kill-atpt)
;;;###autoload
(defun email-kill-atpt (&optional arg) 
  "Kills email at point if any. "
  (interactive "*P")
  (th-kill 'email arg))

(defalias 'parentize-email-atpt 'email-parentize-atpt)
;;;###autoload
(defun email-parentize-atpt (&optional arg) 
  "Parentizes 'email at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'email arg))

(defalias 'separate-email-atpt 'email-separate-atpt)
;;;###autoload
(defun email-separate-atpt (&optional arg) 
  "Separates 'email at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'email arg))

(defalias 'singlequote-email-atpt 'email-singlequote-atpt)
;;;###autoload
(defun email-singlequote-atpt (&optional arg) 
  "Singlequotes 'email at point if any. "
  (interactive "*p")
    (th-singlequote 'email arg))

(defalias 'slash-email-atpt 'email-slash-atpt)
;;;###autoload
(defun email-slash-atpt (&optional arg) 
  "Slash 'email at point if any. "
  (interactive "*p")
    (th-slash 'email arg))

(defalias 'trim-email-atpt 'email-trim-atpt)
;;;###autoload
(defun email-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'email t t))

(defalias 'trim-email-left-atpt 'email-left-trim-atpt)
;;;###autoload
(defun email-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'email t nil))

(defalias 'trim-email-right-atpt 'email-right-trim-atpt)
;;;###autoload
(defun email-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'email nil t))

(defalias 'forward-email-atpt 'email-forward-atpt)
;;;###autoload
(defun email-forward-atpt (&optional arg) 
  "Moves forward over email at point if any, does nothing otherwise.
Returns bounds of email as a list. "
  (interactive "p")
  (th-forward 'email arg))

(defalias 'backward-email-atpt 'email-backward-atpt)
;;;###autoload
(defun email-backward-atpt (&optional arg) 
  "Moves backward over email before point if any, does nothing otherwise.
Returns bounds of email as a list. "
  (interactive "p")
  (th-backward 'email arg))

(defalias 'transpose-email-atpt 'email-transpose-atpt)
;;;###autoload
(defun email-transpose-atpt (&optional arg) 
  "Transposes email with email before point if any. "
  (interactive "*P")
  (th-transpose 'email arg))

(defalias 'sort-email-atpt 'email-sort-atpt)
;;;###autoload
(defun email-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts email's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'email reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-email-atpt 'email-check-atpt)
;;;###autoload
(defun email-check-atpt (&optional arg) 
  " email at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "email" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "email" "-end-position-atpt")))))
    (th-check 'email arg)))
   

;;;###autoload
(defun count-alnum-in-email-atpt (&optional beg end) 
    "Counts matches of alnum in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'email beg end (interactive-p)))

(defun ratio-alnum-in-email-atpt (&optional beg end) 
    "Relation of alnum in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'email beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-email-atpt (&optional beg end) 
    "Counts matches of alpha in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'email beg end (interactive-p)))

(defun ratio-alpha-in-email-atpt (&optional beg end) 
    "Relation of alpha in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'email beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-email-atpt (&optional beg end) 
    "Counts matches of ascii in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'email beg end (interactive-p)))

(defun ratio-ascii-in-email-atpt (&optional beg end) 
    "Relation of ascii in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'email beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-email-atpt (&optional beg end) 
    "Counts matches of digit in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'email beg end (interactive-p)))

(defun ratio-digit-in-email-atpt (&optional beg end) 
    "Relation of digit in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'email beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-email-atpt (&optional beg end) 
    "Counts matches of graph in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'email beg end (interactive-p)))

(defun ratio-graph-in-email-atpt (&optional beg end) 
    "Relation of graph in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'email beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-email-atpt (&optional beg end) 
    "Counts matches of lower in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'email beg end (interactive-p)))

(defun ratio-lower-in-email-atpt (&optional beg end) 
    "Relation of lower in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'email beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-email-atpt (&optional beg end) 
    "Counts matches of punct in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'email beg end (interactive-p)))

(defun ratio-punct-in-email-atpt (&optional beg end) 
    "Relation of punct in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'email beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-email-atpt (&optional beg end) 
    "Counts matches of upper in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'email beg end (interactive-p)))

(defun ratio-upper-in-email-atpt (&optional beg end) 
    "Relation of upper in email if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'email beg end t (interactive-p)))



;;;###autoload
(defun filename-atpt (&optional arg) 
  "Returns filename at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'filename arg))

(defalias 'bounds-of-filename-atpt 'filename-bounds-atpt)
;;;###autoload
(defun filename-bounds-atpt (&optional arg) 
  "Returns borders of filename at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'filename arg))

;;;###autoload
(defun filename-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'filename arg))

;;;###autoload
(defun filename-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-end 'filename arg))

(defalias 'beginning-of-filename-atpt 'filename-beginning-atpt)
;;;###autoload
(defun filename-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'filename arg))

(defalias 'end-of-filename-atpt 'filename-end-atpt)
;;;###autoload
(defun filename-end-atpt (&optional arg) 
  "Goto end of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'filename arg))

(defalias 'length-of-filename-atpt 'filename-length-atpt)
;;;###autoload
(defun filename-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-length 'filename arg))

(defalias 'copy-filename-atpt 'filename-copy-atpt)
;;;###autoload
(defun filename-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class filename at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'filename arg))

(defalias 'escape-filename-atpt 'filename-escape-atpt)
;;;###autoload
(defun filename-escape-atpt () 
  "Makes a copy of symbol or char-class filename at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'filename))

(defalias 'doublequote-filename-atpt 'filename-doublequote-atpt)
;;;###autoload
(defun filename-doublequote-atpt (&optional arg) 
  "Doublequotes 'filename at point if any. "
  (interactive "*p")
    (th-doublequote 'filename arg))

(defalias 'dollar-filename-atpt 'filename-dollar-atpt)
;;;###autoload
(defun filename-dollar-atpt (&optional arg) 
  "Doublequotes 'filename at point if any. "
  (interactive "*p")
    (th-dollar 'filename arg))

(defalias 'angle-greater-filename-atpt 'filename-greater-angle-atpt)
;;;###autoload
(defun filename-greater-angle-atpt (&optional arg) 
  "Sets angles for 'filename after point if any. "
  (interactive "*p")
    (th-greater-angle 'filename arg))

(defalias 'angle-lower-filename-atpt 'filename-lower-angle-atpt)
;;;###autoload
(defun filename-lower-angle-atpt (&optional arg) 
  "Sets angles for 'filename after point if any. "
  (interactive "*p")
    (th-lower-angle 'filename arg))

(defalias 'backslash-filename-atpt 'filename-backslash-atpt)
;;;###autoload
(defun filename-backslash-atpt (&optional arg) 
  "Backslash 'filename at point if any. "
  (interactive "*p")
    (th-backslash 'filename arg))   

(defalias 'brace-filename-atpt 'filename-brace-atpt)
;;;###autoload
(defun filename-brace-atpt (&optional arg) 
  "Braces 'filename at point if any. "
  (interactive "*p")
    (th-brace 'filename arg))

(defalias 'bracket-filename-atpt 'filename-bracket-atpt)
;;;###autoload
(defun filename-bracket-atpt (&optional arg) 
  "Brackets 'filename after point if any. "
  (interactive "*p")
    (th-bracket 'filename arg))

;; (defalias 'comment-filename-atpt 'filename-comment-atpt)
;;;###autoload
(defun filename-comment-atpt (&optional arg) 
  "Comments 'filename at point if any. "
  (interactive "*p")
  (th-comment 'filename arg))

(defalias 'mark-filename-atpt 'filename-mark-atpt)
;;;###autoload
(defun filename-mark-atpt () 
  "Marks 'filename at point if any. "
  (interactive)
  (th-mark 'filename))

(defalias 'hide-filename-atpt 'filename-hide-atpt)
;;;###autoload
(defun filename-hide-atpt () 
  "Toggles hide 'filename in region or buffer if any. "
  (interactive)
  (th-hide 'filename))

(defalias 'highlight-filename-atpt-mode 'filename-highlight-atpt-mode)
(defun filename-highlight-atpt-mode (&optional arg) 
  "Toggles filename-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'filename arg))

(defalias 'kill-filename-atpt 'filename-kill-atpt)
;;;###autoload
(defun filename-kill-atpt (&optional arg) 
  "Kills filename at point if any. "
  (interactive "*P")
  (th-kill 'filename arg))

(defalias 'parentize-filename-atpt 'filename-parentize-atpt)
;;;###autoload
(defun filename-parentize-atpt (&optional arg) 
  "Parentizes 'filename at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'filename arg))

(defalias 'separate-filename-atpt 'filename-separate-atpt)
;;;###autoload
(defun filename-separate-atpt (&optional arg) 
  "Separates 'filename at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'filename arg))

(defalias 'singlequote-filename-atpt 'filename-singlequote-atpt)
;;;###autoload
(defun filename-singlequote-atpt (&optional arg) 
  "Singlequotes 'filename at point if any. "
  (interactive "*p")
    (th-singlequote 'filename arg))

(defalias 'slash-filename-atpt 'filename-slash-atpt)
;;;###autoload
(defun filename-slash-atpt (&optional arg) 
  "Slash 'filename at point if any. "
  (interactive "*p")
    (th-slash 'filename arg))

(defalias 'trim-filename-atpt 'filename-trim-atpt)
;;;###autoload
(defun filename-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'filename t t))

(defalias 'trim-filename-left-atpt 'filename-left-trim-atpt)
;;;###autoload
(defun filename-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'filename t nil))

(defalias 'trim-filename-right-atpt 'filename-right-trim-atpt)
;;;###autoload
(defun filename-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'filename nil t))

(defalias 'forward-filename-atpt 'filename-forward-atpt)
;;;###autoload
(defun filename-forward-atpt (&optional arg) 
  "Moves forward over filename at point if any, does nothing otherwise.
Returns bounds of filename as a list. "
  (interactive "p")
  (th-forward 'filename arg))

(defalias 'backward-filename-atpt 'filename-backward-atpt)
;;;###autoload
(defun filename-backward-atpt (&optional arg) 
  "Moves backward over filename before point if any, does nothing otherwise.
Returns bounds of filename as a list. "
  (interactive "p")
  (th-backward 'filename arg))

(defalias 'transpose-filename-atpt 'filename-transpose-atpt)
;;;###autoload
(defun filename-transpose-atpt (&optional arg) 
  "Transposes filename with filename before point if any. "
  (interactive "*P")
  (th-transpose 'filename arg))

(defalias 'sort-filename-atpt 'filename-sort-atpt)
;;;###autoload
(defun filename-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts filename's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'filename reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-filename-atpt 'filename-check-atpt)
;;;###autoload
(defun filename-check-atpt (&optional arg) 
  " filename at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "filename" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "filename" "-end-position-atpt")))))
    (th-check 'filename arg)))
   

;;;###autoload
(defun count-alnum-in-filename-atpt (&optional beg end) 
    "Counts matches of alnum in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'filename beg end (interactive-p)))

(defun ratio-alnum-in-filename-atpt (&optional beg end) 
    "Relation of alnum in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-filename-atpt (&optional beg end) 
    "Counts matches of alpha in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'filename beg end (interactive-p)))

(defun ratio-alpha-in-filename-atpt (&optional beg end) 
    "Relation of alpha in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-filename-atpt (&optional beg end) 
    "Counts matches of ascii in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'filename beg end (interactive-p)))

(defun ratio-ascii-in-filename-atpt (&optional beg end) 
    "Relation of ascii in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-filename-atpt (&optional beg end) 
    "Counts matches of digit in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'filename beg end (interactive-p)))

(defun ratio-digit-in-filename-atpt (&optional beg end) 
    "Relation of digit in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-filename-atpt (&optional beg end) 
    "Counts matches of graph in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'filename beg end (interactive-p)))

(defun ratio-graph-in-filename-atpt (&optional beg end) 
    "Relation of graph in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-filename-atpt (&optional beg end) 
    "Counts matches of lower in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'filename beg end (interactive-p)))

(defun ratio-lower-in-filename-atpt (&optional beg end) 
    "Relation of lower in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-filename-atpt (&optional beg end) 
    "Counts matches of punct in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'filename beg end (interactive-p)))

(defun ratio-punct-in-filename-atpt (&optional beg end) 
    "Relation of punct in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'filename beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-filename-atpt (&optional beg end) 
    "Counts matches of upper in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'filename beg end (interactive-p)))

(defun ratio-upper-in-filename-atpt (&optional beg end) 
    "Relation of upper in filename if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'filename beg end t (interactive-p)))



;;;###autoload
(defun float-atpt (&optional arg) 
  "Returns float at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'float arg))

(defalias 'bounds-of-float-atpt 'float-bounds-atpt)
;;;###autoload
(defun float-bounds-atpt (&optional arg) 
  "Returns borders of float at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'float arg))

;;;###autoload
(defun float-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'float arg))

;;;###autoload
(defun float-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-end 'float arg))

(defalias 'beginning-of-float-atpt 'float-beginning-atpt)
;;;###autoload
(defun float-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'float arg))

(defalias 'end-of-float-atpt 'float-end-atpt)
;;;###autoload
(defun float-end-atpt (&optional arg) 
  "Goto end of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'float arg))

(defalias 'length-of-float-atpt 'float-length-atpt)
;;;###autoload
(defun float-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-length 'float arg))

(defalias 'copy-float-atpt 'float-copy-atpt)
;;;###autoload
(defun float-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class float at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'float arg))

(defalias 'escape-float-atpt 'float-escape-atpt)
;;;###autoload
(defun float-escape-atpt () 
  "Makes a copy of symbol or char-class float at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'float))

(defalias 'doublequote-float-atpt 'float-doublequote-atpt)
;;;###autoload
(defun float-doublequote-atpt (&optional arg) 
  "Doublequotes 'float at point if any. "
  (interactive "*p")
    (th-doublequote 'float arg))

(defalias 'dollar-float-atpt 'float-dollar-atpt)
;;;###autoload
(defun float-dollar-atpt (&optional arg) 
  "Doublequotes 'float at point if any. "
  (interactive "*p")
    (th-dollar 'float arg))

(defalias 'angle-greater-float-atpt 'float-greater-angle-atpt)
;;;###autoload
(defun float-greater-angle-atpt (&optional arg) 
  "Sets angles for 'float after point if any. "
  (interactive "*p")
    (th-greater-angle 'float arg))

(defalias 'angle-lower-float-atpt 'float-lower-angle-atpt)
;;;###autoload
(defun float-lower-angle-atpt (&optional arg) 
  "Sets angles for 'float after point if any. "
  (interactive "*p")
    (th-lower-angle 'float arg))

(defalias 'backslash-float-atpt 'float-backslash-atpt)
;;;###autoload
(defun float-backslash-atpt (&optional arg) 
  "Backslash 'float at point if any. "
  (interactive "*p")
    (th-backslash 'float arg))   

(defalias 'brace-float-atpt 'float-brace-atpt)
;;;###autoload
(defun float-brace-atpt (&optional arg) 
  "Braces 'float at point if any. "
  (interactive "*p")
    (th-brace 'float arg))

(defalias 'bracket-float-atpt 'float-bracket-atpt)
;;;###autoload
(defun float-bracket-atpt (&optional arg) 
  "Brackets 'float after point if any. "
  (interactive "*p")
    (th-bracket 'float arg))

;; (defalias 'comment-float-atpt 'float-comment-atpt)
;;;###autoload
(defun float-comment-atpt (&optional arg) 
  "Comments 'float at point if any. "
  (interactive "*p")
  (th-comment 'float arg))

(defalias 'mark-float-atpt 'float-mark-atpt)
;;;###autoload
(defun float-mark-atpt () 
  "Marks 'float at point if any. "
  (interactive)
  (th-mark 'float))

(defalias 'hide-float-atpt 'float-hide-atpt)
;;;###autoload
(defun float-hide-atpt () 
  "Toggles hide 'float in region or buffer if any. "
  (interactive)
  (th-hide 'float))

(defalias 'highlight-float-atpt-mode 'float-highlight-atpt-mode)
(defun float-highlight-atpt-mode (&optional arg) 
  "Toggles float-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'float arg))

(defalias 'kill-float-atpt 'float-kill-atpt)
;;;###autoload
(defun float-kill-atpt (&optional arg) 
  "Kills float at point if any. "
  (interactive "*P")
  (th-kill 'float arg))

(defalias 'parentize-float-atpt 'float-parentize-atpt)
;;;###autoload
(defun float-parentize-atpt (&optional arg) 
  "Parentizes 'float at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'float arg))

(defalias 'separate-float-atpt 'float-separate-atpt)
;;;###autoload
(defun float-separate-atpt (&optional arg) 
  "Separates 'float at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'float arg))

(defalias 'singlequote-float-atpt 'float-singlequote-atpt)
;;;###autoload
(defun float-singlequote-atpt (&optional arg) 
  "Singlequotes 'float at point if any. "
  (interactive "*p")
    (th-singlequote 'float arg))

(defalias 'slash-float-atpt 'float-slash-atpt)
;;;###autoload
(defun float-slash-atpt (&optional arg) 
  "Slash 'float at point if any. "
  (interactive "*p")
    (th-slash 'float arg))

(defalias 'trim-float-atpt 'float-trim-atpt)
;;;###autoload
(defun float-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'float t t))

(defalias 'trim-float-left-atpt 'float-left-trim-atpt)
;;;###autoload
(defun float-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'float t nil))

(defalias 'trim-float-right-atpt 'float-right-trim-atpt)
;;;###autoload
(defun float-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'float nil t))

(defalias 'forward-float-atpt 'float-forward-atpt)
;;;###autoload
(defun float-forward-atpt (&optional arg) 
  "Moves forward over float at point if any, does nothing otherwise.
Returns bounds of float as a list. "
  (interactive "p")
  (th-forward 'float arg))

(defalias 'backward-float-atpt 'float-backward-atpt)
;;;###autoload
(defun float-backward-atpt (&optional arg) 
  "Moves backward over float before point if any, does nothing otherwise.
Returns bounds of float as a list. "
  (interactive "p")
  (th-backward 'float arg))

(defalias 'transpose-float-atpt 'float-transpose-atpt)
;;;###autoload
(defun float-transpose-atpt (&optional arg) 
  "Transposes float with float before point if any. "
  (interactive "*P")
  (th-transpose 'float arg))

(defalias 'sort-float-atpt 'float-sort-atpt)
;;;###autoload
(defun float-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts float's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'float reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-float-atpt 'float-check-atpt)
;;;###autoload
(defun float-check-atpt (&optional arg) 
  " float at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "float" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "float" "-end-position-atpt")))))
    (th-check 'float arg)))
   

;;;###autoload
(defun count-alnum-in-float-atpt (&optional beg end) 
    "Counts matches of alnum in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'float beg end (interactive-p)))

(defun ratio-alnum-in-float-atpt (&optional beg end) 
    "Relation of alnum in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'float beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-float-atpt (&optional beg end) 
    "Counts matches of alpha in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'float beg end (interactive-p)))

(defun ratio-alpha-in-float-atpt (&optional beg end) 
    "Relation of alpha in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'float beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-float-atpt (&optional beg end) 
    "Counts matches of ascii in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'float beg end (interactive-p)))

(defun ratio-ascii-in-float-atpt (&optional beg end) 
    "Relation of ascii in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'float beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-float-atpt (&optional beg end) 
    "Counts matches of digit in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'float beg end (interactive-p)))

(defun ratio-digit-in-float-atpt (&optional beg end) 
    "Relation of digit in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'float beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-float-atpt (&optional beg end) 
    "Counts matches of graph in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'float beg end (interactive-p)))

(defun ratio-graph-in-float-atpt (&optional beg end) 
    "Relation of graph in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'float beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-float-atpt (&optional beg end) 
    "Counts matches of lower in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'float beg end (interactive-p)))

(defun ratio-lower-in-float-atpt (&optional beg end) 
    "Relation of lower in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'float beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-float-atpt (&optional beg end) 
    "Counts matches of punct in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'float beg end (interactive-p)))

(defun ratio-punct-in-float-atpt (&optional beg end) 
    "Relation of punct in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'float beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-float-atpt (&optional beg end) 
    "Counts matches of upper in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'float beg end (interactive-p)))

(defun ratio-upper-in-float-atpt (&optional beg end) 
    "Relation of upper in float if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'float beg end t (interactive-p)))



;;;###autoload
(defun function-atpt (&optional arg) 
  "Returns function at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'function arg))

(defalias 'bounds-of-function-atpt 'function-bounds-atpt)
;;;###autoload
(defun function-bounds-atpt (&optional arg) 
  "Returns borders of function at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'function arg))

;;;###autoload
(defun function-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'function arg))

;;;###autoload
(defun function-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-end 'function arg))

(defalias 'beginning-of-function-atpt 'function-beginning-atpt)
;;;###autoload
(defun function-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'function arg))

(defalias 'end-of-function-atpt 'function-end-atpt)
;;;###autoload
(defun function-end-atpt (&optional arg) 
  "Goto end of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'function arg))

(defalias 'length-of-function-atpt 'function-length-atpt)
;;;###autoload
(defun function-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-length 'function arg))

(defalias 'copy-function-atpt 'function-copy-atpt)
;;;###autoload
(defun function-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class function at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'function arg))

(defalias 'escape-function-atpt 'function-escape-atpt)
;;;###autoload
(defun function-escape-atpt () 
  "Makes a copy of symbol or char-class function at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'function))

(defalias 'doublequote-function-atpt 'function-doublequote-atpt)
;;;###autoload
(defun function-doublequote-atpt (&optional arg) 
  "Doublequotes 'function at point if any. "
  (interactive "*p")
    (th-doublequote 'function arg))

(defalias 'dollar-function-atpt 'function-dollar-atpt)
;;;###autoload
(defun function-dollar-atpt (&optional arg) 
  "Doublequotes 'function at point if any. "
  (interactive "*p")
    (th-dollar 'function arg))

(defalias 'angle-greater-function-atpt 'function-greater-angle-atpt)
;;;###autoload
(defun function-greater-angle-atpt (&optional arg) 
  "Sets angles for 'function after point if any. "
  (interactive "*p")
    (th-greater-angle 'function arg))

(defalias 'angle-lower-function-atpt 'function-lower-angle-atpt)
;;;###autoload
(defun function-lower-angle-atpt (&optional arg) 
  "Sets angles for 'function after point if any. "
  (interactive "*p")
    (th-lower-angle 'function arg))

(defalias 'backslash-function-atpt 'function-backslash-atpt)
;;;###autoload
(defun function-backslash-atpt (&optional arg) 
  "Backslash 'function at point if any. "
  (interactive "*p")
    (th-backslash 'function arg))   

(defalias 'brace-function-atpt 'function-brace-atpt)
;;;###autoload
(defun function-brace-atpt (&optional arg) 
  "Braces 'function at point if any. "
  (interactive "*p")
    (th-brace 'function arg))

(defalias 'bracket-function-atpt 'function-bracket-atpt)
;;;###autoload
(defun function-bracket-atpt (&optional arg) 
  "Brackets 'function after point if any. "
  (interactive "*p")
    (th-bracket 'function arg))

;; (defalias 'comment-function-atpt 'function-comment-atpt)
;;;###autoload
(defun function-comment-atpt (&optional arg) 
  "Comments 'function at point if any. "
  (interactive "*p")
  (th-comment 'function arg))

(defalias 'mark-function-atpt 'function-mark-atpt)
;;;###autoload
(defun function-mark-atpt () 
  "Marks 'function at point if any. "
  (interactive)
  (th-mark 'function))

(defalias 'hide-function-atpt 'function-hide-atpt)
;;;###autoload
(defun function-hide-atpt () 
  "Toggles hide 'function in region or buffer if any. "
  (interactive)
  (th-hide 'function))

(defalias 'highlight-function-atpt-mode 'function-highlight-atpt-mode)
(defun function-highlight-atpt-mode (&optional arg) 
  "Toggles function-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'function arg))

(defalias 'kill-function-atpt 'function-kill-atpt)
;;;###autoload
(defun function-kill-atpt (&optional arg) 
  "Kills function at point if any. "
  (interactive "*P")
  (th-kill 'function arg))

(defalias 'parentize-function-atpt 'function-parentize-atpt)
;;;###autoload
(defun function-parentize-atpt (&optional arg) 
  "Parentizes 'function at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'function arg))

(defalias 'separate-function-atpt 'function-separate-atpt)
;;;###autoload
(defun function-separate-atpt (&optional arg) 
  "Separates 'function at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'function arg))

(defalias 'singlequote-function-atpt 'function-singlequote-atpt)
;;;###autoload
(defun function-singlequote-atpt (&optional arg) 
  "Singlequotes 'function at point if any. "
  (interactive "*p")
    (th-singlequote 'function arg))

(defalias 'slash-function-atpt 'function-slash-atpt)
;;;###autoload
(defun function-slash-atpt (&optional arg) 
  "Slash 'function at point if any. "
  (interactive "*p")
    (th-slash 'function arg))

(defalias 'trim-function-atpt 'function-trim-atpt)
;;;###autoload
(defun function-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'function t t))

(defalias 'trim-function-left-atpt 'function-left-trim-atpt)
;;;###autoload
(defun function-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'function t nil))

(defalias 'trim-function-right-atpt 'function-right-trim-atpt)
;;;###autoload
(defun function-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'function nil t))

(defalias 'forward-function-atpt 'function-forward-atpt)
;;;###autoload
(defun function-forward-atpt (&optional arg) 
  "Moves forward over function at point if any, does nothing otherwise.
Returns bounds of function as a list. "
  (interactive "p")
  (th-forward 'function arg))

(defalias 'backward-function-atpt 'function-backward-atpt)
;;;###autoload
(defun function-backward-atpt (&optional arg) 
  "Moves backward over function before point if any, does nothing otherwise.
Returns bounds of function as a list. "
  (interactive "p")
  (th-backward 'function arg))

(defalias 'transpose-function-atpt 'function-transpose-atpt)
;;;###autoload
(defun function-transpose-atpt (&optional arg) 
  "Transposes function with function before point if any. "
  (interactive "*P")
  (th-transpose 'function arg))

(defalias 'sort-function-atpt 'function-sort-atpt)
;;;###autoload
(defun function-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts function's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'function reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-function-atpt 'function-check-atpt)
;;;###autoload
(defun function-check-atpt (&optional arg) 
  " function at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "function" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "function" "-end-position-atpt")))))
    (th-check 'function arg)))
   

;;;###autoload
(defun count-alnum-in-function-atpt (&optional beg end) 
    "Counts matches of alnum in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'function beg end (interactive-p)))

(defun ratio-alnum-in-function-atpt (&optional beg end) 
    "Relation of alnum in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'function beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-function-atpt (&optional beg end) 
    "Counts matches of alpha in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'function beg end (interactive-p)))

(defun ratio-alpha-in-function-atpt (&optional beg end) 
    "Relation of alpha in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'function beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-function-atpt (&optional beg end) 
    "Counts matches of ascii in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'function beg end (interactive-p)))

(defun ratio-ascii-in-function-atpt (&optional beg end) 
    "Relation of ascii in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'function beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-function-atpt (&optional beg end) 
    "Counts matches of digit in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'function beg end (interactive-p)))

(defun ratio-digit-in-function-atpt (&optional beg end) 
    "Relation of digit in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'function beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-function-atpt (&optional beg end) 
    "Counts matches of graph in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'function beg end (interactive-p)))

(defun ratio-graph-in-function-atpt (&optional beg end) 
    "Relation of graph in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'function beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-function-atpt (&optional beg end) 
    "Counts matches of lower in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'function beg end (interactive-p)))

(defun ratio-lower-in-function-atpt (&optional beg end) 
    "Relation of lower in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'function beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-function-atpt (&optional beg end) 
    "Counts matches of punct in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'function beg end (interactive-p)))

(defun ratio-punct-in-function-atpt (&optional beg end) 
    "Relation of punct in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'function beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-function-atpt (&optional beg end) 
    "Counts matches of upper in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'function beg end (interactive-p)))

(defun ratio-upper-in-function-atpt (&optional beg end) 
    "Relation of upper in function if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'function beg end t (interactive-p)))



;;;###autoload
(defun graph-atpt (&optional arg) 
  "Returns graph at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'graph arg))

(defalias 'bounds-of-graph-atpt 'graph-bounds-atpt)
;;;###autoload
(defun graph-bounds-atpt (&optional arg) 
  "Returns borders of graph at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'graph arg))

;;;###autoload
(defun graph-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'graph arg))

;;;###autoload
(defun graph-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-end 'graph arg))

(defalias 'beginning-of-graph-atpt 'graph-beginning-atpt)
;;;###autoload
(defun graph-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'graph arg))

(defalias 'end-of-graph-atpt 'graph-end-atpt)
;;;###autoload
(defun graph-end-atpt (&optional arg) 
  "Goto end of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'graph arg))

(defalias 'length-of-graph-atpt 'graph-length-atpt)
;;;###autoload
(defun graph-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-length 'graph arg))

(defalias 'copy-graph-atpt 'graph-copy-atpt)
;;;###autoload
(defun graph-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class graph at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'graph arg))

(defalias 'escape-graph-atpt 'graph-escape-atpt)
;;;###autoload
(defun graph-escape-atpt () 
  "Makes a copy of symbol or char-class graph at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'graph))

(defalias 'doublequote-graph-atpt 'graph-doublequote-atpt)
;;;###autoload
(defun graph-doublequote-atpt (&optional arg) 
  "Doublequotes 'graph at point if any. "
  (interactive "*p")
    (th-doublequote 'graph arg))

(defalias 'dollar-graph-atpt 'graph-dollar-atpt)
;;;###autoload
(defun graph-dollar-atpt (&optional arg) 
  "Doublequotes 'graph at point if any. "
  (interactive "*p")
    (th-dollar 'graph arg))

(defalias 'angle-greater-graph-atpt 'graph-greater-angle-atpt)
;;;###autoload
(defun graph-greater-angle-atpt (&optional arg) 
  "Sets angles for 'graph after point if any. "
  (interactive "*p")
    (th-greater-angle 'graph arg))

(defalias 'angle-lower-graph-atpt 'graph-lower-angle-atpt)
;;;###autoload
(defun graph-lower-angle-atpt (&optional arg) 
  "Sets angles for 'graph after point if any. "
  (interactive "*p")
    (th-lower-angle 'graph arg))

(defalias 'backslash-graph-atpt 'graph-backslash-atpt)
;;;###autoload
(defun graph-backslash-atpt (&optional arg) 
  "Backslash 'graph at point if any. "
  (interactive "*p")
    (th-backslash 'graph arg))   

(defalias 'brace-graph-atpt 'graph-brace-atpt)
;;;###autoload
(defun graph-brace-atpt (&optional arg) 
  "Braces 'graph at point if any. "
  (interactive "*p")
    (th-brace 'graph arg))

(defalias 'bracket-graph-atpt 'graph-bracket-atpt)
;;;###autoload
(defun graph-bracket-atpt (&optional arg) 
  "Brackets 'graph after point if any. "
  (interactive "*p")
    (th-bracket 'graph arg))

;; (defalias 'comment-graph-atpt 'graph-comment-atpt)
;;;###autoload
(defun graph-comment-atpt (&optional arg) 
  "Comments 'graph at point if any. "
  (interactive "*p")
  (th-comment 'graph arg))

(defalias 'mark-graph-atpt 'graph-mark-atpt)
;;;###autoload
(defun graph-mark-atpt () 
  "Marks 'graph at point if any. "
  (interactive)
  (th-mark 'graph))

(defalias 'hide-graph-atpt 'graph-hide-atpt)
;;;###autoload
(defun graph-hide-atpt () 
  "Toggles hide 'graph in region or buffer if any. "
  (interactive)
  (th-hide 'graph))

(defalias 'highlight-graph-atpt-mode 'graph-highlight-atpt-mode)
(defun graph-highlight-atpt-mode (&optional arg) 
  "Toggles graph-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'graph arg))

(defalias 'kill-graph-atpt 'graph-kill-atpt)
;;;###autoload
(defun graph-kill-atpt (&optional arg) 
  "Kills graph at point if any. "
  (interactive "*P")
  (th-kill 'graph arg))

(defalias 'parentize-graph-atpt 'graph-parentize-atpt)
;;;###autoload
(defun graph-parentize-atpt (&optional arg) 
  "Parentizes 'graph at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'graph arg))

(defalias 'separate-graph-atpt 'graph-separate-atpt)
;;;###autoload
(defun graph-separate-atpt (&optional arg) 
  "Separates 'graph at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'graph arg))

(defalias 'singlequote-graph-atpt 'graph-singlequote-atpt)
;;;###autoload
(defun graph-singlequote-atpt (&optional arg) 
  "Singlequotes 'graph at point if any. "
  (interactive "*p")
    (th-singlequote 'graph arg))

(defalias 'slash-graph-atpt 'graph-slash-atpt)
;;;###autoload
(defun graph-slash-atpt (&optional arg) 
  "Slash 'graph at point if any. "
  (interactive "*p")
    (th-slash 'graph arg))

(defalias 'trim-graph-atpt 'graph-trim-atpt)
;;;###autoload
(defun graph-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'graph t t))

(defalias 'trim-graph-left-atpt 'graph-left-trim-atpt)
;;;###autoload
(defun graph-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'graph t nil))

(defalias 'trim-graph-right-atpt 'graph-right-trim-atpt)
;;;###autoload
(defun graph-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'graph nil t))

(defalias 'forward-graph-atpt 'graph-forward-atpt)
;;;###autoload
(defun graph-forward-atpt (&optional arg) 
  "Moves forward over graph at point if any, does nothing otherwise.
Returns bounds of graph as a list. "
  (interactive "p")
  (th-forward 'graph arg))

(defalias 'backward-graph-atpt 'graph-backward-atpt)
;;;###autoload
(defun graph-backward-atpt (&optional arg) 
  "Moves backward over graph before point if any, does nothing otherwise.
Returns bounds of graph as a list. "
  (interactive "p")
  (th-backward 'graph arg))

(defalias 'transpose-graph-atpt 'graph-transpose-atpt)
;;;###autoload
(defun graph-transpose-atpt (&optional arg) 
  "Transposes graph with graph before point if any. "
  (interactive "*P")
  (th-transpose 'graph arg))

(defalias 'sort-graph-atpt 'graph-sort-atpt)
;;;###autoload
(defun graph-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts graph's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'graph reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-graph-atpt 'graph-check-atpt)
;;;###autoload
(defun graph-check-atpt (&optional arg) 
  " graph at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "graph" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "graph" "-end-position-atpt")))))
    (th-check 'graph arg)))
   

;;;###autoload
(defun count-alnum-in-graph-atpt (&optional beg end) 
    "Counts matches of alnum in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'graph beg end (interactive-p)))

(defun ratio-alnum-in-graph-atpt (&optional beg end) 
    "Relation of alnum in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-graph-atpt (&optional beg end) 
    "Counts matches of alpha in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'graph beg end (interactive-p)))

(defun ratio-alpha-in-graph-atpt (&optional beg end) 
    "Relation of alpha in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-graph-atpt (&optional beg end) 
    "Counts matches of ascii in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'graph beg end (interactive-p)))

(defun ratio-ascii-in-graph-atpt (&optional beg end) 
    "Relation of ascii in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-graph-atpt (&optional beg end) 
    "Counts matches of digit in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'graph beg end (interactive-p)))

(defun ratio-digit-in-graph-atpt (&optional beg end) 
    "Relation of digit in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-graph-atpt (&optional beg end) 
    "Counts matches of graph at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'graph beg end (interactive-p)))


;;;###autoload
(defun count-lower-in-graph-atpt (&optional beg end) 
    "Counts matches of lower in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'graph beg end (interactive-p)))

(defun ratio-lower-in-graph-atpt (&optional beg end) 
    "Relation of lower in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-graph-atpt (&optional beg end) 
    "Counts matches of punct in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'graph beg end (interactive-p)))

(defun ratio-punct-in-graph-atpt (&optional beg end) 
    "Relation of punct in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'graph beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-graph-atpt (&optional beg end) 
    "Counts matches of upper in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'graph beg end (interactive-p)))

(defun ratio-upper-in-graph-atpt (&optional beg end) 
    "Relation of upper in graph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'graph beg end t (interactive-p)))



;;;###autoload
(defun ip-atpt (&optional arg) 
  "Returns ip at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'ip arg))

(defalias 'bounds-of-ip-atpt 'ip-bounds-atpt)
;;;###autoload
(defun ip-bounds-atpt (&optional arg) 
  "Returns borders of ip at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'ip arg))

;;;###autoload
(defun ip-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'ip arg))

;;;###autoload
(defun ip-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-end 'ip arg))

(defalias 'beginning-of-ip-atpt 'ip-beginning-atpt)
;;;###autoload
(defun ip-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'ip arg))

(defalias 'end-of-ip-atpt 'ip-end-atpt)
;;;###autoload
(defun ip-end-atpt (&optional arg) 
  "Goto end of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'ip arg))

(defalias 'length-of-ip-atpt 'ip-length-atpt)
;;;###autoload
(defun ip-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-length 'ip arg))

(defalias 'copy-ip-atpt 'ip-copy-atpt)
;;;###autoload
(defun ip-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class ip at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'ip arg))

(defalias 'escape-ip-atpt 'ip-escape-atpt)
;;;###autoload
(defun ip-escape-atpt () 
  "Makes a copy of symbol or char-class ip at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'ip))

(defalias 'doublequote-ip-atpt 'ip-doublequote-atpt)
;;;###autoload
(defun ip-doublequote-atpt (&optional arg) 
  "Doublequotes 'ip at point if any. "
  (interactive "*p")
    (th-doublequote 'ip arg))

(defalias 'dollar-ip-atpt 'ip-dollar-atpt)
;;;###autoload
(defun ip-dollar-atpt (&optional arg) 
  "Doublequotes 'ip at point if any. "
  (interactive "*p")
    (th-dollar 'ip arg))

(defalias 'angle-greater-ip-atpt 'ip-greater-angle-atpt)
;;;###autoload
(defun ip-greater-angle-atpt (&optional arg) 
  "Sets angles for 'ip after point if any. "
  (interactive "*p")
    (th-greater-angle 'ip arg))

(defalias 'angle-lower-ip-atpt 'ip-lower-angle-atpt)
;;;###autoload
(defun ip-lower-angle-atpt (&optional arg) 
  "Sets angles for 'ip after point if any. "
  (interactive "*p")
    (th-lower-angle 'ip arg))

(defalias 'backslash-ip-atpt 'ip-backslash-atpt)
;;;###autoload
(defun ip-backslash-atpt (&optional arg) 
  "Backslash 'ip at point if any. "
  (interactive "*p")
    (th-backslash 'ip arg))   

(defalias 'brace-ip-atpt 'ip-brace-atpt)
;;;###autoload
(defun ip-brace-atpt (&optional arg) 
  "Braces 'ip at point if any. "
  (interactive "*p")
    (th-brace 'ip arg))

(defalias 'bracket-ip-atpt 'ip-bracket-atpt)
;;;###autoload
(defun ip-bracket-atpt (&optional arg) 
  "Brackets 'ip after point if any. "
  (interactive "*p")
    (th-bracket 'ip arg))

;; (defalias 'comment-ip-atpt 'ip-comment-atpt)
;;;###autoload
(defun ip-comment-atpt (&optional arg) 
  "Comments 'ip at point if any. "
  (interactive "*p")
  (th-comment 'ip arg))

(defalias 'mark-ip-atpt 'ip-mark-atpt)
;;;###autoload
(defun ip-mark-atpt () 
  "Marks 'ip at point if any. "
  (interactive)
  (th-mark 'ip))

(defalias 'hide-ip-atpt 'ip-hide-atpt)
;;;###autoload
(defun ip-hide-atpt () 
  "Toggles hide 'ip in region or buffer if any. "
  (interactive)
  (th-hide 'ip))

(defalias 'highlight-ip-atpt-mode 'ip-highlight-atpt-mode)
(defun ip-highlight-atpt-mode (&optional arg) 
  "Toggles ip-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'ip arg))

(defalias 'kill-ip-atpt 'ip-kill-atpt)
;;;###autoload
(defun ip-kill-atpt (&optional arg) 
  "Kills ip at point if any. "
  (interactive "*P")
  (th-kill 'ip arg))

(defalias 'parentize-ip-atpt 'ip-parentize-atpt)
;;;###autoload
(defun ip-parentize-atpt (&optional arg) 
  "Parentizes 'ip at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'ip arg))

(defalias 'separate-ip-atpt 'ip-separate-atpt)
;;;###autoload
(defun ip-separate-atpt (&optional arg) 
  "Separates 'ip at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'ip arg))

(defalias 'singlequote-ip-atpt 'ip-singlequote-atpt)
;;;###autoload
(defun ip-singlequote-atpt (&optional arg) 
  "Singlequotes 'ip at point if any. "
  (interactive "*p")
    (th-singlequote 'ip arg))

(defalias 'slash-ip-atpt 'ip-slash-atpt)
;;;###autoload
(defun ip-slash-atpt (&optional arg) 
  "Slash 'ip at point if any. "
  (interactive "*p")
    (th-slash 'ip arg))

(defalias 'trim-ip-atpt 'ip-trim-atpt)
;;;###autoload
(defun ip-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ip t t))

(defalias 'trim-ip-left-atpt 'ip-left-trim-atpt)
;;;###autoload
(defun ip-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ip t nil))

(defalias 'trim-ip-right-atpt 'ip-right-trim-atpt)
;;;###autoload
(defun ip-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ip nil t))

(defalias 'forward-ip-atpt 'ip-forward-atpt)
;;;###autoload
(defun ip-forward-atpt (&optional arg) 
  "Moves forward over ip at point if any, does nothing otherwise.
Returns bounds of ip as a list. "
  (interactive "p")
  (th-forward 'ip arg))

(defalias 'backward-ip-atpt 'ip-backward-atpt)
;;;###autoload
(defun ip-backward-atpt (&optional arg) 
  "Moves backward over ip before point if any, does nothing otherwise.
Returns bounds of ip as a list. "
  (interactive "p")
  (th-backward 'ip arg))

(defalias 'transpose-ip-atpt 'ip-transpose-atpt)
;;;###autoload
(defun ip-transpose-atpt (&optional arg) 
  "Transposes ip with ip before point if any. "
  (interactive "*P")
  (th-transpose 'ip arg))

(defalias 'sort-ip-atpt 'ip-sort-atpt)
;;;###autoload
(defun ip-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts ip's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'ip reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-ip-atpt 'ip-check-atpt)
;;;###autoload
(defun ip-check-atpt (&optional arg) 
  " ip at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "ip" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "ip" "-end-position-atpt")))))
    (th-check 'ip arg)))
   

;;;###autoload
(defun count-alnum-in-ip-atpt (&optional beg end) 
    "Counts matches of alnum in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ip beg end (interactive-p)))

(defun ratio-alnum-in-ip-atpt (&optional beg end) 
    "Relation of alnum in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-ip-atpt (&optional beg end) 
    "Counts matches of alpha in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ip beg end (interactive-p)))

(defun ratio-alpha-in-ip-atpt (&optional beg end) 
    "Relation of alpha in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-ip-atpt (&optional beg end) 
    "Counts matches of ascii in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'ip beg end (interactive-p)))

(defun ratio-ascii-in-ip-atpt (&optional beg end) 
    "Relation of ascii in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-ip-atpt (&optional beg end) 
    "Counts matches of digit in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ip beg end (interactive-p)))

(defun ratio-digit-in-ip-atpt (&optional beg end) 
    "Relation of digit in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-ip-atpt (&optional beg end) 
    "Counts matches of graph in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ip beg end (interactive-p)))

(defun ratio-graph-in-ip-atpt (&optional beg end) 
    "Relation of graph in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-ip-atpt (&optional beg end) 
    "Counts matches of lower in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ip beg end (interactive-p)))

(defun ratio-lower-in-ip-atpt (&optional beg end) 
    "Relation of lower in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-ip-atpt (&optional beg end) 
    "Counts matches of punct in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ip beg end (interactive-p)))

(defun ratio-punct-in-ip-atpt (&optional beg end) 
    "Relation of punct in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ip beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-ip-atpt (&optional beg end) 
    "Counts matches of upper in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ip beg end (interactive-p)))

(defun ratio-upper-in-ip-atpt (&optional beg end) 
    "Relation of upper in ip if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ip beg end t (interactive-p)))



;;;###autoload
(defun isbn-atpt (&optional arg) 
  "Returns isbn at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'isbn arg))

(defalias 'bounds-of-isbn-atpt 'isbn-bounds-atpt)
;;;###autoload
(defun isbn-bounds-atpt (&optional arg) 
  "Returns borders of isbn at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'isbn arg))

;;;###autoload
(defun isbn-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'isbn arg))

;;;###autoload
(defun isbn-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-end 'isbn arg))

(defalias 'beginning-of-isbn-atpt 'isbn-beginning-atpt)
;;;###autoload
(defun isbn-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'isbn arg))

(defalias 'end-of-isbn-atpt 'isbn-end-atpt)
;;;###autoload
(defun isbn-end-atpt (&optional arg) 
  "Goto end of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'isbn arg))

(defalias 'length-of-isbn-atpt 'isbn-length-atpt)
;;;###autoload
(defun isbn-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-length 'isbn arg))

(defalias 'copy-isbn-atpt 'isbn-copy-atpt)
;;;###autoload
(defun isbn-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'isbn arg))

(defalias 'escape-isbn-atpt 'isbn-escape-atpt)
;;;###autoload
(defun isbn-escape-atpt () 
  "Makes a copy of symbol or char-class isbn at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'isbn))

(defalias 'doublequote-isbn-atpt 'isbn-doublequote-atpt)
;;;###autoload
(defun isbn-doublequote-atpt (&optional arg) 
  "Doublequotes 'isbn at point if any. "
  (interactive "*p")
    (th-doublequote 'isbn arg))

(defalias 'dollar-isbn-atpt 'isbn-dollar-atpt)
;;;###autoload
(defun isbn-dollar-atpt (&optional arg) 
  "Doublequotes 'isbn at point if any. "
  (interactive "*p")
    (th-dollar 'isbn arg))

(defalias 'angle-greater-isbn-atpt 'isbn-greater-angle-atpt)
;;;###autoload
(defun isbn-greater-angle-atpt (&optional arg) 
  "Sets angles for 'isbn after point if any. "
  (interactive "*p")
    (th-greater-angle 'isbn arg))

(defalias 'angle-lower-isbn-atpt 'isbn-lower-angle-atpt)
;;;###autoload
(defun isbn-lower-angle-atpt (&optional arg) 
  "Sets angles for 'isbn after point if any. "
  (interactive "*p")
    (th-lower-angle 'isbn arg))

(defalias 'backslash-isbn-atpt 'isbn-backslash-atpt)
;;;###autoload
(defun isbn-backslash-atpt (&optional arg) 
  "Backslash 'isbn at point if any. "
  (interactive "*p")
    (th-backslash 'isbn arg))   

(defalias 'brace-isbn-atpt 'isbn-brace-atpt)
;;;###autoload
(defun isbn-brace-atpt (&optional arg) 
  "Braces 'isbn at point if any. "
  (interactive "*p")
    (th-brace 'isbn arg))

(defalias 'bracket-isbn-atpt 'isbn-bracket-atpt)
;;;###autoload
(defun isbn-bracket-atpt (&optional arg) 
  "Brackets 'isbn after point if any. "
  (interactive "*p")
    (th-bracket 'isbn arg))

;; (defalias 'comment-isbn-atpt 'isbn-comment-atpt)
;;;###autoload
(defun isbn-comment-atpt (&optional arg) 
  "Comments 'isbn at point if any. "
  (interactive "*p")
  (th-comment 'isbn arg))

(defalias 'mark-isbn-atpt 'isbn-mark-atpt)
;;;###autoload
(defun isbn-mark-atpt () 
  "Marks 'isbn at point if any. "
  (interactive)
  (th-mark 'isbn))

(defalias 'hide-isbn-atpt 'isbn-hide-atpt)
;;;###autoload
(defun isbn-hide-atpt () 
  "Toggles hide 'isbn in region or buffer if any. "
  (interactive)
  (th-hide 'isbn))

(defalias 'highlight-isbn-atpt-mode 'isbn-highlight-atpt-mode)
(defun isbn-highlight-atpt-mode (&optional arg) 
  "Toggles isbn-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'isbn arg))

(defalias 'kill-isbn-atpt 'isbn-kill-atpt)
;;;###autoload
(defun isbn-kill-atpt (&optional arg) 
  "Kills isbn at point if any. "
  (interactive "*P")
  (th-kill 'isbn arg))

(defalias 'parentize-isbn-atpt 'isbn-parentize-atpt)
;;;###autoload
(defun isbn-parentize-atpt (&optional arg) 
  "Parentizes 'isbn at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'isbn arg))

(defalias 'separate-isbn-atpt 'isbn-separate-atpt)
;;;###autoload
(defun isbn-separate-atpt (&optional arg) 
  "Separates 'isbn at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'isbn arg))

(defalias 'singlequote-isbn-atpt 'isbn-singlequote-atpt)
;;;###autoload
(defun isbn-singlequote-atpt (&optional arg) 
  "Singlequotes 'isbn at point if any. "
  (interactive "*p")
    (th-singlequote 'isbn arg))

(defalias 'slash-isbn-atpt 'isbn-slash-atpt)
;;;###autoload
(defun isbn-slash-atpt (&optional arg) 
  "Slash 'isbn at point if any. "
  (interactive "*p")
    (th-slash 'isbn arg))

(defalias 'trim-isbn-atpt 'isbn-trim-atpt)
;;;###autoload
(defun isbn-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'isbn t t))

(defalias 'trim-isbn-left-atpt 'isbn-left-trim-atpt)
;;;###autoload
(defun isbn-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'isbn t nil))

(defalias 'trim-isbn-right-atpt 'isbn-right-trim-atpt)
;;;###autoload
(defun isbn-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'isbn nil t))

(defalias 'forward-isbn-atpt 'isbn-forward-atpt)
;;;###autoload
(defun isbn-forward-atpt (&optional arg) 
  "Moves forward over isbn at point if any, does nothing otherwise.
Returns bounds of isbn as a list. "
  (interactive "p")
  (th-forward 'isbn arg))

(defalias 'backward-isbn-atpt 'isbn-backward-atpt)
;;;###autoload
(defun isbn-backward-atpt (&optional arg) 
  "Moves backward over isbn before point if any, does nothing otherwise.
Returns bounds of isbn as a list. "
  (interactive "p")
  (th-backward 'isbn arg))

(defalias 'transpose-isbn-atpt 'isbn-transpose-atpt)
;;;###autoload
(defun isbn-transpose-atpt (&optional arg) 
  "Transposes isbn with isbn before point if any. "
  (interactive "*P")
  (th-transpose 'isbn arg))

(defalias 'sort-isbn-atpt 'isbn-sort-atpt)
;;;###autoload
(defun isbn-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts isbn's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'isbn reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-isbn-atpt 'isbn-check-atpt)
;;;###autoload
(defun isbn-check-atpt (&optional arg) 
  " isbn at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "isbn" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "isbn" "-end-position-atpt")))))
    (th-check 'isbn arg)))
   

;;;###autoload
(defun count-alnum-in-isbn-atpt (&optional beg end) 
    "Counts matches of alnum in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'isbn beg end (interactive-p)))

(defun ratio-alnum-in-isbn-atpt (&optional beg end) 
    "Relation of alnum in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-isbn-atpt (&optional beg end) 
    "Counts matches of alpha in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'isbn beg end (interactive-p)))

(defun ratio-alpha-in-isbn-atpt (&optional beg end) 
    "Relation of alpha in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-isbn-atpt (&optional beg end) 
    "Counts matches of ascii in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'isbn beg end (interactive-p)))

(defun ratio-ascii-in-isbn-atpt (&optional beg end) 
    "Relation of ascii in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-isbn-atpt (&optional beg end) 
    "Counts matches of digit in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'isbn beg end (interactive-p)))

(defun ratio-digit-in-isbn-atpt (&optional beg end) 
    "Relation of digit in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-isbn-atpt (&optional beg end) 
    "Counts matches of graph in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'isbn beg end (interactive-p)))

(defun ratio-graph-in-isbn-atpt (&optional beg end) 
    "Relation of graph in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-isbn-atpt (&optional beg end) 
    "Counts matches of lower in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'isbn beg end (interactive-p)))

(defun ratio-lower-in-isbn-atpt (&optional beg end) 
    "Relation of lower in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-isbn-atpt (&optional beg end) 
    "Counts matches of punct in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'isbn beg end (interactive-p)))

(defun ratio-punct-in-isbn-atpt (&optional beg end) 
    "Relation of punct in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'isbn beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-isbn-atpt (&optional beg end) 
    "Counts matches of upper in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'isbn beg end (interactive-p)))

(defun ratio-upper-in-isbn-atpt (&optional beg end) 
    "Relation of upper in isbn if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'isbn beg end t (interactive-p)))



;;;###autoload
(defun line-atpt (&optional arg) 
  "Returns line at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'line arg))

(defalias 'bounds-of-line-atpt 'line-bounds-atpt)
;;;###autoload
(defun line-bounds-atpt (&optional arg) 
  "Returns borders of line at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'line arg))

;;;###autoload
(defun line-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'line arg))

;;;###autoload
(defun line-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-end 'line arg))

(defalias 'beginning-of-line-atpt 'line-beginning-atpt)
;;;###autoload
(defun line-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'line arg))

(defalias 'end-of-line-atpt 'line-end-atpt)
;;;###autoload
(defun line-end-atpt (&optional arg) 
  "Goto end of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'line arg))

(defalias 'length-of-line-atpt 'line-length-atpt)
;;;###autoload
(defun line-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-length 'line arg))

(defalias 'copy-line-atpt 'line-copy-atpt)
;;;###autoload
(defun line-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class line at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'line arg))

(defalias 'escape-line-atpt 'line-escape-atpt)
;;;###autoload
(defun line-escape-atpt () 
  "Makes a copy of symbol or char-class line at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'line))

(defalias 'doublequote-line-atpt 'line-doublequote-atpt)
;;;###autoload
(defun line-doublequote-atpt (&optional arg) 
  "Doublequotes 'line at point if any. "
  (interactive "*p")
    (th-doublequote 'line arg))

(defalias 'dollar-line-atpt 'line-dollar-atpt)
;;;###autoload
(defun line-dollar-atpt (&optional arg) 
  "Doublequotes 'line at point if any. "
  (interactive "*p")
    (th-dollar 'line arg))

(defalias 'angle-greater-line-atpt 'line-greater-angle-atpt)
;;;###autoload
(defun line-greater-angle-atpt (&optional arg) 
  "Sets angles for 'line after point if any. "
  (interactive "*p")
    (th-greater-angle 'line arg))

(defalias 'angle-lower-line-atpt 'line-lower-angle-atpt)
;;;###autoload
(defun line-lower-angle-atpt (&optional arg) 
  "Sets angles for 'line after point if any. "
  (interactive "*p")
    (th-lower-angle 'line arg))

(defalias 'backslash-line-atpt 'line-backslash-atpt)
;;;###autoload
(defun line-backslash-atpt (&optional arg) 
  "Backslash 'line at point if any. "
  (interactive "*p")
    (th-backslash 'line arg))   

(defalias 'brace-line-atpt 'line-brace-atpt)
;;;###autoload
(defun line-brace-atpt (&optional arg) 
  "Braces 'line at point if any. "
  (interactive "*p")
    (th-brace 'line arg))

(defalias 'bracket-line-atpt 'line-bracket-atpt)
;;;###autoload
(defun line-bracket-atpt (&optional arg) 
  "Brackets 'line after point if any. "
  (interactive "*p")
    (th-bracket 'line arg))

;; (defalias 'comment-line-atpt 'line-comment-atpt)
;;;###autoload
(defun line-comment-atpt (&optional arg) 
  "Comments 'line at point if any. "
  (interactive "*p")
  (th-comment 'line arg))

(defalias 'mark-line-atpt 'line-mark-atpt)
;;;###autoload
(defun line-mark-atpt () 
  "Marks 'line at point if any. "
  (interactive)
  (th-mark 'line))

(defalias 'hide-line-atpt 'line-hide-atpt)
;;;###autoload
(defun line-hide-atpt () 
  "Toggles hide 'line in region or buffer if any. "
  (interactive)
  (th-hide 'line))

(defalias 'highlight-line-atpt-mode 'line-highlight-atpt-mode)
(defun line-highlight-atpt-mode (&optional arg) 
  "Toggles line-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'line arg))

(defalias 'kill-line-atpt 'line-kill-atpt)
;;;###autoload
(defun line-kill-atpt (&optional arg) 
  "Kills line at point if any. "
  (interactive "*P")
  (th-kill 'line arg))

(defalias 'parentize-line-atpt 'line-parentize-atpt)
;;;###autoload
(defun line-parentize-atpt (&optional arg) 
  "Parentizes 'line at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'line arg))

(defalias 'separate-line-atpt 'line-separate-atpt)
;;;###autoload
(defun line-separate-atpt (&optional arg) 
  "Separates 'line at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'line arg))

(defalias 'singlequote-line-atpt 'line-singlequote-atpt)
;;;###autoload
(defun line-singlequote-atpt (&optional arg) 
  "Singlequotes 'line at point if any. "
  (interactive "*p")
    (th-singlequote 'line arg))

(defalias 'slash-line-atpt 'line-slash-atpt)
;;;###autoload
(defun line-slash-atpt (&optional arg) 
  "Slash 'line at point if any. "
  (interactive "*p")
    (th-slash 'line arg))

(defalias 'trim-line-atpt 'line-trim-atpt)
;;;###autoload
(defun line-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'line t t))

(defalias 'trim-line-left-atpt 'line-left-trim-atpt)
;;;###autoload
(defun line-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'line t nil))

(defalias 'trim-line-right-atpt 'line-right-trim-atpt)
;;;###autoload
(defun line-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'line nil t))

(defalias 'forward-line-atpt 'line-forward-atpt)
;;;###autoload
(defun line-forward-atpt (&optional arg) 
  "Moves forward over line at point if any, does nothing otherwise.
Returns bounds of line as a list. "
  (interactive "p")
  (th-forward 'line arg))

(defalias 'backward-line-atpt 'line-backward-atpt)
;;;###autoload
(defun line-backward-atpt (&optional arg) 
  "Moves backward over line before point if any, does nothing otherwise.
Returns bounds of line as a list. "
  (interactive "p")
  (th-backward 'line arg))

(defalias 'transpose-line-atpt 'line-transpose-atpt)
;;;###autoload
(defun line-transpose-atpt (&optional arg) 
  "Transposes line with line before point if any. "
  (interactive "*P")
  (th-transpose 'line arg))

(defalias 'sort-line-atpt 'line-sort-atpt)
;;;###autoload
(defun line-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts line's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'line reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-line-atpt 'line-check-atpt)
;;;###autoload
(defun line-check-atpt (&optional arg) 
  " line at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "line" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "line" "-end-position-atpt")))))
    (th-check 'line arg)))
   

;;;###autoload
(defun count-alnum-in-line-atpt (&optional beg end) 
    "Counts matches of alnum in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'line beg end (interactive-p)))

(defun ratio-alnum-in-line-atpt (&optional beg end) 
    "Relation of alnum in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'line beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-line-atpt (&optional beg end) 
    "Counts matches of alpha in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'line beg end (interactive-p)))

(defun ratio-alpha-in-line-atpt (&optional beg end) 
    "Relation of alpha in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'line beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-line-atpt (&optional beg end) 
    "Counts matches of ascii in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'line beg end (interactive-p)))

(defun ratio-ascii-in-line-atpt (&optional beg end) 
    "Relation of ascii in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'line beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-line-atpt (&optional beg end) 
    "Counts matches of digit in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'line beg end (interactive-p)))

(defun ratio-digit-in-line-atpt (&optional beg end) 
    "Relation of digit in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'line beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-line-atpt (&optional beg end) 
    "Counts matches of graph in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'line beg end (interactive-p)))

(defun ratio-graph-in-line-atpt (&optional beg end) 
    "Relation of graph in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'line beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-line-atpt (&optional beg end) 
    "Counts matches of lower in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'line beg end (interactive-p)))

(defun ratio-lower-in-line-atpt (&optional beg end) 
    "Relation of lower in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'line beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-line-atpt (&optional beg end) 
    "Counts matches of punct in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'line beg end (interactive-p)))

(defun ratio-punct-in-line-atpt (&optional beg end) 
    "Relation of punct in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'line beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-line-atpt (&optional beg end) 
    "Counts matches of upper in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'line beg end (interactive-p)))

(defun ratio-upper-in-line-atpt (&optional beg end) 
    "Relation of upper in line if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'line beg end t (interactive-p)))



;;;###autoload
(defun list-atpt (&optional arg) 
  "Returns list at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'list arg))

(defalias 'bounds-of-list-atpt 'list-bounds-atpt)
;;;###autoload
(defun list-bounds-atpt (&optional arg) 
  "Returns borders of list at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'list arg))

;;;###autoload
(defun list-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'list arg))

;;;###autoload
(defun list-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-end 'list arg))

(defalias 'beginning-of-list-atpt 'list-beginning-atpt)
;;;###autoload
(defun list-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'list arg))

(defalias 'end-of-list-atpt 'list-end-atpt)
;;;###autoload
(defun list-end-atpt (&optional arg) 
  "Goto end of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'list arg))

(defalias 'length-of-list-atpt 'list-length-atpt)
;;;###autoload
(defun list-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-length 'list arg))

(defalias 'copy-list-atpt 'list-copy-atpt)
;;;###autoload
(defun list-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class list at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'list arg))

(defalias 'escape-list-atpt 'list-escape-atpt)
;;;###autoload
(defun list-escape-atpt () 
  "Makes a copy of symbol or char-class list at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'list))

(defalias 'doublequote-list-atpt 'list-doublequote-atpt)
;;;###autoload
(defun list-doublequote-atpt (&optional arg) 
  "Doublequotes 'list at point if any. "
  (interactive "*p")
    (th-doublequote 'list arg))

(defalias 'dollar-list-atpt 'list-dollar-atpt)
;;;###autoload
(defun list-dollar-atpt (&optional arg) 
  "Doublequotes 'list at point if any. "
  (interactive "*p")
    (th-dollar 'list arg))

(defalias 'angle-greater-list-atpt 'list-greater-angle-atpt)
;;;###autoload
(defun list-greater-angle-atpt (&optional arg) 
  "Sets angles for 'list after point if any. "
  (interactive "*p")
    (th-greater-angle 'list arg))

(defalias 'angle-lower-list-atpt 'list-lower-angle-atpt)
;;;###autoload
(defun list-lower-angle-atpt (&optional arg) 
  "Sets angles for 'list after point if any. "
  (interactive "*p")
    (th-lower-angle 'list arg))

(defalias 'backslash-list-atpt 'list-backslash-atpt)
;;;###autoload
(defun list-backslash-atpt (&optional arg) 
  "Backslash 'list at point if any. "
  (interactive "*p")
    (th-backslash 'list arg))   

(defalias 'brace-list-atpt 'list-brace-atpt)
;;;###autoload
(defun list-brace-atpt (&optional arg) 
  "Braces 'list at point if any. "
  (interactive "*p")
    (th-brace 'list arg))

(defalias 'bracket-list-atpt 'list-bracket-atpt)
;;;###autoload
(defun list-bracket-atpt (&optional arg) 
  "Brackets 'list after point if any. "
  (interactive "*p")
    (th-bracket 'list arg))

;; (defalias 'comment-list-atpt 'list-comment-atpt)
;;;###autoload
(defun list-comment-atpt (&optional arg) 
  "Comments 'list at point if any. "
  (interactive "*p")
  (th-comment 'list arg))

(defalias 'mark-list-atpt 'list-mark-atpt)
;;;###autoload
(defun list-mark-atpt () 
  "Marks 'list at point if any. "
  (interactive)
  (th-mark 'list))

(defalias 'hide-list-atpt 'list-hide-atpt)
;;;###autoload
(defun list-hide-atpt () 
  "Toggles hide 'list in region or buffer if any. "
  (interactive)
  (th-hide 'list))

(defalias 'highlight-list-atpt-mode 'list-highlight-atpt-mode)
(defun list-highlight-atpt-mode (&optional arg) 
  "Toggles list-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'list arg))

(defalias 'kill-list-atpt 'list-kill-atpt)
;;;###autoload
(defun list-kill-atpt (&optional arg) 
  "Kills list at point if any. "
  (interactive "*P")
  (th-kill 'list arg))

(defalias 'parentize-list-atpt 'list-parentize-atpt)
;;;###autoload
(defun list-parentize-atpt (&optional arg) 
  "Parentizes 'list at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'list arg))

(defalias 'separate-list-atpt 'list-separate-atpt)
;;;###autoload
(defun list-separate-atpt (&optional arg) 
  "Separates 'list at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'list arg))

(defalias 'singlequote-list-atpt 'list-singlequote-atpt)
;;;###autoload
(defun list-singlequote-atpt (&optional arg) 
  "Singlequotes 'list at point if any. "
  (interactive "*p")
    (th-singlequote 'list arg))

(defalias 'slash-list-atpt 'list-slash-atpt)
;;;###autoload
(defun list-slash-atpt (&optional arg) 
  "Slash 'list at point if any. "
  (interactive "*p")
    (th-slash 'list arg))

(defalias 'trim-list-atpt 'list-trim-atpt)
;;;###autoload
(defun list-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'list t t))

(defalias 'trim-list-left-atpt 'list-left-trim-atpt)
;;;###autoload
(defun list-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'list t nil))

(defalias 'trim-list-right-atpt 'list-right-trim-atpt)
;;;###autoload
(defun list-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'list nil t))

(defalias 'forward-list-atpt 'list-forward-atpt)
;;;###autoload
(defun list-forward-atpt (&optional arg) 
  "Moves forward over list at point if any, does nothing otherwise.
Returns bounds of list as a list. "
  (interactive "p")
  (th-forward 'list arg))

(defalias 'backward-list-atpt 'list-backward-atpt)
;;;###autoload
(defun list-backward-atpt (&optional arg) 
  "Moves backward over list before point if any, does nothing otherwise.
Returns bounds of list as a list. "
  (interactive "p")
  (th-backward 'list arg))

(defalias 'transpose-list-atpt 'list-transpose-atpt)
;;;###autoload
(defun list-transpose-atpt (&optional arg) 
  "Transposes list with list before point if any. "
  (interactive "*P")
  (th-transpose 'list arg))

(defalias 'sort-list-atpt 'list-sort-atpt)
;;;###autoload
(defun list-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts list's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'list reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-list-atpt 'list-check-atpt)
;;;###autoload
(defun list-check-atpt (&optional arg) 
  " list at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "list" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "list" "-end-position-atpt")))))
    (th-check 'list arg)))
   

;;;###autoload
(defun count-alnum-in-list-atpt (&optional beg end) 
    "Counts matches of alnum in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'list beg end (interactive-p)))

(defun ratio-alnum-in-list-atpt (&optional beg end) 
    "Relation of alnum in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'list beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-list-atpt (&optional beg end) 
    "Counts matches of alpha in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'list beg end (interactive-p)))

(defun ratio-alpha-in-list-atpt (&optional beg end) 
    "Relation of alpha in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'list beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-list-atpt (&optional beg end) 
    "Counts matches of ascii in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'list beg end (interactive-p)))

(defun ratio-ascii-in-list-atpt (&optional beg end) 
    "Relation of ascii in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'list beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-list-atpt (&optional beg end) 
    "Counts matches of digit in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'list beg end (interactive-p)))

(defun ratio-digit-in-list-atpt (&optional beg end) 
    "Relation of digit in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'list beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-list-atpt (&optional beg end) 
    "Counts matches of graph in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'list beg end (interactive-p)))

(defun ratio-graph-in-list-atpt (&optional beg end) 
    "Relation of graph in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'list beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-list-atpt (&optional beg end) 
    "Counts matches of lower in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'list beg end (interactive-p)))

(defun ratio-lower-in-list-atpt (&optional beg end) 
    "Relation of lower in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'list beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-list-atpt (&optional beg end) 
    "Counts matches of punct in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'list beg end (interactive-p)))

(defun ratio-punct-in-list-atpt (&optional beg end) 
    "Relation of punct in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'list beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-list-atpt (&optional beg end) 
    "Counts matches of upper in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'list beg end (interactive-p)))

(defun ratio-upper-in-list-atpt (&optional beg end) 
    "Relation of upper in list if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'list beg end t (interactive-p)))



;;;###autoload
(defun lower-atpt (&optional arg) 
  "Returns lower at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'lower arg))

(defalias 'bounds-of-lower-atpt 'lower-bounds-atpt)
;;;###autoload
(defun lower-bounds-atpt (&optional arg) 
  "Returns borders of lower at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'lower arg))

;;;###autoload
(defun lower-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'lower arg))

;;;###autoload
(defun lower-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-end 'lower arg))

(defalias 'beginning-of-lower-atpt 'lower-beginning-atpt)
;;;###autoload
(defun lower-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'lower arg))

(defalias 'end-of-lower-atpt 'lower-end-atpt)
;;;###autoload
(defun lower-end-atpt (&optional arg) 
  "Goto end of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'lower arg))

(defalias 'length-of-lower-atpt 'lower-length-atpt)
;;;###autoload
(defun lower-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-length 'lower arg))

(defalias 'copy-lower-atpt 'lower-copy-atpt)
;;;###autoload
(defun lower-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class lower at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'lower arg))

(defalias 'escape-lower-atpt 'lower-escape-atpt)
;;;###autoload
(defun lower-escape-atpt () 
  "Makes a copy of symbol or char-class lower at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'lower))

(defalias 'doublequote-lower-atpt 'lower-doublequote-atpt)
;;;###autoload
(defun lower-doublequote-atpt (&optional arg) 
  "Doublequotes 'lower at point if any. "
  (interactive "*p")
    (th-doublequote 'lower arg))

(defalias 'dollar-lower-atpt 'lower-dollar-atpt)
;;;###autoload
(defun lower-dollar-atpt (&optional arg) 
  "Doublequotes 'lower at point if any. "
  (interactive "*p")
    (th-dollar 'lower arg))

(defalias 'angle-greater-lower-atpt 'lower-greater-angle-atpt)
;;;###autoload
(defun lower-greater-angle-atpt (&optional arg) 
  "Sets angles for 'lower after point if any. "
  (interactive "*p")
    (th-greater-angle 'lower arg))

(defalias 'angle-lower-lower-atpt 'lower-lower-angle-atpt)
;;;###autoload
(defun lower-lower-angle-atpt (&optional arg) 
  "Sets angles for 'lower after point if any. "
  (interactive "*p")
    (th-lower-angle 'lower arg))

(defalias 'backslash-lower-atpt 'lower-backslash-atpt)
;;;###autoload
(defun lower-backslash-atpt (&optional arg) 
  "Backslash 'lower at point if any. "
  (interactive "*p")
    (th-backslash 'lower arg))   

(defalias 'brace-lower-atpt 'lower-brace-atpt)
;;;###autoload
(defun lower-brace-atpt (&optional arg) 
  "Braces 'lower at point if any. "
  (interactive "*p")
    (th-brace 'lower arg))

(defalias 'bracket-lower-atpt 'lower-bracket-atpt)
;;;###autoload
(defun lower-bracket-atpt (&optional arg) 
  "Brackets 'lower after point if any. "
  (interactive "*p")
    (th-bracket 'lower arg))

;; (defalias 'comment-lower-atpt 'lower-comment-atpt)
;;;###autoload
(defun lower-comment-atpt (&optional arg) 
  "Comments 'lower at point if any. "
  (interactive "*p")
  (th-comment 'lower arg))

(defalias 'mark-lower-atpt 'lower-mark-atpt)
;;;###autoload
(defun lower-mark-atpt () 
  "Marks 'lower at point if any. "
  (interactive)
  (th-mark 'lower))

(defalias 'hide-lower-atpt 'lower-hide-atpt)
;;;###autoload
(defun lower-hide-atpt () 
  "Toggles hide 'lower in region or buffer if any. "
  (interactive)
  (th-hide 'lower))

(defalias 'highlight-lower-atpt-mode 'lower-highlight-atpt-mode)
(defun lower-highlight-atpt-mode (&optional arg) 
  "Toggles lower-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'lower arg))

(defalias 'kill-lower-atpt 'lower-kill-atpt)
;;;###autoload
(defun lower-kill-atpt (&optional arg) 
  "Kills lower at point if any. "
  (interactive "*P")
  (th-kill 'lower arg))

(defalias 'parentize-lower-atpt 'lower-parentize-atpt)
;;;###autoload
(defun lower-parentize-atpt (&optional arg) 
  "Parentizes 'lower at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'lower arg))

(defalias 'separate-lower-atpt 'lower-separate-atpt)
;;;###autoload
(defun lower-separate-atpt (&optional arg) 
  "Separates 'lower at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'lower arg))

(defalias 'singlequote-lower-atpt 'lower-singlequote-atpt)
;;;###autoload
(defun lower-singlequote-atpt (&optional arg) 
  "Singlequotes 'lower at point if any. "
  (interactive "*p")
    (th-singlequote 'lower arg))

(defalias 'slash-lower-atpt 'lower-slash-atpt)
;;;###autoload
(defun lower-slash-atpt (&optional arg) 
  "Slash 'lower at point if any. "
  (interactive "*p")
    (th-slash 'lower arg))

(defalias 'trim-lower-atpt 'lower-trim-atpt)
;;;###autoload
(defun lower-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'lower t t))

(defalias 'trim-lower-left-atpt 'lower-left-trim-atpt)
;;;###autoload
(defun lower-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'lower t nil))

(defalias 'trim-lower-right-atpt 'lower-right-trim-atpt)
;;;###autoload
(defun lower-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'lower nil t))

(defalias 'forward-lower-atpt 'lower-forward-atpt)
;;;###autoload
(defun lower-forward-atpt (&optional arg) 
  "Moves forward over lower at point if any, does nothing otherwise.
Returns bounds of lower as a list. "
  (interactive "p")
  (th-forward 'lower arg))

(defalias 'backward-lower-atpt 'lower-backward-atpt)
;;;###autoload
(defun lower-backward-atpt (&optional arg) 
  "Moves backward over lower before point if any, does nothing otherwise.
Returns bounds of lower as a list. "
  (interactive "p")
  (th-backward 'lower arg))

(defalias 'transpose-lower-atpt 'lower-transpose-atpt)
;;;###autoload
(defun lower-transpose-atpt (&optional arg) 
  "Transposes lower with lower before point if any. "
  (interactive "*P")
  (th-transpose 'lower arg))

(defalias 'sort-lower-atpt 'lower-sort-atpt)
;;;###autoload
(defun lower-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts lower's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'lower reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-lower-atpt 'lower-check-atpt)
;;;###autoload
(defun lower-check-atpt (&optional arg) 
  " lower at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "lower" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "lower" "-end-position-atpt")))))
    (th-check 'lower arg)))
   

;;;###autoload
(defun count-alnum-in-lower-atpt (&optional beg end) 
    "Counts matches of alnum in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'lower beg end (interactive-p)))

(defun ratio-alnum-in-lower-atpt (&optional beg end) 
    "Relation of alnum in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-lower-atpt (&optional beg end) 
    "Counts matches of alpha in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'lower beg end (interactive-p)))

(defun ratio-alpha-in-lower-atpt (&optional beg end) 
    "Relation of alpha in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-lower-atpt (&optional beg end) 
    "Counts matches of ascii in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'lower beg end (interactive-p)))

(defun ratio-ascii-in-lower-atpt (&optional beg end) 
    "Relation of ascii in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-lower-atpt (&optional beg end) 
    "Counts matches of digit in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'lower beg end (interactive-p)))

(defun ratio-digit-in-lower-atpt (&optional beg end) 
    "Relation of digit in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-lower-atpt (&optional beg end) 
    "Counts matches of graph in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'lower beg end (interactive-p)))

(defun ratio-graph-in-lower-atpt (&optional beg end) 
    "Relation of graph in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-lower-atpt (&optional beg end) 
    "Counts matches of lower at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'lower beg end (interactive-p)))


;;;###autoload
(defun count-punct-in-lower-atpt (&optional beg end) 
    "Counts matches of punct in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'lower beg end (interactive-p)))

(defun ratio-punct-in-lower-atpt (&optional beg end) 
    "Relation of punct in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'lower beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-lower-atpt (&optional beg end) 
    "Counts matches of upper in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'lower beg end (interactive-p)))

(defun ratio-upper-in-lower-atpt (&optional beg end) 
    "Relation of upper in lower if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'lower beg end t (interactive-p)))



;;;###autoload
(defun ml-text-atpt (&optional arg) 
  "Returns ml-text at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'ml-text arg))

(defalias 'bounds-of-ml-text-atpt 'ml-text-bounds-atpt)
;;;###autoload
(defun ml-text-bounds-atpt (&optional arg) 
  "Returns borders of ml-text at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'ml-text arg))

;;;###autoload
(defun ml-text-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'ml-text arg))

;;;###autoload
(defun ml-text-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-end 'ml-text arg))

(defalias 'beginning-of-ml-text-atpt 'ml-text-beginning-atpt)
;;;###autoload
(defun ml-text-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'ml-text arg))

(defalias 'end-of-ml-text-atpt 'ml-text-end-atpt)
;;;###autoload
(defun ml-text-end-atpt (&optional arg) 
  "Goto end of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'ml-text arg))

(defalias 'length-of-ml-text-atpt 'ml-text-length-atpt)
;;;###autoload
(defun ml-text-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-length 'ml-text arg))

(defalias 'copy-ml-text-atpt 'ml-text-copy-atpt)
;;;###autoload
(defun ml-text-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'ml-text arg))

(defalias 'escape-ml-text-atpt 'ml-text-escape-atpt)
;;;###autoload
(defun ml-text-escape-atpt () 
  "Makes a copy of symbol or char-class ml-text at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'ml-text))

(defalias 'doublequote-ml-text-atpt 'ml-text-doublequote-atpt)
;;;###autoload
(defun ml-text-doublequote-atpt (&optional arg) 
  "Doublequotes 'ml-text at point if any. "
  (interactive "*p")
    (th-doublequote 'ml-text arg))

(defalias 'dollar-ml-text-atpt 'ml-text-dollar-atpt)
;;;###autoload
(defun ml-text-dollar-atpt (&optional arg) 
  "Doublequotes 'ml-text at point if any. "
  (interactive "*p")
    (th-dollar 'ml-text arg))

(defalias 'angle-greater-ml-text-atpt 'ml-text-greater-angle-atpt)
;;;###autoload
(defun ml-text-greater-angle-atpt (&optional arg) 
  "Sets angles for 'ml-text after point if any. "
  (interactive "*p")
    (th-greater-angle 'ml-text arg))

(defalias 'angle-lower-ml-text-atpt 'ml-text-lower-angle-atpt)
;;;###autoload
(defun ml-text-lower-angle-atpt (&optional arg) 
  "Sets angles for 'ml-text after point if any. "
  (interactive "*p")
    (th-lower-angle 'ml-text arg))

(defalias 'backslash-ml-text-atpt 'ml-text-backslash-atpt)
;;;###autoload
(defun ml-text-backslash-atpt (&optional arg) 
  "Backslash 'ml-text at point if any. "
  (interactive "*p")
    (th-backslash 'ml-text arg))   

(defalias 'brace-ml-text-atpt 'ml-text-brace-atpt)
;;;###autoload
(defun ml-text-brace-atpt (&optional arg) 
  "Braces 'ml-text at point if any. "
  (interactive "*p")
    (th-brace 'ml-text arg))

(defalias 'bracket-ml-text-atpt 'ml-text-bracket-atpt)
;;;###autoload
(defun ml-text-bracket-atpt (&optional arg) 
  "Brackets 'ml-text after point if any. "
  (interactive "*p")
    (th-bracket 'ml-text arg))

;; (defalias 'comment-ml-text-atpt 'ml-text-comment-atpt)
;;;###autoload
(defun ml-text-comment-atpt (&optional arg) 
  "Comments 'ml-text at point if any. "
  (interactive "*p")
  (th-comment 'ml-text arg))

(defalias 'mark-ml-text-atpt 'ml-text-mark-atpt)
;;;###autoload
(defun ml-text-mark-atpt () 
  "Marks 'ml-text at point if any. "
  (interactive)
  (th-mark 'ml-text))

(defalias 'hide-ml-text-atpt 'ml-text-hide-atpt)
;;;###autoload
(defun ml-text-hide-atpt () 
  "Toggles hide 'ml-text in region or buffer if any. "
  (interactive)
  (th-hide 'ml-text))

(defalias 'highlight-ml-text-atpt-mode 'ml-text-highlight-atpt-mode)
(defun ml-text-highlight-atpt-mode (&optional arg) 
  "Toggles ml-text-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'ml-text arg))

(defalias 'kill-ml-text-atpt 'ml-text-kill-atpt)
;;;###autoload
(defun ml-text-kill-atpt (&optional arg) 
  "Kills ml-text at point if any. "
  (interactive "*P")
  (th-kill 'ml-text arg))

(defalias 'parentize-ml-text-atpt 'ml-text-parentize-atpt)
;;;###autoload
(defun ml-text-parentize-atpt (&optional arg) 
  "Parentizes 'ml-text at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'ml-text arg))

(defalias 'separate-ml-text-atpt 'ml-text-separate-atpt)
;;;###autoload
(defun ml-text-separate-atpt (&optional arg) 
  "Separates 'ml-text at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'ml-text arg))

(defalias 'singlequote-ml-text-atpt 'ml-text-singlequote-atpt)
;;;###autoload
(defun ml-text-singlequote-atpt (&optional arg) 
  "Singlequotes 'ml-text at point if any. "
  (interactive "*p")
    (th-singlequote 'ml-text arg))

(defalias 'slash-ml-text-atpt 'ml-text-slash-atpt)
;;;###autoload
(defun ml-text-slash-atpt (&optional arg) 
  "Slash 'ml-text at point if any. "
  (interactive "*p")
    (th-slash 'ml-text arg))

(defalias 'trim-ml-text-atpt 'ml-text-trim-atpt)
;;;###autoload
(defun ml-text-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ml-text t t))

(defalias 'trim-ml-text-left-atpt 'ml-text-left-trim-atpt)
;;;###autoload
(defun ml-text-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ml-text t nil))

(defalias 'trim-ml-text-right-atpt 'ml-text-right-trim-atpt)
;;;###autoload
(defun ml-text-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'ml-text nil t))

(defalias 'forward-ml-text-atpt 'ml-text-forward-atpt)
;;;###autoload
(defun ml-text-forward-atpt (&optional arg) 
  "Moves forward over ml-text at point if any, does nothing otherwise.
Returns bounds of ml-text as a list. "
  (interactive "p")
  (th-forward 'ml-text arg))

(defalias 'backward-ml-text-atpt 'ml-text-backward-atpt)
;;;###autoload
(defun ml-text-backward-atpt (&optional arg) 
  "Moves backward over ml-text before point if any, does nothing otherwise.
Returns bounds of ml-text as a list. "
  (interactive "p")
  (th-backward 'ml-text arg))

(defalias 'transpose-ml-text-atpt 'ml-text-transpose-atpt)
;;;###autoload
(defun ml-text-transpose-atpt (&optional arg) 
  "Transposes ml-text with ml-text before point if any. "
  (interactive "*P")
  (th-transpose 'ml-text arg))

(defalias 'sort-ml-text-atpt 'ml-text-sort-atpt)
;;;###autoload
(defun ml-text-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts ml-text's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'ml-text reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-ml-text-atpt 'ml-text-check-atpt)
;;;###autoload
(defun ml-text-check-atpt (&optional arg) 
  " ml-text at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "ml-text" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "ml-text" "-end-position-atpt")))))
    (th-check 'ml-text arg)))
   

;;;###autoload
(defun count-alnum-in-ml-text-atpt (&optional beg end) 
    "Counts matches of alnum in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ml-text beg end (interactive-p)))

(defun ratio-alnum-in-ml-text-atpt (&optional beg end) 
    "Relation of alnum in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-ml-text-atpt (&optional beg end) 
    "Counts matches of alpha in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ml-text beg end (interactive-p)))

(defun ratio-alpha-in-ml-text-atpt (&optional beg end) 
    "Relation of alpha in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-ml-text-atpt (&optional beg end) 
    "Counts matches of ascii in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'ml-text beg end (interactive-p)))

(defun ratio-ascii-in-ml-text-atpt (&optional beg end) 
    "Relation of ascii in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-ml-text-atpt (&optional beg end) 
    "Counts matches of digit in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ml-text beg end (interactive-p)))

(defun ratio-digit-in-ml-text-atpt (&optional beg end) 
    "Relation of digit in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-ml-text-atpt (&optional beg end) 
    "Counts matches of graph in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ml-text beg end (interactive-p)))

(defun ratio-graph-in-ml-text-atpt (&optional beg end) 
    "Relation of graph in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-ml-text-atpt (&optional beg end) 
    "Counts matches of lower in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ml-text beg end (interactive-p)))

(defun ratio-lower-in-ml-text-atpt (&optional beg end) 
    "Relation of lower in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-ml-text-atpt (&optional beg end) 
    "Counts matches of punct in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ml-text beg end (interactive-p)))

(defun ratio-punct-in-ml-text-atpt (&optional beg end) 
    "Relation of punct in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-ml-text-atpt (&optional beg end) 
    "Counts matches of upper in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ml-text beg end (interactive-p)))

(defun ratio-upper-in-ml-text-atpt (&optional beg end) 
    "Relation of upper in ml-text if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'ml-text beg end t (interactive-p)))



;;;###autoload
(defun multibyte-atpt (&optional arg) 
  "Returns multibyte at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'multibyte arg))

(defalias 'bounds-of-multibyte-atpt 'multibyte-bounds-atpt)
;;;###autoload
(defun multibyte-bounds-atpt (&optional arg) 
  "Returns borders of multibyte at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'multibyte arg))

;;;###autoload
(defun multibyte-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'multibyte arg))

;;;###autoload
(defun multibyte-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-end 'multibyte arg))

(defalias 'beginning-of-multibyte-atpt 'multibyte-beginning-atpt)
;;;###autoload
(defun multibyte-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'multibyte arg))

(defalias 'end-of-multibyte-atpt 'multibyte-end-atpt)
;;;###autoload
(defun multibyte-end-atpt (&optional arg) 
  "Goto end of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'multibyte arg))

(defalias 'length-of-multibyte-atpt 'multibyte-length-atpt)
;;;###autoload
(defun multibyte-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-length 'multibyte arg))

(defalias 'copy-multibyte-atpt 'multibyte-copy-atpt)
;;;###autoload
(defun multibyte-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'multibyte arg))

(defalias 'escape-multibyte-atpt 'multibyte-escape-atpt)
;;;###autoload
(defun multibyte-escape-atpt () 
  "Makes a copy of symbol or char-class multibyte at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'multibyte))

(defalias 'doublequote-multibyte-atpt 'multibyte-doublequote-atpt)
;;;###autoload
(defun multibyte-doublequote-atpt (&optional arg) 
  "Doublequotes 'multibyte at point if any. "
  (interactive "*p")
    (th-doublequote 'multibyte arg))

(defalias 'dollar-multibyte-atpt 'multibyte-dollar-atpt)
;;;###autoload
(defun multibyte-dollar-atpt (&optional arg) 
  "Doublequotes 'multibyte at point if any. "
  (interactive "*p")
    (th-dollar 'multibyte arg))

(defalias 'angle-greater-multibyte-atpt 'multibyte-greater-angle-atpt)
;;;###autoload
(defun multibyte-greater-angle-atpt (&optional arg) 
  "Sets angles for 'multibyte after point if any. "
  (interactive "*p")
    (th-greater-angle 'multibyte arg))

(defalias 'angle-lower-multibyte-atpt 'multibyte-lower-angle-atpt)
;;;###autoload
(defun multibyte-lower-angle-atpt (&optional arg) 
  "Sets angles for 'multibyte after point if any. "
  (interactive "*p")
    (th-lower-angle 'multibyte arg))

(defalias 'backslash-multibyte-atpt 'multibyte-backslash-atpt)
;;;###autoload
(defun multibyte-backslash-atpt (&optional arg) 
  "Backslash 'multibyte at point if any. "
  (interactive "*p")
    (th-backslash 'multibyte arg))   

(defalias 'brace-multibyte-atpt 'multibyte-brace-atpt)
;;;###autoload
(defun multibyte-brace-atpt (&optional arg) 
  "Braces 'multibyte at point if any. "
  (interactive "*p")
    (th-brace 'multibyte arg))

(defalias 'bracket-multibyte-atpt 'multibyte-bracket-atpt)
;;;###autoload
(defun multibyte-bracket-atpt (&optional arg) 
  "Brackets 'multibyte after point if any. "
  (interactive "*p")
    (th-bracket 'multibyte arg))

;; (defalias 'comment-multibyte-atpt 'multibyte-comment-atpt)
;;;###autoload
(defun multibyte-comment-atpt (&optional arg) 
  "Comments 'multibyte at point if any. "
  (interactive "*p")
  (th-comment 'multibyte arg))

(defalias 'mark-multibyte-atpt 'multibyte-mark-atpt)
;;;###autoload
(defun multibyte-mark-atpt () 
  "Marks 'multibyte at point if any. "
  (interactive)
  (th-mark 'multibyte))

(defalias 'hide-multibyte-atpt 'multibyte-hide-atpt)
;;;###autoload
(defun multibyte-hide-atpt () 
  "Toggles hide 'multibyte in region or buffer if any. "
  (interactive)
  (th-hide 'multibyte))

(defalias 'highlight-multibyte-atpt-mode 'multibyte-highlight-atpt-mode)
(defun multibyte-highlight-atpt-mode (&optional arg) 
  "Toggles multibyte-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'multibyte arg))

(defalias 'kill-multibyte-atpt 'multibyte-kill-atpt)
;;;###autoload
(defun multibyte-kill-atpt (&optional arg) 
  "Kills multibyte at point if any. "
  (interactive "*P")
  (th-kill 'multibyte arg))

(defalias 'parentize-multibyte-atpt 'multibyte-parentize-atpt)
;;;###autoload
(defun multibyte-parentize-atpt (&optional arg) 
  "Parentizes 'multibyte at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'multibyte arg))

(defalias 'separate-multibyte-atpt 'multibyte-separate-atpt)
;;;###autoload
(defun multibyte-separate-atpt (&optional arg) 
  "Separates 'multibyte at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'multibyte arg))

(defalias 'singlequote-multibyte-atpt 'multibyte-singlequote-atpt)
;;;###autoload
(defun multibyte-singlequote-atpt (&optional arg) 
  "Singlequotes 'multibyte at point if any. "
  (interactive "*p")
    (th-singlequote 'multibyte arg))

(defalias 'slash-multibyte-atpt 'multibyte-slash-atpt)
;;;###autoload
(defun multibyte-slash-atpt (&optional arg) 
  "Slash 'multibyte at point if any. "
  (interactive "*p")
    (th-slash 'multibyte arg))

(defalias 'trim-multibyte-atpt 'multibyte-trim-atpt)
;;;###autoload
(defun multibyte-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'multibyte t t))

(defalias 'trim-multibyte-left-atpt 'multibyte-left-trim-atpt)
;;;###autoload
(defun multibyte-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'multibyte t nil))

(defalias 'trim-multibyte-right-atpt 'multibyte-right-trim-atpt)
;;;###autoload
(defun multibyte-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'multibyte nil t))

(defalias 'forward-multibyte-atpt 'multibyte-forward-atpt)
;;;###autoload
(defun multibyte-forward-atpt (&optional arg) 
  "Moves forward over multibyte at point if any, does nothing otherwise.
Returns bounds of multibyte as a list. "
  (interactive "p")
  (th-forward 'multibyte arg))

(defalias 'backward-multibyte-atpt 'multibyte-backward-atpt)
;;;###autoload
(defun multibyte-backward-atpt (&optional arg) 
  "Moves backward over multibyte before point if any, does nothing otherwise.
Returns bounds of multibyte as a list. "
  (interactive "p")
  (th-backward 'multibyte arg))

(defalias 'transpose-multibyte-atpt 'multibyte-transpose-atpt)
;;;###autoload
(defun multibyte-transpose-atpt (&optional arg) 
  "Transposes multibyte with multibyte before point if any. "
  (interactive "*P")
  (th-transpose 'multibyte arg))

(defalias 'sort-multibyte-atpt 'multibyte-sort-atpt)
;;;###autoload
(defun multibyte-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts multibyte's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'multibyte reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-multibyte-atpt 'multibyte-check-atpt)
;;;###autoload
(defun multibyte-check-atpt (&optional arg) 
  " multibyte at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "multibyte" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "multibyte" "-end-position-atpt")))))
    (th-check 'multibyte arg)))
   

;;;###autoload
(defun nonascii-atpt (&optional arg) 
  "Returns nonascii at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'nonascii arg))

(defalias 'bounds-of-nonascii-atpt 'nonascii-bounds-atpt)
;;;###autoload
(defun nonascii-bounds-atpt (&optional arg) 
  "Returns borders of nonascii at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'nonascii arg))

;;;###autoload
(defun nonascii-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'nonascii arg))

;;;###autoload
(defun nonascii-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-end 'nonascii arg))

(defalias 'beginning-of-nonascii-atpt 'nonascii-beginning-atpt)
;;;###autoload
(defun nonascii-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'nonascii arg))

(defalias 'end-of-nonascii-atpt 'nonascii-end-atpt)
;;;###autoload
(defun nonascii-end-atpt (&optional arg) 
  "Goto end of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'nonascii arg))

(defalias 'length-of-nonascii-atpt 'nonascii-length-atpt)
;;;###autoload
(defun nonascii-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-length 'nonascii arg))

(defalias 'copy-nonascii-atpt 'nonascii-copy-atpt)
;;;###autoload
(defun nonascii-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'nonascii arg))

(defalias 'escape-nonascii-atpt 'nonascii-escape-atpt)
;;;###autoload
(defun nonascii-escape-atpt () 
  "Makes a copy of symbol or char-class nonascii at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'nonascii))

(defalias 'doublequote-nonascii-atpt 'nonascii-doublequote-atpt)
;;;###autoload
(defun nonascii-doublequote-atpt (&optional arg) 
  "Doublequotes 'nonascii at point if any. "
  (interactive "*p")
    (th-doublequote 'nonascii arg))

(defalias 'dollar-nonascii-atpt 'nonascii-dollar-atpt)
;;;###autoload
(defun nonascii-dollar-atpt (&optional arg) 
  "Doublequotes 'nonascii at point if any. "
  (interactive "*p")
    (th-dollar 'nonascii arg))

(defalias 'angle-greater-nonascii-atpt 'nonascii-greater-angle-atpt)
;;;###autoload
(defun nonascii-greater-angle-atpt (&optional arg) 
  "Sets angles for 'nonascii after point if any. "
  (interactive "*p")
    (th-greater-angle 'nonascii arg))

(defalias 'angle-lower-nonascii-atpt 'nonascii-lower-angle-atpt)
;;;###autoload
(defun nonascii-lower-angle-atpt (&optional arg) 
  "Sets angles for 'nonascii after point if any. "
  (interactive "*p")
    (th-lower-angle 'nonascii arg))

(defalias 'backslash-nonascii-atpt 'nonascii-backslash-atpt)
;;;###autoload
(defun nonascii-backslash-atpt (&optional arg) 
  "Backslash 'nonascii at point if any. "
  (interactive "*p")
    (th-backslash 'nonascii arg))   

(defalias 'brace-nonascii-atpt 'nonascii-brace-atpt)
;;;###autoload
(defun nonascii-brace-atpt (&optional arg) 
  "Braces 'nonascii at point if any. "
  (interactive "*p")
    (th-brace 'nonascii arg))

(defalias 'bracket-nonascii-atpt 'nonascii-bracket-atpt)
;;;###autoload
(defun nonascii-bracket-atpt (&optional arg) 
  "Brackets 'nonascii after point if any. "
  (interactive "*p")
    (th-bracket 'nonascii arg))

;; (defalias 'comment-nonascii-atpt 'nonascii-comment-atpt)
;;;###autoload
(defun nonascii-comment-atpt (&optional arg) 
  "Comments 'nonascii at point if any. "
  (interactive "*p")
  (th-comment 'nonascii arg))

(defalias 'mark-nonascii-atpt 'nonascii-mark-atpt)
;;;###autoload
(defun nonascii-mark-atpt () 
  "Marks 'nonascii at point if any. "
  (interactive)
  (th-mark 'nonascii))

(defalias 'hide-nonascii-atpt 'nonascii-hide-atpt)
;;;###autoload
(defun nonascii-hide-atpt () 
  "Toggles hide 'nonascii in region or buffer if any. "
  (interactive)
  (th-hide 'nonascii))

(defalias 'highlight-nonascii-atpt-mode 'nonascii-highlight-atpt-mode)
(defun nonascii-highlight-atpt-mode (&optional arg) 
  "Toggles nonascii-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'nonascii arg))

(defalias 'kill-nonascii-atpt 'nonascii-kill-atpt)
;;;###autoload
(defun nonascii-kill-atpt (&optional arg) 
  "Kills nonascii at point if any. "
  (interactive "*P")
  (th-kill 'nonascii arg))

(defalias 'parentize-nonascii-atpt 'nonascii-parentize-atpt)
;;;###autoload
(defun nonascii-parentize-atpt (&optional arg) 
  "Parentizes 'nonascii at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'nonascii arg))

(defalias 'separate-nonascii-atpt 'nonascii-separate-atpt)
;;;###autoload
(defun nonascii-separate-atpt (&optional arg) 
  "Separates 'nonascii at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'nonascii arg))

(defalias 'singlequote-nonascii-atpt 'nonascii-singlequote-atpt)
;;;###autoload
(defun nonascii-singlequote-atpt (&optional arg) 
  "Singlequotes 'nonascii at point if any. "
  (interactive "*p")
    (th-singlequote 'nonascii arg))

(defalias 'slash-nonascii-atpt 'nonascii-slash-atpt)
;;;###autoload
(defun nonascii-slash-atpt (&optional arg) 
  "Slash 'nonascii at point if any. "
  (interactive "*p")
    (th-slash 'nonascii arg))

(defalias 'trim-nonascii-atpt 'nonascii-trim-atpt)
;;;###autoload
(defun nonascii-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'nonascii t t))

(defalias 'trim-nonascii-left-atpt 'nonascii-left-trim-atpt)
;;;###autoload
(defun nonascii-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'nonascii t nil))

(defalias 'trim-nonascii-right-atpt 'nonascii-right-trim-atpt)
;;;###autoload
(defun nonascii-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'nonascii nil t))

(defalias 'forward-nonascii-atpt 'nonascii-forward-atpt)
;;;###autoload
(defun nonascii-forward-atpt (&optional arg) 
  "Moves forward over nonascii at point if any, does nothing otherwise.
Returns bounds of nonascii as a list. "
  (interactive "p")
  (th-forward 'nonascii arg))

(defalias 'backward-nonascii-atpt 'nonascii-backward-atpt)
;;;###autoload
(defun nonascii-backward-atpt (&optional arg) 
  "Moves backward over nonascii before point if any, does nothing otherwise.
Returns bounds of nonascii as a list. "
  (interactive "p")
  (th-backward 'nonascii arg))

(defalias 'transpose-nonascii-atpt 'nonascii-transpose-atpt)
;;;###autoload
(defun nonascii-transpose-atpt (&optional arg) 
  "Transposes nonascii with nonascii before point if any. "
  (interactive "*P")
  (th-transpose 'nonascii arg))

(defalias 'sort-nonascii-atpt 'nonascii-sort-atpt)
;;;###autoload
(defun nonascii-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts nonascii's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'nonascii reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-nonascii-atpt 'nonascii-check-atpt)
;;;###autoload
(defun nonascii-check-atpt (&optional arg) 
  " nonascii at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "nonascii" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "nonascii" "-end-position-atpt")))))
    (th-check 'nonascii arg)))
   

;;;###autoload
(defun number-atpt (&optional arg) 
  "Returns number at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'number arg))

(defalias 'bounds-of-number-atpt 'number-bounds-atpt)
;;;###autoload
(defun number-bounds-atpt (&optional arg) 
  "Returns borders of number at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'number arg))

;;;###autoload
(defun number-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'number arg))

;;;###autoload
(defun number-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-end 'number arg))

(defalias 'beginning-of-number-atpt 'number-beginning-atpt)
;;;###autoload
(defun number-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'number arg))

(defalias 'end-of-number-atpt 'number-end-atpt)
;;;###autoload
(defun number-end-atpt (&optional arg) 
  "Goto end of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'number arg))

(defalias 'length-of-number-atpt 'number-length-atpt)
;;;###autoload
(defun number-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-length 'number arg))

(defalias 'copy-number-atpt 'number-copy-atpt)
;;;###autoload
(defun number-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class number at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'number arg))

(defalias 'escape-number-atpt 'number-escape-atpt)
;;;###autoload
(defun number-escape-atpt () 
  "Makes a copy of symbol or char-class number at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'number))

(defalias 'doublequote-number-atpt 'number-doublequote-atpt)
;;;###autoload
(defun number-doublequote-atpt (&optional arg) 
  "Doublequotes 'number at point if any. "
  (interactive "*p")
    (th-doublequote 'number arg))

(defalias 'dollar-number-atpt 'number-dollar-atpt)
;;;###autoload
(defun number-dollar-atpt (&optional arg) 
  "Doublequotes 'number at point if any. "
  (interactive "*p")
    (th-dollar 'number arg))

(defalias 'angle-greater-number-atpt 'number-greater-angle-atpt)
;;;###autoload
(defun number-greater-angle-atpt (&optional arg) 
  "Sets angles for 'number after point if any. "
  (interactive "*p")
    (th-greater-angle 'number arg))

(defalias 'angle-lower-number-atpt 'number-lower-angle-atpt)
;;;###autoload
(defun number-lower-angle-atpt (&optional arg) 
  "Sets angles for 'number after point if any. "
  (interactive "*p")
    (th-lower-angle 'number arg))

(defalias 'backslash-number-atpt 'number-backslash-atpt)
;;;###autoload
(defun number-backslash-atpt (&optional arg) 
  "Backslash 'number at point if any. "
  (interactive "*p")
    (th-backslash 'number arg))   

(defalias 'brace-number-atpt 'number-brace-atpt)
;;;###autoload
(defun number-brace-atpt (&optional arg) 
  "Braces 'number at point if any. "
  (interactive "*p")
    (th-brace 'number arg))

(defalias 'bracket-number-atpt 'number-bracket-atpt)
;;;###autoload
(defun number-bracket-atpt (&optional arg) 
  "Brackets 'number after point if any. "
  (interactive "*p")
    (th-bracket 'number arg))

;; (defalias 'comment-number-atpt 'number-comment-atpt)
;;;###autoload
(defun number-comment-atpt (&optional arg) 
  "Comments 'number at point if any. "
  (interactive "*p")
  (th-comment 'number arg))

(defalias 'mark-number-atpt 'number-mark-atpt)
;;;###autoload
(defun number-mark-atpt () 
  "Marks 'number at point if any. "
  (interactive)
  (th-mark 'number))

(defalias 'hide-number-atpt 'number-hide-atpt)
;;;###autoload
(defun number-hide-atpt () 
  "Toggles hide 'number in region or buffer if any. "
  (interactive)
  (th-hide 'number))

(defalias 'highlight-number-atpt-mode 'number-highlight-atpt-mode)
(defun number-highlight-atpt-mode (&optional arg) 
  "Toggles number-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'number arg))

(defalias 'kill-number-atpt 'number-kill-atpt)
;;;###autoload
(defun number-kill-atpt (&optional arg) 
  "Kills number at point if any. "
  (interactive "*P")
  (th-kill 'number arg))

(defalias 'parentize-number-atpt 'number-parentize-atpt)
;;;###autoload
(defun number-parentize-atpt (&optional arg) 
  "Parentizes 'number at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'number arg))

(defalias 'separate-number-atpt 'number-separate-atpt)
;;;###autoload
(defun number-separate-atpt (&optional arg) 
  "Separates 'number at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'number arg))

(defalias 'singlequote-number-atpt 'number-singlequote-atpt)
;;;###autoload
(defun number-singlequote-atpt (&optional arg) 
  "Singlequotes 'number at point if any. "
  (interactive "*p")
    (th-singlequote 'number arg))

(defalias 'slash-number-atpt 'number-slash-atpt)
;;;###autoload
(defun number-slash-atpt (&optional arg) 
  "Slash 'number at point if any. "
  (interactive "*p")
    (th-slash 'number arg))

(defalias 'trim-number-atpt 'number-trim-atpt)
;;;###autoload
(defun number-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'number t t))

(defalias 'trim-number-left-atpt 'number-left-trim-atpt)
;;;###autoload
(defun number-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'number t nil))

(defalias 'trim-number-right-atpt 'number-right-trim-atpt)
;;;###autoload
(defun number-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'number nil t))

(defalias 'forward-number-atpt 'number-forward-atpt)
;;;###autoload
(defun number-forward-atpt (&optional arg) 
  "Moves forward over number at point if any, does nothing otherwise.
Returns bounds of number as a list. "
  (interactive "p")
  (th-forward 'number arg))

(defalias 'backward-number-atpt 'number-backward-atpt)
;;;###autoload
(defun number-backward-atpt (&optional arg) 
  "Moves backward over number before point if any, does nothing otherwise.
Returns bounds of number as a list. "
  (interactive "p")
  (th-backward 'number arg))

(defalias 'transpose-number-atpt 'number-transpose-atpt)
;;;###autoload
(defun number-transpose-atpt (&optional arg) 
  "Transposes number with number before point if any. "
  (interactive "*P")
  (th-transpose 'number arg))

(defalias 'sort-number-atpt 'number-sort-atpt)
;;;###autoload
(defun number-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts number's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'number reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-number-atpt 'number-check-atpt)
;;;###autoload
(defun number-check-atpt (&optional arg) 
  " number at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "number" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "number" "-end-position-atpt")))))
    (th-check 'number arg)))
   

;;;###autoload
(defun count-alnum-in-number-atpt (&optional beg end) 
    "Counts matches of alnum in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'number beg end (interactive-p)))

(defun ratio-alnum-in-number-atpt (&optional beg end) 
    "Relation of alnum in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'number beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-number-atpt (&optional beg end) 
    "Counts matches of alpha in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'number beg end (interactive-p)))

(defun ratio-alpha-in-number-atpt (&optional beg end) 
    "Relation of alpha in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'number beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-number-atpt (&optional beg end) 
    "Counts matches of ascii in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'number beg end (interactive-p)))

(defun ratio-ascii-in-number-atpt (&optional beg end) 
    "Relation of ascii in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'number beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-number-atpt (&optional beg end) 
    "Counts matches of digit in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'number beg end (interactive-p)))

(defun ratio-digit-in-number-atpt (&optional beg end) 
    "Relation of digit in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'number beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-number-atpt (&optional beg end) 
    "Counts matches of graph in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'number beg end (interactive-p)))

(defun ratio-graph-in-number-atpt (&optional beg end) 
    "Relation of graph in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'number beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-number-atpt (&optional beg end) 
    "Counts matches of lower in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'number beg end (interactive-p)))

(defun ratio-lower-in-number-atpt (&optional beg end) 
    "Relation of lower in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'number beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-number-atpt (&optional beg end) 
    "Counts matches of punct in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'number beg end (interactive-p)))

(defun ratio-punct-in-number-atpt (&optional beg end) 
    "Relation of punct in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'number beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-number-atpt (&optional beg end) 
    "Counts matches of upper in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'number beg end (interactive-p)))

(defun ratio-upper-in-number-atpt (&optional beg end) 
    "Relation of upper in number if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'number beg end t (interactive-p)))



;;;###autoload
(defun page-atpt (&optional arg) 
  "Returns page at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'page arg))

(defalias 'bounds-of-page-atpt 'page-bounds-atpt)
;;;###autoload
(defun page-bounds-atpt (&optional arg) 
  "Returns borders of page at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'page arg))

;;;###autoload
(defun page-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'page arg))

;;;###autoload
(defun page-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-end 'page arg))

(defalias 'beginning-of-page-atpt 'page-beginning-atpt)
;;;###autoload
(defun page-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'page arg))

(defalias 'end-of-page-atpt 'page-end-atpt)
;;;###autoload
(defun page-end-atpt (&optional arg) 
  "Goto end of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'page arg))

(defalias 'length-of-page-atpt 'page-length-atpt)
;;;###autoload
(defun page-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-length 'page arg))

(defalias 'copy-page-atpt 'page-copy-atpt)
;;;###autoload
(defun page-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class page at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'page arg))

(defalias 'escape-page-atpt 'page-escape-atpt)
;;;###autoload
(defun page-escape-atpt () 
  "Makes a copy of symbol or char-class page at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'page))

(defalias 'doublequote-page-atpt 'page-doublequote-atpt)
;;;###autoload
(defun page-doublequote-atpt (&optional arg) 
  "Doublequotes 'page at point if any. "
  (interactive "*p")
    (th-doublequote 'page arg))

(defalias 'dollar-page-atpt 'page-dollar-atpt)
;;;###autoload
(defun page-dollar-atpt (&optional arg) 
  "Doublequotes 'page at point if any. "
  (interactive "*p")
    (th-dollar 'page arg))

(defalias 'angle-greater-page-atpt 'page-greater-angle-atpt)
;;;###autoload
(defun page-greater-angle-atpt (&optional arg) 
  "Sets angles for 'page after point if any. "
  (interactive "*p")
    (th-greater-angle 'page arg))

(defalias 'angle-lower-page-atpt 'page-lower-angle-atpt)
;;;###autoload
(defun page-lower-angle-atpt (&optional arg) 
  "Sets angles for 'page after point if any. "
  (interactive "*p")
    (th-lower-angle 'page arg))

(defalias 'backslash-page-atpt 'page-backslash-atpt)
;;;###autoload
(defun page-backslash-atpt (&optional arg) 
  "Backslash 'page at point if any. "
  (interactive "*p")
    (th-backslash 'page arg))   

(defalias 'brace-page-atpt 'page-brace-atpt)
;;;###autoload
(defun page-brace-atpt (&optional arg) 
  "Braces 'page at point if any. "
  (interactive "*p")
    (th-brace 'page arg))

(defalias 'bracket-page-atpt 'page-bracket-atpt)
;;;###autoload
(defun page-bracket-atpt (&optional arg) 
  "Brackets 'page after point if any. "
  (interactive "*p")
    (th-bracket 'page arg))

;; (defalias 'comment-page-atpt 'page-comment-atpt)
;;;###autoload
(defun page-comment-atpt (&optional arg) 
  "Comments 'page at point if any. "
  (interactive "*p")
  (th-comment 'page arg))

(defalias 'mark-page-atpt 'page-mark-atpt)
;;;###autoload
(defun page-mark-atpt () 
  "Marks 'page at point if any. "
  (interactive)
  (th-mark 'page))

(defalias 'hide-page-atpt 'page-hide-atpt)
;;;###autoload
(defun page-hide-atpt () 
  "Toggles hide 'page in region or buffer if any. "
  (interactive)
  (th-hide 'page))

(defalias 'highlight-page-atpt-mode 'page-highlight-atpt-mode)
(defun page-highlight-atpt-mode (&optional arg) 
  "Toggles page-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'page arg))

(defalias 'kill-page-atpt 'page-kill-atpt)
;;;###autoload
(defun page-kill-atpt (&optional arg) 
  "Kills page at point if any. "
  (interactive "*P")
  (th-kill 'page arg))

(defalias 'parentize-page-atpt 'page-parentize-atpt)
;;;###autoload
(defun page-parentize-atpt (&optional arg) 
  "Parentizes 'page at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'page arg))

(defalias 'separate-page-atpt 'page-separate-atpt)
;;;###autoload
(defun page-separate-atpt (&optional arg) 
  "Separates 'page at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'page arg))

(defalias 'singlequote-page-atpt 'page-singlequote-atpt)
;;;###autoload
(defun page-singlequote-atpt (&optional arg) 
  "Singlequotes 'page at point if any. "
  (interactive "*p")
    (th-singlequote 'page arg))

(defalias 'slash-page-atpt 'page-slash-atpt)
;;;###autoload
(defun page-slash-atpt (&optional arg) 
  "Slash 'page at point if any. "
  (interactive "*p")
    (th-slash 'page arg))

(defalias 'trim-page-atpt 'page-trim-atpt)
;;;###autoload
(defun page-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'page t t))

(defalias 'trim-page-left-atpt 'page-left-trim-atpt)
;;;###autoload
(defun page-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'page t nil))

(defalias 'trim-page-right-atpt 'page-right-trim-atpt)
;;;###autoload
(defun page-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'page nil t))

(defalias 'forward-page-atpt 'page-forward-atpt)
;;;###autoload
(defun page-forward-atpt (&optional arg) 
  "Moves forward over page at point if any, does nothing otherwise.
Returns bounds of page as a list. "
  (interactive "p")
  (th-forward 'page arg))

(defalias 'backward-page-atpt 'page-backward-atpt)
;;;###autoload
(defun page-backward-atpt (&optional arg) 
  "Moves backward over page before point if any, does nothing otherwise.
Returns bounds of page as a list. "
  (interactive "p")
  (th-backward 'page arg))

(defalias 'transpose-page-atpt 'page-transpose-atpt)
;;;###autoload
(defun page-transpose-atpt (&optional arg) 
  "Transposes page with page before point if any. "
  (interactive "*P")
  (th-transpose 'page arg))

(defalias 'sort-page-atpt 'page-sort-atpt)
;;;###autoload
(defun page-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts page's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'page reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-page-atpt 'page-check-atpt)
;;;###autoload
(defun page-check-atpt (&optional arg) 
  " page at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "page" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "page" "-end-position-atpt")))))
    (th-check 'page arg)))
   

;;;###autoload
(defun count-alnum-in-page-atpt (&optional beg end) 
    "Counts matches of alnum in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'page beg end (interactive-p)))

(defun ratio-alnum-in-page-atpt (&optional beg end) 
    "Relation of alnum in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'page beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-page-atpt (&optional beg end) 
    "Counts matches of alpha in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'page beg end (interactive-p)))

(defun ratio-alpha-in-page-atpt (&optional beg end) 
    "Relation of alpha in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'page beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-page-atpt (&optional beg end) 
    "Counts matches of ascii in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'page beg end (interactive-p)))

(defun ratio-ascii-in-page-atpt (&optional beg end) 
    "Relation of ascii in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'page beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-page-atpt (&optional beg end) 
    "Counts matches of digit in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'page beg end (interactive-p)))

(defun ratio-digit-in-page-atpt (&optional beg end) 
    "Relation of digit in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'page beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-page-atpt (&optional beg end) 
    "Counts matches of graph in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'page beg end (interactive-p)))

(defun ratio-graph-in-page-atpt (&optional beg end) 
    "Relation of graph in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'page beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-page-atpt (&optional beg end) 
    "Counts matches of lower in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'page beg end (interactive-p)))

(defun ratio-lower-in-page-atpt (&optional beg end) 
    "Relation of lower in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'page beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-page-atpt (&optional beg end) 
    "Counts matches of punct in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'page beg end (interactive-p)))

(defun ratio-punct-in-page-atpt (&optional beg end) 
    "Relation of punct in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'page beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-page-atpt (&optional beg end) 
    "Counts matches of upper in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'page beg end (interactive-p)))

(defun ratio-upper-in-page-atpt (&optional beg end) 
    "Relation of upper in page if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'page beg end t (interactive-p)))



;;;###autoload
(defun paragraph-atpt (&optional arg) 
  "Returns paragraph at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'paragraph arg))

(defalias 'bounds-of-paragraph-atpt 'paragraph-bounds-atpt)
;;;###autoload
(defun paragraph-bounds-atpt (&optional arg) 
  "Returns borders of paragraph at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'paragraph arg))

;;;###autoload
(defun paragraph-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'paragraph arg))

;;;###autoload
(defun paragraph-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-end 'paragraph arg))

(defalias 'beginning-of-paragraph-atpt 'paragraph-beginning-atpt)
;;;###autoload
(defun paragraph-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'paragraph arg))

(defalias 'end-of-paragraph-atpt 'paragraph-end-atpt)
;;;###autoload
(defun paragraph-end-atpt (&optional arg) 
  "Goto end of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'paragraph arg))

(defalias 'length-of-paragraph-atpt 'paragraph-length-atpt)
;;;###autoload
(defun paragraph-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-length 'paragraph arg))

(defalias 'copy-paragraph-atpt 'paragraph-copy-atpt)
;;;###autoload
(defun paragraph-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'paragraph arg))

(defalias 'escape-paragraph-atpt 'paragraph-escape-atpt)
;;;###autoload
(defun paragraph-escape-atpt () 
  "Makes a copy of symbol or char-class paragraph at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'paragraph))

(defalias 'doublequote-paragraph-atpt 'paragraph-doublequote-atpt)
;;;###autoload
(defun paragraph-doublequote-atpt (&optional arg) 
  "Doublequotes 'paragraph at point if any. "
  (interactive "*p")
    (th-doublequote 'paragraph arg))

(defalias 'dollar-paragraph-atpt 'paragraph-dollar-atpt)
;;;###autoload
(defun paragraph-dollar-atpt (&optional arg) 
  "Doublequotes 'paragraph at point if any. "
  (interactive "*p")
    (th-dollar 'paragraph arg))

(defalias 'angle-greater-paragraph-atpt 'paragraph-greater-angle-atpt)
;;;###autoload
(defun paragraph-greater-angle-atpt (&optional arg) 
  "Sets angles for 'paragraph after point if any. "
  (interactive "*p")
    (th-greater-angle 'paragraph arg))

(defalias 'angle-lower-paragraph-atpt 'paragraph-lower-angle-atpt)
;;;###autoload
(defun paragraph-lower-angle-atpt (&optional arg) 
  "Sets angles for 'paragraph after point if any. "
  (interactive "*p")
    (th-lower-angle 'paragraph arg))

(defalias 'backslash-paragraph-atpt 'paragraph-backslash-atpt)
;;;###autoload
(defun paragraph-backslash-atpt (&optional arg) 
  "Backslash 'paragraph at point if any. "
  (interactive "*p")
    (th-backslash 'paragraph arg))   

(defalias 'brace-paragraph-atpt 'paragraph-brace-atpt)
;;;###autoload
(defun paragraph-brace-atpt (&optional arg) 
  "Braces 'paragraph at point if any. "
  (interactive "*p")
    (th-brace 'paragraph arg))

(defalias 'bracket-paragraph-atpt 'paragraph-bracket-atpt)
;;;###autoload
(defun paragraph-bracket-atpt (&optional arg) 
  "Brackets 'paragraph after point if any. "
  (interactive "*p")
    (th-bracket 'paragraph arg))

;; (defalias 'comment-paragraph-atpt 'paragraph-comment-atpt)
;;;###autoload
(defun paragraph-comment-atpt (&optional arg) 
  "Comments 'paragraph at point if any. "
  (interactive "*p")
  (th-comment 'paragraph arg))

(defalias 'mark-paragraph-atpt 'paragraph-mark-atpt)
;;;###autoload
(defun paragraph-mark-atpt () 
  "Marks 'paragraph at point if any. "
  (interactive)
  (th-mark 'paragraph))

(defalias 'hide-paragraph-atpt 'paragraph-hide-atpt)
;;;###autoload
(defun paragraph-hide-atpt () 
  "Toggles hide 'paragraph in region or buffer if any. "
  (interactive)
  (th-hide 'paragraph))

(defalias 'highlight-paragraph-atpt-mode 'paragraph-highlight-atpt-mode)
(defun paragraph-highlight-atpt-mode (&optional arg) 
  "Toggles paragraph-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'paragraph arg))

(defalias 'kill-paragraph-atpt 'paragraph-kill-atpt)
;;;###autoload
(defun paragraph-kill-atpt (&optional arg) 
  "Kills paragraph at point if any. "
  (interactive "*P")
  (th-kill 'paragraph arg))

(defalias 'parentize-paragraph-atpt 'paragraph-parentize-atpt)
;;;###autoload
(defun paragraph-parentize-atpt (&optional arg) 
  "Parentizes 'paragraph at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'paragraph arg))

(defalias 'separate-paragraph-atpt 'paragraph-separate-atpt)
;;;###autoload
(defun paragraph-separate-atpt (&optional arg) 
  "Separates 'paragraph at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'paragraph arg))

(defalias 'singlequote-paragraph-atpt 'paragraph-singlequote-atpt)
;;;###autoload
(defun paragraph-singlequote-atpt (&optional arg) 
  "Singlequotes 'paragraph at point if any. "
  (interactive "*p")
    (th-singlequote 'paragraph arg))

(defalias 'slash-paragraph-atpt 'paragraph-slash-atpt)
;;;###autoload
(defun paragraph-slash-atpt (&optional arg) 
  "Slash 'paragraph at point if any. "
  (interactive "*p")
    (th-slash 'paragraph arg))

(defalias 'trim-paragraph-atpt 'paragraph-trim-atpt)
;;;###autoload
(defun paragraph-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paragraph t t))

(defalias 'trim-paragraph-left-atpt 'paragraph-left-trim-atpt)
;;;###autoload
(defun paragraph-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paragraph t nil))

(defalias 'trim-paragraph-right-atpt 'paragraph-right-trim-atpt)
;;;###autoload
(defun paragraph-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paragraph nil t))

(defalias 'forward-paragraph-atpt 'paragraph-forward-atpt)
;;;###autoload
(defun paragraph-forward-atpt (&optional arg) 
  "Moves forward over paragraph at point if any, does nothing otherwise.
Returns bounds of paragraph as a list. "
  (interactive "p")
  (th-forward 'paragraph arg))

(defalias 'backward-paragraph-atpt 'paragraph-backward-atpt)
;;;###autoload
(defun paragraph-backward-atpt (&optional arg) 
  "Moves backward over paragraph before point if any, does nothing otherwise.
Returns bounds of paragraph as a list. "
  (interactive "p")
  (th-backward 'paragraph arg))

(defalias 'transpose-paragraph-atpt 'paragraph-transpose-atpt)
;;;###autoload
(defun paragraph-transpose-atpt (&optional arg) 
  "Transposes paragraph with paragraph before point if any. "
  (interactive "*P")
  (th-transpose 'paragraph arg))

(defalias 'sort-paragraph-atpt 'paragraph-sort-atpt)
;;;###autoload
(defun paragraph-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts paragraph's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'paragraph reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-paragraph-atpt 'paragraph-check-atpt)
;;;###autoload
(defun paragraph-check-atpt (&optional arg) 
  " paragraph at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "paragraph" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "paragraph" "-end-position-atpt")))))
    (th-check 'paragraph arg)))
   

;;;###autoload
(defun count-alnum-in-paragraph-atpt (&optional beg end) 
    "Counts matches of alnum in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'paragraph beg end (interactive-p)))

(defun ratio-alnum-in-paragraph-atpt (&optional beg end) 
    "Relation of alnum in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-paragraph-atpt (&optional beg end) 
    "Counts matches of alpha in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'paragraph beg end (interactive-p)))

(defun ratio-alpha-in-paragraph-atpt (&optional beg end) 
    "Relation of alpha in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-paragraph-atpt (&optional beg end) 
    "Counts matches of ascii in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'paragraph beg end (interactive-p)))

(defun ratio-ascii-in-paragraph-atpt (&optional beg end) 
    "Relation of ascii in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-paragraph-atpt (&optional beg end) 
    "Counts matches of digit in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'paragraph beg end (interactive-p)))

(defun ratio-digit-in-paragraph-atpt (&optional beg end) 
    "Relation of digit in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-paragraph-atpt (&optional beg end) 
    "Counts matches of graph in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'paragraph beg end (interactive-p)))

(defun ratio-graph-in-paragraph-atpt (&optional beg end) 
    "Relation of graph in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-paragraph-atpt (&optional beg end) 
    "Counts matches of lower in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'paragraph beg end (interactive-p)))

(defun ratio-lower-in-paragraph-atpt (&optional beg end) 
    "Relation of lower in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-paragraph-atpt (&optional beg end) 
    "Counts matches of punct in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'paragraph beg end (interactive-p)))

(defun ratio-punct-in-paragraph-atpt (&optional beg end) 
    "Relation of punct in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-paragraph-atpt (&optional beg end) 
    "Counts matches of upper in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'paragraph beg end (interactive-p)))

(defun ratio-upper-in-paragraph-atpt (&optional beg end) 
    "Relation of upper in paragraph if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'paragraph beg end t (interactive-p)))



;;;###autoload
(defun paren-atpt (&optional arg) 
  "Returns paren at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'paren arg))

(defalias 'bounds-of-paren-atpt 'paren-bounds-atpt)
;;;###autoload
(defun paren-bounds-atpt (&optional arg) 
  "Returns borders of paren at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'paren arg))

;;;###autoload
(defun paren-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'paren arg))

;;;###autoload
(defun paren-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-end 'paren arg))

(defalias 'beginning-of-paren-atpt 'paren-beginning-atpt)
;;;###autoload
(defun paren-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'paren arg))

(defalias 'end-of-paren-atpt 'paren-end-atpt)
;;;###autoload
(defun paren-end-atpt (&optional arg) 
  "Goto end of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'paren arg))

(defalias 'length-of-paren-atpt 'paren-length-atpt)
;;;###autoload
(defun paren-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-length 'paren arg))

(defalias 'copy-paren-atpt 'paren-copy-atpt)
;;;###autoload
(defun paren-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class paren at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'paren arg))

(defalias 'escape-paren-atpt 'paren-escape-atpt)
;;;###autoload
(defun paren-escape-atpt () 
  "Makes a copy of symbol or char-class paren at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'paren))

(defalias 'doublequote-paren-atpt 'paren-doublequote-atpt)
;;;###autoload
(defun paren-doublequote-atpt (&optional arg) 
  "Doublequotes 'paren at point if any. "
  (interactive "*p")
    (th-doublequote 'paren arg))

(defalias 'dollar-paren-atpt 'paren-dollar-atpt)
;;;###autoload
(defun paren-dollar-atpt (&optional arg) 
  "Doublequotes 'paren at point if any. "
  (interactive "*p")
    (th-dollar 'paren arg))

(defalias 'angle-greater-paren-atpt 'paren-greater-angle-atpt)
;;;###autoload
(defun paren-greater-angle-atpt (&optional arg) 
  "Sets angles for 'paren after point if any. "
  (interactive "*p")
    (th-greater-angle 'paren arg))

(defalias 'angle-lower-paren-atpt 'paren-lower-angle-atpt)
;;;###autoload
(defun paren-lower-angle-atpt (&optional arg) 
  "Sets angles for 'paren after point if any. "
  (interactive "*p")
    (th-lower-angle 'paren arg))

(defalias 'backslash-paren-atpt 'paren-backslash-atpt)
;;;###autoload
(defun paren-backslash-atpt (&optional arg) 
  "Backslash 'paren at point if any. "
  (interactive "*p")
    (th-backslash 'paren arg))   

(defalias 'brace-paren-atpt 'paren-brace-atpt)
;;;###autoload
(defun paren-brace-atpt (&optional arg) 
  "Braces 'paren at point if any. "
  (interactive "*p")
    (th-brace 'paren arg))

(defalias 'bracket-paren-atpt 'paren-bracket-atpt)
;;;###autoload
(defun paren-bracket-atpt (&optional arg) 
  "Brackets 'paren after point if any. "
  (interactive "*p")
    (th-bracket 'paren arg))

;; (defalias 'comment-paren-atpt 'paren-comment-atpt)
;;;###autoload
(defun paren-comment-atpt (&optional arg) 
  "Comments 'paren at point if any. "
  (interactive "*p")
  (th-comment 'paren arg))

(defalias 'mark-paren-atpt 'paren-mark-atpt)
;;;###autoload
(defun paren-mark-atpt () 
  "Marks 'paren at point if any. "
  (interactive)
  (th-mark 'paren))

(defalias 'hide-paren-atpt 'paren-hide-atpt)
;;;###autoload
(defun paren-hide-atpt () 
  "Toggles hide 'paren in region or buffer if any. "
  (interactive)
  (th-hide 'paren))

(defalias 'highlight-paren-atpt-mode 'paren-highlight-atpt-mode)
(defun paren-highlight-atpt-mode (&optional arg) 
  "Toggles paren-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'paren arg))

(defalias 'kill-paren-atpt 'paren-kill-atpt)
;;;###autoload
(defun paren-kill-atpt (&optional arg) 
  "Kills paren at point if any. "
  (interactive "*P")
  (th-kill 'paren arg))

(defalias 'parentize-paren-atpt 'paren-parentize-atpt)
;;;###autoload
(defun paren-parentize-atpt (&optional arg) 
  "Parentizes 'paren at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'paren arg))

(defalias 'separate-paren-atpt 'paren-separate-atpt)
;;;###autoload
(defun paren-separate-atpt (&optional arg) 
  "Separates 'paren at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'paren arg))

(defalias 'singlequote-paren-atpt 'paren-singlequote-atpt)
;;;###autoload
(defun paren-singlequote-atpt (&optional arg) 
  "Singlequotes 'paren at point if any. "
  (interactive "*p")
    (th-singlequote 'paren arg))

(defalias 'slash-paren-atpt 'paren-slash-atpt)
;;;###autoload
(defun paren-slash-atpt (&optional arg) 
  "Slash 'paren at point if any. "
  (interactive "*p")
    (th-slash 'paren arg))

(defalias 'trim-paren-atpt 'paren-trim-atpt)
;;;###autoload
(defun paren-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paren t t))

(defalias 'trim-paren-left-atpt 'paren-left-trim-atpt)
;;;###autoload
(defun paren-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paren t nil))

(defalias 'trim-paren-right-atpt 'paren-right-trim-atpt)
;;;###autoload
(defun paren-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'paren nil t))

(defalias 'forward-paren-atpt 'paren-forward-atpt)
;;;###autoload
(defun paren-forward-atpt (&optional arg) 
  "Moves forward over paren at point if any, does nothing otherwise.
Returns bounds of paren as a list. "
  (interactive "p")
  (th-forward 'paren arg))

(defalias 'backward-paren-atpt 'paren-backward-atpt)
;;;###autoload
(defun paren-backward-atpt (&optional arg) 
  "Moves backward over paren before point if any, does nothing otherwise.
Returns bounds of paren as a list. "
  (interactive "p")
  (th-backward 'paren arg))

(defalias 'transpose-paren-atpt 'paren-transpose-atpt)
;;;###autoload
(defun paren-transpose-atpt (&optional arg) 
  "Transposes paren with paren before point if any. "
  (interactive "*P")
  (th-transpose 'paren arg))

(defalias 'sort-paren-atpt 'paren-sort-atpt)
;;;###autoload
(defun paren-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts paren's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'paren reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-paren-atpt 'paren-check-atpt)
;;;###autoload
(defun paren-check-atpt (&optional arg) 
  " paren at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "paren" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "paren" "-end-position-atpt")))))
    (th-check 'paren arg)))
   

;;;###autoload
(defun count-alnum-in-paren-atpt (&optional beg end) 
    "Counts matches of alnum in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'paren beg end (interactive-p)))

(defun ratio-alnum-in-paren-atpt (&optional beg end) 
    "Relation of alnum in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-paren-atpt (&optional beg end) 
    "Counts matches of alpha in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'paren beg end (interactive-p)))

(defun ratio-alpha-in-paren-atpt (&optional beg end) 
    "Relation of alpha in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-paren-atpt (&optional beg end) 
    "Counts matches of ascii in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'paren beg end (interactive-p)))

(defun ratio-ascii-in-paren-atpt (&optional beg end) 
    "Relation of ascii in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-paren-atpt (&optional beg end) 
    "Counts matches of digit in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'paren beg end (interactive-p)))

(defun ratio-digit-in-paren-atpt (&optional beg end) 
    "Relation of digit in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-paren-atpt (&optional beg end) 
    "Counts matches of graph in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'paren beg end (interactive-p)))

(defun ratio-graph-in-paren-atpt (&optional beg end) 
    "Relation of graph in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-paren-atpt (&optional beg end) 
    "Counts matches of lower in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'paren beg end (interactive-p)))

(defun ratio-lower-in-paren-atpt (&optional beg end) 
    "Relation of lower in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-paren-atpt (&optional beg end) 
    "Counts matches of punct in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'paren beg end (interactive-p)))

(defun ratio-punct-in-paren-atpt (&optional beg end) 
    "Relation of punct in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'paren beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-paren-atpt (&optional beg end) 
    "Counts matches of upper in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'paren beg end (interactive-p)))

(defun ratio-upper-in-paren-atpt (&optional beg end) 
    "Relation of upper in paren if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'paren beg end t (interactive-p)))



;;;###autoload
(defun parentized-atpt (&optional arg) 
  "Returns parentized at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'parentized arg))

(defalias 'bounds-of-parentized-atpt 'parentized-bounds-atpt)
;;;###autoload
(defun parentized-bounds-atpt (&optional arg) 
  "Returns borders of parentized at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'parentized arg))

;;;###autoload
(defun parentized-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'parentized arg))

;;;###autoload
(defun parentized-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-end 'parentized arg))

(defalias 'beginning-of-parentized-atpt 'parentized-beginning-atpt)
;;;###autoload
(defun parentized-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'parentized arg))

(defalias 'end-of-parentized-atpt 'parentized-end-atpt)
;;;###autoload
(defun parentized-end-atpt (&optional arg) 
  "Goto end of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'parentized arg))

(defalias 'length-of-parentized-atpt 'parentized-length-atpt)
;;;###autoload
(defun parentized-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-length 'parentized arg))

(defalias 'copy-parentized-atpt 'parentized-copy-atpt)
;;;###autoload
(defun parentized-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'parentized arg))

(defalias 'escape-parentized-atpt 'parentized-escape-atpt)
;;;###autoload
(defun parentized-escape-atpt () 
  "Makes a copy of symbol or char-class parentized at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'parentized))

(defalias 'doublequote-parentized-atpt 'parentized-doublequote-atpt)
;;;###autoload
(defun parentized-doublequote-atpt (&optional arg) 
  "Doublequotes 'parentized at point if any. "
  (interactive "*p")
    (th-doublequote 'parentized arg))

(defalias 'dollar-parentized-atpt 'parentized-dollar-atpt)
;;;###autoload
(defun parentized-dollar-atpt (&optional arg) 
  "Doublequotes 'parentized at point if any. "
  (interactive "*p")
    (th-dollar 'parentized arg))

(defalias 'angle-greater-parentized-atpt 'parentized-greater-angle-atpt)
;;;###autoload
(defun parentized-greater-angle-atpt (&optional arg) 
  "Sets angles for 'parentized after point if any. "
  (interactive "*p")
    (th-greater-angle 'parentized arg))

(defalias 'angle-lower-parentized-atpt 'parentized-lower-angle-atpt)
;;;###autoload
(defun parentized-lower-angle-atpt (&optional arg) 
  "Sets angles for 'parentized after point if any. "
  (interactive "*p")
    (th-lower-angle 'parentized arg))

(defalias 'backslash-parentized-atpt 'parentized-backslash-atpt)
;;;###autoload
(defun parentized-backslash-atpt (&optional arg) 
  "Backslash 'parentized at point if any. "
  (interactive "*p")
    (th-backslash 'parentized arg))   

(defalias 'brace-parentized-atpt 'parentized-brace-atpt)
;;;###autoload
(defun parentized-brace-atpt (&optional arg) 
  "Braces 'parentized at point if any. "
  (interactive "*p")
    (th-brace 'parentized arg))

(defalias 'bracket-parentized-atpt 'parentized-bracket-atpt)
;;;###autoload
(defun parentized-bracket-atpt (&optional arg) 
  "Brackets 'parentized after point if any. "
  (interactive "*p")
    (th-bracket 'parentized arg))

;; (defalias 'comment-parentized-atpt 'parentized-comment-atpt)
;;;###autoload
(defun parentized-comment-atpt (&optional arg) 
  "Comments 'parentized at point if any. "
  (interactive "*p")
  (th-comment 'parentized arg))

(defalias 'mark-parentized-atpt 'parentized-mark-atpt)
;;;###autoload
(defun parentized-mark-atpt () 
  "Marks 'parentized at point if any. "
  (interactive)
  (th-mark 'parentized))

(defalias 'hide-parentized-atpt 'parentized-hide-atpt)
;;;###autoload
(defun parentized-hide-atpt () 
  "Toggles hide 'parentized in region or buffer if any. "
  (interactive)
  (th-hide 'parentized))

(defalias 'highlight-parentized-atpt-mode 'parentized-highlight-atpt-mode)
(defun parentized-highlight-atpt-mode (&optional arg) 
  "Toggles parentized-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'parentized arg))

(defalias 'kill-parentized-atpt 'parentized-kill-atpt)
;;;###autoload
(defun parentized-kill-atpt (&optional arg) 
  "Kills parentized at point if any. "
  (interactive "*P")
  (th-kill 'parentized arg))

(defalias 'parentize-parentized-atpt 'parentized-parentize-atpt)
;;;###autoload
(defun parentized-parentize-atpt (&optional arg) 
  "Parentizes 'parentized at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'parentized arg))

(defalias 'separate-parentized-atpt 'parentized-separate-atpt)
;;;###autoload
(defun parentized-separate-atpt (&optional arg) 
  "Separates 'parentized at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'parentized arg))

(defalias 'singlequote-parentized-atpt 'parentized-singlequote-atpt)
;;;###autoload
(defun parentized-singlequote-atpt (&optional arg) 
  "Singlequotes 'parentized at point if any. "
  (interactive "*p")
    (th-singlequote 'parentized arg))

(defalias 'slash-parentized-atpt 'parentized-slash-atpt)
;;;###autoload
(defun parentized-slash-atpt (&optional arg) 
  "Slash 'parentized at point if any. "
  (interactive "*p")
    (th-slash 'parentized arg))

(defalias 'trim-parentized-atpt 'parentized-trim-atpt)
;;;###autoload
(defun parentized-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'parentized t t))

(defalias 'trim-parentized-left-atpt 'parentized-left-trim-atpt)
;;;###autoload
(defun parentized-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'parentized t nil))

(defalias 'trim-parentized-right-atpt 'parentized-right-trim-atpt)
;;;###autoload
(defun parentized-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'parentized nil t))

(defalias 'forward-parentized-atpt 'parentized-forward-atpt)
;;;###autoload
(defun parentized-forward-atpt (&optional arg) 
  "Moves forward over parentized at point if any, does nothing otherwise.
Returns bounds of parentized as a list. "
  (interactive "p")
  (th-forward 'parentized arg))

(defalias 'backward-parentized-atpt 'parentized-backward-atpt)
;;;###autoload
(defun parentized-backward-atpt (&optional arg) 
  "Moves backward over parentized before point if any, does nothing otherwise.
Returns bounds of parentized as a list. "
  (interactive "p")
  (th-backward 'parentized arg))

(defalias 'transpose-parentized-atpt 'parentized-transpose-atpt)
;;;###autoload
(defun parentized-transpose-atpt (&optional arg) 
  "Transposes parentized with parentized before point if any. "
  (interactive "*P")
  (th-transpose 'parentized arg))

(defalias 'sort-parentized-atpt 'parentized-sort-atpt)
;;;###autoload
(defun parentized-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts parentized's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'parentized reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-parentized-atpt 'parentized-check-atpt)
;;;###autoload
(defun parentized-check-atpt (&optional arg) 
  " parentized at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "parentized" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "parentized" "-end-position-atpt")))))
    (th-check 'parentized arg)))
   

;;;###autoload
(defun count-alnum-in-parentized-atpt (&optional beg end) 
    "Counts matches of alnum in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'parentized beg end (interactive-p)))

(defun ratio-alnum-in-parentized-atpt (&optional beg end) 
    "Relation of alnum in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-parentized-atpt (&optional beg end) 
    "Counts matches of alpha in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'parentized beg end (interactive-p)))

(defun ratio-alpha-in-parentized-atpt (&optional beg end) 
    "Relation of alpha in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-parentized-atpt (&optional beg end) 
    "Counts matches of ascii in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'parentized beg end (interactive-p)))

(defun ratio-ascii-in-parentized-atpt (&optional beg end) 
    "Relation of ascii in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-parentized-atpt (&optional beg end) 
    "Counts matches of digit in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'parentized beg end (interactive-p)))

(defun ratio-digit-in-parentized-atpt (&optional beg end) 
    "Relation of digit in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-parentized-atpt (&optional beg end) 
    "Counts matches of graph in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'parentized beg end (interactive-p)))

(defun ratio-graph-in-parentized-atpt (&optional beg end) 
    "Relation of graph in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-parentized-atpt (&optional beg end) 
    "Counts matches of lower in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'parentized beg end (interactive-p)))

(defun ratio-lower-in-parentized-atpt (&optional beg end) 
    "Relation of lower in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-parentized-atpt (&optional beg end) 
    "Counts matches of punct in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'parentized beg end (interactive-p)))

(defun ratio-punct-in-parentized-atpt (&optional beg end) 
    "Relation of punct in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'parentized beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-parentized-atpt (&optional beg end) 
    "Counts matches of upper in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'parentized beg end (interactive-p)))

(defun ratio-upper-in-parentized-atpt (&optional beg end) 
    "Relation of upper in parentized if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'parentized beg end t (interactive-p)))



;;;###autoload
(defun phone-atpt (&optional arg) 
  "Returns phone at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'phone arg))

(defalias 'bounds-of-phone-atpt 'phone-bounds-atpt)
;;;###autoload
(defun phone-bounds-atpt (&optional arg) 
  "Returns borders of phone at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'phone arg))

;;;###autoload
(defun phone-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'phone arg))

;;;###autoload
(defun phone-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-end 'phone arg))

(defalias 'beginning-of-phone-atpt 'phone-beginning-atpt)
;;;###autoload
(defun phone-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'phone arg))

(defalias 'end-of-phone-atpt 'phone-end-atpt)
;;;###autoload
(defun phone-end-atpt (&optional arg) 
  "Goto end of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'phone arg))

(defalias 'length-of-phone-atpt 'phone-length-atpt)
;;;###autoload
(defun phone-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-length 'phone arg))

(defalias 'copy-phone-atpt 'phone-copy-atpt)
;;;###autoload
(defun phone-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class phone at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'phone arg))

(defalias 'escape-phone-atpt 'phone-escape-atpt)
;;;###autoload
(defun phone-escape-atpt () 
  "Makes a copy of symbol or char-class phone at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'phone))

(defalias 'doublequote-phone-atpt 'phone-doublequote-atpt)
;;;###autoload
(defun phone-doublequote-atpt (&optional arg) 
  "Doublequotes 'phone at point if any. "
  (interactive "*p")
    (th-doublequote 'phone arg))

(defalias 'dollar-phone-atpt 'phone-dollar-atpt)
;;;###autoload
(defun phone-dollar-atpt (&optional arg) 
  "Doublequotes 'phone at point if any. "
  (interactive "*p")
    (th-dollar 'phone arg))

(defalias 'angle-greater-phone-atpt 'phone-greater-angle-atpt)
;;;###autoload
(defun phone-greater-angle-atpt (&optional arg) 
  "Sets angles for 'phone after point if any. "
  (interactive "*p")
    (th-greater-angle 'phone arg))

(defalias 'angle-lower-phone-atpt 'phone-lower-angle-atpt)
;;;###autoload
(defun phone-lower-angle-atpt (&optional arg) 
  "Sets angles for 'phone after point if any. "
  (interactive "*p")
    (th-lower-angle 'phone arg))

(defalias 'backslash-phone-atpt 'phone-backslash-atpt)
;;;###autoload
(defun phone-backslash-atpt (&optional arg) 
  "Backslash 'phone at point if any. "
  (interactive "*p")
    (th-backslash 'phone arg))   

(defalias 'brace-phone-atpt 'phone-brace-atpt)
;;;###autoload
(defun phone-brace-atpt (&optional arg) 
  "Braces 'phone at point if any. "
  (interactive "*p")
    (th-brace 'phone arg))

(defalias 'bracket-phone-atpt 'phone-bracket-atpt)
;;;###autoload
(defun phone-bracket-atpt (&optional arg) 
  "Brackets 'phone after point if any. "
  (interactive "*p")
    (th-bracket 'phone arg))

;; (defalias 'comment-phone-atpt 'phone-comment-atpt)
;;;###autoload
(defun phone-comment-atpt (&optional arg) 
  "Comments 'phone at point if any. "
  (interactive "*p")
  (th-comment 'phone arg))

(defalias 'mark-phone-atpt 'phone-mark-atpt)
;;;###autoload
(defun phone-mark-atpt () 
  "Marks 'phone at point if any. "
  (interactive)
  (th-mark 'phone))

(defalias 'hide-phone-atpt 'phone-hide-atpt)
;;;###autoload
(defun phone-hide-atpt () 
  "Toggles hide 'phone in region or buffer if any. "
  (interactive)
  (th-hide 'phone))

(defalias 'highlight-phone-atpt-mode 'phone-highlight-atpt-mode)
(defun phone-highlight-atpt-mode (&optional arg) 
  "Toggles phone-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'phone arg))

(defalias 'kill-phone-atpt 'phone-kill-atpt)
;;;###autoload
(defun phone-kill-atpt (&optional arg) 
  "Kills phone at point if any. "
  (interactive "*P")
  (th-kill 'phone arg))

(defalias 'parentize-phone-atpt 'phone-parentize-atpt)
;;;###autoload
(defun phone-parentize-atpt (&optional arg) 
  "Parentizes 'phone at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'phone arg))

(defalias 'separate-phone-atpt 'phone-separate-atpt)
;;;###autoload
(defun phone-separate-atpt (&optional arg) 
  "Separates 'phone at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'phone arg))

(defalias 'singlequote-phone-atpt 'phone-singlequote-atpt)
;;;###autoload
(defun phone-singlequote-atpt (&optional arg) 
  "Singlequotes 'phone at point if any. "
  (interactive "*p")
    (th-singlequote 'phone arg))

(defalias 'slash-phone-atpt 'phone-slash-atpt)
;;;###autoload
(defun phone-slash-atpt (&optional arg) 
  "Slash 'phone at point if any. "
  (interactive "*p")
    (th-slash 'phone arg))

(defalias 'trim-phone-atpt 'phone-trim-atpt)
;;;###autoload
(defun phone-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'phone t t))

(defalias 'trim-phone-left-atpt 'phone-left-trim-atpt)
;;;###autoload
(defun phone-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'phone t nil))

(defalias 'trim-phone-right-atpt 'phone-right-trim-atpt)
;;;###autoload
(defun phone-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'phone nil t))

(defalias 'forward-phone-atpt 'phone-forward-atpt)
;;;###autoload
(defun phone-forward-atpt (&optional arg) 
  "Moves forward over phone at point if any, does nothing otherwise.
Returns bounds of phone as a list. "
  (interactive "p")
  (th-forward 'phone arg))

(defalias 'backward-phone-atpt 'phone-backward-atpt)
;;;###autoload
(defun phone-backward-atpt (&optional arg) 
  "Moves backward over phone before point if any, does nothing otherwise.
Returns bounds of phone as a list. "
  (interactive "p")
  (th-backward 'phone arg))

(defalias 'transpose-phone-atpt 'phone-transpose-atpt)
;;;###autoload
(defun phone-transpose-atpt (&optional arg) 
  "Transposes phone with phone before point if any. "
  (interactive "*P")
  (th-transpose 'phone arg))

(defalias 'sort-phone-atpt 'phone-sort-atpt)
;;;###autoload
(defun phone-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts phone's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'phone reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-phone-atpt 'phone-check-atpt)
;;;###autoload
(defun phone-check-atpt (&optional arg) 
  " phone at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "phone" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "phone" "-end-position-atpt")))))
    (th-check 'phone arg)))
   

;;;###autoload
(defun count-alnum-in-phone-atpt (&optional beg end) 
    "Counts matches of alnum in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'phone beg end (interactive-p)))

(defun ratio-alnum-in-phone-atpt (&optional beg end) 
    "Relation of alnum in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-phone-atpt (&optional beg end) 
    "Counts matches of alpha in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'phone beg end (interactive-p)))

(defun ratio-alpha-in-phone-atpt (&optional beg end) 
    "Relation of alpha in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-phone-atpt (&optional beg end) 
    "Counts matches of ascii in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'phone beg end (interactive-p)))

(defun ratio-ascii-in-phone-atpt (&optional beg end) 
    "Relation of ascii in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-phone-atpt (&optional beg end) 
    "Counts matches of digit in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'phone beg end (interactive-p)))

(defun ratio-digit-in-phone-atpt (&optional beg end) 
    "Relation of digit in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-phone-atpt (&optional beg end) 
    "Counts matches of graph in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'phone beg end (interactive-p)))

(defun ratio-graph-in-phone-atpt (&optional beg end) 
    "Relation of graph in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-phone-atpt (&optional beg end) 
    "Counts matches of lower in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'phone beg end (interactive-p)))

(defun ratio-lower-in-phone-atpt (&optional beg end) 
    "Relation of lower in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-phone-atpt (&optional beg end) 
    "Counts matches of punct in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'phone beg end (interactive-p)))

(defun ratio-punct-in-phone-atpt (&optional beg end) 
    "Relation of punct in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'phone beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-phone-atpt (&optional beg end) 
    "Counts matches of upper in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'phone beg end (interactive-p)))

(defun ratio-upper-in-phone-atpt (&optional beg end) 
    "Relation of upper in phone if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'phone beg end t (interactive-p)))



;;;###autoload
(defun print-atpt (&optional arg) 
  "Returns print at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'print arg))

(defalias 'bounds-of-print-atpt 'print-bounds-atpt)
;;;###autoload
(defun print-bounds-atpt (&optional arg) 
  "Returns borders of print at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'print arg))

;;;###autoload
(defun print-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'print arg))

;;;###autoload
(defun print-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-end 'print arg))

(defalias 'beginning-of-print-atpt 'print-beginning-atpt)
;;;###autoload
(defun print-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'print arg))

(defalias 'end-of-print-atpt 'print-end-atpt)
;;;###autoload
(defun print-end-atpt (&optional arg) 
  "Goto end of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'print arg))

(defalias 'length-of-print-atpt 'print-length-atpt)
;;;###autoload
(defun print-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-length 'print arg))

(defalias 'copy-print-atpt 'print-copy-atpt)
;;;###autoload
(defun print-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class print at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'print arg))

(defalias 'escape-print-atpt 'print-escape-atpt)
;;;###autoload
(defun print-escape-atpt () 
  "Makes a copy of symbol or char-class print at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'print))

(defalias 'doublequote-print-atpt 'print-doublequote-atpt)
;;;###autoload
(defun print-doublequote-atpt (&optional arg) 
  "Doublequotes 'print at point if any. "
  (interactive "*p")
    (th-doublequote 'print arg))

(defalias 'dollar-print-atpt 'print-dollar-atpt)
;;;###autoload
(defun print-dollar-atpt (&optional arg) 
  "Doublequotes 'print at point if any. "
  (interactive "*p")
    (th-dollar 'print arg))

(defalias 'angle-greater-print-atpt 'print-greater-angle-atpt)
;;;###autoload
(defun print-greater-angle-atpt (&optional arg) 
  "Sets angles for 'print after point if any. "
  (interactive "*p")
    (th-greater-angle 'print arg))

(defalias 'angle-lower-print-atpt 'print-lower-angle-atpt)
;;;###autoload
(defun print-lower-angle-atpt (&optional arg) 
  "Sets angles for 'print after point if any. "
  (interactive "*p")
    (th-lower-angle 'print arg))

(defalias 'backslash-print-atpt 'print-backslash-atpt)
;;;###autoload
(defun print-backslash-atpt (&optional arg) 
  "Backslash 'print at point if any. "
  (interactive "*p")
    (th-backslash 'print arg))   

(defalias 'brace-print-atpt 'print-brace-atpt)
;;;###autoload
(defun print-brace-atpt (&optional arg) 
  "Braces 'print at point if any. "
  (interactive "*p")
    (th-brace 'print arg))

(defalias 'bracket-print-atpt 'print-bracket-atpt)
;;;###autoload
(defun print-bracket-atpt (&optional arg) 
  "Brackets 'print after point if any. "
  (interactive "*p")
    (th-bracket 'print arg))

;; (defalias 'comment-print-atpt 'print-comment-atpt)
;;;###autoload
(defun print-comment-atpt (&optional arg) 
  "Comments 'print at point if any. "
  (interactive "*p")
  (th-comment 'print arg))

(defalias 'mark-print-atpt 'print-mark-atpt)
;;;###autoload
(defun print-mark-atpt () 
  "Marks 'print at point if any. "
  (interactive)
  (th-mark 'print))

(defalias 'hide-print-atpt 'print-hide-atpt)
;;;###autoload
(defun print-hide-atpt () 
  "Toggles hide 'print in region or buffer if any. "
  (interactive)
  (th-hide 'print))

(defalias 'highlight-print-atpt-mode 'print-highlight-atpt-mode)
(defun print-highlight-atpt-mode (&optional arg) 
  "Toggles print-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'print arg))

(defalias 'kill-print-atpt 'print-kill-atpt)
;;;###autoload
(defun print-kill-atpt (&optional arg) 
  "Kills print at point if any. "
  (interactive "*P")
  (th-kill 'print arg))

(defalias 'parentize-print-atpt 'print-parentize-atpt)
;;;###autoload
(defun print-parentize-atpt (&optional arg) 
  "Parentizes 'print at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'print arg))

(defalias 'separate-print-atpt 'print-separate-atpt)
;;;###autoload
(defun print-separate-atpt (&optional arg) 
  "Separates 'print at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'print arg))

(defalias 'singlequote-print-atpt 'print-singlequote-atpt)
;;;###autoload
(defun print-singlequote-atpt (&optional arg) 
  "Singlequotes 'print at point if any. "
  (interactive "*p")
    (th-singlequote 'print arg))

(defalias 'slash-print-atpt 'print-slash-atpt)
;;;###autoload
(defun print-slash-atpt (&optional arg) 
  "Slash 'print at point if any. "
  (interactive "*p")
    (th-slash 'print arg))

(defalias 'trim-print-atpt 'print-trim-atpt)
;;;###autoload
(defun print-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'print t t))

(defalias 'trim-print-left-atpt 'print-left-trim-atpt)
;;;###autoload
(defun print-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'print t nil))

(defalias 'trim-print-right-atpt 'print-right-trim-atpt)
;;;###autoload
(defun print-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'print nil t))

(defalias 'forward-print-atpt 'print-forward-atpt)
;;;###autoload
(defun print-forward-atpt (&optional arg) 
  "Moves forward over print at point if any, does nothing otherwise.
Returns bounds of print as a list. "
  (interactive "p")
  (th-forward 'print arg))

(defalias 'backward-print-atpt 'print-backward-atpt)
;;;###autoload
(defun print-backward-atpt (&optional arg) 
  "Moves backward over print before point if any, does nothing otherwise.
Returns bounds of print as a list. "
  (interactive "p")
  (th-backward 'print arg))

(defalias 'transpose-print-atpt 'print-transpose-atpt)
;;;###autoload
(defun print-transpose-atpt (&optional arg) 
  "Transposes print with print before point if any. "
  (interactive "*P")
  (th-transpose 'print arg))

(defalias 'sort-print-atpt 'print-sort-atpt)
;;;###autoload
(defun print-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts print's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'print reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-print-atpt 'print-check-atpt)
;;;###autoload
(defun print-check-atpt (&optional arg) 
  " print at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "print" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "print" "-end-position-atpt")))))
    (th-check 'print arg)))
   

;;;###autoload
(defun count-alnum-in-print-atpt (&optional beg end) 
    "Counts matches of alnum in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'print beg end (interactive-p)))

(defun ratio-alnum-in-print-atpt (&optional beg end) 
    "Relation of alnum in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'print beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-print-atpt (&optional beg end) 
    "Counts matches of alpha in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'print beg end (interactive-p)))

(defun ratio-alpha-in-print-atpt (&optional beg end) 
    "Relation of alpha in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'print beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-print-atpt (&optional beg end) 
    "Counts matches of ascii in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'print beg end (interactive-p)))

(defun ratio-ascii-in-print-atpt (&optional beg end) 
    "Relation of ascii in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'print beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-print-atpt (&optional beg end) 
    "Counts matches of digit in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'print beg end (interactive-p)))

(defun ratio-digit-in-print-atpt (&optional beg end) 
    "Relation of digit in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'print beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-print-atpt (&optional beg end) 
    "Counts matches of graph in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'print beg end (interactive-p)))

(defun ratio-graph-in-print-atpt (&optional beg end) 
    "Relation of graph in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'print beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-print-atpt (&optional beg end) 
    "Counts matches of lower in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'print beg end (interactive-p)))

(defun ratio-lower-in-print-atpt (&optional beg end) 
    "Relation of lower in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'print beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-print-atpt (&optional beg end) 
    "Counts matches of punct in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'print beg end (interactive-p)))

(defun ratio-punct-in-print-atpt (&optional beg end) 
    "Relation of punct in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'print beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-print-atpt (&optional beg end) 
    "Counts matches of upper in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'print beg end (interactive-p)))

(defun ratio-upper-in-print-atpt (&optional beg end) 
    "Relation of upper in print if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'print beg end t (interactive-p)))



;;;###autoload
(defun punct-atpt (&optional arg) 
  "Returns punct at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'punct arg))

(defalias 'bounds-of-punct-atpt 'punct-bounds-atpt)
;;;###autoload
(defun punct-bounds-atpt (&optional arg) 
  "Returns borders of punct at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'punct arg))

;;;###autoload
(defun punct-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'punct arg))

;;;###autoload
(defun punct-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-end 'punct arg))

(defalias 'beginning-of-punct-atpt 'punct-beginning-atpt)
;;;###autoload
(defun punct-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'punct arg))

(defalias 'end-of-punct-atpt 'punct-end-atpt)
;;;###autoload
(defun punct-end-atpt (&optional arg) 
  "Goto end of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'punct arg))

(defalias 'length-of-punct-atpt 'punct-length-atpt)
;;;###autoload
(defun punct-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-length 'punct arg))

(defalias 'copy-punct-atpt 'punct-copy-atpt)
;;;###autoload
(defun punct-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class punct at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'punct arg))

(defalias 'escape-punct-atpt 'punct-escape-atpt)
;;;###autoload
(defun punct-escape-atpt () 
  "Makes a copy of symbol or char-class punct at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'punct))

(defalias 'doublequote-punct-atpt 'punct-doublequote-atpt)
;;;###autoload
(defun punct-doublequote-atpt (&optional arg) 
  "Doublequotes 'punct at point if any. "
  (interactive "*p")
    (th-doublequote 'punct arg))

(defalias 'dollar-punct-atpt 'punct-dollar-atpt)
;;;###autoload
(defun punct-dollar-atpt (&optional arg) 
  "Doublequotes 'punct at point if any. "
  (interactive "*p")
    (th-dollar 'punct arg))

(defalias 'angle-greater-punct-atpt 'punct-greater-angle-atpt)
;;;###autoload
(defun punct-greater-angle-atpt (&optional arg) 
  "Sets angles for 'punct after point if any. "
  (interactive "*p")
    (th-greater-angle 'punct arg))

(defalias 'angle-lower-punct-atpt 'punct-lower-angle-atpt)
;;;###autoload
(defun punct-lower-angle-atpt (&optional arg) 
  "Sets angles for 'punct after point if any. "
  (interactive "*p")
    (th-lower-angle 'punct arg))

(defalias 'backslash-punct-atpt 'punct-backslash-atpt)
;;;###autoload
(defun punct-backslash-atpt (&optional arg) 
  "Backslash 'punct at point if any. "
  (interactive "*p")
    (th-backslash 'punct arg))   

(defalias 'brace-punct-atpt 'punct-brace-atpt)
;;;###autoload
(defun punct-brace-atpt (&optional arg) 
  "Braces 'punct at point if any. "
  (interactive "*p")
    (th-brace 'punct arg))

(defalias 'bracket-punct-atpt 'punct-bracket-atpt)
;;;###autoload
(defun punct-bracket-atpt (&optional arg) 
  "Brackets 'punct after point if any. "
  (interactive "*p")
    (th-bracket 'punct arg))

;; (defalias 'comment-punct-atpt 'punct-comment-atpt)
;;;###autoload
(defun punct-comment-atpt (&optional arg) 
  "Comments 'punct at point if any. "
  (interactive "*p")
  (th-comment 'punct arg))

(defalias 'mark-punct-atpt 'punct-mark-atpt)
;;;###autoload
(defun punct-mark-atpt () 
  "Marks 'punct at point if any. "
  (interactive)
  (th-mark 'punct))

(defalias 'hide-punct-atpt 'punct-hide-atpt)
;;;###autoload
(defun punct-hide-atpt () 
  "Toggles hide 'punct in region or buffer if any. "
  (interactive)
  (th-hide 'punct))

(defalias 'highlight-punct-atpt-mode 'punct-highlight-atpt-mode)
(defun punct-highlight-atpt-mode (&optional arg) 
  "Toggles punct-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'punct arg))

(defalias 'kill-punct-atpt 'punct-kill-atpt)
;;;###autoload
(defun punct-kill-atpt (&optional arg) 
  "Kills punct at point if any. "
  (interactive "*P")
  (th-kill 'punct arg))

(defalias 'parentize-punct-atpt 'punct-parentize-atpt)
;;;###autoload
(defun punct-parentize-atpt (&optional arg) 
  "Parentizes 'punct at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'punct arg))

(defalias 'separate-punct-atpt 'punct-separate-atpt)
;;;###autoload
(defun punct-separate-atpt (&optional arg) 
  "Separates 'punct at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'punct arg))

(defalias 'singlequote-punct-atpt 'punct-singlequote-atpt)
;;;###autoload
(defun punct-singlequote-atpt (&optional arg) 
  "Singlequotes 'punct at point if any. "
  (interactive "*p")
    (th-singlequote 'punct arg))

(defalias 'slash-punct-atpt 'punct-slash-atpt)
;;;###autoload
(defun punct-slash-atpt (&optional arg) 
  "Slash 'punct at point if any. "
  (interactive "*p")
    (th-slash 'punct arg))

(defalias 'trim-punct-atpt 'punct-trim-atpt)
;;;###autoload
(defun punct-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'punct t t))

(defalias 'trim-punct-left-atpt 'punct-left-trim-atpt)
;;;###autoload
(defun punct-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'punct t nil))

(defalias 'trim-punct-right-atpt 'punct-right-trim-atpt)
;;;###autoload
(defun punct-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'punct nil t))

(defalias 'forward-punct-atpt 'punct-forward-atpt)
;;;###autoload
(defun punct-forward-atpt (&optional arg) 
  "Moves forward over punct at point if any, does nothing otherwise.
Returns bounds of punct as a list. "
  (interactive "p")
  (th-forward 'punct arg))

(defalias 'backward-punct-atpt 'punct-backward-atpt)
;;;###autoload
(defun punct-backward-atpt (&optional arg) 
  "Moves backward over punct before point if any, does nothing otherwise.
Returns bounds of punct as a list. "
  (interactive "p")
  (th-backward 'punct arg))

(defalias 'transpose-punct-atpt 'punct-transpose-atpt)
;;;###autoload
(defun punct-transpose-atpt (&optional arg) 
  "Transposes punct with punct before point if any. "
  (interactive "*P")
  (th-transpose 'punct arg))

(defalias 'sort-punct-atpt 'punct-sort-atpt)
;;;###autoload
(defun punct-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts punct's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'punct reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-punct-atpt 'punct-check-atpt)
;;;###autoload
(defun punct-check-atpt (&optional arg) 
  " punct at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "punct" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "punct" "-end-position-atpt")))))
    (th-check 'punct arg)))
   

;;;###autoload
(defun count-alnum-in-punct-atpt (&optional beg end) 
    "Counts matches of alnum in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'punct beg end (interactive-p)))

(defun ratio-alnum-in-punct-atpt (&optional beg end) 
    "Relation of alnum in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-punct-atpt (&optional beg end) 
    "Counts matches of alpha in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'punct beg end (interactive-p)))

(defun ratio-alpha-in-punct-atpt (&optional beg end) 
    "Relation of alpha in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-punct-atpt (&optional beg end) 
    "Counts matches of ascii in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'punct beg end (interactive-p)))

(defun ratio-ascii-in-punct-atpt (&optional beg end) 
    "Relation of ascii in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-punct-atpt (&optional beg end) 
    "Counts matches of digit in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'punct beg end (interactive-p)))

(defun ratio-digit-in-punct-atpt (&optional beg end) 
    "Relation of digit in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-punct-atpt (&optional beg end) 
    "Counts matches of graph in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'punct beg end (interactive-p)))

(defun ratio-graph-in-punct-atpt (&optional beg end) 
    "Relation of graph in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-punct-atpt (&optional beg end) 
    "Counts matches of lower in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'punct beg end (interactive-p)))

(defun ratio-lower-in-punct-atpt (&optional beg end) 
    "Relation of lower in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'punct beg end t (interactive-p)))



;;;###autoload
(defun count-punct-atpt (&optional beg end) 
    "Counts matches of punct at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'punct beg end (interactive-p)))


;;;###autoload
(defun count-upper-in-punct-atpt (&optional beg end) 
    "Counts matches of upper in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'punct beg end (interactive-p)))

(defun ratio-upper-in-punct-atpt (&optional beg end) 
    "Relation of upper in punct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'punct beg end t (interactive-p)))



;;;###autoload
(defun region-atpt (&optional arg) 
  "Returns region at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'region arg))

(defalias 'bounds-of-region-atpt 'region-bounds-atpt)
;;;###autoload
(defun region-bounds-atpt (&optional arg) 
  "Returns borders of region at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'region arg))

;;;###autoload
(defun region-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'region arg))

;;;###autoload
(defun region-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-end 'region arg))

(defalias 'beginning-of-region-atpt 'region-beginning-atpt)
;;;###autoload
(defun region-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'region arg))

(defalias 'end-of-region-atpt 'region-end-atpt)
;;;###autoload
(defun region-end-atpt (&optional arg) 
  "Goto end of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'region arg))

(defalias 'length-of-region-atpt 'region-length-atpt)
;;;###autoload
(defun region-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-length 'region arg))

(defalias 'copy-region-atpt 'region-copy-atpt)
;;;###autoload
(defun region-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class region at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'region arg))

(defalias 'escape-region-atpt 'region-escape-atpt)
;;;###autoload
(defun region-escape-atpt () 
  "Makes a copy of symbol or char-class region at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'region))

(defalias 'doublequote-region-atpt 'region-doublequote-atpt)
;;;###autoload
(defun region-doublequote-atpt (&optional arg) 
  "Doublequotes 'region at point if any. "
  (interactive "*p")
    (th-doublequote 'region arg))

(defalias 'dollar-region-atpt 'region-dollar-atpt)
;;;###autoload
(defun region-dollar-atpt (&optional arg) 
  "Doublequotes 'region at point if any. "
  (interactive "*p")
    (th-dollar 'region arg))

(defalias 'angle-greater-region-atpt 'region-greater-angle-atpt)
;;;###autoload
(defun region-greater-angle-atpt (&optional arg) 
  "Sets angles for 'region after point if any. "
  (interactive "*p")
    (th-greater-angle 'region arg))

(defalias 'angle-lower-region-atpt 'region-lower-angle-atpt)
;;;###autoload
(defun region-lower-angle-atpt (&optional arg) 
  "Sets angles for 'region after point if any. "
  (interactive "*p")
    (th-lower-angle 'region arg))

(defalias 'backslash-region-atpt 'region-backslash-atpt)
;;;###autoload
(defun region-backslash-atpt (&optional arg) 
  "Backslash 'region at point if any. "
  (interactive "*p")
    (th-backslash 'region arg))   

(defalias 'brace-region-atpt 'region-brace-atpt)
;;;###autoload
(defun region-brace-atpt (&optional arg) 
  "Braces 'region at point if any. "
  (interactive "*p")
    (th-brace 'region arg))

(defalias 'bracket-region-atpt 'region-bracket-atpt)
;;;###autoload
(defun region-bracket-atpt (&optional arg) 
  "Brackets 'region after point if any. "
  (interactive "*p")
    (th-bracket 'region arg))

;; (defalias 'comment-region-atpt 'region-comment-atpt)
;;;###autoload
(defun region-comment-atpt (&optional arg) 
  "Comments 'region at point if any. "
  (interactive "*p")
  (th-comment 'region arg))

(defalias 'mark-region-atpt 'region-mark-atpt)
;;;###autoload
(defun region-mark-atpt () 
  "Marks 'region at point if any. "
  (interactive)
  (th-mark 'region))

(defalias 'hide-region-atpt 'region-hide-atpt)
;;;###autoload
(defun region-hide-atpt () 
  "Toggles hide 'region in region or buffer if any. "
  (interactive)
  (th-hide 'region))

(defalias 'highlight-region-atpt-mode 'region-highlight-atpt-mode)
(defun region-highlight-atpt-mode (&optional arg) 
  "Toggles region-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'region arg))

(defalias 'kill-region-atpt 'region-kill-atpt)
;;;###autoload
(defun region-kill-atpt (&optional arg) 
  "Kills region at point if any. "
  (interactive "*P")
  (th-kill 'region arg))

(defalias 'parentize-region-atpt 'region-parentize-atpt)
;;;###autoload
(defun region-parentize-atpt (&optional arg) 
  "Parentizes 'region at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'region arg))

(defalias 'separate-region-atpt 'region-separate-atpt)
;;;###autoload
(defun region-separate-atpt (&optional arg) 
  "Separates 'region at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'region arg))

(defalias 'singlequote-region-atpt 'region-singlequote-atpt)
;;;###autoload
(defun region-singlequote-atpt (&optional arg) 
  "Singlequotes 'region at point if any. "
  (interactive "*p")
    (th-singlequote 'region arg))

(defalias 'slash-region-atpt 'region-slash-atpt)
;;;###autoload
(defun region-slash-atpt (&optional arg) 
  "Slash 'region at point if any. "
  (interactive "*p")
    (th-slash 'region arg))

(defalias 'trim-region-atpt 'region-trim-atpt)
;;;###autoload
(defun region-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'region t t))

(defalias 'trim-region-left-atpt 'region-left-trim-atpt)
;;;###autoload
(defun region-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'region t nil))

(defalias 'trim-region-right-atpt 'region-right-trim-atpt)
;;;###autoload
(defun region-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'region nil t))

(defalias 'forward-region-atpt 'region-forward-atpt)
;;;###autoload
(defun region-forward-atpt (&optional arg) 
  "Moves forward over region at point if any, does nothing otherwise.
Returns bounds of region as a list. "
  (interactive "p")
  (th-forward 'region arg))

(defalias 'backward-region-atpt 'region-backward-atpt)
;;;###autoload
(defun region-backward-atpt (&optional arg) 
  "Moves backward over region before point if any, does nothing otherwise.
Returns bounds of region as a list. "
  (interactive "p")
  (th-backward 'region arg))

(defalias 'transpose-region-atpt 'region-transpose-atpt)
;;;###autoload
(defun region-transpose-atpt (&optional arg) 
  "Transposes region with region before point if any. "
  (interactive "*P")
  (th-transpose 'region arg))

(defalias 'sort-region-atpt 'region-sort-atpt)
;;;###autoload
(defun region-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts region's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'region reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-region-atpt 'region-check-atpt)
;;;###autoload
(defun region-check-atpt (&optional arg) 
  " region at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "region" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "region" "-end-position-atpt")))))
    (th-check 'region arg)))
   

;;;###autoload
(defun count-alnum-in-region-atpt (&optional beg end) 
    "Counts matches of alnum in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'region beg end (interactive-p)))

(defun ratio-alnum-in-region-atpt (&optional beg end) 
    "Relation of alnum in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'region beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-region-atpt (&optional beg end) 
    "Counts matches of alpha in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'region beg end (interactive-p)))

(defun ratio-alpha-in-region-atpt (&optional beg end) 
    "Relation of alpha in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'region beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-region-atpt (&optional beg end) 
    "Counts matches of ascii in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'region beg end (interactive-p)))

(defun ratio-ascii-in-region-atpt (&optional beg end) 
    "Relation of ascii in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'region beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-region-atpt (&optional beg end) 
    "Counts matches of digit in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'region beg end (interactive-p)))

(defun ratio-digit-in-region-atpt (&optional beg end) 
    "Relation of digit in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'region beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-region-atpt (&optional beg end) 
    "Counts matches of graph in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'region beg end (interactive-p)))

(defun ratio-graph-in-region-atpt (&optional beg end) 
    "Relation of graph in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'region beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-region-atpt (&optional beg end) 
    "Counts matches of lower in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'region beg end (interactive-p)))

(defun ratio-lower-in-region-atpt (&optional beg end) 
    "Relation of lower in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'region beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-region-atpt (&optional beg end) 
    "Counts matches of punct in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'region beg end (interactive-p)))

(defun ratio-punct-in-region-atpt (&optional beg end) 
    "Relation of punct in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'region beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-region-atpt (&optional beg end) 
    "Counts matches of upper in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'region beg end (interactive-p)))

(defun ratio-upper-in-region-atpt (&optional beg end) 
    "Relation of upper in region if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'region beg end t (interactive-p)))



;;;###autoload
(defun sentence-atpt (&optional arg) 
  "Returns sentence at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'sentence arg))

(defalias 'bounds-of-sentence-atpt 'sentence-bounds-atpt)
;;;###autoload
(defun sentence-bounds-atpt (&optional arg) 
  "Returns borders of sentence at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'sentence arg))

;;;###autoload
(defun sentence-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'sentence arg))

;;;###autoload
(defun sentence-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-end 'sentence arg))

(defalias 'beginning-of-sentence-atpt 'sentence-beginning-atpt)
;;;###autoload
(defun sentence-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'sentence arg))

(defalias 'end-of-sentence-atpt 'sentence-end-atpt)
;;;###autoload
(defun sentence-end-atpt (&optional arg) 
  "Goto end of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'sentence arg))

(defalias 'length-of-sentence-atpt 'sentence-length-atpt)
;;;###autoload
(defun sentence-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-length 'sentence arg))

(defalias 'copy-sentence-atpt 'sentence-copy-atpt)
;;;###autoload
(defun sentence-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'sentence arg))

(defalias 'escape-sentence-atpt 'sentence-escape-atpt)
;;;###autoload
(defun sentence-escape-atpt () 
  "Makes a copy of symbol or char-class sentence at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'sentence))

(defalias 'doublequote-sentence-atpt 'sentence-doublequote-atpt)
;;;###autoload
(defun sentence-doublequote-atpt (&optional arg) 
  "Doublequotes 'sentence at point if any. "
  (interactive "*p")
    (th-doublequote 'sentence arg))

(defalias 'dollar-sentence-atpt 'sentence-dollar-atpt)
;;;###autoload
(defun sentence-dollar-atpt (&optional arg) 
  "Doublequotes 'sentence at point if any. "
  (interactive "*p")
    (th-dollar 'sentence arg))

(defalias 'angle-greater-sentence-atpt 'sentence-greater-angle-atpt)
;;;###autoload
(defun sentence-greater-angle-atpt (&optional arg) 
  "Sets angles for 'sentence after point if any. "
  (interactive "*p")
    (th-greater-angle 'sentence arg))

(defalias 'angle-lower-sentence-atpt 'sentence-lower-angle-atpt)
;;;###autoload
(defun sentence-lower-angle-atpt (&optional arg) 
  "Sets angles for 'sentence after point if any. "
  (interactive "*p")
    (th-lower-angle 'sentence arg))

(defalias 'backslash-sentence-atpt 'sentence-backslash-atpt)
;;;###autoload
(defun sentence-backslash-atpt (&optional arg) 
  "Backslash 'sentence at point if any. "
  (interactive "*p")
    (th-backslash 'sentence arg))   

(defalias 'brace-sentence-atpt 'sentence-brace-atpt)
;;;###autoload
(defun sentence-brace-atpt (&optional arg) 
  "Braces 'sentence at point if any. "
  (interactive "*p")
    (th-brace 'sentence arg))

(defalias 'bracket-sentence-atpt 'sentence-bracket-atpt)
;;;###autoload
(defun sentence-bracket-atpt (&optional arg) 
  "Brackets 'sentence after point if any. "
  (interactive "*p")
    (th-bracket 'sentence arg))

;; (defalias 'comment-sentence-atpt 'sentence-comment-atpt)
;;;###autoload
(defun sentence-comment-atpt (&optional arg) 
  "Comments 'sentence at point if any. "
  (interactive "*p")
  (th-comment 'sentence arg))

(defalias 'mark-sentence-atpt 'sentence-mark-atpt)
;;;###autoload
(defun sentence-mark-atpt () 
  "Marks 'sentence at point if any. "
  (interactive)
  (th-mark 'sentence))

(defalias 'hide-sentence-atpt 'sentence-hide-atpt)
;;;###autoload
(defun sentence-hide-atpt () 
  "Toggles hide 'sentence in region or buffer if any. "
  (interactive)
  (th-hide 'sentence))

(defalias 'highlight-sentence-atpt-mode 'sentence-highlight-atpt-mode)
(defun sentence-highlight-atpt-mode (&optional arg) 
  "Toggles sentence-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'sentence arg))

(defalias 'kill-sentence-atpt 'sentence-kill-atpt)
;;;###autoload
(defun sentence-kill-atpt (&optional arg) 
  "Kills sentence at point if any. "
  (interactive "*P")
  (th-kill 'sentence arg))

(defalias 'parentize-sentence-atpt 'sentence-parentize-atpt)
;;;###autoload
(defun sentence-parentize-atpt (&optional arg) 
  "Parentizes 'sentence at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'sentence arg))

(defalias 'separate-sentence-atpt 'sentence-separate-atpt)
;;;###autoload
(defun sentence-separate-atpt (&optional arg) 
  "Separates 'sentence at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'sentence arg))

(defalias 'singlequote-sentence-atpt 'sentence-singlequote-atpt)
;;;###autoload
(defun sentence-singlequote-atpt (&optional arg) 
  "Singlequotes 'sentence at point if any. "
  (interactive "*p")
    (th-singlequote 'sentence arg))

(defalias 'slash-sentence-atpt 'sentence-slash-atpt)
;;;###autoload
(defun sentence-slash-atpt (&optional arg) 
  "Slash 'sentence at point if any. "
  (interactive "*p")
    (th-slash 'sentence arg))

(defalias 'trim-sentence-atpt 'sentence-trim-atpt)
;;;###autoload
(defun sentence-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sentence t t))

(defalias 'trim-sentence-left-atpt 'sentence-left-trim-atpt)
;;;###autoload
(defun sentence-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sentence t nil))

(defalias 'trim-sentence-right-atpt 'sentence-right-trim-atpt)
;;;###autoload
(defun sentence-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sentence nil t))

(defalias 'forward-sentence-atpt 'sentence-forward-atpt)
;;;###autoload
(defun sentence-forward-atpt (&optional arg) 
  "Moves forward over sentence at point if any, does nothing otherwise.
Returns bounds of sentence as a list. "
  (interactive "p")
  (th-forward 'sentence arg))

(defalias 'backward-sentence-atpt 'sentence-backward-atpt)
;;;###autoload
(defun sentence-backward-atpt (&optional arg) 
  "Moves backward over sentence before point if any, does nothing otherwise.
Returns bounds of sentence as a list. "
  (interactive "p")
  (th-backward 'sentence arg))

(defalias 'transpose-sentence-atpt 'sentence-transpose-atpt)
;;;###autoload
(defun sentence-transpose-atpt (&optional arg) 
  "Transposes sentence with sentence before point if any. "
  (interactive "*P")
  (th-transpose 'sentence arg))

(defalias 'sort-sentence-atpt 'sentence-sort-atpt)
;;;###autoload
(defun sentence-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts sentence's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'sentence reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-sentence-atpt 'sentence-check-atpt)
;;;###autoload
(defun sentence-check-atpt (&optional arg) 
  " sentence at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "sentence" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "sentence" "-end-position-atpt")))))
    (th-check 'sentence arg)))
   

;;;###autoload
(defun count-alnum-in-sentence-atpt (&optional beg end) 
    "Counts matches of alnum in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sentence beg end (interactive-p)))

(defun ratio-alnum-in-sentence-atpt (&optional beg end) 
    "Relation of alnum in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-sentence-atpt (&optional beg end) 
    "Counts matches of alpha in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sentence beg end (interactive-p)))

(defun ratio-alpha-in-sentence-atpt (&optional beg end) 
    "Relation of alpha in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-sentence-atpt (&optional beg end) 
    "Counts matches of ascii in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sentence beg end (interactive-p)))

(defun ratio-ascii-in-sentence-atpt (&optional beg end) 
    "Relation of ascii in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-sentence-atpt (&optional beg end) 
    "Counts matches of digit in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sentence beg end (interactive-p)))

(defun ratio-digit-in-sentence-atpt (&optional beg end) 
    "Relation of digit in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-sentence-atpt (&optional beg end) 
    "Counts matches of graph in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sentence beg end (interactive-p)))

(defun ratio-graph-in-sentence-atpt (&optional beg end) 
    "Relation of graph in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-sentence-atpt (&optional beg end) 
    "Counts matches of lower in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sentence beg end (interactive-p)))

(defun ratio-lower-in-sentence-atpt (&optional beg end) 
    "Relation of lower in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-sentence-atpt (&optional beg end) 
    "Counts matches of punct in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sentence beg end (interactive-p)))

(defun ratio-punct-in-sentence-atpt (&optional beg end) 
    "Relation of punct in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sentence beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-sentence-atpt (&optional beg end) 
    "Counts matches of upper in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sentence beg end (interactive-p)))

(defun ratio-upper-in-sentence-atpt (&optional beg end) 
    "Relation of upper in sentence if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sentence beg end t (interactive-p)))



;;;###autoload
(defun sexp-atpt (&optional arg) 
  "Returns sexp at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'sexp arg))

(defalias 'bounds-of-sexp-atpt 'sexp-bounds-atpt)
;;;###autoload
(defun sexp-bounds-atpt (&optional arg) 
  "Returns borders of sexp at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'sexp arg))

;;;###autoload
(defun sexp-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'sexp arg))

;;;###autoload
(defun sexp-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-end 'sexp arg))

(defalias 'beginning-of-sexp-atpt 'sexp-beginning-atpt)
;;;###autoload
(defun sexp-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'sexp arg))

(defalias 'end-of-sexp-atpt 'sexp-end-atpt)
;;;###autoload
(defun sexp-end-atpt (&optional arg) 
  "Goto end of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'sexp arg))

(defalias 'length-of-sexp-atpt 'sexp-length-atpt)
;;;###autoload
(defun sexp-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-length 'sexp arg))

(defalias 'copy-sexp-atpt 'sexp-copy-atpt)
;;;###autoload
(defun sexp-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'sexp arg))

(defalias 'escape-sexp-atpt 'sexp-escape-atpt)
;;;###autoload
(defun sexp-escape-atpt () 
  "Makes a copy of symbol or char-class sexp at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'sexp))

(defalias 'doublequote-sexp-atpt 'sexp-doublequote-atpt)
;;;###autoload
(defun sexp-doublequote-atpt (&optional arg) 
  "Doublequotes 'sexp at point if any. "
  (interactive "*p")
    (th-doublequote 'sexp arg))

(defalias 'dollar-sexp-atpt 'sexp-dollar-atpt)
;;;###autoload
(defun sexp-dollar-atpt (&optional arg) 
  "Doublequotes 'sexp at point if any. "
  (interactive "*p")
    (th-dollar 'sexp arg))

(defalias 'angle-greater-sexp-atpt 'sexp-greater-angle-atpt)
;;;###autoload
(defun sexp-greater-angle-atpt (&optional arg) 
  "Sets angles for 'sexp after point if any. "
  (interactive "*p")
    (th-greater-angle 'sexp arg))

(defalias 'angle-lower-sexp-atpt 'sexp-lower-angle-atpt)
;;;###autoload
(defun sexp-lower-angle-atpt (&optional arg) 
  "Sets angles for 'sexp after point if any. "
  (interactive "*p")
    (th-lower-angle 'sexp arg))

(defalias 'backslash-sexp-atpt 'sexp-backslash-atpt)
;;;###autoload
(defun sexp-backslash-atpt (&optional arg) 
  "Backslash 'sexp at point if any. "
  (interactive "*p")
    (th-backslash 'sexp arg))   

(defalias 'brace-sexp-atpt 'sexp-brace-atpt)
;;;###autoload
(defun sexp-brace-atpt (&optional arg) 
  "Braces 'sexp at point if any. "
  (interactive "*p")
    (th-brace 'sexp arg))

(defalias 'bracket-sexp-atpt 'sexp-bracket-atpt)
;;;###autoload
(defun sexp-bracket-atpt (&optional arg) 
  "Brackets 'sexp after point if any. "
  (interactive "*p")
    (th-bracket 'sexp arg))

;; (defalias 'comment-sexp-atpt 'sexp-comment-atpt)
;;;###autoload
(defun sexp-comment-atpt (&optional arg) 
  "Comments 'sexp at point if any. "
  (interactive "*p")
  (th-comment 'sexp arg))

(defalias 'mark-sexp-atpt 'sexp-mark-atpt)
;;;###autoload
(defun sexp-mark-atpt () 
  "Marks 'sexp at point if any. "
  (interactive)
  (th-mark 'sexp))

(defalias 'hide-sexp-atpt 'sexp-hide-atpt)
;;;###autoload
(defun sexp-hide-atpt () 
  "Toggles hide 'sexp in region or buffer if any. "
  (interactive)
  (th-hide 'sexp))

(defalias 'highlight-sexp-atpt-mode 'sexp-highlight-atpt-mode)
(defun sexp-highlight-atpt-mode (&optional arg) 
  "Toggles sexp-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'sexp arg))

(defalias 'kill-sexp-atpt 'sexp-kill-atpt)
;;;###autoload
(defun sexp-kill-atpt (&optional arg) 
  "Kills sexp at point if any. "
  (interactive "*P")
  (th-kill 'sexp arg))

(defalias 'parentize-sexp-atpt 'sexp-parentize-atpt)
;;;###autoload
(defun sexp-parentize-atpt (&optional arg) 
  "Parentizes 'sexp at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'sexp arg))

(defalias 'separate-sexp-atpt 'sexp-separate-atpt)
;;;###autoload
(defun sexp-separate-atpt (&optional arg) 
  "Separates 'sexp at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'sexp arg))

(defalias 'singlequote-sexp-atpt 'sexp-singlequote-atpt)
;;;###autoload
(defun sexp-singlequote-atpt (&optional arg) 
  "Singlequotes 'sexp at point if any. "
  (interactive "*p")
    (th-singlequote 'sexp arg))

(defalias 'slash-sexp-atpt 'sexp-slash-atpt)
;;;###autoload
(defun sexp-slash-atpt (&optional arg) 
  "Slash 'sexp at point if any. "
  (interactive "*p")
    (th-slash 'sexp arg))

(defalias 'trim-sexp-atpt 'sexp-trim-atpt)
;;;###autoload
(defun sexp-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sexp t t))

(defalias 'trim-sexp-left-atpt 'sexp-left-trim-atpt)
;;;###autoload
(defun sexp-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sexp t nil))

(defalias 'trim-sexp-right-atpt 'sexp-right-trim-atpt)
;;;###autoload
(defun sexp-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sexp nil t))

(defalias 'forward-sexp-atpt 'sexp-forward-atpt)
;;;###autoload
(defun sexp-forward-atpt (&optional arg) 
  "Moves forward over sexp at point if any, does nothing otherwise.
Returns bounds of sexp as a list. "
  (interactive "p")
  (th-forward 'sexp arg))

(defalias 'backward-sexp-atpt 'sexp-backward-atpt)
;;;###autoload
(defun sexp-backward-atpt (&optional arg) 
  "Moves backward over sexp before point if any, does nothing otherwise.
Returns bounds of sexp as a list. "
  (interactive "p")
  (th-backward 'sexp arg))

(defalias 'transpose-sexp-atpt 'sexp-transpose-atpt)
;;;###autoload
(defun sexp-transpose-atpt (&optional arg) 
  "Transposes sexp with sexp before point if any. "
  (interactive "*P")
  (th-transpose 'sexp arg))

(defalias 'sort-sexp-atpt 'sexp-sort-atpt)
;;;###autoload
(defun sexp-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts sexp's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'sexp reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-sexp-atpt 'sexp-check-atpt)
;;;###autoload
(defun sexp-check-atpt (&optional arg) 
  " sexp at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "sexp" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "sexp" "-end-position-atpt")))))
    (th-check 'sexp arg)))
   

;;;###autoload
(defun count-alnum-in-sexp-atpt (&optional beg end) 
    "Counts matches of alnum in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sexp beg end (interactive-p)))

(defun ratio-alnum-in-sexp-atpt (&optional beg end) 
    "Relation of alnum in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-sexp-atpt (&optional beg end) 
    "Counts matches of alpha in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sexp beg end (interactive-p)))

(defun ratio-alpha-in-sexp-atpt (&optional beg end) 
    "Relation of alpha in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-sexp-atpt (&optional beg end) 
    "Counts matches of ascii in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sexp beg end (interactive-p)))

(defun ratio-ascii-in-sexp-atpt (&optional beg end) 
    "Relation of ascii in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-sexp-atpt (&optional beg end) 
    "Counts matches of digit in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sexp beg end (interactive-p)))

(defun ratio-digit-in-sexp-atpt (&optional beg end) 
    "Relation of digit in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-sexp-atpt (&optional beg end) 
    "Counts matches of graph in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sexp beg end (interactive-p)))

(defun ratio-graph-in-sexp-atpt (&optional beg end) 
    "Relation of graph in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-sexp-atpt (&optional beg end) 
    "Counts matches of lower in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sexp beg end (interactive-p)))

(defun ratio-lower-in-sexp-atpt (&optional beg end) 
    "Relation of lower in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-sexp-atpt (&optional beg end) 
    "Counts matches of punct in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sexp beg end (interactive-p)))

(defun ratio-punct-in-sexp-atpt (&optional beg end) 
    "Relation of punct in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sexp beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-sexp-atpt (&optional beg end) 
    "Counts matches of upper in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sexp beg end (interactive-p)))

(defun ratio-upper-in-sexp-atpt (&optional beg end) 
    "Relation of upper in sexp if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sexp beg end t (interactive-p)))



;;;###autoload
(defun singlequoted-atpt (&optional arg) 
  "Returns singlequoted at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'singlequoted arg))

(defalias 'bounds-of-singlequoted-atpt 'singlequoted-bounds-atpt)
;;;###autoload
(defun singlequoted-bounds-atpt (&optional arg) 
  "Returns borders of singlequoted at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'singlequoted arg))

;;;###autoload
(defun singlequoted-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'singlequoted arg))

;;;###autoload
(defun singlequoted-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-end 'singlequoted arg))

(defalias 'beginning-of-singlequoted-atpt 'singlequoted-beginning-atpt)
;;;###autoload
(defun singlequoted-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'singlequoted arg))

(defalias 'end-of-singlequoted-atpt 'singlequoted-end-atpt)
;;;###autoload
(defun singlequoted-end-atpt (&optional arg) 
  "Goto end of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'singlequoted arg))

(defalias 'length-of-singlequoted-atpt 'singlequoted-length-atpt)
;;;###autoload
(defun singlequoted-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-length 'singlequoted arg))

(defalias 'copy-singlequoted-atpt 'singlequoted-copy-atpt)
;;;###autoload
(defun singlequoted-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'singlequoted arg))

(defalias 'escape-singlequoted-atpt 'singlequoted-escape-atpt)
;;;###autoload
(defun singlequoted-escape-atpt () 
  "Makes a copy of symbol or char-class singlequoted at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'singlequoted))

(defalias 'doublequote-singlequoted-atpt 'singlequoted-doublequote-atpt)
;;;###autoload
(defun singlequoted-doublequote-atpt (&optional arg) 
  "Doublequotes 'singlequoted at point if any. "
  (interactive "*p")
    (th-doublequote 'singlequoted arg))

(defalias 'dollar-singlequoted-atpt 'singlequoted-dollar-atpt)
;;;###autoload
(defun singlequoted-dollar-atpt (&optional arg) 
  "Doublequotes 'singlequoted at point if any. "
  (interactive "*p")
    (th-dollar 'singlequoted arg))

(defalias 'angle-greater-singlequoted-atpt 'singlequoted-greater-angle-atpt)
;;;###autoload
(defun singlequoted-greater-angle-atpt (&optional arg) 
  "Sets angles for 'singlequoted after point if any. "
  (interactive "*p")
    (th-greater-angle 'singlequoted arg))

(defalias 'angle-lower-singlequoted-atpt 'singlequoted-lower-angle-atpt)
;;;###autoload
(defun singlequoted-lower-angle-atpt (&optional arg) 
  "Sets angles for 'singlequoted after point if any. "
  (interactive "*p")
    (th-lower-angle 'singlequoted arg))

(defalias 'backslash-singlequoted-atpt 'singlequoted-backslash-atpt)
;;;###autoload
(defun singlequoted-backslash-atpt (&optional arg) 
  "Backslash 'singlequoted at point if any. "
  (interactive "*p")
    (th-backslash 'singlequoted arg))   

(defalias 'brace-singlequoted-atpt 'singlequoted-brace-atpt)
;;;###autoload
(defun singlequoted-brace-atpt (&optional arg) 
  "Braces 'singlequoted at point if any. "
  (interactive "*p")
    (th-brace 'singlequoted arg))

(defalias 'bracket-singlequoted-atpt 'singlequoted-bracket-atpt)
;;;###autoload
(defun singlequoted-bracket-atpt (&optional arg) 
  "Brackets 'singlequoted after point if any. "
  (interactive "*p")
    (th-bracket 'singlequoted arg))

;; (defalias 'comment-singlequoted-atpt 'singlequoted-comment-atpt)
;;;###autoload
(defun singlequoted-comment-atpt (&optional arg) 
  "Comments 'singlequoted at point if any. "
  (interactive "*p")
  (th-comment 'singlequoted arg))

(defalias 'mark-singlequoted-atpt 'singlequoted-mark-atpt)
;;;###autoload
(defun singlequoted-mark-atpt () 
  "Marks 'singlequoted at point if any. "
  (interactive)
  (th-mark 'singlequoted))

(defalias 'hide-singlequoted-atpt 'singlequoted-hide-atpt)
;;;###autoload
(defun singlequoted-hide-atpt () 
  "Toggles hide 'singlequoted in region or buffer if any. "
  (interactive)
  (th-hide 'singlequoted))

(defalias 'highlight-singlequoted-atpt-mode 'singlequoted-highlight-atpt-mode)
(defun singlequoted-highlight-atpt-mode (&optional arg) 
  "Toggles singlequoted-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'singlequoted arg))

(defalias 'kill-singlequoted-atpt 'singlequoted-kill-atpt)
;;;###autoload
(defun singlequoted-kill-atpt (&optional arg) 
  "Kills singlequoted at point if any. "
  (interactive "*P")
  (th-kill 'singlequoted arg))

(defalias 'parentize-singlequoted-atpt 'singlequoted-parentize-atpt)
;;;###autoload
(defun singlequoted-parentize-atpt (&optional arg) 
  "Parentizes 'singlequoted at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'singlequoted arg))

(defalias 'separate-singlequoted-atpt 'singlequoted-separate-atpt)
;;;###autoload
(defun singlequoted-separate-atpt (&optional arg) 
  "Separates 'singlequoted at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'singlequoted arg))

(defalias 'singlequote-singlequoted-atpt 'singlequoted-singlequote-atpt)
;;;###autoload
(defun singlequoted-singlequote-atpt (&optional arg) 
  "Singlequotes 'singlequoted at point if any. "
  (interactive "*p")
    (th-singlequote 'singlequoted arg))

(defalias 'slash-singlequoted-atpt 'singlequoted-slash-atpt)
;;;###autoload
(defun singlequoted-slash-atpt (&optional arg) 
  "Slash 'singlequoted at point if any. "
  (interactive "*p")
    (th-slash 'singlequoted arg))

(defalias 'trim-singlequoted-atpt 'singlequoted-trim-atpt)
;;;###autoload
(defun singlequoted-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'singlequoted t t))

(defalias 'trim-singlequoted-left-atpt 'singlequoted-left-trim-atpt)
;;;###autoload
(defun singlequoted-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'singlequoted t nil))

(defalias 'trim-singlequoted-right-atpt 'singlequoted-right-trim-atpt)
;;;###autoload
(defun singlequoted-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'singlequoted nil t))

(defalias 'forward-singlequoted-atpt 'singlequoted-forward-atpt)
;;;###autoload
(defun singlequoted-forward-atpt (&optional arg) 
  "Moves forward over singlequoted at point if any, does nothing otherwise.
Returns bounds of singlequoted as a list. "
  (interactive "p")
  (th-forward 'singlequoted arg))

(defalias 'backward-singlequoted-atpt 'singlequoted-backward-atpt)
;;;###autoload
(defun singlequoted-backward-atpt (&optional arg) 
  "Moves backward over singlequoted before point if any, does nothing otherwise.
Returns bounds of singlequoted as a list. "
  (interactive "p")
  (th-backward 'singlequoted arg))

(defalias 'transpose-singlequoted-atpt 'singlequoted-transpose-atpt)
;;;###autoload
(defun singlequoted-transpose-atpt (&optional arg) 
  "Transposes singlequoted with singlequoted before point if any. "
  (interactive "*P")
  (th-transpose 'singlequoted arg))

(defalias 'sort-singlequoted-atpt 'singlequoted-sort-atpt)
;;;###autoload
(defun singlequoted-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts singlequoted's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'singlequoted reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-singlequoted-atpt 'singlequoted-check-atpt)
;;;###autoload
(defun singlequoted-check-atpt (&optional arg) 
  " singlequoted at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "singlequoted" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "singlequoted" "-end-position-atpt")))))
    (th-check 'singlequoted arg)))
   

;;;###autoload
(defun count-alnum-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of alnum in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'singlequoted beg end (interactive-p)))

(defun ratio-alnum-in-singlequoted-atpt (&optional beg end) 
    "Relation of alnum in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of alpha in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'singlequoted beg end (interactive-p)))

(defun ratio-alpha-in-singlequoted-atpt (&optional beg end) 
    "Relation of alpha in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of ascii in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'singlequoted beg end (interactive-p)))

(defun ratio-ascii-in-singlequoted-atpt (&optional beg end) 
    "Relation of ascii in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of digit in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'singlequoted beg end (interactive-p)))

(defun ratio-digit-in-singlequoted-atpt (&optional beg end) 
    "Relation of digit in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of graph in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'singlequoted beg end (interactive-p)))

(defun ratio-graph-in-singlequoted-atpt (&optional beg end) 
    "Relation of graph in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of lower in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'singlequoted beg end (interactive-p)))

(defun ratio-lower-in-singlequoted-atpt (&optional beg end) 
    "Relation of lower in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of punct in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'singlequoted beg end (interactive-p)))

(defun ratio-punct-in-singlequoted-atpt (&optional beg end) 
    "Relation of punct in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-singlequoted-atpt (&optional beg end) 
    "Counts matches of upper in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'singlequoted beg end (interactive-p)))

(defun ratio-upper-in-singlequoted-atpt (&optional beg end) 
    "Relation of upper in singlequoted if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'singlequoted beg end t (interactive-p)))



;;;###autoload
(defun slashed-atpt (&optional arg) 
  "Returns slashed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'slashed arg))

(defalias 'bounds-of-slashed-atpt 'slashed-bounds-atpt)
;;;###autoload
(defun slashed-bounds-atpt (&optional arg) 
  "Returns borders of slashed at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'slashed arg))

;;;###autoload
(defun slashed-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'slashed arg))

;;;###autoload
(defun slashed-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-end 'slashed arg))

(defalias 'beginning-of-slashed-atpt 'slashed-beginning-atpt)
;;;###autoload
(defun slashed-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'slashed arg))

(defalias 'end-of-slashed-atpt 'slashed-end-atpt)
;;;###autoload
(defun slashed-end-atpt (&optional arg) 
  "Goto end of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'slashed arg))

(defalias 'length-of-slashed-atpt 'slashed-length-atpt)
;;;###autoload
(defun slashed-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-length 'slashed arg))

(defalias 'copy-slashed-atpt 'slashed-copy-atpt)
;;;###autoload
(defun slashed-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'slashed arg))

(defalias 'escape-slashed-atpt 'slashed-escape-atpt)
;;;###autoload
(defun slashed-escape-atpt () 
  "Makes a copy of symbol or char-class slashed at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'slashed))

(defalias 'doublequote-slashed-atpt 'slashed-doublequote-atpt)
;;;###autoload
(defun slashed-doublequote-atpt (&optional arg) 
  "Doublequotes 'slashed at point if any. "
  (interactive "*p")
    (th-doublequote 'slashed arg))

(defalias 'dollar-slashed-atpt 'slashed-dollar-atpt)
;;;###autoload
(defun slashed-dollar-atpt (&optional arg) 
  "Doublequotes 'slashed at point if any. "
  (interactive "*p")
    (th-dollar 'slashed arg))

(defalias 'angle-greater-slashed-atpt 'slashed-greater-angle-atpt)
;;;###autoload
(defun slashed-greater-angle-atpt (&optional arg) 
  "Sets angles for 'slashed after point if any. "
  (interactive "*p")
    (th-greater-angle 'slashed arg))

(defalias 'angle-lower-slashed-atpt 'slashed-lower-angle-atpt)
;;;###autoload
(defun slashed-lower-angle-atpt (&optional arg) 
  "Sets angles for 'slashed after point if any. "
  (interactive "*p")
    (th-lower-angle 'slashed arg))

(defalias 'backslash-slashed-atpt 'slashed-backslash-atpt)
;;;###autoload
(defun slashed-backslash-atpt (&optional arg) 
  "Backslash 'slashed at point if any. "
  (interactive "*p")
    (th-backslash 'slashed arg))   

(defalias 'brace-slashed-atpt 'slashed-brace-atpt)
;;;###autoload
(defun slashed-brace-atpt (&optional arg) 
  "Braces 'slashed at point if any. "
  (interactive "*p")
    (th-brace 'slashed arg))

(defalias 'bracket-slashed-atpt 'slashed-bracket-atpt)
;;;###autoload
(defun slashed-bracket-atpt (&optional arg) 
  "Brackets 'slashed after point if any. "
  (interactive "*p")
    (th-bracket 'slashed arg))

;; (defalias 'comment-slashed-atpt 'slashed-comment-atpt)
;;;###autoload
(defun slashed-comment-atpt (&optional arg) 
  "Comments 'slashed at point if any. "
  (interactive "*p")
  (th-comment 'slashed arg))

(defalias 'mark-slashed-atpt 'slashed-mark-atpt)
;;;###autoload
(defun slashed-mark-atpt () 
  "Marks 'slashed at point if any. "
  (interactive)
  (th-mark 'slashed))

(defalias 'hide-slashed-atpt 'slashed-hide-atpt)
;;;###autoload
(defun slashed-hide-atpt () 
  "Toggles hide 'slashed in region or buffer if any. "
  (interactive)
  (th-hide 'slashed))

(defalias 'highlight-slashed-atpt-mode 'slashed-highlight-atpt-mode)
(defun slashed-highlight-atpt-mode (&optional arg) 
  "Toggles slashed-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'slashed arg))

(defalias 'kill-slashed-atpt 'slashed-kill-atpt)
;;;###autoload
(defun slashed-kill-atpt (&optional arg) 
  "Kills slashed at point if any. "
  (interactive "*P")
  (th-kill 'slashed arg))

(defalias 'parentize-slashed-atpt 'slashed-parentize-atpt)
;;;###autoload
(defun slashed-parentize-atpt (&optional arg) 
  "Parentizes 'slashed at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'slashed arg))

(defalias 'separate-slashed-atpt 'slashed-separate-atpt)
;;;###autoload
(defun slashed-separate-atpt (&optional arg) 
  "Separates 'slashed at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'slashed arg))

(defalias 'singlequote-slashed-atpt 'slashed-singlequote-atpt)
;;;###autoload
(defun slashed-singlequote-atpt (&optional arg) 
  "Singlequotes 'slashed at point if any. "
  (interactive "*p")
    (th-singlequote 'slashed arg))

(defalias 'slash-slashed-atpt 'slashed-slash-atpt)
;;;###autoload
(defun slashed-slash-atpt (&optional arg) 
  "Slash 'slashed at point if any. "
  (interactive "*p")
    (th-slash 'slashed arg))

(defalias 'trim-slashed-atpt 'slashed-trim-atpt)
;;;###autoload
(defun slashed-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'slashed t t))

(defalias 'trim-slashed-left-atpt 'slashed-left-trim-atpt)
;;;###autoload
(defun slashed-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'slashed t nil))

(defalias 'trim-slashed-right-atpt 'slashed-right-trim-atpt)
;;;###autoload
(defun slashed-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'slashed nil t))

(defalias 'forward-slashed-atpt 'slashed-forward-atpt)
;;;###autoload
(defun slashed-forward-atpt (&optional arg) 
  "Moves forward over slashed at point if any, does nothing otherwise.
Returns bounds of slashed as a list. "
  (interactive "p")
  (th-forward 'slashed arg))

(defalias 'backward-slashed-atpt 'slashed-backward-atpt)
;;;###autoload
(defun slashed-backward-atpt (&optional arg) 
  "Moves backward over slashed before point if any, does nothing otherwise.
Returns bounds of slashed as a list. "
  (interactive "p")
  (th-backward 'slashed arg))

(defalias 'transpose-slashed-atpt 'slashed-transpose-atpt)
;;;###autoload
(defun slashed-transpose-atpt (&optional arg) 
  "Transposes slashed with slashed before point if any. "
  (interactive "*P")
  (th-transpose 'slashed arg))

(defalias 'sort-slashed-atpt 'slashed-sort-atpt)
;;;###autoload
(defun slashed-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts slashed's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'slashed reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-slashed-atpt 'slashed-check-atpt)
;;;###autoload
(defun slashed-check-atpt (&optional arg) 
  " slashed at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "slashed" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "slashed" "-end-position-atpt")))))
    (th-check 'slashed arg)))
   

;;;###autoload
(defun count-alnum-in-slashed-atpt (&optional beg end) 
    "Counts matches of alnum in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'slashed beg end (interactive-p)))

(defun ratio-alnum-in-slashed-atpt (&optional beg end) 
    "Relation of alnum in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-slashed-atpt (&optional beg end) 
    "Counts matches of alpha in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'slashed beg end (interactive-p)))

(defun ratio-alpha-in-slashed-atpt (&optional beg end) 
    "Relation of alpha in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-slashed-atpt (&optional beg end) 
    "Counts matches of ascii in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'slashed beg end (interactive-p)))

(defun ratio-ascii-in-slashed-atpt (&optional beg end) 
    "Relation of ascii in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-slashed-atpt (&optional beg end) 
    "Counts matches of digit in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'slashed beg end (interactive-p)))

(defun ratio-digit-in-slashed-atpt (&optional beg end) 
    "Relation of digit in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-slashed-atpt (&optional beg end) 
    "Counts matches of graph in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'slashed beg end (interactive-p)))

(defun ratio-graph-in-slashed-atpt (&optional beg end) 
    "Relation of graph in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-slashed-atpt (&optional beg end) 
    "Counts matches of lower in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'slashed beg end (interactive-p)))

(defun ratio-lower-in-slashed-atpt (&optional beg end) 
    "Relation of lower in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-slashed-atpt (&optional beg end) 
    "Counts matches of punct in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'slashed beg end (interactive-p)))

(defun ratio-punct-in-slashed-atpt (&optional beg end) 
    "Relation of punct in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'slashed beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-slashed-atpt (&optional beg end) 
    "Counts matches of upper in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'slashed beg end (interactive-p)))

(defun ratio-upper-in-slashed-atpt (&optional beg end) 
    "Relation of upper in slashed if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'slashed beg end t (interactive-p)))



;;;###autoload
(defun space-atpt (&optional arg) 
  "Returns space at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'space arg))

(defalias 'bounds-of-space-atpt 'space-bounds-atpt)
;;;###autoload
(defun space-bounds-atpt (&optional arg) 
  "Returns borders of space at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'space arg))

;;;###autoload
(defun space-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'space arg))

;;;###autoload
(defun space-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-end 'space arg))

(defalias 'beginning-of-space-atpt 'space-beginning-atpt)
;;;###autoload
(defun space-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'space arg))

(defalias 'end-of-space-atpt 'space-end-atpt)
;;;###autoload
(defun space-end-atpt (&optional arg) 
  "Goto end of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'space arg))

(defalias 'length-of-space-atpt 'space-length-atpt)
;;;###autoload
(defun space-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-length 'space arg))

(defalias 'copy-space-atpt 'space-copy-atpt)
;;;###autoload
(defun space-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class space at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'space arg))

(defalias 'escape-space-atpt 'space-escape-atpt)
;;;###autoload
(defun space-escape-atpt () 
  "Makes a copy of symbol or char-class space at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'space))

(defalias 'doublequote-space-atpt 'space-doublequote-atpt)
;;;###autoload
(defun space-doublequote-atpt (&optional arg) 
  "Doublequotes 'space at point if any. "
  (interactive "*p")
    (th-doublequote 'space arg))

(defalias 'dollar-space-atpt 'space-dollar-atpt)
;;;###autoload
(defun space-dollar-atpt (&optional arg) 
  "Doublequotes 'space at point if any. "
  (interactive "*p")
    (th-dollar 'space arg))

(defalias 'angle-greater-space-atpt 'space-greater-angle-atpt)
;;;###autoload
(defun space-greater-angle-atpt (&optional arg) 
  "Sets angles for 'space after point if any. "
  (interactive "*p")
    (th-greater-angle 'space arg))

(defalias 'angle-lower-space-atpt 'space-lower-angle-atpt)
;;;###autoload
(defun space-lower-angle-atpt (&optional arg) 
  "Sets angles for 'space after point if any. "
  (interactive "*p")
    (th-lower-angle 'space arg))

(defalias 'backslash-space-atpt 'space-backslash-atpt)
;;;###autoload
(defun space-backslash-atpt (&optional arg) 
  "Backslash 'space at point if any. "
  (interactive "*p")
    (th-backslash 'space arg))   

(defalias 'brace-space-atpt 'space-brace-atpt)
;;;###autoload
(defun space-brace-atpt (&optional arg) 
  "Braces 'space at point if any. "
  (interactive "*p")
    (th-brace 'space arg))

(defalias 'bracket-space-atpt 'space-bracket-atpt)
;;;###autoload
(defun space-bracket-atpt (&optional arg) 
  "Brackets 'space after point if any. "
  (interactive "*p")
    (th-bracket 'space arg))

;; (defalias 'comment-space-atpt 'space-comment-atpt)
;;;###autoload
(defun space-comment-atpt (&optional arg) 
  "Comments 'space at point if any. "
  (interactive "*p")
  (th-comment 'space arg))

(defalias 'mark-space-atpt 'space-mark-atpt)
;;;###autoload
(defun space-mark-atpt () 
  "Marks 'space at point if any. "
  (interactive)
  (th-mark 'space))

(defalias 'hide-space-atpt 'space-hide-atpt)
;;;###autoload
(defun space-hide-atpt () 
  "Toggles hide 'space in region or buffer if any. "
  (interactive)
  (th-hide 'space))

(defalias 'highlight-space-atpt-mode 'space-highlight-atpt-mode)
(defun space-highlight-atpt-mode (&optional arg) 
  "Toggles space-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'space arg))

(defalias 'kill-space-atpt 'space-kill-atpt)
;;;###autoload
(defun space-kill-atpt (&optional arg) 
  "Kills space at point if any. "
  (interactive "*P")
  (th-kill 'space arg))

(defalias 'parentize-space-atpt 'space-parentize-atpt)
;;;###autoload
(defun space-parentize-atpt (&optional arg) 
  "Parentizes 'space at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'space arg))

(defalias 'separate-space-atpt 'space-separate-atpt)
;;;###autoload
(defun space-separate-atpt (&optional arg) 
  "Separates 'space at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'space arg))

(defalias 'singlequote-space-atpt 'space-singlequote-atpt)
;;;###autoload
(defun space-singlequote-atpt (&optional arg) 
  "Singlequotes 'space at point if any. "
  (interactive "*p")
    (th-singlequote 'space arg))

(defalias 'slash-space-atpt 'space-slash-atpt)
;;;###autoload
(defun space-slash-atpt (&optional arg) 
  "Slash 'space at point if any. "
  (interactive "*p")
    (th-slash 'space arg))

(defalias 'trim-space-atpt 'space-trim-atpt)
;;;###autoload
(defun space-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'space t t))

(defalias 'trim-space-left-atpt 'space-left-trim-atpt)
;;;###autoload
(defun space-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'space t nil))

(defalias 'trim-space-right-atpt 'space-right-trim-atpt)
;;;###autoload
(defun space-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'space nil t))

(defalias 'forward-space-atpt 'space-forward-atpt)
;;;###autoload
(defun space-forward-atpt (&optional arg) 
  "Moves forward over space at point if any, does nothing otherwise.
Returns bounds of space as a list. "
  (interactive "p")
  (th-forward 'space arg))

(defalias 'backward-space-atpt 'space-backward-atpt)
;;;###autoload
(defun space-backward-atpt (&optional arg) 
  "Moves backward over space before point if any, does nothing otherwise.
Returns bounds of space as a list. "
  (interactive "p")
  (th-backward 'space arg))

(defalias 'transpose-space-atpt 'space-transpose-atpt)
;;;###autoload
(defun space-transpose-atpt (&optional arg) 
  "Transposes space with space before point if any. "
  (interactive "*P")
  (th-transpose 'space arg))

(defalias 'sort-space-atpt 'space-sort-atpt)
;;;###autoload
(defun space-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts space's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'space reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-space-atpt 'space-check-atpt)
;;;###autoload
(defun space-check-atpt (&optional arg) 
  " space at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "space" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "space" "-end-position-atpt")))))
    (th-check 'space arg)))
   

;;;###autoload
(defun string-atpt (&optional arg) 
  "Returns string at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'string arg))

(defalias 'bounds-of-string-atpt 'string-bounds-atpt)
;;;###autoload
(defun string-bounds-atpt (&optional arg) 
  "Returns borders of string at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'string arg))

;;;###autoload
(defun string-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'string arg))

;;;###autoload
(defun string-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-end 'string arg))

(defalias 'beginning-of-string-atpt 'string-beginning-atpt)
;;;###autoload
(defun string-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'string arg))

(defalias 'end-of-string-atpt 'string-end-atpt)
;;;###autoload
(defun string-end-atpt (&optional arg) 
  "Goto end of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'string arg))

(defalias 'length-of-string-atpt 'string-length-atpt)
;;;###autoload
(defun string-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-length 'string arg))

(defalias 'copy-string-atpt 'string-copy-atpt)
;;;###autoload
(defun string-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class string at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'string arg))

(defalias 'escape-string-atpt 'string-escape-atpt)
;;;###autoload
(defun string-escape-atpt () 
  "Makes a copy of symbol or char-class string at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'string))

(defalias 'doublequote-string-atpt 'string-doublequote-atpt)
;;;###autoload
(defun string-doublequote-atpt (&optional arg) 
  "Doublequotes 'string at point if any. "
  (interactive "*p")
    (th-doublequote 'string arg))

(defalias 'dollar-string-atpt 'string-dollar-atpt)
;;;###autoload
(defun string-dollar-atpt (&optional arg) 
  "Doublequotes 'string at point if any. "
  (interactive "*p")
    (th-dollar 'string arg))

(defalias 'angle-greater-string-atpt 'string-greater-angle-atpt)
;;;###autoload
(defun string-greater-angle-atpt (&optional arg) 
  "Sets angles for 'string after point if any. "
  (interactive "*p")
    (th-greater-angle 'string arg))

(defalias 'angle-lower-string-atpt 'string-lower-angle-atpt)
;;;###autoload
(defun string-lower-angle-atpt (&optional arg) 
  "Sets angles for 'string after point if any. "
  (interactive "*p")
    (th-lower-angle 'string arg))

(defalias 'backslash-string-atpt 'string-backslash-atpt)
;;;###autoload
(defun string-backslash-atpt (&optional arg) 
  "Backslash 'string at point if any. "
  (interactive "*p")
    (th-backslash 'string arg))   

(defalias 'brace-string-atpt 'string-brace-atpt)
;;;###autoload
(defun string-brace-atpt (&optional arg) 
  "Braces 'string at point if any. "
  (interactive "*p")
    (th-brace 'string arg))

(defalias 'bracket-string-atpt 'string-bracket-atpt)
;;;###autoload
(defun string-bracket-atpt (&optional arg) 
  "Brackets 'string after point if any. "
  (interactive "*p")
    (th-bracket 'string arg))

;; (defalias 'comment-string-atpt 'string-comment-atpt)
;;;###autoload
(defun string-comment-atpt (&optional arg) 
  "Comments 'string at point if any. "
  (interactive "*p")
  (th-comment 'string arg))

(defalias 'mark-string-atpt 'string-mark-atpt)
;;;###autoload
(defun string-mark-atpt () 
  "Marks 'string at point if any. "
  (interactive)
  (th-mark 'string))

(defalias 'hide-string-atpt 'string-hide-atpt)
;;;###autoload
(defun string-hide-atpt () 
  "Toggles hide 'string in region or buffer if any. "
  (interactive)
  (th-hide 'string))

(defalias 'highlight-string-atpt-mode 'string-highlight-atpt-mode)
(defun string-highlight-atpt-mode (&optional arg) 
  "Toggles string-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'string arg))

(defalias 'kill-string-atpt 'string-kill-atpt)
;;;###autoload
(defun string-kill-atpt (&optional arg) 
  "Kills string at point if any. "
  (interactive "*P")
  (th-kill 'string arg))

(defalias 'parentize-string-atpt 'string-parentize-atpt)
;;;###autoload
(defun string-parentize-atpt (&optional arg) 
  "Parentizes 'string at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'string arg))

(defalias 'separate-string-atpt 'string-separate-atpt)
;;;###autoload
(defun string-separate-atpt (&optional arg) 
  "Separates 'string at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'string arg))

(defalias 'singlequote-string-atpt 'string-singlequote-atpt)
;;;###autoload
(defun string-singlequote-atpt (&optional arg) 
  "Singlequotes 'string at point if any. "
  (interactive "*p")
    (th-singlequote 'string arg))

(defalias 'slash-string-atpt 'string-slash-atpt)
;;;###autoload
(defun string-slash-atpt (&optional arg) 
  "Slash 'string at point if any. "
  (interactive "*p")
    (th-slash 'string arg))

(defalias 'trim-string-atpt 'string-trim-atpt)
;;;###autoload
(defun string-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'string t t))

(defalias 'trim-string-left-atpt 'string-left-trim-atpt)
;;;###autoload
(defun string-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'string t nil))

(defalias 'trim-string-right-atpt 'string-right-trim-atpt)
;;;###autoload
(defun string-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'string nil t))

(defalias 'forward-string-atpt 'string-forward-atpt)
;;;###autoload
(defun string-forward-atpt (&optional arg) 
  "Moves forward over string at point if any, does nothing otherwise.
Returns bounds of string as a list. "
  (interactive "p")
  (th-forward 'string arg))

(defalias 'backward-string-atpt 'string-backward-atpt)
;;;###autoload
(defun string-backward-atpt (&optional arg) 
  "Moves backward over string before point if any, does nothing otherwise.
Returns bounds of string as a list. "
  (interactive "p")
  (th-backward 'string arg))

(defalias 'transpose-string-atpt 'string-transpose-atpt)
;;;###autoload
(defun string-transpose-atpt (&optional arg) 
  "Transposes string with string before point if any. "
  (interactive "*P")
  (th-transpose 'string arg))

(defalias 'sort-string-atpt 'string-sort-atpt)
;;;###autoload
(defun string-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts string's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'string reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-string-atpt 'string-check-atpt)
;;;###autoload
(defun string-check-atpt (&optional arg) 
  " string at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "string" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "string" "-end-position-atpt")))))
    (th-check 'string arg)))
   

;;;###autoload
(defun count-alnum-in-string-atpt (&optional beg end) 
    "Counts matches of alnum in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'string beg end (interactive-p)))

(defun ratio-alnum-in-string-atpt (&optional beg end) 
    "Relation of alnum in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'string beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-string-atpt (&optional beg end) 
    "Counts matches of alpha in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'string beg end (interactive-p)))

(defun ratio-alpha-in-string-atpt (&optional beg end) 
    "Relation of alpha in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'string beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-string-atpt (&optional beg end) 
    "Counts matches of ascii in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'string beg end (interactive-p)))

(defun ratio-ascii-in-string-atpt (&optional beg end) 
    "Relation of ascii in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'string beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-string-atpt (&optional beg end) 
    "Counts matches of digit in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'string beg end (interactive-p)))

(defun ratio-digit-in-string-atpt (&optional beg end) 
    "Relation of digit in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'string beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-string-atpt (&optional beg end) 
    "Counts matches of graph in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'string beg end (interactive-p)))

(defun ratio-graph-in-string-atpt (&optional beg end) 
    "Relation of graph in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'string beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-string-atpt (&optional beg end) 
    "Counts matches of lower in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'string beg end (interactive-p)))

(defun ratio-lower-in-string-atpt (&optional beg end) 
    "Relation of lower in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'string beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-string-atpt (&optional beg end) 
    "Counts matches of punct in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'string beg end (interactive-p)))

(defun ratio-punct-in-string-atpt (&optional beg end) 
    "Relation of punct in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'string beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-string-atpt (&optional beg end) 
    "Counts matches of upper in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'string beg end (interactive-p)))

(defun ratio-upper-in-string-atpt (&optional beg end) 
    "Relation of upper in string if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'string beg end t (interactive-p)))



;;;###autoload
(defun sh-struct-atpt (&optional arg) 
  "Returns sh-struct at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'sh-struct arg))

(defalias 'bounds-of-sh-struct-atpt 'sh-struct-bounds-atpt)
;;;###autoload
(defun sh-struct-bounds-atpt (&optional arg) 
  "Returns borders of sh-struct at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'sh-struct arg))

;;;###autoload
(defun sh-struct-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'sh-struct arg))

;;;###autoload
(defun sh-struct-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-end 'sh-struct arg))

(defalias 'beginning-of-sh-struct-atpt 'sh-struct-beginning-atpt)
;;;###autoload
(defun sh-struct-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'sh-struct arg))

(defalias 'end-of-sh-struct-atpt 'sh-struct-end-atpt)
;;;###autoload
(defun sh-struct-end-atpt (&optional arg) 
  "Goto end of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'sh-struct arg))

(defalias 'length-of-sh-struct-atpt 'sh-struct-length-atpt)
;;;###autoload
(defun sh-struct-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-length 'sh-struct arg))

(defalias 'copy-sh-struct-atpt 'sh-struct-copy-atpt)
;;;###autoload
(defun sh-struct-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'sh-struct arg))

(defalias 'escape-sh-struct-atpt 'sh-struct-escape-atpt)
;;;###autoload
(defun sh-struct-escape-atpt () 
  "Makes a copy of symbol or char-class sh-struct at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'sh-struct))

(defalias 'doublequote-sh-struct-atpt 'sh-struct-doublequote-atpt)
;;;###autoload
(defun sh-struct-doublequote-atpt (&optional arg) 
  "Doublequotes 'sh-struct at point if any. "
  (interactive "*p")
    (th-doublequote 'sh-struct arg))

(defalias 'dollar-sh-struct-atpt 'sh-struct-dollar-atpt)
;;;###autoload
(defun sh-struct-dollar-atpt (&optional arg) 
  "Doublequotes 'sh-struct at point if any. "
  (interactive "*p")
    (th-dollar 'sh-struct arg))

(defalias 'angle-greater-sh-struct-atpt 'sh-struct-greater-angle-atpt)
;;;###autoload
(defun sh-struct-greater-angle-atpt (&optional arg) 
  "Sets angles for 'sh-struct after point if any. "
  (interactive "*p")
    (th-greater-angle 'sh-struct arg))

(defalias 'angle-lower-sh-struct-atpt 'sh-struct-lower-angle-atpt)
;;;###autoload
(defun sh-struct-lower-angle-atpt (&optional arg) 
  "Sets angles for 'sh-struct after point if any. "
  (interactive "*p")
    (th-lower-angle 'sh-struct arg))

(defalias 'backslash-sh-struct-atpt 'sh-struct-backslash-atpt)
;;;###autoload
(defun sh-struct-backslash-atpt (&optional arg) 
  "Backslash 'sh-struct at point if any. "
  (interactive "*p")
    (th-backslash 'sh-struct arg))   

(defalias 'brace-sh-struct-atpt 'sh-struct-brace-atpt)
;;;###autoload
(defun sh-struct-brace-atpt (&optional arg) 
  "Braces 'sh-struct at point if any. "
  (interactive "*p")
    (th-brace 'sh-struct arg))

(defalias 'bracket-sh-struct-atpt 'sh-struct-bracket-atpt)
;;;###autoload
(defun sh-struct-bracket-atpt (&optional arg) 
  "Brackets 'sh-struct after point if any. "
  (interactive "*p")
    (th-bracket 'sh-struct arg))

;; (defalias 'comment-sh-struct-atpt 'sh-struct-comment-atpt)
;;;###autoload
(defun sh-struct-comment-atpt (&optional arg) 
  "Comments 'sh-struct at point if any. "
  (interactive "*p")
  (th-comment 'sh-struct arg))

(defalias 'mark-sh-struct-atpt 'sh-struct-mark-atpt)
;;;###autoload
(defun sh-struct-mark-atpt () 
  "Marks 'sh-struct at point if any. "
  (interactive)
  (th-mark 'sh-struct))

(defalias 'hide-sh-struct-atpt 'sh-struct-hide-atpt)
;;;###autoload
(defun sh-struct-hide-atpt () 
  "Toggles hide 'sh-struct in region or buffer if any. "
  (interactive)
  (th-hide 'sh-struct))

(defalias 'highlight-sh-struct-atpt-mode 'sh-struct-highlight-atpt-mode)
(defun sh-struct-highlight-atpt-mode (&optional arg) 
  "Toggles sh-struct-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'sh-struct arg))

(defalias 'kill-sh-struct-atpt 'sh-struct-kill-atpt)
;;;###autoload
(defun sh-struct-kill-atpt (&optional arg) 
  "Kills sh-struct at point if any. "
  (interactive "*P")
  (th-kill 'sh-struct arg))

(defalias 'parentize-sh-struct-atpt 'sh-struct-parentize-atpt)
;;;###autoload
(defun sh-struct-parentize-atpt (&optional arg) 
  "Parentizes 'sh-struct at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'sh-struct arg))

(defalias 'separate-sh-struct-atpt 'sh-struct-separate-atpt)
;;;###autoload
(defun sh-struct-separate-atpt (&optional arg) 
  "Separates 'sh-struct at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'sh-struct arg))

(defalias 'singlequote-sh-struct-atpt 'sh-struct-singlequote-atpt)
;;;###autoload
(defun sh-struct-singlequote-atpt (&optional arg) 
  "Singlequotes 'sh-struct at point if any. "
  (interactive "*p")
    (th-singlequote 'sh-struct arg))

(defalias 'slash-sh-struct-atpt 'sh-struct-slash-atpt)
;;;###autoload
(defun sh-struct-slash-atpt (&optional arg) 
  "Slash 'sh-struct at point if any. "
  (interactive "*p")
    (th-slash 'sh-struct arg))

(defalias 'trim-sh-struct-atpt 'sh-struct-trim-atpt)
;;;###autoload
(defun sh-struct-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sh-struct t t))

(defalias 'trim-sh-struct-left-atpt 'sh-struct-left-trim-atpt)
;;;###autoload
(defun sh-struct-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sh-struct t nil))

(defalias 'trim-sh-struct-right-atpt 'sh-struct-right-trim-atpt)
;;;###autoload
(defun sh-struct-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'sh-struct nil t))

(defalias 'forward-sh-struct-atpt 'sh-struct-forward-atpt)
;;;###autoload
(defun sh-struct-forward-atpt (&optional arg) 
  "Moves forward over sh-struct at point if any, does nothing otherwise.
Returns bounds of sh-struct as a list. "
  (interactive "p")
  (th-forward 'sh-struct arg))

(defalias 'backward-sh-struct-atpt 'sh-struct-backward-atpt)
;;;###autoload
(defun sh-struct-backward-atpt (&optional arg) 
  "Moves backward over sh-struct before point if any, does nothing otherwise.
Returns bounds of sh-struct as a list. "
  (interactive "p")
  (th-backward 'sh-struct arg))

(defalias 'transpose-sh-struct-atpt 'sh-struct-transpose-atpt)
;;;###autoload
(defun sh-struct-transpose-atpt (&optional arg) 
  "Transposes sh-struct with sh-struct before point if any. "
  (interactive "*P")
  (th-transpose 'sh-struct arg))

(defalias 'sort-sh-struct-atpt 'sh-struct-sort-atpt)
;;;###autoload
(defun sh-struct-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts sh-struct's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'sh-struct reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-sh-struct-atpt 'sh-struct-check-atpt)
;;;###autoload
(defun sh-struct-check-atpt (&optional arg) 
  " sh-struct at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "sh-struct" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "sh-struct" "-end-position-atpt")))))
    (th-check 'sh-struct arg)))
   

;;;###autoload
(defun count-alnum-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of alnum in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sh-struct beg end (interactive-p)))

(defun ratio-alnum-in-sh-struct-atpt (&optional beg end) 
    "Relation of alnum in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of alpha in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sh-struct beg end (interactive-p)))

(defun ratio-alpha-in-sh-struct-atpt (&optional beg end) 
    "Relation of alpha in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of ascii in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sh-struct beg end (interactive-p)))

(defun ratio-ascii-in-sh-struct-atpt (&optional beg end) 
    "Relation of ascii in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of digit in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sh-struct beg end (interactive-p)))

(defun ratio-digit-in-sh-struct-atpt (&optional beg end) 
    "Relation of digit in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of graph in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sh-struct beg end (interactive-p)))

(defun ratio-graph-in-sh-struct-atpt (&optional beg end) 
    "Relation of graph in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of lower in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sh-struct beg end (interactive-p)))

(defun ratio-lower-in-sh-struct-atpt (&optional beg end) 
    "Relation of lower in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of punct in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sh-struct beg end (interactive-p)))

(defun ratio-punct-in-sh-struct-atpt (&optional beg end) 
    "Relation of punct in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-sh-struct-atpt (&optional beg end) 
    "Counts matches of upper in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sh-struct beg end (interactive-p)))

(defun ratio-upper-in-sh-struct-atpt (&optional beg end) 
    "Relation of upper in sh-struct if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'sh-struct beg end t (interactive-p)))



;;;###autoload
(defun symbol-atpt (&optional arg) 
  "Returns symbol at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'symbol arg))

(defalias 'bounds-of-symbol-atpt 'symbol-bounds-atpt)
;;;###autoload
(defun symbol-bounds-atpt (&optional arg) 
  "Returns borders of symbol at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'symbol arg))

;;;###autoload
(defun symbol-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'symbol arg))

;;;###autoload
(defun symbol-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-end 'symbol arg))

(defalias 'beginning-of-symbol-atpt 'symbol-beginning-atpt)
;;;###autoload
(defun symbol-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'symbol arg))

(defalias 'end-of-symbol-atpt 'symbol-end-atpt)
;;;###autoload
(defun symbol-end-atpt (&optional arg) 
  "Goto end of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'symbol arg))

(defalias 'length-of-symbol-atpt 'symbol-length-atpt)
;;;###autoload
(defun symbol-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-length 'symbol arg))

(defalias 'copy-symbol-atpt 'symbol-copy-atpt)
;;;###autoload
(defun symbol-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'symbol arg))

(defalias 'escape-symbol-atpt 'symbol-escape-atpt)
;;;###autoload
(defun symbol-escape-atpt () 
  "Makes a copy of symbol or char-class symbol at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'symbol))

(defalias 'doublequote-symbol-atpt 'symbol-doublequote-atpt)
;;;###autoload
(defun symbol-doublequote-atpt (&optional arg) 
  "Doublequotes 'symbol at point if any. "
  (interactive "*p")
    (th-doublequote 'symbol arg))

(defalias 'dollar-symbol-atpt 'symbol-dollar-atpt)
;;;###autoload
(defun symbol-dollar-atpt (&optional arg) 
  "Doublequotes 'symbol at point if any. "
  (interactive "*p")
    (th-dollar 'symbol arg))

(defalias 'angle-greater-symbol-atpt 'symbol-greater-angle-atpt)
;;;###autoload
(defun symbol-greater-angle-atpt (&optional arg) 
  "Sets angles for 'symbol after point if any. "
  (interactive "*p")
    (th-greater-angle 'symbol arg))

(defalias 'angle-lower-symbol-atpt 'symbol-lower-angle-atpt)
;;;###autoload
(defun symbol-lower-angle-atpt (&optional arg) 
  "Sets angles for 'symbol after point if any. "
  (interactive "*p")
    (th-lower-angle 'symbol arg))

(defalias 'backslash-symbol-atpt 'symbol-backslash-atpt)
;;;###autoload
(defun symbol-backslash-atpt (&optional arg) 
  "Backslash 'symbol at point if any. "
  (interactive "*p")
    (th-backslash 'symbol arg))   

(defalias 'brace-symbol-atpt 'symbol-brace-atpt)
;;;###autoload
(defun symbol-brace-atpt (&optional arg) 
  "Braces 'symbol at point if any. "
  (interactive "*p")
    (th-brace 'symbol arg))

(defalias 'bracket-symbol-atpt 'symbol-bracket-atpt)
;;;###autoload
(defun symbol-bracket-atpt (&optional arg) 
  "Brackets 'symbol after point if any. "
  (interactive "*p")
    (th-bracket 'symbol arg))

;; (defalias 'comment-symbol-atpt 'symbol-comment-atpt)
;;;###autoload
(defun symbol-comment-atpt (&optional arg) 
  "Comments 'symbol at point if any. "
  (interactive "*p")
  (th-comment 'symbol arg))

(defalias 'mark-symbol-atpt 'symbol-mark-atpt)
;;;###autoload
(defun symbol-mark-atpt () 
  "Marks 'symbol at point if any. "
  (interactive)
  (th-mark 'symbol))

(defalias 'hide-symbol-atpt 'symbol-hide-atpt)
;;;###autoload
(defun symbol-hide-atpt () 
  "Toggles hide 'symbol in region or buffer if any. "
  (interactive)
  (th-hide 'symbol))

(defalias 'highlight-symbol-atpt-mode 'symbol-highlight-atpt-mode)
(defun symbol-highlight-atpt-mode (&optional arg) 
  "Toggles symbol-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'symbol arg))

(defalias 'kill-symbol-atpt 'symbol-kill-atpt)
;;;###autoload
(defun symbol-kill-atpt (&optional arg) 
  "Kills symbol at point if any. "
  (interactive "*P")
  (th-kill 'symbol arg))

(defalias 'parentize-symbol-atpt 'symbol-parentize-atpt)
;;;###autoload
(defun symbol-parentize-atpt (&optional arg) 
  "Parentizes 'symbol at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'symbol arg))

(defalias 'separate-symbol-atpt 'symbol-separate-atpt)
;;;###autoload
(defun symbol-separate-atpt (&optional arg) 
  "Separates 'symbol at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'symbol arg))

(defalias 'singlequote-symbol-atpt 'symbol-singlequote-atpt)
;;;###autoload
(defun symbol-singlequote-atpt (&optional arg) 
  "Singlequotes 'symbol at point if any. "
  (interactive "*p")
    (th-singlequote 'symbol arg))

(defalias 'slash-symbol-atpt 'symbol-slash-atpt)
;;;###autoload
(defun symbol-slash-atpt (&optional arg) 
  "Slash 'symbol at point if any. "
  (interactive "*p")
    (th-slash 'symbol arg))

(defalias 'trim-symbol-atpt 'symbol-trim-atpt)
;;;###autoload
(defun symbol-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'symbol t t))

(defalias 'trim-symbol-left-atpt 'symbol-left-trim-atpt)
;;;###autoload
(defun symbol-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'symbol t nil))

(defalias 'trim-symbol-right-atpt 'symbol-right-trim-atpt)
;;;###autoload
(defun symbol-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'symbol nil t))

(defalias 'forward-symbol-atpt 'symbol-forward-atpt)
;;;###autoload
(defun symbol-forward-atpt (&optional arg) 
  "Moves forward over symbol at point if any, does nothing otherwise.
Returns bounds of symbol as a list. "
  (interactive "p")
  (th-forward 'symbol arg))

(defalias 'backward-symbol-atpt 'symbol-backward-atpt)
;;;###autoload
(defun symbol-backward-atpt (&optional arg) 
  "Moves backward over symbol before point if any, does nothing otherwise.
Returns bounds of symbol as a list. "
  (interactive "p")
  (th-backward 'symbol arg))

(defalias 'transpose-symbol-atpt 'symbol-transpose-atpt)
;;;###autoload
(defun symbol-transpose-atpt (&optional arg) 
  "Transposes symbol with symbol before point if any. "
  (interactive "*P")
  (th-transpose 'symbol arg))

(defalias 'sort-symbol-atpt 'symbol-sort-atpt)
;;;###autoload
(defun symbol-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts symbol's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'symbol reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-symbol-atpt 'symbol-check-atpt)
;;;###autoload
(defun symbol-check-atpt (&optional arg) 
  " symbol at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "symbol" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "symbol" "-end-position-atpt")))))
    (th-check 'symbol arg)))
   

;;;###autoload
(defun count-alnum-in-symbol-atpt (&optional beg end) 
    "Counts matches of alnum in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'symbol beg end (interactive-p)))

(defun ratio-alnum-in-symbol-atpt (&optional beg end) 
    "Relation of alnum in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-symbol-atpt (&optional beg end) 
    "Counts matches of alpha in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'symbol beg end (interactive-p)))

(defun ratio-alpha-in-symbol-atpt (&optional beg end) 
    "Relation of alpha in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-symbol-atpt (&optional beg end) 
    "Counts matches of ascii in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'symbol beg end (interactive-p)))

(defun ratio-ascii-in-symbol-atpt (&optional beg end) 
    "Relation of ascii in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-symbol-atpt (&optional beg end) 
    "Counts matches of digit in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'symbol beg end (interactive-p)))

(defun ratio-digit-in-symbol-atpt (&optional beg end) 
    "Relation of digit in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-symbol-atpt (&optional beg end) 
    "Counts matches of graph in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'symbol beg end (interactive-p)))

(defun ratio-graph-in-symbol-atpt (&optional beg end) 
    "Relation of graph in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-symbol-atpt (&optional beg end) 
    "Counts matches of lower in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'symbol beg end (interactive-p)))

(defun ratio-lower-in-symbol-atpt (&optional beg end) 
    "Relation of lower in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-symbol-atpt (&optional beg end) 
    "Counts matches of punct in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'symbol beg end (interactive-p)))

(defun ratio-punct-in-symbol-atpt (&optional beg end) 
    "Relation of punct in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'symbol beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-symbol-atpt (&optional beg end) 
    "Counts matches of upper in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'symbol beg end (interactive-p)))

(defun ratio-upper-in-symbol-atpt (&optional beg end) 
    "Relation of upper in symbol if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'symbol beg end t (interactive-p)))



;;;###autoload
(defun td-atpt (&optional arg) 
  "Returns td at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'td arg))

(defalias 'bounds-of-td-atpt 'td-bounds-atpt)
;;;###autoload
(defun td-bounds-atpt (&optional arg) 
  "Returns borders of td at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'td arg))

;;;###autoload
(defun td-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'td arg))

;;;###autoload
(defun td-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-end 'td arg))

(defalias 'beginning-of-td-atpt 'td-beginning-atpt)
;;;###autoload
(defun td-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'td arg))

(defalias 'end-of-td-atpt 'td-end-atpt)
;;;###autoload
(defun td-end-atpt (&optional arg) 
  "Goto end of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'td arg))

(defalias 'length-of-td-atpt 'td-length-atpt)
;;;###autoload
(defun td-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-length 'td arg))

(defalias 'copy-td-atpt 'td-copy-atpt)
;;;###autoload
(defun td-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class td at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'td arg))

(defalias 'escape-td-atpt 'td-escape-atpt)
;;;###autoload
(defun td-escape-atpt () 
  "Makes a copy of symbol or char-class td at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'td))

(defalias 'doublequote-td-atpt 'td-doublequote-atpt)
;;;###autoload
(defun td-doublequote-atpt (&optional arg) 
  "Doublequotes 'td at point if any. "
  (interactive "*p")
    (th-doublequote 'td arg))

(defalias 'dollar-td-atpt 'td-dollar-atpt)
;;;###autoload
(defun td-dollar-atpt (&optional arg) 
  "Doublequotes 'td at point if any. "
  (interactive "*p")
    (th-dollar 'td arg))

(defalias 'angle-greater-td-atpt 'td-greater-angle-atpt)
;;;###autoload
(defun td-greater-angle-atpt (&optional arg) 
  "Sets angles for 'td after point if any. "
  (interactive "*p")
    (th-greater-angle 'td arg))

(defalias 'angle-lower-td-atpt 'td-lower-angle-atpt)
;;;###autoload
(defun td-lower-angle-atpt (&optional arg) 
  "Sets angles for 'td after point if any. "
  (interactive "*p")
    (th-lower-angle 'td arg))

(defalias 'backslash-td-atpt 'td-backslash-atpt)
;;;###autoload
(defun td-backslash-atpt (&optional arg) 
  "Backslash 'td at point if any. "
  (interactive "*p")
    (th-backslash 'td arg))   

(defalias 'brace-td-atpt 'td-brace-atpt)
;;;###autoload
(defun td-brace-atpt (&optional arg) 
  "Braces 'td at point if any. "
  (interactive "*p")
    (th-brace 'td arg))

(defalias 'bracket-td-atpt 'td-bracket-atpt)
;;;###autoload
(defun td-bracket-atpt (&optional arg) 
  "Brackets 'td after point if any. "
  (interactive "*p")
    (th-bracket 'td arg))

;; (defalias 'comment-td-atpt 'td-comment-atpt)
;;;###autoload
(defun td-comment-atpt (&optional arg) 
  "Comments 'td at point if any. "
  (interactive "*p")
  (th-comment 'td arg))

(defalias 'mark-td-atpt 'td-mark-atpt)
;;;###autoload
(defun td-mark-atpt () 
  "Marks 'td at point if any. "
  (interactive)
  (th-mark 'td))

(defalias 'hide-td-atpt 'td-hide-atpt)
;;;###autoload
(defun td-hide-atpt () 
  "Toggles hide 'td in region or buffer if any. "
  (interactive)
  (th-hide 'td))

(defalias 'highlight-td-atpt-mode 'td-highlight-atpt-mode)
(defun td-highlight-atpt-mode (&optional arg) 
  "Toggles td-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'td arg))

(defalias 'kill-td-atpt 'td-kill-atpt)
;;;###autoload
(defun td-kill-atpt (&optional arg) 
  "Kills td at point if any. "
  (interactive "*P")
  (th-kill 'td arg))

(defalias 'parentize-td-atpt 'td-parentize-atpt)
;;;###autoload
(defun td-parentize-atpt (&optional arg) 
  "Parentizes 'td at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'td arg))

(defalias 'separate-td-atpt 'td-separate-atpt)
;;;###autoload
(defun td-separate-atpt (&optional arg) 
  "Separates 'td at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'td arg))

(defalias 'singlequote-td-atpt 'td-singlequote-atpt)
;;;###autoload
(defun td-singlequote-atpt (&optional arg) 
  "Singlequotes 'td at point if any. "
  (interactive "*p")
    (th-singlequote 'td arg))

(defalias 'slash-td-atpt 'td-slash-atpt)
;;;###autoload
(defun td-slash-atpt (&optional arg) 
  "Slash 'td at point if any. "
  (interactive "*p")
    (th-slash 'td arg))

(defalias 'trim-td-atpt 'td-trim-atpt)
;;;###autoload
(defun td-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'td t t))

(defalias 'trim-td-left-atpt 'td-left-trim-atpt)
;;;###autoload
(defun td-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'td t nil))

(defalias 'trim-td-right-atpt 'td-right-trim-atpt)
;;;###autoload
(defun td-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'td nil t))

(defalias 'forward-td-atpt 'td-forward-atpt)
;;;###autoload
(defun td-forward-atpt (&optional arg) 
  "Moves forward over td at point if any, does nothing otherwise.
Returns bounds of td as a list. "
  (interactive "p")
  (th-forward 'td arg))

(defalias 'backward-td-atpt 'td-backward-atpt)
;;;###autoload
(defun td-backward-atpt (&optional arg) 
  "Moves backward over td before point if any, does nothing otherwise.
Returns bounds of td as a list. "
  (interactive "p")
  (th-backward 'td arg))

(defalias 'transpose-td-atpt 'td-transpose-atpt)
;;;###autoload
(defun td-transpose-atpt (&optional arg) 
  "Transposes td with td before point if any. "
  (interactive "*P")
  (th-transpose 'td arg))

(defalias 'sort-td-atpt 'td-sort-atpt)
;;;###autoload
(defun td-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts td's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'td reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-td-atpt 'td-check-atpt)
;;;###autoload
(defun td-check-atpt (&optional arg) 
  " td at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "td" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "td" "-end-position-atpt")))))
    (th-check 'td arg)))
   

;;;###autoload
(defun count-alnum-in-td-atpt (&optional beg end) 
    "Counts matches of alnum in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'td beg end (interactive-p)))

(defun ratio-alnum-in-td-atpt (&optional beg end) 
    "Relation of alnum in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'td beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-td-atpt (&optional beg end) 
    "Counts matches of alpha in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'td beg end (interactive-p)))

(defun ratio-alpha-in-td-atpt (&optional beg end) 
    "Relation of alpha in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'td beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-td-atpt (&optional beg end) 
    "Counts matches of ascii in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'td beg end (interactive-p)))

(defun ratio-ascii-in-td-atpt (&optional beg end) 
    "Relation of ascii in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'td beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-td-atpt (&optional beg end) 
    "Counts matches of digit in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'td beg end (interactive-p)))

(defun ratio-digit-in-td-atpt (&optional beg end) 
    "Relation of digit in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'td beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-td-atpt (&optional beg end) 
    "Counts matches of graph in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'td beg end (interactive-p)))

(defun ratio-graph-in-td-atpt (&optional beg end) 
    "Relation of graph in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'td beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-td-atpt (&optional beg end) 
    "Counts matches of lower in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'td beg end (interactive-p)))

(defun ratio-lower-in-td-atpt (&optional beg end) 
    "Relation of lower in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'td beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-td-atpt (&optional beg end) 
    "Counts matches of punct in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'td beg end (interactive-p)))

(defun ratio-punct-in-td-atpt (&optional beg end) 
    "Relation of punct in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'td beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-td-atpt (&optional beg end) 
    "Counts matches of upper in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'td beg end (interactive-p)))

(defun ratio-upper-in-td-atpt (&optional beg end) 
    "Relation of upper in td if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'td beg end t (interactive-p)))



;;;###autoload
(defun unibyte-atpt (&optional arg) 
  "Returns unibyte at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'unibyte arg))

(defalias 'bounds-of-unibyte-atpt 'unibyte-bounds-atpt)
;;;###autoload
(defun unibyte-bounds-atpt (&optional arg) 
  "Returns borders of unibyte at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'unibyte arg))

;;;###autoload
(defun unibyte-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'unibyte arg))

;;;###autoload
(defun unibyte-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-end 'unibyte arg))

(defalias 'beginning-of-unibyte-atpt 'unibyte-beginning-atpt)
;;;###autoload
(defun unibyte-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'unibyte arg))

(defalias 'end-of-unibyte-atpt 'unibyte-end-atpt)
;;;###autoload
(defun unibyte-end-atpt (&optional arg) 
  "Goto end of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'unibyte arg))

(defalias 'length-of-unibyte-atpt 'unibyte-length-atpt)
;;;###autoload
(defun unibyte-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-length 'unibyte arg))

(defalias 'copy-unibyte-atpt 'unibyte-copy-atpt)
;;;###autoload
(defun unibyte-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'unibyte arg))

(defalias 'escape-unibyte-atpt 'unibyte-escape-atpt)
;;;###autoload
(defun unibyte-escape-atpt () 
  "Makes a copy of symbol or char-class unibyte at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'unibyte))

(defalias 'doublequote-unibyte-atpt 'unibyte-doublequote-atpt)
;;;###autoload
(defun unibyte-doublequote-atpt (&optional arg) 
  "Doublequotes 'unibyte at point if any. "
  (interactive "*p")
    (th-doublequote 'unibyte arg))

(defalias 'dollar-unibyte-atpt 'unibyte-dollar-atpt)
;;;###autoload
(defun unibyte-dollar-atpt (&optional arg) 
  "Doublequotes 'unibyte at point if any. "
  (interactive "*p")
    (th-dollar 'unibyte arg))

(defalias 'angle-greater-unibyte-atpt 'unibyte-greater-angle-atpt)
;;;###autoload
(defun unibyte-greater-angle-atpt (&optional arg) 
  "Sets angles for 'unibyte after point if any. "
  (interactive "*p")
    (th-greater-angle 'unibyte arg))

(defalias 'angle-lower-unibyte-atpt 'unibyte-lower-angle-atpt)
;;;###autoload
(defun unibyte-lower-angle-atpt (&optional arg) 
  "Sets angles for 'unibyte after point if any. "
  (interactive "*p")
    (th-lower-angle 'unibyte arg))

(defalias 'backslash-unibyte-atpt 'unibyte-backslash-atpt)
;;;###autoload
(defun unibyte-backslash-atpt (&optional arg) 
  "Backslash 'unibyte at point if any. "
  (interactive "*p")
    (th-backslash 'unibyte arg))   

(defalias 'brace-unibyte-atpt 'unibyte-brace-atpt)
;;;###autoload
(defun unibyte-brace-atpt (&optional arg) 
  "Braces 'unibyte at point if any. "
  (interactive "*p")
    (th-brace 'unibyte arg))

(defalias 'bracket-unibyte-atpt 'unibyte-bracket-atpt)
;;;###autoload
(defun unibyte-bracket-atpt (&optional arg) 
  "Brackets 'unibyte after point if any. "
  (interactive "*p")
    (th-bracket 'unibyte arg))

;; (defalias 'comment-unibyte-atpt 'unibyte-comment-atpt)
;;;###autoload
(defun unibyte-comment-atpt (&optional arg) 
  "Comments 'unibyte at point if any. "
  (interactive "*p")
  (th-comment 'unibyte arg))

(defalias 'mark-unibyte-atpt 'unibyte-mark-atpt)
;;;###autoload
(defun unibyte-mark-atpt () 
  "Marks 'unibyte at point if any. "
  (interactive)
  (th-mark 'unibyte))

(defalias 'hide-unibyte-atpt 'unibyte-hide-atpt)
;;;###autoload
(defun unibyte-hide-atpt () 
  "Toggles hide 'unibyte in region or buffer if any. "
  (interactive)
  (th-hide 'unibyte))

(defalias 'highlight-unibyte-atpt-mode 'unibyte-highlight-atpt-mode)
(defun unibyte-highlight-atpt-mode (&optional arg) 
  "Toggles unibyte-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'unibyte arg))

(defalias 'kill-unibyte-atpt 'unibyte-kill-atpt)
;;;###autoload
(defun unibyte-kill-atpt (&optional arg) 
  "Kills unibyte at point if any. "
  (interactive "*P")
  (th-kill 'unibyte arg))

(defalias 'parentize-unibyte-atpt 'unibyte-parentize-atpt)
;;;###autoload
(defun unibyte-parentize-atpt (&optional arg) 
  "Parentizes 'unibyte at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'unibyte arg))

(defalias 'separate-unibyte-atpt 'unibyte-separate-atpt)
;;;###autoload
(defun unibyte-separate-atpt (&optional arg) 
  "Separates 'unibyte at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'unibyte arg))

(defalias 'singlequote-unibyte-atpt 'unibyte-singlequote-atpt)
;;;###autoload
(defun unibyte-singlequote-atpt (&optional arg) 
  "Singlequotes 'unibyte at point if any. "
  (interactive "*p")
    (th-singlequote 'unibyte arg))

(defalias 'slash-unibyte-atpt 'unibyte-slash-atpt)
;;;###autoload
(defun unibyte-slash-atpt (&optional arg) 
  "Slash 'unibyte at point if any. "
  (interactive "*p")
    (th-slash 'unibyte arg))

(defalias 'trim-unibyte-atpt 'unibyte-trim-atpt)
;;;###autoload
(defun unibyte-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'unibyte t t))

(defalias 'trim-unibyte-left-atpt 'unibyte-left-trim-atpt)
;;;###autoload
(defun unibyte-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'unibyte t nil))

(defalias 'trim-unibyte-right-atpt 'unibyte-right-trim-atpt)
;;;###autoload
(defun unibyte-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'unibyte nil t))

(defalias 'forward-unibyte-atpt 'unibyte-forward-atpt)
;;;###autoload
(defun unibyte-forward-atpt (&optional arg) 
  "Moves forward over unibyte at point if any, does nothing otherwise.
Returns bounds of unibyte as a list. "
  (interactive "p")
  (th-forward 'unibyte arg))

(defalias 'backward-unibyte-atpt 'unibyte-backward-atpt)
;;;###autoload
(defun unibyte-backward-atpt (&optional arg) 
  "Moves backward over unibyte before point if any, does nothing otherwise.
Returns bounds of unibyte as a list. "
  (interactive "p")
  (th-backward 'unibyte arg))

(defalias 'transpose-unibyte-atpt 'unibyte-transpose-atpt)
;;;###autoload
(defun unibyte-transpose-atpt (&optional arg) 
  "Transposes unibyte with unibyte before point if any. "
  (interactive "*P")
  (th-transpose 'unibyte arg))

(defalias 'sort-unibyte-atpt 'unibyte-sort-atpt)
;;;###autoload
(defun unibyte-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts unibyte's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'unibyte reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-unibyte-atpt 'unibyte-check-atpt)
;;;###autoload
(defun unibyte-check-atpt (&optional arg) 
  " unibyte at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "unibyte" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "unibyte" "-end-position-atpt")))))
    (th-check 'unibyte arg)))
   

;;;###autoload
(defun upper-atpt (&optional arg) 
  "Returns upper at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'upper arg))

(defalias 'bounds-of-upper-atpt 'upper-bounds-atpt)
;;;###autoload
(defun upper-bounds-atpt (&optional arg) 
  "Returns borders of upper at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'upper arg))

;;;###autoload
(defun upper-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'upper arg))

;;;###autoload
(defun upper-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-end 'upper arg))

(defalias 'beginning-of-upper-atpt 'upper-beginning-atpt)
;;;###autoload
(defun upper-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'upper arg))

(defalias 'end-of-upper-atpt 'upper-end-atpt)
;;;###autoload
(defun upper-end-atpt (&optional arg) 
  "Goto end of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'upper arg))

(defalias 'length-of-upper-atpt 'upper-length-atpt)
;;;###autoload
(defun upper-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-length 'upper arg))

(defalias 'copy-upper-atpt 'upper-copy-atpt)
;;;###autoload
(defun upper-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class upper at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'upper arg))

(defalias 'escape-upper-atpt 'upper-escape-atpt)
;;;###autoload
(defun upper-escape-atpt () 
  "Makes a copy of symbol or char-class upper at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'upper))

(defalias 'doublequote-upper-atpt 'upper-doublequote-atpt)
;;;###autoload
(defun upper-doublequote-atpt (&optional arg) 
  "Doublequotes 'upper at point if any. "
  (interactive "*p")
    (th-doublequote 'upper arg))

(defalias 'dollar-upper-atpt 'upper-dollar-atpt)
;;;###autoload
(defun upper-dollar-atpt (&optional arg) 
  "Doublequotes 'upper at point if any. "
  (interactive "*p")
    (th-dollar 'upper arg))

(defalias 'angle-greater-upper-atpt 'upper-greater-angle-atpt)
;;;###autoload
(defun upper-greater-angle-atpt (&optional arg) 
  "Sets angles for 'upper after point if any. "
  (interactive "*p")
    (th-greater-angle 'upper arg))

(defalias 'angle-lower-upper-atpt 'upper-lower-angle-atpt)
;;;###autoload
(defun upper-lower-angle-atpt (&optional arg) 
  "Sets angles for 'upper after point if any. "
  (interactive "*p")
    (th-lower-angle 'upper arg))

(defalias 'backslash-upper-atpt 'upper-backslash-atpt)
;;;###autoload
(defun upper-backslash-atpt (&optional arg) 
  "Backslash 'upper at point if any. "
  (interactive "*p")
    (th-backslash 'upper arg))   

(defalias 'brace-upper-atpt 'upper-brace-atpt)
;;;###autoload
(defun upper-brace-atpt (&optional arg) 
  "Braces 'upper at point if any. "
  (interactive "*p")
    (th-brace 'upper arg))

(defalias 'bracket-upper-atpt 'upper-bracket-atpt)
;;;###autoload
(defun upper-bracket-atpt (&optional arg) 
  "Brackets 'upper after point if any. "
  (interactive "*p")
    (th-bracket 'upper arg))

;; (defalias 'comment-upper-atpt 'upper-comment-atpt)
;;;###autoload
(defun upper-comment-atpt (&optional arg) 
  "Comments 'upper at point if any. "
  (interactive "*p")
  (th-comment 'upper arg))

(defalias 'mark-upper-atpt 'upper-mark-atpt)
;;;###autoload
(defun upper-mark-atpt () 
  "Marks 'upper at point if any. "
  (interactive)
  (th-mark 'upper))

(defalias 'hide-upper-atpt 'upper-hide-atpt)
;;;###autoload
(defun upper-hide-atpt () 
  "Toggles hide 'upper in region or buffer if any. "
  (interactive)
  (th-hide 'upper))

(defalias 'highlight-upper-atpt-mode 'upper-highlight-atpt-mode)
(defun upper-highlight-atpt-mode (&optional arg) 
  "Toggles upper-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'upper arg))

(defalias 'kill-upper-atpt 'upper-kill-atpt)
;;;###autoload
(defun upper-kill-atpt (&optional arg) 
  "Kills upper at point if any. "
  (interactive "*P")
  (th-kill 'upper arg))

(defalias 'parentize-upper-atpt 'upper-parentize-atpt)
;;;###autoload
(defun upper-parentize-atpt (&optional arg) 
  "Parentizes 'upper at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'upper arg))

(defalias 'separate-upper-atpt 'upper-separate-atpt)
;;;###autoload
(defun upper-separate-atpt (&optional arg) 
  "Separates 'upper at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'upper arg))

(defalias 'singlequote-upper-atpt 'upper-singlequote-atpt)
;;;###autoload
(defun upper-singlequote-atpt (&optional arg) 
  "Singlequotes 'upper at point if any. "
  (interactive "*p")
    (th-singlequote 'upper arg))

(defalias 'slash-upper-atpt 'upper-slash-atpt)
;;;###autoload
(defun upper-slash-atpt (&optional arg) 
  "Slash 'upper at point if any. "
  (interactive "*p")
    (th-slash 'upper arg))

(defalias 'trim-upper-atpt 'upper-trim-atpt)
;;;###autoload
(defun upper-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'upper t t))

(defalias 'trim-upper-left-atpt 'upper-left-trim-atpt)
;;;###autoload
(defun upper-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'upper t nil))

(defalias 'trim-upper-right-atpt 'upper-right-trim-atpt)
;;;###autoload
(defun upper-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'upper nil t))

(defalias 'forward-upper-atpt 'upper-forward-atpt)
;;;###autoload
(defun upper-forward-atpt (&optional arg) 
  "Moves forward over upper at point if any, does nothing otherwise.
Returns bounds of upper as a list. "
  (interactive "p")
  (th-forward 'upper arg))

(defalias 'backward-upper-atpt 'upper-backward-atpt)
;;;###autoload
(defun upper-backward-atpt (&optional arg) 
  "Moves backward over upper before point if any, does nothing otherwise.
Returns bounds of upper as a list. "
  (interactive "p")
  (th-backward 'upper arg))

(defalias 'transpose-upper-atpt 'upper-transpose-atpt)
;;;###autoload
(defun upper-transpose-atpt (&optional arg) 
  "Transposes upper with upper before point if any. "
  (interactive "*P")
  (th-transpose 'upper arg))

(defalias 'sort-upper-atpt 'upper-sort-atpt)
;;;###autoload
(defun upper-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts upper's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'upper reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-upper-atpt 'upper-check-atpt)
;;;###autoload
(defun upper-check-atpt (&optional arg) 
  " upper at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "upper" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "upper" "-end-position-atpt")))))
    (th-check 'upper arg)))
   

;;;###autoload
(defun count-alnum-in-upper-atpt (&optional beg end) 
    "Counts matches of alnum in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'upper beg end (interactive-p)))

(defun ratio-alnum-in-upper-atpt (&optional beg end) 
    "Relation of alnum in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-upper-atpt (&optional beg end) 
    "Counts matches of alpha in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'upper beg end (interactive-p)))

(defun ratio-alpha-in-upper-atpt (&optional beg end) 
    "Relation of alpha in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-upper-atpt (&optional beg end) 
    "Counts matches of ascii in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'upper beg end (interactive-p)))

(defun ratio-ascii-in-upper-atpt (&optional beg end) 
    "Relation of ascii in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-upper-atpt (&optional beg end) 
    "Counts matches of digit in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'upper beg end (interactive-p)))

(defun ratio-digit-in-upper-atpt (&optional beg end) 
    "Relation of digit in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-upper-atpt (&optional beg end) 
    "Counts matches of graph in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'upper beg end (interactive-p)))

(defun ratio-graph-in-upper-atpt (&optional beg end) 
    "Relation of graph in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-upper-atpt (&optional beg end) 
    "Counts matches of lower in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'upper beg end (interactive-p)))

(defun ratio-lower-in-upper-atpt (&optional beg end) 
    "Relation of lower in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-upper-atpt (&optional beg end) 
    "Counts matches of punct in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'upper beg end (interactive-p)))

(defun ratio-punct-in-upper-atpt (&optional beg end) 
    "Relation of punct in upper if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'upper beg end t (interactive-p)))



;;;###autoload
(defun count-upper-atpt (&optional beg end) 
    "Counts matches of upper at point if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'upper beg end (interactive-p)))


;;;###autoload
(defun url-atpt (&optional arg) 
  "Returns url at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'url arg))

(defalias 'bounds-of-url-atpt 'url-bounds-atpt)
;;;###autoload
(defun url-bounds-atpt (&optional arg) 
  "Returns borders of url at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'url arg))

;;;###autoload
(defun url-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'url arg))

;;;###autoload
(defun url-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-end 'url arg))

(defalias 'beginning-of-url-atpt 'url-beginning-atpt)
;;;###autoload
(defun url-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'url arg))

(defalias 'end-of-url-atpt 'url-end-atpt)
;;;###autoload
(defun url-end-atpt (&optional arg) 
  "Goto end of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'url arg))

(defalias 'length-of-url-atpt 'url-length-atpt)
;;;###autoload
(defun url-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-length 'url arg))

(defalias 'copy-url-atpt 'url-copy-atpt)
;;;###autoload
(defun url-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class url at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'url arg))

(defalias 'escape-url-atpt 'url-escape-atpt)
;;;###autoload
(defun url-escape-atpt () 
  "Makes a copy of symbol or char-class url at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'url))

(defalias 'doublequote-url-atpt 'url-doublequote-atpt)
;;;###autoload
(defun url-doublequote-atpt (&optional arg) 
  "Doublequotes 'url at point if any. "
  (interactive "*p")
    (th-doublequote 'url arg))

(defalias 'dollar-url-atpt 'url-dollar-atpt)
;;;###autoload
(defun url-dollar-atpt (&optional arg) 
  "Doublequotes 'url at point if any. "
  (interactive "*p")
    (th-dollar 'url arg))

(defalias 'angle-greater-url-atpt 'url-greater-angle-atpt)
;;;###autoload
(defun url-greater-angle-atpt (&optional arg) 
  "Sets angles for 'url after point if any. "
  (interactive "*p")
    (th-greater-angle 'url arg))

(defalias 'angle-lower-url-atpt 'url-lower-angle-atpt)
;;;###autoload
(defun url-lower-angle-atpt (&optional arg) 
  "Sets angles for 'url after point if any. "
  (interactive "*p")
    (th-lower-angle 'url arg))

(defalias 'backslash-url-atpt 'url-backslash-atpt)
;;;###autoload
(defun url-backslash-atpt (&optional arg) 
  "Backslash 'url at point if any. "
  (interactive "*p")
    (th-backslash 'url arg))   

(defalias 'brace-url-atpt 'url-brace-atpt)
;;;###autoload
(defun url-brace-atpt (&optional arg) 
  "Braces 'url at point if any. "
  (interactive "*p")
    (th-brace 'url arg))

(defalias 'bracket-url-atpt 'url-bracket-atpt)
;;;###autoload
(defun url-bracket-atpt (&optional arg) 
  "Brackets 'url after point if any. "
  (interactive "*p")
    (th-bracket 'url arg))

;; (defalias 'comment-url-atpt 'url-comment-atpt)
;;;###autoload
(defun url-comment-atpt (&optional arg) 
  "Comments 'url at point if any. "
  (interactive "*p")
  (th-comment 'url arg))

(defalias 'mark-url-atpt 'url-mark-atpt)
;;;###autoload
(defun url-mark-atpt () 
  "Marks 'url at point if any. "
  (interactive)
  (th-mark 'url))

(defalias 'hide-url-atpt 'url-hide-atpt)
;;;###autoload
(defun url-hide-atpt () 
  "Toggles hide 'url in region or buffer if any. "
  (interactive)
  (th-hide 'url))

(defalias 'highlight-url-atpt-mode 'url-highlight-atpt-mode)
(defun url-highlight-atpt-mode (&optional arg) 
  "Toggles url-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'url arg))

(defalias 'kill-url-atpt 'url-kill-atpt)
;;;###autoload
(defun url-kill-atpt (&optional arg) 
  "Kills url at point if any. "
  (interactive "*P")
  (th-kill 'url arg))

(defalias 'parentize-url-atpt 'url-parentize-atpt)
;;;###autoload
(defun url-parentize-atpt (&optional arg) 
  "Parentizes 'url at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'url arg))

(defalias 'separate-url-atpt 'url-separate-atpt)
;;;###autoload
(defun url-separate-atpt (&optional arg) 
  "Separates 'url at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'url arg))

(defalias 'singlequote-url-atpt 'url-singlequote-atpt)
;;;###autoload
(defun url-singlequote-atpt (&optional arg) 
  "Singlequotes 'url at point if any. "
  (interactive "*p")
    (th-singlequote 'url arg))

(defalias 'slash-url-atpt 'url-slash-atpt)
;;;###autoload
(defun url-slash-atpt (&optional arg) 
  "Slash 'url at point if any. "
  (interactive "*p")
    (th-slash 'url arg))

(defalias 'trim-url-atpt 'url-trim-atpt)
;;;###autoload
(defun url-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'url t t))

(defalias 'trim-url-left-atpt 'url-left-trim-atpt)
;;;###autoload
(defun url-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'url t nil))

(defalias 'trim-url-right-atpt 'url-right-trim-atpt)
;;;###autoload
(defun url-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'url nil t))

(defalias 'forward-url-atpt 'url-forward-atpt)
;;;###autoload
(defun url-forward-atpt (&optional arg) 
  "Moves forward over url at point if any, does nothing otherwise.
Returns bounds of url as a list. "
  (interactive "p")
  (th-forward 'url arg))

(defalias 'backward-url-atpt 'url-backward-atpt)
;;;###autoload
(defun url-backward-atpt (&optional arg) 
  "Moves backward over url before point if any, does nothing otherwise.
Returns bounds of url as a list. "
  (interactive "p")
  (th-backward 'url arg))

(defalias 'transpose-url-atpt 'url-transpose-atpt)
;;;###autoload
(defun url-transpose-atpt (&optional arg) 
  "Transposes url with url before point if any. "
  (interactive "*P")
  (th-transpose 'url arg))

(defalias 'sort-url-atpt 'url-sort-atpt)
;;;###autoload
(defun url-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts url's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'url reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-url-atpt 'url-check-atpt)
;;;###autoload
(defun url-check-atpt (&optional arg) 
  " url at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "url" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "url" "-end-position-atpt")))))
    (th-check 'url arg)))
   

;;;###autoload
(defun count-alnum-in-url-atpt (&optional beg end) 
    "Counts matches of alnum in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'url beg end (interactive-p)))

(defun ratio-alnum-in-url-atpt (&optional beg end) 
    "Relation of alnum in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'url beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-url-atpt (&optional beg end) 
    "Counts matches of alpha in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'url beg end (interactive-p)))

(defun ratio-alpha-in-url-atpt (&optional beg end) 
    "Relation of alpha in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'url beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-url-atpt (&optional beg end) 
    "Counts matches of ascii in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'url beg end (interactive-p)))

(defun ratio-ascii-in-url-atpt (&optional beg end) 
    "Relation of ascii in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'url beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-url-atpt (&optional beg end) 
    "Counts matches of digit in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'url beg end (interactive-p)))

(defun ratio-digit-in-url-atpt (&optional beg end) 
    "Relation of digit in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'url beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-url-atpt (&optional beg end) 
    "Counts matches of graph in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'url beg end (interactive-p)))

(defun ratio-graph-in-url-atpt (&optional beg end) 
    "Relation of graph in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'url beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-url-atpt (&optional beg end) 
    "Counts matches of lower in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'url beg end (interactive-p)))

(defun ratio-lower-in-url-atpt (&optional beg end) 
    "Relation of lower in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'url beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-url-atpt (&optional beg end) 
    "Counts matches of punct in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'url beg end (interactive-p)))

(defun ratio-punct-in-url-atpt (&optional beg end) 
    "Relation of punct in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'url beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-url-atpt (&optional beg end) 
    "Counts matches of upper in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'url beg end (interactive-p)))

(defun ratio-upper-in-url-atpt (&optional beg end) 
    "Relation of upper in url if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'url beg end t (interactive-p)))



;;;###autoload
(defun whitespace-atpt (&optional arg) 
  "Returns whitespace at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'whitespace arg))

(defalias 'bounds-of-whitespace-atpt 'whitespace-bounds-atpt)
;;;###autoload
(defun whitespace-bounds-atpt (&optional arg) 
  "Returns borders of whitespace at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'whitespace arg))

;;;###autoload
(defun whitespace-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'whitespace arg))

;;;###autoload
(defun whitespace-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-end 'whitespace arg))

(defalias 'beginning-of-whitespace-atpt 'whitespace-beginning-atpt)
;;;###autoload
(defun whitespace-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'whitespace arg))

(defalias 'end-of-whitespace-atpt 'whitespace-end-atpt)
;;;###autoload
(defun whitespace-end-atpt (&optional arg) 
  "Goto end of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'whitespace arg))

(defalias 'length-of-whitespace-atpt 'whitespace-length-atpt)
;;;###autoload
(defun whitespace-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-length 'whitespace arg))

(defalias 'copy-whitespace-atpt 'whitespace-copy-atpt)
;;;###autoload
(defun whitespace-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'whitespace arg))

(defalias 'escape-whitespace-atpt 'whitespace-escape-atpt)
;;;###autoload
(defun whitespace-escape-atpt () 
  "Makes a copy of symbol or char-class whitespace at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'whitespace))

(defalias 'doublequote-whitespace-atpt 'whitespace-doublequote-atpt)
;;;###autoload
(defun whitespace-doublequote-atpt (&optional arg) 
  "Doublequotes 'whitespace at point if any. "
  (interactive "*p")
    (th-doublequote 'whitespace arg))

(defalias 'dollar-whitespace-atpt 'whitespace-dollar-atpt)
;;;###autoload
(defun whitespace-dollar-atpt (&optional arg) 
  "Doublequotes 'whitespace at point if any. "
  (interactive "*p")
    (th-dollar 'whitespace arg))

(defalias 'angle-greater-whitespace-atpt 'whitespace-greater-angle-atpt)
;;;###autoload
(defun whitespace-greater-angle-atpt (&optional arg) 
  "Sets angles for 'whitespace after point if any. "
  (interactive "*p")
    (th-greater-angle 'whitespace arg))

(defalias 'angle-lower-whitespace-atpt 'whitespace-lower-angle-atpt)
;;;###autoload
(defun whitespace-lower-angle-atpt (&optional arg) 
  "Sets angles for 'whitespace after point if any. "
  (interactive "*p")
    (th-lower-angle 'whitespace arg))

(defalias 'backslash-whitespace-atpt 'whitespace-backslash-atpt)
;;;###autoload
(defun whitespace-backslash-atpt (&optional arg) 
  "Backslash 'whitespace at point if any. "
  (interactive "*p")
    (th-backslash 'whitespace arg))   

(defalias 'brace-whitespace-atpt 'whitespace-brace-atpt)
;;;###autoload
(defun whitespace-brace-atpt (&optional arg) 
  "Braces 'whitespace at point if any. "
  (interactive "*p")
    (th-brace 'whitespace arg))

(defalias 'bracket-whitespace-atpt 'whitespace-bracket-atpt)
;;;###autoload
(defun whitespace-bracket-atpt (&optional arg) 
  "Brackets 'whitespace after point if any. "
  (interactive "*p")
    (th-bracket 'whitespace arg))

;; (defalias 'comment-whitespace-atpt 'whitespace-comment-atpt)
;;;###autoload
(defun whitespace-comment-atpt (&optional arg) 
  "Comments 'whitespace at point if any. "
  (interactive "*p")
  (th-comment 'whitespace arg))

(defalias 'mark-whitespace-atpt 'whitespace-mark-atpt)
;;;###autoload
(defun whitespace-mark-atpt () 
  "Marks 'whitespace at point if any. "
  (interactive)
  (th-mark 'whitespace))

(defalias 'hide-whitespace-atpt 'whitespace-hide-atpt)
;;;###autoload
(defun whitespace-hide-atpt () 
  "Toggles hide 'whitespace in region or buffer if any. "
  (interactive)
  (th-hide 'whitespace))

(defalias 'highlight-whitespace-atpt-mode 'whitespace-highlight-atpt-mode)
(defun whitespace-highlight-atpt-mode (&optional arg) 
  "Toggles whitespace-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'whitespace arg))

(defalias 'kill-whitespace-atpt 'whitespace-kill-atpt)
;;;###autoload
(defun whitespace-kill-atpt (&optional arg) 
  "Kills whitespace at point if any. "
  (interactive "*P")
  (th-kill 'whitespace arg))

(defalias 'parentize-whitespace-atpt 'whitespace-parentize-atpt)
;;;###autoload
(defun whitespace-parentize-atpt (&optional arg) 
  "Parentizes 'whitespace at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'whitespace arg))

(defalias 'separate-whitespace-atpt 'whitespace-separate-atpt)
;;;###autoload
(defun whitespace-separate-atpt (&optional arg) 
  "Separates 'whitespace at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'whitespace arg))

(defalias 'singlequote-whitespace-atpt 'whitespace-singlequote-atpt)
;;;###autoload
(defun whitespace-singlequote-atpt (&optional arg) 
  "Singlequotes 'whitespace at point if any. "
  (interactive "*p")
    (th-singlequote 'whitespace arg))

(defalias 'slash-whitespace-atpt 'whitespace-slash-atpt)
;;;###autoload
(defun whitespace-slash-atpt (&optional arg) 
  "Slash 'whitespace at point if any. "
  (interactive "*p")
    (th-slash 'whitespace arg))

(defalias 'trim-whitespace-atpt 'whitespace-trim-atpt)
;;;###autoload
(defun whitespace-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'whitespace t t))

(defalias 'trim-whitespace-left-atpt 'whitespace-left-trim-atpt)
;;;###autoload
(defun whitespace-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'whitespace t nil))

(defalias 'trim-whitespace-right-atpt 'whitespace-right-trim-atpt)
;;;###autoload
(defun whitespace-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'whitespace nil t))

(defalias 'forward-whitespace-atpt 'whitespace-forward-atpt)
;;;###autoload
(defun whitespace-forward-atpt (&optional arg) 
  "Moves forward over whitespace at point if any, does nothing otherwise.
Returns bounds of whitespace as a list. "
  (interactive "p")
  (th-forward 'whitespace arg))

(defalias 'backward-whitespace-atpt 'whitespace-backward-atpt)
;;;###autoload
(defun whitespace-backward-atpt (&optional arg) 
  "Moves backward over whitespace before point if any, does nothing otherwise.
Returns bounds of whitespace as a list. "
  (interactive "p")
  (th-backward 'whitespace arg))

(defalias 'transpose-whitespace-atpt 'whitespace-transpose-atpt)
;;;###autoload
(defun whitespace-transpose-atpt (&optional arg) 
  "Transposes whitespace with whitespace before point if any. "
  (interactive "*P")
  (th-transpose 'whitespace arg))

(defalias 'sort-whitespace-atpt 'whitespace-sort-atpt)
;;;###autoload
(defun whitespace-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts whitespace's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'whitespace reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-whitespace-atpt 'whitespace-check-atpt)
;;;###autoload
(defun whitespace-check-atpt (&optional arg) 
  " whitespace at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "whitespace" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "whitespace" "-end-position-atpt")))))
    (th-check 'whitespace arg)))
   

;;;###autoload
(defun word-atpt (&optional arg) 
  "Returns word at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'word arg))

(defalias 'bounds-of-word-atpt 'word-bounds-atpt)
;;;###autoload
(defun word-bounds-atpt (&optional arg) 
  "Returns borders of word at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'word arg))

;;;###autoload
(defun word-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'word arg))

;;;###autoload
(defun word-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-end 'word arg))

(defalias 'beginning-of-word-atpt 'word-beginning-atpt)
;;;###autoload
(defun word-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'word arg))

(defalias 'end-of-word-atpt 'word-end-atpt)
;;;###autoload
(defun word-end-atpt (&optional arg) 
  "Goto end of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'word arg))

(defalias 'length-of-word-atpt 'word-length-atpt)
;;;###autoload
(defun word-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-length 'word arg))

(defalias 'copy-word-atpt 'word-copy-atpt)
;;;###autoload
(defun word-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class word at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'word arg))

(defalias 'escape-word-atpt 'word-escape-atpt)
;;;###autoload
(defun word-escape-atpt () 
  "Makes a copy of symbol or char-class word at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'word))

(defalias 'doublequote-word-atpt 'word-doublequote-atpt)
;;;###autoload
(defun word-doublequote-atpt (&optional arg) 
  "Doublequotes 'word at point if any. "
  (interactive "*p")
    (th-doublequote 'word arg))

(defalias 'dollar-word-atpt 'word-dollar-atpt)
;;;###autoload
(defun word-dollar-atpt (&optional arg) 
  "Doublequotes 'word at point if any. "
  (interactive "*p")
    (th-dollar 'word arg))

(defalias 'angle-greater-word-atpt 'word-greater-angle-atpt)
;;;###autoload
(defun word-greater-angle-atpt (&optional arg) 
  "Sets angles for 'word after point if any. "
  (interactive "*p")
    (th-greater-angle 'word arg))

(defalias 'angle-lower-word-atpt 'word-lower-angle-atpt)
;;;###autoload
(defun word-lower-angle-atpt (&optional arg) 
  "Sets angles for 'word after point if any. "
  (interactive "*p")
    (th-lower-angle 'word arg))

(defalias 'backslash-word-atpt 'word-backslash-atpt)
;;;###autoload
(defun word-backslash-atpt (&optional arg) 
  "Backslash 'word at point if any. "
  (interactive "*p")
    (th-backslash 'word arg))   

(defalias 'brace-word-atpt 'word-brace-atpt)
;;;###autoload
(defun word-brace-atpt (&optional arg) 
  "Braces 'word at point if any. "
  (interactive "*p")
    (th-brace 'word arg))

(defalias 'bracket-word-atpt 'word-bracket-atpt)
;;;###autoload
(defun word-bracket-atpt (&optional arg) 
  "Brackets 'word after point if any. "
  (interactive "*p")
    (th-bracket 'word arg))

;; (defalias 'comment-word-atpt 'word-comment-atpt)
;;;###autoload
(defun word-comment-atpt (&optional arg) 
  "Comments 'word at point if any. "
  (interactive "*p")
  (th-comment 'word arg))

(defalias 'mark-word-atpt 'word-mark-atpt)
;;;###autoload
(defun word-mark-atpt () 
  "Marks 'word at point if any. "
  (interactive)
  (th-mark 'word))

(defalias 'hide-word-atpt 'word-hide-atpt)
;;;###autoload
(defun word-hide-atpt () 
  "Toggles hide 'word in region or buffer if any. "
  (interactive)
  (th-hide 'word))

(defalias 'highlight-word-atpt-mode 'word-highlight-atpt-mode)
(defun word-highlight-atpt-mode (&optional arg) 
  "Toggles word-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'word arg))

(defalias 'kill-word-atpt 'word-kill-atpt)
;;;###autoload
(defun word-kill-atpt (&optional arg) 
  "Kills word at point if any. "
  (interactive "*P")
  (th-kill 'word arg))

(defalias 'parentize-word-atpt 'word-parentize-atpt)
;;;###autoload
(defun word-parentize-atpt (&optional arg) 
  "Parentizes 'word at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'word arg))

(defalias 'separate-word-atpt 'word-separate-atpt)
;;;###autoload
(defun word-separate-atpt (&optional arg) 
  "Separates 'word at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'word arg))

(defalias 'singlequote-word-atpt 'word-singlequote-atpt)
;;;###autoload
(defun word-singlequote-atpt (&optional arg) 
  "Singlequotes 'word at point if any. "
  (interactive "*p")
    (th-singlequote 'word arg))

(defalias 'slash-word-atpt 'word-slash-atpt)
;;;###autoload
(defun word-slash-atpt (&optional arg) 
  "Slash 'word at point if any. "
  (interactive "*p")
    (th-slash 'word arg))

(defalias 'trim-word-atpt 'word-trim-atpt)
;;;###autoload
(defun word-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'word t t))

(defalias 'trim-word-left-atpt 'word-left-trim-atpt)
;;;###autoload
(defun word-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'word t nil))

(defalias 'trim-word-right-atpt 'word-right-trim-atpt)
;;;###autoload
(defun word-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'word nil t))

(defalias 'forward-word-atpt 'word-forward-atpt)
;;;###autoload
(defun word-forward-atpt (&optional arg) 
  "Moves forward over word at point if any, does nothing otherwise.
Returns bounds of word as a list. "
  (interactive "p")
  (th-forward 'word arg))

(defalias 'backward-word-atpt 'word-backward-atpt)
;;;###autoload
(defun word-backward-atpt (&optional arg) 
  "Moves backward over word before point if any, does nothing otherwise.
Returns bounds of word as a list. "
  (interactive "p")
  (th-backward 'word arg))

(defalias 'transpose-word-atpt 'word-transpose-atpt)
;;;###autoload
(defun word-transpose-atpt (&optional arg) 
  "Transposes word with word before point if any. "
  (interactive "*P")
  (th-transpose 'word arg))

(defalias 'sort-word-atpt 'word-sort-atpt)
;;;###autoload
(defun word-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts word's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'word reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-word-atpt 'word-check-atpt)
;;;###autoload
(defun word-check-atpt (&optional arg) 
  " word at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "word" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "word" "-end-position-atpt")))))
    (th-check 'word arg)))
   

;;;###autoload
(defun count-alnum-in-word-atpt (&optional beg end) 
    "Counts matches of alnum in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'word beg end (interactive-p)))

(defun ratio-alnum-in-word-atpt (&optional beg end) 
    "Relation of alnum in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alnum 'word beg end t (interactive-p)))



;;;###autoload
(defun count-alpha-in-word-atpt (&optional beg end) 
    "Counts matches of alpha in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'word beg end (interactive-p)))

(defun ratio-alpha-in-word-atpt (&optional beg end) 
    "Relation of alpha in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'alpha 'word beg end t (interactive-p)))



;;;###autoload
(defun count-ascii-in-word-atpt (&optional beg end) 
    "Counts matches of ascii in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'word beg end (interactive-p)))

(defun ratio-ascii-in-word-atpt (&optional beg end) 
    "Relation of ascii in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'ascii 'word beg end t (interactive-p)))



;;;###autoload
(defun count-digit-in-word-atpt (&optional beg end) 
    "Counts matches of digit in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'word beg end (interactive-p)))

(defun ratio-digit-in-word-atpt (&optional beg end) 
    "Relation of digit in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'digit 'word beg end t (interactive-p)))



;;;###autoload
(defun count-graph-in-word-atpt (&optional beg end) 
    "Counts matches of graph in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'word beg end (interactive-p)))

(defun ratio-graph-in-word-atpt (&optional beg end) 
    "Relation of graph in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'graph 'word beg end t (interactive-p)))



;;;###autoload
(defun count-lower-in-word-atpt (&optional beg end) 
    "Counts matches of lower in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'word beg end (interactive-p)))

(defun ratio-lower-in-word-atpt (&optional beg end) 
    "Relation of lower in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'lower 'word beg end t (interactive-p)))



;;;###autoload
(defun count-punct-in-word-atpt (&optional beg end) 
    "Counts matches of punct in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'word beg end (interactive-p)))

(defun ratio-punct-in-word-atpt (&optional beg end) 
    "Relation of punct in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'punct 'word beg end t (interactive-p)))



;;;###autoload
(defun count-upper-in-word-atpt (&optional beg end) 
    "Counts matches of upper in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'word beg end (interactive-p)))

(defun ratio-upper-in-word-atpt (&optional beg end) 
    "Relation of upper in word if useful, returns nil otherwise "
  (interactive)
  (th-count-base  'upper 'word beg end t (interactive-p)))



;;;###autoload
(defun xdigit-atpt (&optional arg) 
  "Returns xdigit at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
    (th 'xdigit arg))

(defalias 'bounds-of-xdigit-atpt 'xdigit-bounds-atpt)
;;;###autoload
(defun xdigit-bounds-atpt (&optional arg) 
  "Returns borders of xdigit at point if any, nil otherwise. 
With \C-u customization of `with-delimiters-at' is reverted. "
  (interactive "p")
  (th-bounds 'xdigit arg))

;;;###autoload
(defun xdigit-beginning-position-atpt (&optional arg) 
  "Returns beginning of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-beg 'xdigit arg))

;;;###autoload
(defun xdigit-end-position-atpt (&optional arg) 
  "Returns end position of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-end 'xdigit arg))

(defalias 'beginning-of-xdigit-atpt 'xdigit-beginning-atpt)
;;;###autoload
(defun xdigit-beginning-atpt (&optional arg) 
  "Goto beginning of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-gotobeg 'xdigit arg))

(defalias 'end-of-xdigit-atpt 'xdigit-end-atpt)
;;;###autoload
(defun xdigit-end-atpt (&optional arg) 
  "Goto end of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-gotoend 'xdigit arg))

(defalias 'length-of-xdigit-atpt 'xdigit-length-atpt)
;;;###autoload
(defun xdigit-length-atpt (&optional arg) 
  "Returns beginning of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-length 'xdigit arg))

(defalias 'copy-xdigit-atpt 'xdigit-copy-atpt)
;;;###autoload
(defun xdigit-copy-atpt (&optional arg) 
  "Makes a copy of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "p")
  (th-copy 'xdigit arg))

(defalias 'escape-xdigit-atpt 'xdigit-escape-atpt)
;;;###autoload
(defun xdigit-escape-atpt () 
  "Makes a copy of symbol or char-class xdigit at point if any, nil otherwise "
  (interactive "*")
  (th-escape 'xdigit))

(defalias 'doublequote-xdigit-atpt 'xdigit-doublequote-atpt)
;;;###autoload
(defun xdigit-doublequote-atpt (&optional arg) 
  "Doublequotes 'xdigit at point if any. "
  (interactive "*p")
    (th-doublequote 'xdigit arg))

(defalias 'dollar-xdigit-atpt 'xdigit-dollar-atpt)
;;;###autoload
(defun xdigit-dollar-atpt (&optional arg) 
  "Doublequotes 'xdigit at point if any. "
  (interactive "*p")
    (th-dollar 'xdigit arg))

(defalias 'angle-greater-xdigit-atpt 'xdigit-greater-angle-atpt)
;;;###autoload
(defun xdigit-greater-angle-atpt (&optional arg) 
  "Sets angles for 'xdigit after point if any. "
  (interactive "*p")
    (th-greater-angle 'xdigit arg))

(defalias 'angle-lower-xdigit-atpt 'xdigit-lower-angle-atpt)
;;;###autoload
(defun xdigit-lower-angle-atpt (&optional arg) 
  "Sets angles for 'xdigit after point if any. "
  (interactive "*p")
    (th-lower-angle 'xdigit arg))

(defalias 'backslash-xdigit-atpt 'xdigit-backslash-atpt)
;;;###autoload
(defun xdigit-backslash-atpt (&optional arg) 
  "Backslash 'xdigit at point if any. "
  (interactive "*p")
    (th-backslash 'xdigit arg))   

(defalias 'brace-xdigit-atpt 'xdigit-brace-atpt)
;;;###autoload
(defun xdigit-brace-atpt (&optional arg) 
  "Braces 'xdigit at point if any. "
  (interactive "*p")
    (th-brace 'xdigit arg))

(defalias 'bracket-xdigit-atpt 'xdigit-bracket-atpt)
;;;###autoload
(defun xdigit-bracket-atpt (&optional arg) 
  "Brackets 'xdigit after point if any. "
  (interactive "*p")
    (th-bracket 'xdigit arg))

;; (defalias 'comment-xdigit-atpt 'xdigit-comment-atpt)
;;;###autoload
(defun xdigit-comment-atpt (&optional arg) 
  "Comments 'xdigit at point if any. "
  (interactive "*p")
  (th-comment 'xdigit arg))

(defalias 'mark-xdigit-atpt 'xdigit-mark-atpt)
;;;###autoload
(defun xdigit-mark-atpt () 
  "Marks 'xdigit at point if any. "
  (interactive)
  (th-mark 'xdigit))

(defalias 'hide-xdigit-atpt 'xdigit-hide-atpt)
;;;###autoload
(defun xdigit-hide-atpt () 
  "Toggles hide 'xdigit in region or buffer if any. "
  (interactive)
  (th-hide 'xdigit))

(defalias 'highlight-xdigit-atpt-mode 'xdigit-highlight-atpt-mode)
(defun xdigit-highlight-atpt-mode (&optional arg) 
  "Toggles xdigit-highlight-atpt-mode "
  (interactive "p")
  (th-highlight 'xdigit arg))

(defalias 'kill-xdigit-atpt 'xdigit-kill-atpt)
;;;###autoload
(defun xdigit-kill-atpt (&optional arg) 
  "Kills xdigit at point if any. "
  (interactive "*P")
  (th-kill 'xdigit arg))

(defalias 'parentize-xdigit-atpt 'xdigit-parentize-atpt)
;;;###autoload
(defun xdigit-parentize-atpt (&optional arg) 
  "Parentizes 'xdigit at point if any, does nothing otherwise"
  (interactive "*p")
    (th-parentize 'xdigit arg))

(defalias 'separate-xdigit-atpt 'xdigit-separate-atpt)
;;;###autoload
(defun xdigit-separate-atpt (&optional arg) 
  "Separates 'xdigit at point if any, does nothing otherwise
  inserts newlines, borders are the beginning or the end of buffer "
  (interactive "*p")
  (th-separate 'xdigit arg))

(defalias 'singlequote-xdigit-atpt 'xdigit-singlequote-atpt)
;;;###autoload
(defun xdigit-singlequote-atpt (&optional arg) 
  "Singlequotes 'xdigit at point if any. "
  (interactive "*p")
    (th-singlequote 'xdigit arg))

(defalias 'slash-xdigit-atpt 'xdigit-slash-atpt)
;;;###autoload
(defun xdigit-slash-atpt (&optional arg) 
  "Slash 'xdigit at point if any. "
  (interactive "*p")
    (th-slash 'xdigit arg))

(defalias 'trim-xdigit-atpt 'xdigit-trim-atpt)
;;;###autoload
(defun xdigit-trim-atpt (&optional arg) 
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'xdigit t t))

(defalias 'trim-xdigit-left-atpt 'xdigit-left-trim-atpt)
;;;###autoload
(defun xdigit-left-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'xdigit t nil))

(defalias 'trim-xdigit-right-atpt 'xdigit-right-trim-atpt)
;;;###autoload
(defun xdigit-right-trim-atpt (&optional arg)
  "Removes leading and trailing char. "
  (interactive "*p")
  (th-trim 'xdigit nil t))

(defalias 'forward-xdigit-atpt 'xdigit-forward-atpt)
;;;###autoload
(defun xdigit-forward-atpt (&optional arg) 
  "Moves forward over xdigit at point if any, does nothing otherwise.
Returns bounds of xdigit as a list. "
  (interactive "p")
  (th-forward 'xdigit arg))

(defalias 'backward-xdigit-atpt 'xdigit-backward-atpt)
;;;###autoload
(defun xdigit-backward-atpt (&optional arg) 
  "Moves backward over xdigit before point if any, does nothing otherwise.
Returns bounds of xdigit as a list. "
  (interactive "p")
  (th-backward 'xdigit arg))

(defalias 'transpose-xdigit-atpt 'xdigit-transpose-atpt)
;;;###autoload
(defun xdigit-transpose-atpt (&optional arg) 
  "Transposes xdigit with xdigit before point if any. "
  (interactive "*P")
  (th-transpose 'xdigit arg))

(defalias 'sort-xdigit-atpt 'xdigit-sort-atpt)
;;;###autoload
(defun xdigit-sort-atpt (reverse beg end &optional startkeyfun endkeyfun predicate) 
  "Sorts xdigit's in region, with ARG in reverse order.
  STARTKEYFUN may be replaced by a function which stops at an alternative beginning.
  ENDKEYFUN might be a function specifying THING's end when sorting.  
  With PREDICATE define a the function to compare. Defaults are `<' for numbers, otherwise `string<'.
  See doku from `sort-subr', for details.
  "
  (interactive "*P\nr")
  (let ((reverse (when reverse)) startkeyfun endkeyfun predicate)
    (unless (region-exists-p) (message "%s" "Region must be active!"))
    (th-sort 'xdigit reverse beg end startkeyfun endkeyfun predicate)))

(defalias 'check-xdigit-atpt 'xdigit-check-atpt)
;;;###autoload
(defun xdigit-check-atpt (&optional arg) 
  " xdigit at point if any, nil otherwise "
  (interactive "p")
  (let ((beg (funcall (intern-soft (concat "xdigit" "-beginning-position-atpt"))))
	(end (funcall (intern-soft (concat "xdigit" "-end-position-atpt")))))
    (th-check 'xdigit arg)))
   

(defun angled-greater-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with angled-greater(s),
otherwise copy angled-greater(ed) at point.
With arg, copy angled-greater(ed) without delimiters. 
With negative argument kill angled-greater(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'angled-greater arg))

(defun angled-lower-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with angled-lower(s),
otherwise copy angled-lower(ed) at point.
With arg, copy angled-lower(ed) without delimiters. 
With negative argument kill angled-lower(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'angled-lower arg))

(defun backslashed-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with backslashed(s),
otherwise copy backslashed(ed) at point.
With arg, copy backslashed(ed) without delimiters. 
With negative argument kill backslashed(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'backslashed arg))

(defun braced-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with braced(s),
otherwise copy braced(ed) at point.
With arg, copy braced(ed) without delimiters. 
With negative argument kill braced(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'braced arg))

(defun bracketed-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with bracketed(s),
otherwise copy bracketed(ed) at point.
With arg, copy bracketed(ed) without delimiters. 
With negative argument kill bracketed(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'bracketed arg))

(defun commented-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with commented(s),
otherwise copy commented(ed) at point.
With arg, copy commented(ed) without delimiters. 
With negative argument kill commented(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'commented arg))

(defun delimited-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with delimited(s),
otherwise copy delimited(ed) at point.
With arg, copy delimited(ed) without delimiters. 
With negative argument kill delimited(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'delimited arg))

(defun dollared-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with dollared(s),
otherwise copy dollared(ed) at point.
With arg, copy dollared(ed) without delimiters. 
With negative argument kill dollared(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'dollared arg))

(defun doublequoted-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with doublequoted(s),
otherwise copy doublequoted(ed) at point.
With arg, copy doublequoted(ed) without delimiters. 
With negative argument kill doublequoted(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'doublequoted arg))

(defun parentized-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with parentized(s),
otherwise copy parentized(ed) at point.
With arg, copy parentized(ed) without delimiters. 
With negative argument kill parentized(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'parentized arg))

(defun singlequoted-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with singlequoted(s),
otherwise copy singlequoted(ed) at point.
With arg, copy singlequoted(ed) without delimiters. 
With negative argument kill singlequoted(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'singlequoted arg))

(defun slashed-or-copy-atpt (&optional arg) 
  "If region is highlighted, provide THING at point with slashed(s),
otherwise copy slashed(ed) at point.
With arg, copy slashed(ed) without delimiters. 
With negative argument kill slashed(ed) at point. "
  (interactive "p")
  (th-base-copy-or 'slashed arg))

(provide 'thing-at-point-utils)
;;; thing-at-point-utils.el ends here
